package provide qcode 14.8

namespace eval qc {
    namespace export pgpass*
}


proc qc::pgpass2ldict {filename} {
    #| Convert the contents of the current user's .pgpass as an ldict
    set ldict [list]
    set lines [split [qc::cat $filename] \n]
    foreach line $lines {
        if { [regexp {^ *#} $line] } {
            continue
        }
        if { [regexp {^([^:]+):([^:]+):([^:]+):([^:]+):(.+)$} $line -> hostname port database username password] } {
            lappend ldict [dict_from hostname port database username password]
        }
    }
    return $ldict
}

proc qc::pgpass_credentials_exist {pgpass_filename db_name} {
    #| Check if credentials for this db_name exist in the users ~/.pgpass file
    if { [file exists $pgpass_filename] } {
        set ldict [qc::pgpass2ldict $pgpass_filename]
        set index [qc::ldict_search ldict database $db_name]
        if { $index >= 0 } {
            return true
        } 
    }
    return false
}
proc qc::pgpass_credentials {pgpass_filename db_name} {
    #| Return the access credentials for this database in the pgpass file.
    set ldict [qc::pgpass2ldict $pgpass_filename]
    set index [qc::ldict_search ldict database $db_name]
    return [lindex $ldict $index]
}   

namespace eval qc {
    namespace export format_*
}

proc qc::format_commify {number} {
    #| Commify number into groups of three 2314 -> 2,314
    while {[regsub {^([-+]?\d+)(\d\d\d)} $number {\1,\2} number]} {}
    return $number
}

proc qc::format_money { value } {
    return [format_commify [qc::round $value 2]]
}

proc qc::format_dec { value dec_places } {
    #| Round and commify decimal
    #| 2314.235 -> 2,314.24
    return [format_commify [qc::round $value $dec_places]]
}

proc qc::format_int {value} {
    # Round to integer and commify
    # 2314.235 -> 2,314
    return [format_commify [qc::round $value 0]]
}

proc qc::format_linebreak {string width} {
    #| Split $string into a list of lines without exceeding $width
    #| Avoid splitting words
    set result {}
    while {[string length $string]>$width} {
	set position [string last " " $string $width]
	if { $position <=0 } {
	    set position $width
	    lappend result [string range $string 0 [expr {$position-1}]]
	    set string [string range $string $position end]
	} else {
	    lappend result [string range $string 0 [expr {$position-1}]]
	    set string [string range $string [expr {$position+1}] end]
	}
      
    }
    lappend result $string
}

proc qc::format_cc {cc_no} {
    #| Format a credit card number in groups of 4 digits
    #| strip non digits
    regsub -all {[^0-9]} $cc_no {} cc_no
    return [join [format_linebreak $cc_no 4] " "]
}

proc qc::format_cc_bin {cc_no {prefix 6}} {
    #| First prefix digits of cc_no
    if { $prefix > 6 } { error "prefix must be less than 6" }
    
    regsub -all {[^0-9\*]} $cc_no {} cc_no
    return [string range $cc_no 0 "$prefix-1"]
}

proc qc::format_cc_tail {cc_no {suffix 4}} {
    #| First prefix digits of cc_no
    if { $suffix > 4 } { error "suffix must be less than 4" }
    
    regsub -all {[^0-9\*]} $cc_no {} cc_no
    return [string range $cc_no "end-[expr {$suffix - 1}]" end]
}

proc qc::format_cc_masked {cc_no {prefix 6} {suffix 4}} {
    if { ![qc::is creditcard $cc_no] } {
	error "Can't mask \"$cc_no\" because its is not a credit card number."
    }
    if { $prefix > 6 } { error "prefix must be less than 6" }
    if { $suffix > 4 } { error "suffix must be less than 4" }

    regsub -all {[^0-9]} $cc_no {} cc_no
    set masked_cc_no [string range $cc_no 0 [expr {$prefix - 1}]]
    append masked_cc_no [string repeat * [expr {[string length $cc_no] - $prefix - $suffix}]]
    append masked_cc_no [string range $cc_no end-[expr {$suffix - 1}] end] 

    return [join [format_linebreak $masked_cc_no 4] " "]
}

proc qc::format_cc_masked_string {string {prefix 6} {suffix 4}} {
    if { $prefix > 6 } { error "prefix must be less than 6" }
    if { $suffix > 4 } { error "suffix must be less than 4" }
    
    set cc_no_pattern {(?:^|[^0-9]\.|[^\.\+0-9])([3456](?:\d[ \-\.]*){11,17}\d)(?:[^0-9]|$)}
    set indices [regexp -inline -indices -all -nocase $cc_no_pattern $string]
    set start 0
    set masked_string ""
   
    foreach {. item} $indices {
	set cc_no_str [string range $string [lindex $item 0] [lindex $item 1]]
	set cc_no [string map {" " "" - "" . ""} $cc_no_str]

	if { [qc::is creditcard $cc_no] } {
	    append masked_string [string range $string $start [lindex $item 0]-1]
	    
	    set masked_cc_no [string range $cc_no 0 [expr {$prefix - 1}]]
	    append masked_cc_no [string repeat * [expr {[string length $cc_no] - $prefix - $suffix}]]
	    append masked_cc_no [string range $cc_no end-[expr {$suffix - 1}] end] 

	    append masked_string [join [format_linebreak $masked_cc_no 4] " "]
	    set start [lindex $item 1]+1
	}
    }    
    append masked_string [string range $string $start end]
    
    return $masked_string
}

proc qc::format_ordinal {number} {
    #| Format number with suffix 23 -> 23rd or 4 -> 4th
    # Taken from TCL Wiki RS
    set suffix th
    if {($number%100)<10 || ($number%100)>20} {
	switch -- [expr abs($number)%10] {
	    1 {set suffix st}
	    2 {set suffix nd}
	    3 {set suffix rd}
	}
    }
    append number $suffix
}

# TODO Could this be named differently? Not sure a user would expect this to truncate
# strings longer than $width
proc qc::format_right0 {string width} {
    #| Right justified padded with leading 0's
    return [format "%0${width}.${width}s" $string]
}

proc qc::format_left {string width} {
    #| Padd with spaces left justified
    return [format "%-${width}.${width}s" $string]
}

proc qc::format_right {string width} {
    # Pad with spaces right justified
    return [format "%${width}.${width}s" $string]
}

proc qc::format_center { string width } {
    #| Pad string with spaces to be aligned centrally
    if { [string length $string]<$width } {
	append string [string repeat " " [expr {($width - [string length $string])/2}]]
    }
    return [format "%${width}.${width}s" $string]
}

proc qc::format_bool { value {true Yes} {false No}} {
    #| Cast boolean and wrap in span tags with style
    if { [string is true -strict $value] } {
	return "<span class=\"true\">$true</span>"
    } else {
	return "<span class=\"false\">$false</span>"
    }
}

proc qc::format_yesno { value } {
    #| Call format_bool with default Yes/No
    return [qc::format_bool $value Yes No]
}

proc qc::format_number {args} {
    #| format a number
    args $args -dp ? -sigfigs ? -commify yes -- value
    if { [info exists sigfigs] } {
	set value [qc::sigfigs $value $sigfigs]
    }
    if { [info exists dp] } {
	set value [qc::round $value $dp]
    }
    if { [true $commify] } {
	set value [qc::format_commify $value]
    } 
    return $value
}

proc qc::format_if_number {args} {
    #| If value is a number then commify
    args $args -dp ? -sigfigs ? -zeros yes -commify yes -- value
    if { [qc::is decimal $value] } {
	if { [info exists sigfigs] && [qc::is integer $sigfigs]} {
	    set value [qc::sigfigs $value $sigfigs]
	}
	if { [info exists dp] && [qc::is integer $dp] } {
	    set value [qc::round $value $dp]
	}
	if { !$zeros && $value==0 } {
	    set value ""
	}
	if { [true $commify] } {
	    set value [format_commify $value]
	} 
    }
    return $value
}

proc qc::format_safe_html {safe_html} {
    #| Formats the given text by removing <root> node if present and converting back to HTML from XML.
    if {[qc::is safe_html $safe_html]} {
        set doc [dom parse $safe_html]
        set html [$doc asHTML -escapeNonASCII -htmlEntities]
        # remove root node if it was present
        regexp {^<root>(.+)</root>$} $html -> html
        $doc delete
        return $html
    } else {
        return -code error  "Cannot format \"[qc::trunc $html 50]...\" as it is not safe HTML."
    }
}

namespace eval qc::oauth2 {

    namespace export \
        authorize_code_url \
        token_request
    namespace ensemble create

    proc authorize_code_url {args} {
        #| Returns a URL to seek authorization code from a service.

        qc::args \
            $args \
            -state ? \
            -scope ? \
            -redirect_uri ? \
            server_url \
            client_id

        set response_type "code"

        return [qc::url \
                    $server_url \
                    ~ \
                    client_id \
                    response_type \
                    redirect_uri \
                    scope \
                    state \
                   ]
    }

    ############################################################################
    #
    # Token request procs to authorise access and obtain a valid access token.
    #
    ############################################################################
    namespace eval token_request {

        namespace export \
            authorization_code \
            password \
            client_credentials \
            refresh
        namespace ensemble create

        # Grant Types
        variable authorization_code_grant_type "authorization_code"
        variable password_grant_type           "password"
        variable client_credentials_grant_type "client_credentials"
        variable refresh_grant_type            "refresh_token"

        proc authorization_code {args} {
            #| Request authorisation using the Authorization Code grant type.

            qc::args \
                $args \
                -client_secret ? \
                -redirect_uri ? \
                -basic_auth true \
                -accept "application/json; charset=utf-8" \
                -valid_response_codes [list 200 400 401] \
                server_url \
                client_id \
                code

            variable authorization_code_grant_type

            set http_post_flags [dict create \
                                     -valid_response_codes $valid_response_codes \
                                    ]

            if { $accept ne "" } {
                dict set http_post_flags -accept $accept
            }

            set data [dict create \
                          grant_type $authorization_code_grant_type \
                          code $code \
                          ]

            if { [info exists client_secret] } {
                if { !$basic_auth } {
                    dict set data client_id $client_id
                    dict set data client_secret $client_secret
                } else {
                    set encoded_credentials [base64::encode \
                                                 -wrapchar "" \
                                                 "${client_id}:${client_secret}" \
                                                ]
                    dict set http_post_flags -authorization "Basic $encoded_credentials"
                }
            } else {
                dict set data client_id $client_id
            }

            if { [info exists redirect_uri] } {
                dict set data redirect_uri $redirect_uri
            }

            return [qc::http_post \
                        -response_headers true \
                        -response_code true \
                        {*}$http_post_flags \
                        $server_url \
                        {*}$data \
                       ]
        }

        proc password {args} {
            #| Request authorisation using the Password grant type.

            qc::args \
                $args \
                -client_id ? \
                -client_secret ? \
                -basic_auth true \
                -scope ? \
                -accept "application/json; charset=utf-8" \
                -valid_response_codes [list 200 400 401] \
                server_url \
                username \
                password

            variable password_grant_type

            set http_post_flags [dict create \
                                     -valid_response_codes $valid_response_codes \
                                    ]

            if { $accept ne "" } {
                dict set http_post_flags -accept $accept
            }

            set data [dict create \
                          grant_type $password_grant_type \
                          username $username \
                          password $password \
                          ]

            if { [info exists client_id]
                 && [info exists client_secret] } {
                if { !$basic_auth } {
                    dict set data client_id $client_id
                    dict set data client_secret $client_secret
                } else {
                    set encoded_credentials [base64::encode \
                                                 -wrapchar "" \
                                                 "${client_id}:${client_secret}" \
                                                ]
                    dict set http_post_flags -authorization "Basic $encoded_credentials"
                }
            }

            if { [info exists scope] } {
                dict set data scope $scope
            }

            return [qc::http_post \
                        -response_headers true \
                        -response_code true \
                        {*}$http_post_flags \
                        $server_url \
                        {*}$data \
                       ]
        }

        proc client_credentials {args} {
            #| Request authorisation using the Client Credentials grant type.

            qc::args \
                $args \
                -basic_auth true \
                -scope ? \
                -accept "application/json; charset=utf-8" \
                -valid_response_codes [list 200 400 401] \
                server_url \
                client_id \
                client_secret

            variable client_credentials_grant_type

            set http_post_flags [dict create \
                                     -valid_response_codes $valid_response_codes \
                                    ]

            if { $accept ne "" } {
                dict set http_post_flags -accept $accept
            }

            set data [dict create \
                          grant_type $client_credentials_grant_type \
                          ]

            if { !$basic_auth } {
                dict set data client_id $client_id
                dict set data client_secret $client_secret
            } else {
                set encoded_credentials [base64::encode \
                                             -wrapchar "" \
                                             "${client_id}:${client_secret}" \
                                            ]
                dict set http_post_flags -authorization "Basic $encoded_credentials"
            }

            if { [info exists scope] } {
                dict set data scope $scope
            }

            return [qc::http_post \
                        -response_headers true \
                        -response_code true \
                        {*}$http_post_flags \
                        $server_url \
                        {*}$data \
                       ]
        }

        proc refresh {args} {
            #| Request authorisation using the Refresh Token grant type.

            qc::args \
                $args \
                -client_id ? \
                -client_secret ? \
                -basic_auth true \
                -scope ? \
                -accept "application/json; charset=utf-8" \
                -valid_response_codes [list 200 400 401] \
                server_url \
                refresh_token

            variable refresh_grant_type

            set http_post_flags [dict create \
                                     -valid_response_codes $valid_response_codes \
                                    ]

            if { $accept ne "" } {
                dict set http_post_flags -accept $accept
            }

            set data [dict create \
                          grant_type $refresh_grant_type \
                          refresh_token $refresh_token \
                          ]

            if { [info exists client_id]
                 && [info exists client_secret] } {
                if { !$basic_auth } {
                    dict set data client_id $client_id
                    dict set data client_secret $client_secret
                } else {
                    set encoded_credentials [base64::encode \
                                                 -wrapchar "" \
                                                 "${client_id}:${client_secret}" \
                                                ]
                    dict set http_post_flags -authorization "Basic $encoded_credentials"
                }
            }

            if { [info exists scope] } {
                dict set data scope $scope
            }

            return [qc::http_post \
                        -response_headers true \
                        -response_code true \
                        {*}$http_post_flags \
                        $server_url \
                        {*}$data \
                       ]
        }
    }
}
namespace eval qc {
    namespace export css_selector2xpath
}

proc qc::css_selector2xpath {selector {xpath ""}} {
    #| Convert a css selector to an XPath
    # (Errors on pseudo-elements and dynamic pseudo-classes such as :hover)
    # (Also does not support any of the x-of-type pseudo-classes unless combined with an explicit element type)

    # A css selector consists of one or more "simple selector sequences", separated by "combinators"
    # This maps almost exactly to XPath, which consists of one or more "Steps".

    set combinator " "
    set sequence_combinator_list [list]
    while { $selector ne "" } {
        switch $combinator {
            " " { append xpath {//} }
            > { append xpath {/} }
            ~ { append xpath {/following-sibling::} }
            + { append xpath {/following-sibling::*[1]/self::} }
        }

        # Shift next simple selector sequence from the selector string
        set depth 0
        set sequence ""
        for {set i 0} {$i < [string length $selector]} {incr i} {
            switch [string index $selector $i] {
                \( - \[ {incr depth}
                \) - \] {incr depth -1}
                " " - > - ~ - + {
                    if { $depth == 0 } {
                        break
                    }
                }
            }
            append sequence [string index $selector $i]               
        }
        if { $depth > 0 } {
            error "Unable to parse selector $selector - unclosed parentheses"
        }
        set selector [string range $selector $i end]

        append xpath [qc::css_simple_selector_sequence2xpath $sequence]

        if { $selector ne "" } {
            # Shift next combinator from selector string
            if { ! [regexp {^\s*([\s>+~])\s*(.+?)$} $selector -> combinator remainder] } {
                error "Unable to parse combinator / sequence from selector string $selector"
            }
            set selector $remainder
        }
    }
    return $xpath
}

proc qc::css_simple_selector_sequence2xpath {sequence} {
    #| Convert a css simple selector sequence to an XPath node test + predicate

    set xpath ""
    # Shift type selector from simple selector sequence
    if { [regexp {^([^\[:.\#]+)(.*?)$} $sequence -> element_type remainder] } {
        set sequence $remainder
    } else {
        set element_type "*"
    }

    # Css type and universal selectors have the same syntax as XPath node tests
    if { [regexp {[^a-zA-Z0-9*]} $element_type] } {
        error "Unable to parse $sequence - Invalid element type"
    }
    append xpath $element_type

    while { $sequence ne "" } {
        # Shift next simple selector from simple selector sequence
        set simple_selector [string index $sequence 0]
        set depth 0
        set end false
        for {set i 1} {$i < [string length $sequence]} {incr i} {
            switch [string index $sequence $i] {
                \( { incr depth }
                \) { incr depth -1 }
                \[ -
                : -
                . -
                \# {
                    if { $depth == 0 } {
                        break
                    }
                }
            }
            append simple_selector [string index $sequence $i]
        }
        if { $depth > 0 } {
            error "Unable to parse $sequence - unclosed parentheses"
        }
        set sequence [string range $sequence $i end]

        set predicate [qc::css_simple_selector2xpath $simple_selector $element_type]
        append xpath "\[$predicate\]"
    }
    return $xpath
}

proc qc::css_simple_selector2xpath {simple_selector {element_type *}} {
    #| Convert a css simple selector (but not a type or universal selector) to an XPath predicate
    switch [string index $simple_selector 0] {
        . { # class selector
            set class [string range $simple_selector 1 end]
            return "contains(concat(' ',normalize-space(@class),' '),[qc::xpath_literal " $class "])"
        }
        \# { # id selector
            set id [string range $simple_selector 1 end]
            return "@id=[qc::xpath_literal $id]"
        }
        : { # pseudo-class
            set pseudo_class [string range $simple_selector 1 end]
            return [qc::css_pseudo_class2xpath $pseudo_class $element_type]
        }
        \[ { #attribute selector
            if { [string index $simple_selector end] ne "\]" } {
                error "Unable to parse $simple_selector"
            }
            set attribute_selector [string range $simple_selector 1 end-1]
            return [qc::css_attribute_selector2xpath $attribute_selector]
        }
    }
}

proc qc::css_attribute_selector2xpath {attribute_selector} {
    #| Convert a css attribute selector to an XPath expression
    if { ! [regexp {^([-a-zA-Z0-9_]+)(?:([~|^$*]?=)"(.*)")?$} $attribute_selector -> attribute test value] } {
        error "Unable to parse $attribute_selector"
    }
    switch $test {
        "" {
            # Element has the attribute
            return "@$attribute"
        }
        = {
            # Element has attribute, exactly equal to value
            return "@${attribute}=[qc::xpath_literal $value]"
        }
        ~= {
            # Treat the attribute value as a space-separated list of words,
            # test whether $value is one of those words
            if { $value eq "" || [string first " " $value] != -1 } {
                return "false()"
            } else {
                return "contains(concat(' ',normalize-space(@${attribute}),' '),[qc::xpath_literal " ${value} "])"
            }
        }
        |= {
            # Attribute is hyphen separated list of values starting with $value
            return "@${attribute}=[qc::xpath_literal $value] or starts-with(@${attribute},[qc::xpath_literal "${value}-"])"
        }
        ^= {
            # Attribute starts with $value
            if { $value eq "" } {
                return "false()"
            } else {
                return "starts-with(@${attribute},[qc::xpath_literal $value])"
            }
        }
        $= {
            # Attribute ends with $value
            if { $value eq "" } {
                return "false()"
            } else {
                return "substring(@${attribute},string-length(@${attribute}) - string-length([qc::xpath_literal $value])) = [qc::xpath_literal $value]"
            }
        }
        *= {
            # Attribute contains substring $value
            if { $value eq "" || [string first " " $value] != -1 } {
                return "false()"
            } else {
                return "contains(@${attribute},[qc::xpath_literal $value])"
            }
        }
        default {
            error "Unknown attribute selector $attribute_selector"
        }
    }
}

proc qc::css_pseudo_class2xpath {pseudo_class {element_type *}} {
    #| Convert a css pseudo-class selector to an XPath expression
    if { [string index $pseudo_class 1] eq ":" } {
        error "Unable to convert pseudo-element :$pseudo_class to xpath"
    }
    # Many pseudo-classes allow optional arguments (eg. :nth-child(2) ).
    if { ! [regexp {^([-a-z]+)(?:\((.+)\))?$} $pseudo_class -> function argument] } {
        error "Unable to parse pseudo-class $pseudo_class"
    }
    if { $function in {nth-child nth-last-child nth-of-type nth-last-of-type lang not} } {
        if { $argument eq "" } {
            error "Pseudo class $function requires an argument"
        }
    } else {
        if { $argument ne "" } {
            error "Pseduo-class $function does not accept arguments"
        }
    }
    switch $function {
        root {
            return "count(ancestor::*)=0"
        }
        nth-child {
            return [qc::nth_term $argument "(count(preceding-sibling::*) + 1)"]
        }
        nth-last-child {
            return [qc::nth_term $argument "(count(following-sibling::*) + 1)"]
        }
        first-child {
            return {count(preceding-sibling::*) = 0}
        }
        last-child {
            return {count(following-sibling::*) = 0}
        }
        only-child {
            return {last()=1}
        }
        empty {
            return {count(*)=0}
        }
        lang {
            return "lang([qc::xpath_literal $argument])"
        }
        enabled {
            return {not(@disabled)}
        }
        disabled {
            return {@disabled}
        }
        checked {
            return {@checked}
        }
        not {
            if { [regexp {[:.#\[]} [string range $argument 1 end]] } {
                error "Error parsing pseudo-class $pseudo_class"
            }
            if { [string index $argument 0] in {\# . : \[} } {
                if { [regexp {^:not\(.*\)$}] } {
                    error "Nested negation not permitted by css specs"
                }
                return "not([qc::css_simple_selector2xpath $argument])"
            } {
                if { regexp {[^a-zA-Z0-9*]} $element_type } {
                    error "Error parsing pseudo-class $pseudo_class"
                }
                return "not($agument)"
            }
        }
        only-of-type {
            if { $element_type eq "*" } {
                error "x-of-type psuedo-classes not supported without explicit type"
            }
            return "count(preceding-sibling::${element_type}) + count(following-sibling::${element_type}) = 0"
        }
        nth-of-type {
            if { $element_type eq "*" } {
                error "x-of-type psuedo-classes not supported without explicit type"
            }
            return [nth_term $argument "(count(preceding-sibling::${element_type}) + 1)"]
        }
        nth-last-of-type {
            if { $element_type eq "*" } {
                error "x-of-type psuedo-classes not supported without explicit type"
            }
            return [nth_term $argument "(count(following-sibling::${element_type}) + 1)"]
        }
        first-of-type {
            if { $element_type eq "*" } {
                error "x-of-type psuedo-classes not supported without explicit type"
            }
            return "count(preceding-sibling::${element_type}) = 0"
        }
        last-of-type {
            if { $element_type eq "*" } {
                error "x-of-type psuedo-classes not supported without explicit type"
            }
            return "count(following-sibling::${element_type}) = 0"
        }

        link -
        visited -
        active -
        hover -
        focus -
        target {
            error "Unable to parse dynamic pseudo-class $function to xpath"
        }
        first-line -
        first-letter -
        before -
        after {
            error "Unable to convert pseudo-element $function to xpath"
        }
    }
}

proc qc::nth_term {expression n} {
    #| Given an XPath expression $n returning the position of the current node in some context,
    # and an expression in the form used by :nth-child in the css3 selector specs,
    # return a matching XPath expression.
    if { [regexp {^\s*([0-9]+)\s*$} $expression -> b] } {
        set a 0
    } else {
        if { ! [regexp {^\s*(?:([-+]?[0-9]+)[nN])?\s*([-+]\s*[0-9]+)?\s*$} $expression -> a b] } {
            if { ! [regexp -nocase {^\s*(odd|even)\s*$} $expression -> keyword] } {
                error "Unable to parse nth-term \"$expression\""
            }
            switch $keyword {
                odd {set a 2; set b 1}
                even {set a 2; set b 0}
            }
        }
    }
    if { $a eq "" && $b eq "" } {
        error "Unable to parse nth-term \"$expression\""
    }
    if { $a eq "" } {
        set a 0
    }
    if { $b eq "" } {
        set b 0
    }
    set b [string map {" " ""} $b]
    set a [string trimleft $a +]
    set b [string trimleft $b +]
    if { $a == 0 } {
        return "$n = $b"
    } elseif { $a > 0 } {
        set mod [expr {$b % $a}]
        if { $b > 0 } {
            return "$n >= $b and ($n mod $a) = $mod"
        } else {
            return "($n mod $a) = $mod"
        }
    } else {
        set mod [expr {$b % - $a}]
        if { $b >= 0 } {
            return "($n <= $b) and ($n mod $a) = $mod"
        } else {
            return "false()"
        }
    }
}

proc qc::xpath_literal {string} {
    if { [string first ' $string] == -1 } {
        return "'$string'"
    }
    if { [string first \" $string] == -1 } {
        return "\"$string\""
    }
    return "concat('[join [split $string '] {',"'",'}]')"
}
namespace eval qc {
    namespace export check checks
}

proc qc::check {args} {
    #| Check that the value of the local variable is of a given type or can be cast into that type.
    #| If the variable cannot be cast into the given type then throw an error of type USER.
    #| Use the error message given or a default message for the given type.
    #| The empty string is treated as a NULL value and always treated as valid unless NOT NULL is specified in types.
    switch -exact -- [llength $args] {
	0 - 
	1 {
	    return -code error "wrong#args: Should be varName type ?type? ?type? ?errorMessage?"
	}
	default {
	    set varName [lindex $args 0]
	    set args [lrange $args 1 end]
	}
    }
   
    # Parse
    array set alias [list INT INTEGER BOOL BOOLEAN STRING VARCHAR NZ NON_ZERO]
    set nulls yes
    set allow_html no
    set allow_creditcards yes
    set TYPES {}
    for {set index 0} {$index<[llength $args]} {incr index} {
	set TYPE [upper [lindex $args $index]]
	set type [lower $TYPE]
	set NEXT_TYPE [upper [lindex $args [expr {$index+1}]]]
	# This is an alias
	if { [info exists alias($TYPE)] } {
	    set TYPE $alias($TYPE)
	    set type [lower $TYPE]
	}
	# Creditcards not allowed in text fields by default
	if { [in {STRING VARCHAR} $TYPE] } {
	    set allow_creditcards no
	}
	# 
	if { $TYPE eq "NOT" && $NEXT_TYPE eq "NULL" } {
	    set nulls no
	    incr index 
	    continue
	} elseif { $TYPE eq "NOT" && $NEXT_TYPE eq "HTML" } {
	    set allow_html no
	    incr index 
	    continue
	} elseif {$TYPE eq "HTML"} {
	    set allow_html yes
	    continue
	} elseif { [in {NO NOT} $TYPE] && $NEXT_TYPE eq "CREDITCARDS" } {
	    set allow_creditcards no
	    incr index 
	    continue
	} elseif {$TYPE eq "CREDITCARDS"} {
	    set allow_creditcards yes
	    continue
	} elseif {$TYPE eq "CREDITCARD"} {
	    lappend TYPES CREDITCARD
	    set allow_creditcards yes
	    continue
	} elseif {[info commands "::qc::is_$type"] ne ""} {
	    lappend TYPES $TYPE
	    set n [llength [info args "::qc::is_$type"]]
	    if {$n>1} {
		set type_args($TYPE) [lrange $args [expr {$index+1}] [expr {$index+$n-1}]]
		incr index [expr {$n-1}]
	    }
	} elseif {[info commands is_$type] ne ""} {
	    lappend TYPES $TYPE
	    set n [llength [info args is_$type]]
	    if {$n>1} {
		set type_args($TYPE) [lrange $args [expr {$index+1}] [expr {$index+$n-1}]]
		incr index [expr {$n-1}]
	    }
	} elseif {$TYPE eq "PNZ"} {
	    lappend TYPES POS NZ
	    set nulls no
	} elseif {$TYPE eq "PRICE"} {
	    lappend TYPES DECIMAL PNZ 
	    set nulls no
	} elseif {$TYPE eq "QTY"} {
	    lappend TYPES INTEGER PNZ
	    set nulls no
	} elseif {$index>0 && $index==[llength $args]-1} {
	    set errorMessage [lindex $args end]
	} else {
	    error "Don't know how to check $TYPE"
	}
    }

    upvar 1 $varName varValue
    # Exists
    if { ![info exists varValue] } {error "No such variable $varName"}
    # NULLs
    if { $nulls && [string equal $varValue ""] } {
	# NULL values are OK
	return true
    }
    if { !$nulls && [string equal $varValue ""] } {
	default errorMessage "$varName is empty"
	error $errorMessage {} USER
    }
    
    foreach TYPE $TYPES {
	set type [lower $TYPE]
	# Try to cast to the type specified if a proc exists
	if { [in {POS NZ} $TYPE] && ![is_decimal $varValue] } {
	    # Implied cast
	    ::try {
                set varValue [qc::cast_decimal $varValue]
            } on error {} {} 
	} elseif { [info commands "::qc::cast_$type"] ne "" } {
	    ::try {
		if { [info exists type_args($TYPE)] } {
		    set varValue ["qc::cast_$type" $varValue {*}$type_args($TYPE)]
		} else {
		    set varValue ["qc::cast_$type" $varValue]
		}
	    } on error {} {}
	} elseif { [info commands cast_$type] ne "" } {
	    ::try {
		if { [info exists type_args($TYPE)] } {
		    set varValue [cast_$type $varValue {*}$type_args($TYPE)]
		} else {
		    set varValue [cast_$type $varValue]
		}
	    } on error {} {}
	}
	# Check
	if {!([info exists type_args($TYPE)] && [info commands "::qc::is_$type"] ne "" && [qc::is_$type $varValue {*}$type_args($TYPE)])
	    && 
	    !(![info exists type_args($TYPE)] && [info commands "::qc::is_$type"] ne "" && [qc::is_$type $varValue])
            &&
            !([info exists type_args($TYPE)] && [info commands "::qc::is_$type"] eq "" && [is_$type $varValue {*}$type_args($TYPE)])
	    && 
	    !(![info exists type_args($TYPE)] && [info commands "::qc::is_$type"] eq "" && [is_$type $varValue])} {
	    # Failed
	    if { [info commands not_$type] ne "" } { 
		if { [info exists type_args($TYPE)] } {
		    default errorMessage [not_$type $varName $varValue {*}$type_args($TYPE)] 
		} else {
		    default errorMessage [not_$type $varName $varValue] 
		}
	    } else {
		default errorMessage "\"[html_escape $varValue]\" is not a valid $type for $varName"
	    }
	    error $errorMessage {} USER
	}
    }

    # HTML Markup OFF by default
    if { !$allow_html && [regexp {[<>]} $varValue] } {
	error "\"[html_escape $varValue]\" contains HTML which is not allowed for $varName" {} USER
    }

    # Check contains creditcard by default
    if { !$allow_creditcards && [contains_creditcard $varValue] } {
	error [expr {[info exists errorMessage] ? $errorMessage : "\"[html_escape $varValue]\" contains a creditcard number which is not allowed for $varName"}] {} USER
    }

    return true
}

proc qc::checks { body } {
    # Call check foreach line of checks in the format
    # varName type ?type? ?type? ?errorMessage?
    global errorInfo errorCode
    set errors {}
    set lines [split $body \n]
    foreach line $lines {
	set line [string trim $line]
	if { [ne $line ""] && [catch [list uplevel 1 [list check {*}$line]] errorMessage]==1 } {
	    if {$errorCode eq "USER"} {
		lappend errors $errorMessage
	    } else {
		error $errorMessage $errorInfo $errorCode
	    }
	}
    }
    if { [llength $errors] > 0 } {
	error [html_list $errors] {} USER
    }
}


namespace eval qc {
    namespace export ns_set_*
}

#| Aolserver specific

proc qc::ns_set_to_vars { set_id {level 0}} {
    #| Take an ns_set with id $set_id from caller 
    #| and place variables in level $level.
    incr level
    foreach {key value} [ns_set array $set_id] {
	upset $level $key $value
    }
}

proc qc::ns_set_to_dict { set_id } {
    #| Take an ns_set with id $set_id from caller 
    #| return a dict
    return [ns_set array $set_id]
}

proc qc::ns_set_to_multimap { set_id } {
    #| Take an ns_set with id $set_id from caller 
    #| return a multimap of key values pairs
    return [ns_set array $set_id]
}

proc qc::ns_set_getall { set_id key } {
    #| Take an ns_set with id $set_id from caller 
    #| return all values for the key given
    set values {}
    if { ![string equal $set_id ""] } {
	set size [ns_set size $set_id]
	for {set i 0} {$i<$size} {incr i} {
	    if {[eq [ns_set key $set_id $i] $key]} {
		lappend values [ns_set value $set_id $i]
	    }
	}
    }
    return $values
}

proc qc::ns_set_keys { set_id } {
    #| Take an ns_set with id $set_id from caller 
    #| return all keys
    set keys {}
    if { ![string equal $set_id ""] } {
	set size [ns_set size $set_id]
	for {set i 0} {$i<$size} {incr i} {
	    lappend keys [ns_set key $set_id $i]
	}
    }
    return $keys
}

proc qc::ns_set_values { set_id } {
    #| Take an ns_set with id $set_id from caller 
    #| return all values
    set values {}
    if { ![string equal $set_id ""] } {
	set size [ns_set size $set_id]
	for {set i 0} {$i<$size} {incr i} {
	    lappend values [ns_set value $set_id $i]
	}
    }
    return $values
}


namespace eval qc {
    namespace export http_* IANAEncoding2TclEncoding
}

proc qc::http_curl {args} {
    #| This is a wrapper for TclCurl
    # If run within AOLserver uses ns_proxy to make thread safe.
    set script {
	proc main {args} {
	    package require TclCurl
	    set curlHandle [curl::init]
            if { [set index [lsearch $args -infovar]]!=-1 } {
                set infovar [lindex $args [expr {$index+1}]]
                set args [lreplace $args $index [expr {$index+1}]]
            }
	    $curlHandle configure {*}$args
	    catch { $curlHandle perform } curlErrorNumber
	    set responsecode [$curlHandle getinfo responsecode]
            if { [info exists infovar] } {
                foreach property {effectiveurl httpcode responsecode filetime totaltime namelookuptime connecttime pretransfertime starttransfertime sizeupload sizedownload speeddownload speedupload headersize requestsize sslverifyresult contentlengthdownload contentlengthupload contenttype} { 
                    dict set $infovar $property [$curlHandle getinfo $property] 
                }
            }
	    $curlHandle cleanup
	    set dict {}
	    foreach var [info locals] {
		switch -- $var {
		    args -
		    dict -
                    infovar -
                    index -
                    property -
		    curlHandle {}
		    default {
			if { [array exists $var] } {
			    lappend dict $var [array get $var]
			} else {
			    lappend dict $var [set $var]
			}
		    }
		}
	    }
	    return $dict
	}
    }
    append script [list main {*}$args]

    if { [info commands ns_proxy] eq "ns_proxy" } {
        # We are running under AOLserver so wrap with ns_proxy

        set handle [ns_proxy get myproxy]

        if { [dict exists $args -timeout] } {
	    # The proxy may timeout first due to a dns lookup so double timeout
	    set timeout [expr {[dict get $args -timeout]*1000*2}]
        } else {
	    # default timeout 60seconds
	    set timeout 60000
        }
        ::try {
	    set dict [ns_proxy eval $handle $script $timeout]
	    ns_proxy release $handle
	    return $dict
        } on error {error_message options} {
	    ns_proxy release $handle
	    error $error_message [dict get $options -errorinfo] [dict get $options -errorcode]
        }
    } else {
        # no ns_proxy so evaluate directly
        return [eval $script]
    }
}

proc qc::http_post {args} {
    #| Perform an HTTP POST
    # args is name value name value ... list
    args \
        $args \
        -noproxy \
        -timeout 60 \
        -sslversion tlsv1 \
        -encoding utf-8 \
        -content-type ? \
        -soapaction ? \
        -accept ? \
        -authorization ? \
        -headers {} \
        -data ? \
        -filename ? \
        -valid_response_codes [list 100 200] \
        -response_headers false \
        -response_code false \
        url \
        args

    if { ![info exists data]} {
	set pairs {}
	foreach {name value} $args {
	    lappend pairs "[qc::url_encode $name $encoding]=[qc::url_encode $value $encoding]"
	}
	set data [join $pairs &]
    }

    set httpheaders {}
    if { [info exists authorization] } {
	lappend httpheaders [qc::http_header "Authorization" $authorization]
    }

    if { [info exists content-type] } {
	lappend httpheaders [qc::http_header "Content-Type" ${content-type}]
    }

    if { [info exists accept] } {
	lappend httpheaders [qc::http_header "Accept" $accept]
    }

    if { [info exists soapaction] } {
	lappend httpheaders [qc::http_header "SOAPAction" $soapaction]
    }
    foreach {name value} $headers {
	lappend httpheaders [qc::http_header $name $value]
    }
   
    set curl_args [list \
                      {*}[qc::iif [info exists noproxy] {-proxy ""} {}] \
                      -headervar return_headers \
                      -url $url \
                      -sslverifypeer 0 \
                      -sslverifyhost 0 \
                      -timeout $timeout \
                      -sslversion $sslversion \
                      -bodyvar html \
                      -post 1 \
                      -httpheader $httpheaders \
                   ]

    if { [info exists content-type] && [string match "multipart/*" ${content-type}] } {
        # eg. multipart/formdata
        foreach mimepart $data {
            lappend curl_args -httppost $mimepart
        }
      
    } else {
        # eg. application/x-www-form-urlencoded
        lappend curl_args -postfields $data
    }

    if { [info exists filename] } {
        lappend curl_args -file $filename
    }

    dict2vars [qc::http_curl {*}$curl_args] return_headers html responsecode curlErrorNumber

    switch $curlErrorNumber {
	0 {
	    if { [in $valid_response_codes $responsecode] } {
		set response_body [encoding convertfrom [qc::http_encoding $return_headers $html] $html]
                if { !$response_headers && !$response_code } {
                    # Only return the response body.
                    return $response_body
                } else {
                    # User has requested response headers and/or code alongside the response body.
                    set response [dict create \
                                      body $response_body \
                                     ]

                    if { $response_code } {
                        dict set response code $responsecode
                    }

                    if { $response_headers } {
                        dict set response headers $return_headers
                    }

                    return $response
                }
	    } else {
		# raise an error
		switch $responsecode {
		    404 {return -code error -errorcode CURL "URL NOT FOUND $url"}
		    500 {return -code error -errorcode CURL "SERVER ERROR $url $html"}
		    default {return -code error -errorcode CURL "RESPONSE $responsecode while contacting $url $html"}
		}
	    }
	}
	28 {
	    return -code error -errorcode TIMEOUT "Timeout after $timeout seconds trying to contact $url"
	}
	default {
	    return -code error -errorcode CURL [curl::easystrerror $curlErrorNumber]
	}
    }
}

proc qc::http_get {args} {
    # usage http_get ?-timeout timeout? ?-headers {name value name value ...}? ?-noproxy? url
    args \
        $args \
        -timeout 60 \
        -sslversion tlsv1 \
        -headers {} \
        -valid_response_codes {200} \
        -noproxy \
        -response_headers false \
        -response_code false \
        -- \
        url

    set httpheaders {}
    foreach {name value} $headers {
	lappend httpheaders [qc::http_header $name $value]
    }
    dict2vars [qc::http_curl {*}[qc::iif [info exists noproxy] {-proxy ""} {}] -headervar return_headers -url $url -sslverifypeer 0 -sslverifyhost 0 -timeout $timeout -sslversion $sslversion -followlocation 1 -httpheader $httpheaders  -bodyvar html] return_headers html responsecode curlErrorNumber

    switch $curlErrorNumber {
	0 {
            if { [in $valid_response_codes $responsecode] } {
                # OK
                set response_body [encoding convertfrom [qc::http_encoding $return_headers $html] $html]
                if { !$response_headers && !$response_code } {
                    # Only return the response body.
                    return $response_body
                } else {
                    # User has requested response headers and/or code alongside the response body.
                    set response [dict create \
                                      body $response_body \
                                     ]

                    if { $response_code } {
                        dict set response code $responsecode
                    }

                    if { $response_headers } {
                        dict set response headers $return_headers
                    }

                    return $response
                }
            } else {
                switch $responsecode {                    
                    404 {return -code error -errorcode CURL "URL NOT FOUND $url"}
                    500 {return -code error -errorcode CURL "SERVER ERROR $url"}
                    default {return -code error -errorcode CURL "RESPONSE $responsecode while contacting $url"}
                }
            }
	}
	28 {
	    return -code error -errorcode TIMEOUT "Timeout after $timeout seconds trying to contact $url"
	}
	default {
	    return -code error -errorcode CURL [curl::easystrerror $curlErrorNumber]
	}
    }
}

proc qc::http_header {name value} {
    #| Return http header. 
    #| Raise an error if the value of the http header contains newline characters.
    if { [regexp {\n} $value] } {
        error "The value of http header, \"$name: $value\", contains newline characters."
    }
    return "$name: $value"
}

proc qc::http_put {args} {
    # usage http_put \
    #   ?-valid_response_codes 200? \
    #   ?-header 0? \
    #   ?-timeout timeout? \
    #   ?-infile infile? \
    #   ?-data data? \
    #   ?-headers {name value name value ...}? \
    #   url
    args $args \
        -header 0 \
        -timeout 60 \
        -valid_response_codes [list 200] \
        -sslversion tlsv1 \
        -headers {} \
        -infile ? \
        -data ? \
        -response_code false \
        -response_headers false \
        url 

    set httpheaders {}
    foreach {name value} $headers {
	lappend httpheaders [qc::http_header $name $value]
    }

    if { [info exists data] && [info exists infile]} {
        error "qc::http:put must have only 1 of -data or -infile specified"
    } elseif { [info exists infile] } {
        dict2vars \
                [qc::http_curl \
                    -header $header \
                    -upload 1 \
                    -infile $infile \
                    -headervar return_headers \
                    -url $url \
                    -sslverifypeer 0 \
                    -sslverifyhost 0 \
                    -timeout $timeout \
                    -sslversion $sslversion \
                    -followlocation 1 \
                    -httpheader $httpheaders \
                    -bodyvar html \
                ] \
                return_headers \
                html \
                responsecode \
                curlErrorNumber

    } elseif { [info exists data] }  {
        dict2vars \
            [qc::http_curl \
                -header $header \
                -customrequest PUT \
                -postfields $data \
                -headervar return_headers \
                -url $url \
                -sslverifypeer 0 \
                -sslverifyhost 0 \
                -timeout $timeout \
                -sslversion $sslversion \
                -followlocation 1 \
                -httpheader $httpheaders \
                -bodyvar html \
            ] \
            return_headers \
            html \
            responsecode \
            curlErrorNumber

    } else {
        error "qc::http:put must have 1 of -data or -infile specified"
    }

    switch $curlErrorNumber {
	0 {
            if { $responsecode in $valid_response_codes } {
                set response_body [encoding convertfrom [qc::http_encoding $return_headers $html] $html]
                if { !$response_headers && !$response_code } {
                    # Only return the response body.
                    return $response_body
                } else {
                    # User has requested response headers and/or code alongside the response body.
                    set response [dict create \
                                      body $response_body \
                                     ]

                    if { $response_code } {
                        dict set response code $responsecode
                    }

                    if { $response_headers } {
                        dict set response headers $return_headers
                    }

                    return $response
                }
            }
	    switch $responsecode {
		404 {
                    return \
                        -code error \
                        -errorcode CURL \
                        "URL NOT FOUND $url"
                    }
		500 {
                    return \
                        -code error \
                        -errorcode CURL \
                        "SERVER ERROR $url"
                }
		default {
                    return \
                        -code error \
                        -errorcode CURL \
                        "RESPONSE $responsecode while contacting $url"
                }
	    }
	}
	28 {
	    return \
                -code error \
                -errorcode TIMEOUT \
                "Timeout after $timeout seconds trying to contact $url"
	}
	default {
	    return \
                -code error \
                -errorcode CURL \
                [curl::easystrerror $curlErrorNumber]
	}
    }
}

proc qc::http_patch {args} {
    # usage http_patch \
    #   ?-valid_response_codes 200? \
    #   ?-header 0? \
    #   ?-timeout timeout? \
    #   ?-data data? \
    #   ?-headers {name value name value ...}? \
    #   url
    args $args \
        -header 0 \
        -timeout 60 \
        -valid_response_codes [list 200] \
        -sslversion tlsv1 \
        -headers {} \
        -data ? \
        -response_headers false \
        -response_code false \
        url

    set httpheaders {}
    foreach {name value} $headers {
        lappend httpheaders [qc::http_header $name $value]
    }

    if { ![info exists data] } {
        error "qc::http:put must have -data specified"
    } else {
        dict2vars \
            [qc::http_curl \
                -header $header \
                -customrequest PATCH \
                -postfields $data \
                -headervar return_headers \
                -url $url \
                -sslverifypeer 0 \
                -sslverifyhost 0 \
                -timeout $timeout \
                -sslversion $sslversion \
                -followlocation 1 \
                -httpheader $httpheaders \
                -bodyvar html \
            ] \
            return_headers \
            html \
            responsecode \
            curlErrorNumber

    }

    switch $curlErrorNumber {
        0 {
            if { $responsecode in $valid_response_codes } {
                set response_body [encoding convertfrom \
                                       [qc::http_encoding $return_headers $html] \
                                       $html]

                if { !$response_headers && !$response_code } {
                    return $response_body
                } else {
                    set response [dict create \
                                      body $response_body \
                                     ]

                    if { $response_code } {
                        dict set response code $responsecode
                    }

                    if { $response_headers } {
                        dict set response headers $return_headers
                    }

                    return $response
                }
            }
            switch $responsecode {
                404 {
                    return \
                        -code error \
                        -errorcode CURL \
                        "URL NOT FOUND $url"
                    }
                500 {
                    return \
                        -code error \
                        -errorcode CURL \
                        "SERVER ERROR $url"
                }
                default {
                    return \
                        -code error \
                        -errorcode CURL \
                        "RESPONSE $responsecode while contacting $url"
                }
            }
        }
        28 {
            return \
                -code error \
                -errorcode TIMEOUT \
                "Timeout after $timeout seconds trying to contact $url"
        }
        default {
            return \
                -code error \
                -errorcode CURL \
                [curl::easystrerror $curlErrorNumber]
        }
    }
}

proc qc::http_encoding {headers body} {
    #| Return the TCL encoding scheme used for http.
    # Try to determine the http encoding from the following sources (in this order):
    #   * HTTP-Header (charset attribute)
    #   * XML declaration (encoding attribute)
    # Otherwise return iso8859-1 as the default http encoding.
    # TODO this defaults to iso8859-1 for non-xml, but xml_encoding defaults to utf-8
    # I suspect this should be updated to default to utf-8 also.
    set encoding [qc::http_header_encoding $headers]
    if { $encoding eq "" } {
	if { [regexp {^\s*<\?xml} $body] } {
	    # Its an XML document
	    set encoding [qc::xml_encoding $body]
	} else {
	    set encoding "iso8859-1"
	}
    }
    return $encoding
}

proc qc::http_header_encoding {dict} {
    #| Return the TCL encoding scheme for a http header dict.
    # Try to determine encoding from the charset attribute specified in the http header dict.
    # Otherwise return "".
    array set return_headers $dict
    set encoding ""
    foreach key {Content-Type content-type} {
        if { [info exists return_headers($key)] && [regexp -nocase {.*;.*charset=(.*)} $return_headers($key) -> charset] } {
	    set encoding [IANAEncoding2TclEncoding [string trim $charset {"' }]]
        }
    }
    return $encoding
}

#----------------------------------------------------------------------------
#   IANAEncoding2TclEncoding
#   From v0.82 tDom lib/tdom.tcl
#----------------------------------------------------------------------------

proc qc::IANAEncoding2TclEncoding {IANAName} {
    
    switch [string tolower $IANAName] {
        "iso-ir-6" -
        "ansi_x3.4-1968" -
        "ansi_x3.4-1986" -
        "iso_646.irv:1991" -
        "ascii" -
        "iso646-us" -
        "us-ascii" -
        "us" -
        "ibm367" -
        "cp367" -
        "csASCII" {return ascii}
        "utf-8"       {return utf-8}
        "utf-16"      {return unicode; # not sure about this}
        "iso-8859-1"  {return iso8859-1}
        "iso-8859-2"  {return iso8859-2}
        "iso-8859-3"  {return iso8859-3}
        "iso-8859-4"  {return iso8859-4}
        "iso-8859-5"  {return iso8859-5}
        "iso-8859-6"  {return iso8859-6}
        "iso-8859-7"  {return iso8859-7}
        "iso-8859-8"  {return iso8859-8}
        "iso-8859-9"  {return iso8859-9}
        "iso-8859-10" {return iso8859-10}
        "iso-8859-13" {return iso8859-13}
        "iso-8859-14" {return iso8859-14}
        "iso-8859-15" {return iso8859-15}
        "iso-8859-16" {return iso8859-16}
        "iso-2022-kr" {return iso2022-kr}
        "euc-kr"      {return euc-kr}
        "iso-2022-jp" {return iso2022-jp}
        "koi8-r"      {return koi8-r}
        "shift_jis"   {return shiftjis}
        "euc-jp"      {return euc-jp}
        "gb2312"      {return gb2312}
        "big5"        {return big5}
        "cp866"       {return cp866}
        
        "windows-1250" -
        "cp1250"      {return cp1250}
        
        "windows-1253" -
        "cp1253"      {return cp1253}

        "windows-1254" -
        "cp1254"      {return cp1254}

        "windows-1255" -
        "cp1255"      {return cp1255}

        "windows-1256" -
        "cp1256"      {return cp1256}

        "windows-1257" -
        "cp1257"      {return cp1257}

        "windows-1251" -
        "cp1251"      {return cp1251}

        "windows-1252" -
        "cp1252"      {return cp1252}    

        "iso_8859-1:1987" -
        "iso-ir-100" -
        "iso_8859-1" -
        "latin1" -
        "l1" -
        "ibm819" -
        "cp819" -
        "csisolatin1" {return iso8859-1}
        
        "iso_8859-2:1987" -
        "iso-ir-101" -
        "iso_8859-2" -
        "iso-8859-2" -
        "latin2" -
        "l2" -
        "csisolatin2" {return iso8859-2}

        "iso_8859-5:1988" -
        "iso-ir-144" -
        "iso_8859-5" -
        "iso-8859-5" -
        "cyrillic" -
        "csisolatincyrillic" {return iso8859-5}

        "ms_kanji" -
        "csshiftjis"  {return shiftjis}
        
        "csiso2022kr" {return iso2022-kr}

        "ibm866" -
        "csibm866"    {return cp866}

        "gbk" {return cp936}

        "iso-ir-149" -
        "ks_c_5601-1987" -
        "ks_c_5601-1989" -
        "ks_c_5601" -
        "korean" -
        "csksc56011987" {return ksc5601}
        
        default {
            error "Unrecognized encoding name '$IANAName'"
        }
    }
}
#----------------------------------------------------------------------------
# End of tDom code
#----------------------------------------------------------------------------

proc qc::http_head {args} {
    #| Return a dict of name value pairs returned by the server in the HTTP header
    # usage http_head ?-timeout timeout? ?-useragent useragent? url
    args $args -headers {} -timeout 60 -useragent ? url
    default useragent "Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.8.0.7) Gecko/20060909 FreeBSD/i386 Firefox/1.5.0.7"
    #
    
    set httpheaders {}
    foreach {name value} $headers {
	lappend httpheaders [qc::http_header $name $value]
    }

    dict2vars [qc::http_curl -nobody 1 -headervar headers -url $url -sslverifypeer 0 -sslverifyhost 0 -timeout $timeout -followlocation 1 -httpheader $httpheaders] headers responsecode curlErrorNumber

    switch $curlErrorNumber {
	0 {
	    switch $responsecode {
		200 { 
		    # OK 
		    return $headers
		}
		404 {return -code error -errorcode CURL "URL NOT FOUND $url"}
		500 {return -code error -errorcode CURL "SERVER ERROR $url"}
		default {return -code error -errorcode CURL "RESPONSE $responsecode while contacting $url"}
	    }
	}
	28 {
	    return -code error -errorcode TIMEOUT "Timeout after $timeout seconds trying to contact $url"
	}
	default {
	    return -code error -errorcode CURL [curl::easystrerror $curlErrorNumber]
	}
    }
}

proc qc::http_exists {args} {
    #| Test if an URL returns a valid response
    # usage http_head ?-timeout timeout? ?-useragent useragent? url
    args $args -timeout 60 -useragent ? -valid_response_codes {100 200} url
    default useragent "Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.8.0.7) Gecko/20060909 FreeBSD/i386 Firefox/1.5.0.7"
    #
    dict2vars [qc::http_curl  -nobody 1 -headervar headers -url $url -sslverifypeer 0 -sslverifyhost 0 -timeout $timeout -followlocation 1] responsecode curlErrorNumber

    if { [in $valid_response_codes $responsecode] } {
	return true
    } else {
	return false
    }
}

proc qc::http_save {args} {
    #| Save the HTTP response to a file.
    args $args \
        -timeout 60 \
        -headers {} \
        -sslversion tlsv1 \
        -sslverifypeer 0 \
        -sslverifyhost 0 \
        -return_headers_var {} \
        -follow_location 0 \
        -- \
        url \
        file

    set httpheaders {}
    foreach {name value} $headers {
	lappend httpheaders [qc::http_header $name $value]
    }

    dict2vars [qc::http_curl \
                -httpheader $httpheaders \
                -headervar return_headers \
                -timeout $timeout \
                -url $url \
                -file $file \
                -sslverifypeer $sslverifypeer \
                -sslverifyhost $sslverifyhost \
                -sslversion $sslversion \
                -followlocation $follow_location \
              ] return_headers responsecode curlErrorNumber

    if { $responsecode != 200 } {
	file delete $file
    }
    switch $responsecode {
	200 { 
	    # OK 
	}
	404 {return -code error -errorcode CURL "URL NOT FOUND $url"}
	500 {return -code error -errorcode CURL "SERVER ERROR $url"}
	default {return -code error -errorcode CURL "RESPONSE $responsecode while contacting $url"}
    }
    switch $curlErrorNumber {
	0 {
	    # OK
            if { $return_headers_var ne {} } {
                upset 1 $return_headers_var $return_headers
            }
	    return 1
	}
	default {
	    file delete $file
	    return -code error -errorcode CURL [curl::easystrerror $curlErrorNumber]
	}
    }
}

proc qc::http_delete {args} {
    #| Send http DELETE request
    args $args -timeout 60 -headers {} -- url

    set httpheaders {}
    foreach {name value} $headers {
	lappend httpheaders [qc::http_header $name $value]
    }

    dict2vars [qc::http_curl -customrequest DELETE -httpheader $httpheaders -timeout $timeout -url $url -sslverifypeer 0 -sslverifyhost 0] responsecode curlErrorNumber
    switch $responsecode {
        204 -
	200 { 
	    # OK 
	}
	404 {return -code error -errorcode CURL "URL NOT FOUND $url"}
	500 {return -code error -errorcode CURL "SERVER ERROR $url"}
	default {return -code error -errorcode CURL "RESPONSE $responsecode while contacting $url"}
    }
    switch $curlErrorNumber {
	0 {
	    # OK
	    return 1
	}
	default {
	    return -code error -errorcode CURL [curl::easystrerror $curlErrorNumber]
	}
    }
}

proc qc::http_url_resolve {args} {
    #| Return resolved url after following redirects
    #
    # Misconfigured clickthrough servers may not redirect HEAD requests so always request a body by default
    args $args -timeout 60 -sslversion tlsv1 -nobody 0 -- url

    dict2vars [qc::http_curl -url $url -sslversion $sslversion -sslverifypeer 0 -sslverifyhost 0 -timeout $timeout -followlocation 1 -nobody $nobody -infovar info -bodyvar html] info curlErrorNumber

    switch $curlErrorNumber {
	0 {
	    # OK
	    return [dict get $info effectiveurl]
	}
	default {
	    return -code error -errorcode CURL [curl::easystrerror $curlErrorNumber]
	}
    }    
}

proc qc::http_header_get {header_name} {
    #| Returns the value for the given header.
    set headers_set_id [ns_conn headers]
    if { [ns_set ifind $headers_set_id $header_name] == -1 } {
        return ""
    } elseif { [string tolower $header_name] in [qc::http_headers_with_value_as_list] } {
        # Header may appear more than once so get all values for the header.
        set header_values {}
        for {set i 0} {$i < [ns_set size $headers_set_id]} {incr i} {
            if { [string equal -nocase [ns_set key $headers_set_id $i] $header_name] } {
                lappend header_values [ns_set value $headers_set_id $i]
            }
        }
        
        return [join $header_values ","]
    } else {
        return [ns_set iget $headers_set_id $header_name]
    }
}

proc qc::http_header_parse {header_name header_value} {
    #| Parses the given header value string into a Tcl structure that is a list of dicts.
    #| E.G. qc::http_header_parse "application/json;q=0.9,text/xml,text/plain;q=0.8"
    #|      -> {token application/json params {q 0.9}} {token text/xml params {}} {token text/plain params {q 0.8}}
    
    # TODO: Header comments.
    # TODO: Quoted strings.
    # TODO: Validation of tokens and params?

    # Get the elements of the header value
    if { [string tolower $header_name] in [qc::http_headers_with_value_as_list] } {
        set elements [qc::map {x {string trim $x}} [split $header_value ","]]
    } else {
        # TODO: Handle other delimiters for other headers.
        set elements [list $header_value]
    }
    
    set result {}
    foreach element $elements {
        set element_dict [dict create]
        # Parse all params for this element
        set params [qc::map {x {string trim $x}} [split $element ";"]]
        # Most elements begin with a token but there are special cases where the element only consists of params.
        if { [string tolower $header_name] in [list "cookie" "set-cookie" "strict-transport-security"] } {
            set token ""
        } else {
            set token [lindex $params 0]
            set params [lrange $params 1 end]
        }

        # Parse param names and their values
        set param_dict [dict create]
        foreach param $params {
            set list [qc::map {x {string trim $x}} [split $param "="]]
            set name [lindex $list 0]
            if { [llength $list]  > 1 } {
                set value [lindex $list 1]
            } else {
                set value ""
            }
            dict set param_dict $name $value
        }
        dict set element_dict token $token
        dict set element_dict params $param_dict
        lappend result $element_dict
    }

    return $result
}

proc qc::http_headers_with_value_as_list {} {
    #| A list of headers whose values are defined as a comma separated list.
    set headers {}
    lappend headers accept
    lappend headers accept-charset
    lappend headers accept-encoding
    lappend headers accept-language
    lappend headers accept-ranges
    lappend headers allow
    lappend headers cache-control
    lappend headers connection
    lappend headers content-encoding
    lappend headers content-language
    lappend headers expect
    lappend headers if-match
    lappend headers if-none-match
    lappend headers pragma
    lappend headers proxy-authenticate
    lappend headers te
    lappend headers trailer
    lappend headers upgrade
    lappend headers vary
    lappend headers via
    lappend headers warning
    lappend headers www-authenticate
    lappend headers x-forwarded-for
    return $headers
}

proc qc::http_header_sort_values_by_weight {values} {
    #| Sorts the header values by weight.
    if { [llength $values] == 0 } {
        return [list]
    }
    set sorted [list [lindex $values 0]]
    foreach value [lrange $values 1 end] {
        if { [dict exists $value params q] } {
            set weight [dict get $value params q]
        } else {
            set weight 1
        }
        set inserted false
        for { set i 0 } { $i < [llength $sorted] } { incr i } {
            set dict [lindex $sorted $i]
            if { [dict exists $dict params q] } {
                set sorted_weight [dict get $dict params q]
            } else {
                set sorted_weight 1
            }
            if { $weight > $sorted_weight } {
                set sorted [linsert $sorted $i $value]
                set inserted true
                break
            }
        }
        
        if { ! $inserted } {
            set sorted [linsert $sorted end $value]
        }
    }
    return $sorted
}

proc qc::http_accept_header_best_mime_type {available_media} {
    #| Returns a mime type from the available media list that is the best match
    #| for the preferences of the Accept header.
    #| If any mime type will suffice then "*/*" is returned.
    #| Returns the empty string if no sufficient match is found.
    
    # TODO specificity.
    set sorted [qc::http_header_sort_values_by_weight [qc::http_header_parse Accept [qc::http_header_get Accept]]]
    if { [llength $sorted] == 0 } {
        return "*/*"
    }
    foreach dict $sorted {
        lassign [split [dict get $dict token] "/"] type subtype
        # If best mime type is "*/*" then let caller know so that they can choose the most suitable type.
        if { $type eq "*" && $subtype eq "*" } {
            return "$type/$subtype"
        }
        foreach item $available_media {
            lassign [split $item "/"] available_type available_subtype
            if { [string match -nocase $type $available_type] && [string match -nocase $subtype $available_subtype] } {
                return $item
            }
        }
    }
    return ""
}

proc qc::http_content_negotiate {available_mime_types} {
    #| Try to find an acceptable mime type to respond to the client with.
    #| Returns a mime type if an acceptable one was found otherwise the empty string.
    set result "" 
    set suffix [string range [file extension [qc::conn_path]] 1 end]
    # Prioritise suffix type because it is a direct request for a resource of that type.
    if { $suffix ne "" } {
        # Check if the suffix type is available and that the client accepts it.
        foreach mime_type $available_mime_types {
            lassign [split $mime_type "/"] type subtype
            if { [string match -nocase $subtype $suffix] && [qc::http_accept_header_best_mime_type [list $mime_type]] ne "" } {
                set result $mime_type
                break
            }
        }
    }

    if { $result eq "" } {
        set result [qc::http_accept_header_best_mime_type $available_mime_types]
    }

    return $result
}

namespace eval qc {
    namespace export {*}{
        image_data
        image_cache_exists
    }
}

proc qc::image_data {args} {
    #| Return dict of width, height, & url of an image. Usage:
    #| ?-autocrop? ?-mime_type */*? -- cache_dir file_id max_width max_height
    set caller_args $args
    qc::args $args {*}{
        -autocrop
        -mime_type */*
        --
        cache_dir
        file_id
        max_width
        max_height
    }

    default \
        autocrop false

    if { ! [qc::image_cache_exists {*}$caller_args] } {
        # Queue the image to be cached and return a placeholder.

        if { $autocrop } {
            qc::image_resize_task_add \
                -autocrop \
                -mime_type $mime_type \
                -cache_dir $cache_dir \
                -- \
                $file_id \
                $max_width \
                $max_height
        } else {
            qc::image_resize_task_add \
                -mime_type $mime_type \
                -cache_dir $cache_dir \
                -- \
                $file_id \
                $max_width \
                $max_height
        }

        set filename "${max_width}x${max_height}.png"
        set url [qc::url "https://via.placeholder.com/:filename" \
                     filename $filename \
                     text "queued for processing" \
                    ]

        return [dict create \
                    width $max_width \
                    height $max_height \
                    url $url \
                   ]
    }

    if { $mime_type eq "*/*" } {
        # We have a cached file for the args but no mime-type specified
        set available_types \
            [qc::_image_cache_available_mime_types {*}$caller_args]
        # Preference to use if no mime-type specified
        foreach preference {
            "image/svg+xml"
            "image/png"
            "image/gif"
            "image/jpeg"
            "image/webp"
        } {
            if { [lsearch -exact $available_types $preference] > -1 } {
                set mime_type $preference
                break
            }
        }
    }

    set cache_data \
        [qc::_image_cache_data \
             $cache_dir $file_id $mime_type $max_width $max_height $autocrop]
    return $cache_data
}

proc qc::image_data2 {args} {
    #| Return dict of width, height, & url of an image. Usage:
    #| ?-autocrop? -- cache_dir file_id mime_type max_width max_height
    set caller_args $args
    qc::args $args {*}{
        -autocrop
        --
        cache_dir
        file_id
        mime_type
        max_width
        max_height
    }
    default autocrop false
    if { ! [qc::image_cache_exists2 {*}$caller_args] } {
        qc::_image_cache_create \
            $cache_dir $file_id $mime_type $max_width $max_height $autocrop
    }
    set cache_data \
        [qc::_image_cache_data \
             $cache_dir $file_id $mime_type $max_width $max_height $autocrop]
    return $cache_data
}

proc qc::image_cache_exists {args} {
    #| Return true if a cached version of the image exists. Usage:
    #| ?-autocrop? ?-mime_type */*? -- cache_dir file_id max_width max_height
    set caller_args $args
    qc::args $args {*}{
        -autocrop
        -mime_type */*
        --
        cache_dir
        file_id
        max_width
        max_height
    }
    default autocrop false
    # Special treatment for svg images
    if { $mime_type in [list "*/*" "image/svg+xml"]
         && [qc::_image_cache_original_exists $cache_dir $file_id]
             && [qc::mime_type_guess \
                     [qc::_image_cache_original_file $cache_dir $file_id]] \
             eq "image/svg+xml" } {
        return true
    }
    
    foreach file [qc::_image_filesystem_cache_glob {*}$caller_args] {
        lassign \
            [qc::_image_filesystem_cache_file2dimensions $file] \
            width height

        if { ( $width == $max_width
               &&
               $height <= $max_height )
             ||
             ( $height == $max_height
               &&
               $width <= $max_width )
         } {
            return true
        }
    }
    return false
}

proc qc::image_cache_exists2 {args} {
    #| Return true if a cached version of the image exists. Usage:
    #| ?-autocrop? -- cache_dir file_id mime_type max_width max_height
    set caller_args $args
    qc::args $args {*}{
        -autocrop
        --
        cache_dir
        file_id
        mime_type
        max_width
        max_height
    }
    default autocrop false
    # Special treatment for svg
    if { $mime_type eq "image/svg+xml"
         && [qc::_image_cache_original_exists $cache_dir $file_id]
             && [qc::mime_type_guess \
                     [qc::_image_cache_original_file $cache_dir $file_id]] \
             eq "image/svg+xml" } {
        return true
    }
    
    foreach file [qc::_image_filesystem_cache_glob2 {*}$caller_args] {
        lassign \
            [qc::_image_filesystem_cache_file2dimensions $file] \
            width height

        if { ( $width == $max_width
               &&
               $height <= $max_height )
             ||
             ( $height == $max_height
               &&
               $width <= $max_width )
         } {
            return true
        }
    }
    return false
}

proc qc::_image_cache_available_mime_types {args} {
    #| Return a list of mime types for matching cache entries
    set caller_args $args
    qc::args $args {*}{
        -autocrop
        -mime_type */*
        --
        cache_dir
        file_id
        max_width
        max_height
    }
    default autocrop false

    set mime_type_flags [dict create]
    # Special treatment for svg
    if { $mime_type eq "image/svg+xml"
         && [qc::_image_cache_original_exists $cache_dir $file_id]
         && [qc::mime_type_guess \
                 [qc::_image_cache_original_file $cache_dir $file_id]] \
             eq "image/svg+xml" } {
        dict set mime_type_flags "image/svg+xml" true
    }
    
    foreach file [qc::_image_filesystem_cache_glob {*}$caller_args] {
        lassign \
            [qc::_image_filesystem_cache_file2dimensions $file] \
            width height

        if { ( $width == $max_width
               &&
               $height <= $max_height )
             ||
             ( $height == $max_height
               &&
               $width <= $max_width )
         } {
            dict set mime_type_flags [qc::mime_type_guess $file] true
        }
    }
    return [dict keys $mime_type_flags]
}

proc qc::_image_cache_data {
    cache_dir file_id mime_type max_width max_height autocrop
} {
    #| Return dict of width, height & url of an image from cache.
    set glob_args [list]
    if { $autocrop } {
        lappend glob_args -autocrop
    }
    lappend glob_args \
        -mime_type $mime_type \
        -- \
        $cache_dir \
        $file_id \
        $max_width \
        $max_height
    # Special treatment for svg
    if { $mime_type in [list "*/*" "image/svg+xml"]
         && [qc::_image_cache_original_exists $cache_dir $file_id]
         && [qc::mime_type_guess \
                 [qc::_image_cache_original_file $cache_dir $file_id]] \
             eq "image/svg+xml" } {
        set file [qc::_image_cache_original_file $cache_dir $file_id]
        lassign [qc::svg_dimensions $file] width height
        if { $width > $max_width
             ||
             $height > $max_height } {
            set scale [expr {min(
                                 1.0 * $max_width / $width,
                                 1.0 * $max_height / $height
                                 )}]
            set width [qc::round [expr {$scale * $width}] 0]
            set height [qc::round [expr {$scale * $height}] 0]
        }        
        set url [qc::file2url $file]
        set data [dict_from width height url]
        return $data
    }
    
    foreach file [qc::_image_filesystem_cache_glob {*}$glob_args] {
        lassign \
            [qc::_image_filesystem_cache_file2dimensions $file] \
            width height

        if { ( $width == $max_width
               &&
               $height <= $max_height )
             ||
             ( $height == $max_height
               &&
               $width <= $max_width )
         } {
            set url [qc::file2url $file]
            set data [dict_from width height url]
            return $data
        }
    }
    error "No cache for $cache_dir $file_id $mime_type $max_width $max_height $autocrop"
}

proc qc::_image_cache_create {
    cache_dir file_id mime_type max_width max_height autocrop
} {
    #| Create a cache of an image with the given requirements
    if { ! [qc::_image_cache_original_exists $cache_dir $file_id] } {
        qc::_image_cache_original_create $cache_dir $file_id
    }
    # The original will suffice for svg
    if { $mime_type eq "image/svg+xml" } {
        return
    }
    set original [qc::_image_cache_original_file $cache_dir $file_id]
    
    set file [qc::image_file_convert \
                 $original $mime_type $max_width $max_height $autocrop]
    
    dict2vars [qc::image_file_info $file] width height

    set ext [qc::mime_file_extension $mime_type]
    set file_path_parts [list $cache_dir]
    if { $autocrop } {
        lappend file_path_parts \
            $file_id \
            autocrop \
            ${width}x${height}
    } else {
        lappend file_path_parts \
            ${file_id}-${width}x${height}
    }
    lappend file_path_parts ${file_id}${ext}

    set cache_file [join $file_path_parts "/"]

    file mkdir [file dirname $cache_file]
    file rename -force $file $cache_file
}

namespace eval qc {
    namespace export xml xml_* xml2dict
}

proc qc::xml { tagName nodeValue {dict_att ""} } {
    #| Constructs xml node from supplied parameters
    set latt {}
    foreach {name value} $dict_att {
	lappend latt "$name=\"[qc::xml_escape $value]\""
    }
    if { [llength $latt] > 0 } {
	return "<$tagName [join $latt]>[qc::xml_escape $nodeValue]</$tagName>"
    } else {
	return "<$tagName>[qc::xml_escape $nodeValue]</$tagName>"
    }
}

proc qc::xml_escape { string } {
    #| Escape reserved characters and converts characters above 127 to entity decimal
    # Escape < > &
    set string [html_escape $string]
    # Escape characters used by subst i.e. []$\ 
    regsub -all {[][$\\]} $string {\\&} string
    # Escape characters with value at or above 127
    regsub -all -- {([\u007F-\u00FF])} $string {[qc::xml_encode_char \1]} string
    
    return [subst $string]
}

proc qc::xml_encode_char {string} {scan $string %c t; return "&#$t\;"}

proc qc::xml_decode_char_NNN {string} {
    #| Convert entity decimal format to ascii character
    # Escape characters used by subst i.e. []$\ 
    regsub -all {[][$\\]} $string {\\&} string
    regsub -all {&\#(\d{1,3});?} $string {[format %c [scan \1 %d tmp;set tmp]]} string
    return [subst $string]
}

proc qc::xml_from { args } {
    #| Return xml structure from the specified local variable names and values
    set list {}
    foreach name $args {
	upvar 1 $name value
	if { [info exists value] } {
	    lappend xml [qc::xml $name $value]
	} else {
	    error "Can't create xml with $name: No such variable"
	}
    }
    return [join $xml \n]
}

proc qc::xml_ldict { li_tag ldict } {
    #| Create xml structure from a list of dicts
    set xml ""
    foreach dict $ldict {
	append xml "<$li_tag>[qc::dict2xml $dict]</$li_tag>"
    }
    return $xml
}

proc qc::xml2dict { xml root_element } {
    #| Converts an XML structure in the form of a text string into a dict.
    set dict ""
    package require tdom
    dom parse $xml doc
    set subtree [$doc getElementsByTagName $root_element]
    if { $subtree eq "" } {
        error "XML parse error"
    } else {
        set node [lindex $subtree 0]
        set nodes [$node childNodes]
        foreach node $nodes {
            if { [llength [$node childNodes]] > 1 \
               || ([llength [$node childNodes]] == 1 \
                  && [ne [[$node firstChild] nodeType] TEXT_NODE] ) } {
                lappend dict [$node nodeName] [xml2dict $xml [$node nodeName]]
            }  elseif { [llength [$node childNodes]] == 0 } {
                # empty node
                lappend dict [$node nodeName] {}
            } else {
                lappend dict [$node nodeName] [$node asText]
            }
        }
    }
    return $dict
}

proc qc::xml_encoding {xml} {
    #| Return the TCL encoding scheme used for an xml document.
    # Try to determine encoding from the encoding attribute in the xml declaration.
    # Otherwise return default encoding "utf-8".
    set regexp1 {^[^>]+encoding=\"([^\"]*)\"}
    set regexp2 {^[^>]+encoding='([^']*)'}

    if { [regexp -nocase -expanded $regexp1 $xml -> encoding] || [regexp -nocase -expanded $regexp2 $xml -> encoding] } {
	return [IANAEncoding2TclEncoding [string trim $encoding]]
    } else {
	return utf-8
    }
}

proc qc::xml_declaration_valid {declaration} {
    #| Determines if the given XML declaration is valid.

    set pattern {^<\?xml\s+}
    append pattern {version\s*=\s*("1\.[0-9]+"|'1\.[0-9]+')}
    set encoding_value {[a-zA-Z]([a-zA-Z0-9.-_]|-)*}
    append pattern [subst {(\\s+encoding\\s*=\\s*("$encoding_value"|'$encoding_value'))?}]
    append pattern {(\s+standalone\s*=\s*("(yes|no)"|'(yes|no)'))?}
    append pattern {\s*\?>}

    return [regexp $pattern $declaration]
}


namespace eval qc {
    namespace export cast_* data_type_error_check
}

proc qc::cast_integer {string} {
    #| Deprecated - see qc::cast integer
    #| Try to cast given string into an integer
    return [qc::cast integer $string]
}

proc qc::cast_int {string} {
    #| Deprecated - see qc::cast integer
    return [qc::cast integer $string]
}

proc qc::cast_decimal {string {precision ""}} {
    #| Deprecated - see qc::cast decimal
    #| Try to cast given string into a decimal value
    if { $precision ne "" } {
        set p [string length $string]
        if { $precision >= $p } {
            set p [expr {$p + $precision}]
        }
        return [qc::cast decimal -precision $p -scale $precision $string]
    } else {
        return [qc::cast decimal $string]
    }
}

proc qc::cast_date {string} {
    #| Deprecated - see qc::cast date
    #| Try to convert the given string into an ISO date.
    return [qc::cast date $string]
}

proc qc::cast_timestamp {string} {
    #| Deprecated - see qc::cast timestamp
    #| Try to convert the given string into an ISO datetime.
    return [qc::cast timestamp $string]
}

proc qc::cast_timestamptz {string} {
    #| Deprecated - see qc::cast timestamptz
    #| Try to convert the given string into an ISO datetime with timezone.
    return [qc::cast timestamptz $string]
}

proc qc::cast_epoch { string } {
    #| Deprecated - see qc::cast epoch
    #| Try to convert the given string into an epoch
    return [qc::cast epoch $string]
}

proc qc::cast_boolean { string {true t} {false f} } {
    #| Deprecated - see qc::cast boolean
    #| Cast a string as a boolean
    return [qc::cast boolean $string $true $false]
}

proc qc::cast_bool { string {true t} {false f} } {
    #| Deprecated - see qc::cast boolean
    return [qc::cast boolean $string $true $false]
}

proc qc::cast_postcode { postcode } {
    #| Deprecated - see qc::cast postcode
    #| Try to cast a string into UK Postcode form
    return [qc::cast postcode $postcode]
}

proc qc::cast_creditcard { no } {
    #| Deprecated - see qc::cast creditcard
    return [qc::cast creditcard $no]
}

proc qc::cast_period {string} {
    #| Deprecated - see qc::cast period
    #| Return a pair of dates defining the period.
    return [qc::cast period $string]
}

proc qc::cast_url {string} {
    #| Deprecated - see qc::cast url
    #| (See also qc::is url)
    return [qc::cast url $string]
}

proc qc::cast_relative_url {string} {
    #| Deprecated - see qc::cast url
    #| (See also qc::is url)
    return [qc::cast relative_url $string]
}

proc qc::is_period {string} {
    #| Deprecated - see qc::is period
    #| Test if string can be casted to a pair of dates defining a period.
    return [qc::is period $string]
}

proc qc::cast_value2model {name value} {
    #| Return the value cast to the data model.
    #| Name can be a partial or fully qualified column identifier.
    
    # Resolve name to column, table, and schema
    lassign [qc::memoize qc::db_resolve_field_name $name] {*}{
        schema
        table
        column
    }

    set data_type [qc::memoize qc::db_column_type \
                       -qualified -- $schema $table $column]
    set nullable [qc::memoize qc::db_column_nullable $schema $table $column]

    # Check if nullable
    if {! $nullable && $value eq ""} {
        error "$column cannot be empty."
    } elseif {$nullable && $value eq ""} {
        return $value
    }

    # Check value against data type
    if {[qc::castable $data_type $value]} {
        return [qc::cast $data_type $value]
    } else {
        error [qc::data_type_error_check $data_type $value]
    }
}

proc qc::cast_values2model {args} {
    #| Check the data types of the values against the definitions for these names.
    #| Returns a new list of values after casting to appropriate type.
    #| Throws an error if type-checking fails.
    if { [llength $args]%2 != 0 } {
        return -code error "usage cast_values2model name value ?name value?"
    }
    set casted_dict {}
    set errors {}
    
    dict for {name value} $args {

        # Resolve name to column, table, and schema
        lassign [qc::memoize qc::db_resolve_field_name $name] {*}{
            schema
            table
            column
        }
        
        set data_type [qc::memoize qc::db_column_type \
                           -qualified -- $schema $table $column]
        set nullable [qc::memoize qc::db_column_nullable $schema $table $column]

        # Check if nullable
        if {! $nullable && $value eq ""} {
            lappend errors "$column cannot be empty."
            continue
        } elseif {$nullable && $value eq ""} {
            lappend casted_dict $name $value
            continue
        }

        # Check value against data type
        if {[qc::castable $data_type $value]} {
            lappend casted_dict $name [qc::cast $data_type $value]
        } else {           
            lappend errors [qc::html_escape [qc::data_type_error_check $data_type $value]]
        }
    }
    
    if {[llength $errors] > 0} {
        return -code error -errorcode USER [qc::html_list $errors]
    } else {
        return $casted_dict
    }
}

proc qc::data_type_error_check {data_type value} {
    #| Checks the given value against the data type and reports any error.
    switch -regexp -matchvar matches -- $data_type {
        {^varchar(\(([0-9]+)\))?$} {
            set length [string range [lindex $matches 1] 1 [expr {[string length [lindex $matches 1]] - 2}]]              
            if {! [qc::is varchar $length $value]} {
                return "\"[qc::trunc $value 100]...\" is too long. Must be $length characters or less."
            }
        }
        {^char(\(([0-9]+)\))?$} {
            set length [string range [lindex $matches 1] 1 [expr {[string length [lindex $matches 1]] - 2}]]
            set chars "characters"
            if {$length eq ""} {
                set length 1
                set chars "character"
            }
            if {! [qc::is char $length $value]} {
                if {[string length $value] < $length} {
                    return "\"$value\" is too short. Must be exactly $length $chars."
                } elseif {[string length $value] > $length} {
                    return "\"[qc::trunc $value 100]...\" is too long. Must be exactly $length $chars."
                }
            }
        }
        ^int4$ {
            if {! [qc::is integer $value] && ! [qc::castable integer $value]} {
                return "\"$value\" is not a valid integer. It must be a number between -2147483648 and 2147483647."
            }
        }
        ^int8$ {
            if {! [qc::is bigint $value] && ! [qc::castable bigint $value]} {
                return "\"$value\"is not a valid big int. It must be a number between -9223372036854775808 and 9223372036854775807."
            } 
        }
        ^int2$ {
            if {! [qc::is smallint $value] && ![qc::castable smallint $value]} {
                return "\"$value\"is not a valid small int. It must be a number between -32768 and 32767."
            }  
        }
        ^bool$ {
            if {! [qc::is boolean $value] && ! [qc::castable boolean $value]} {
                return "\"$value\"is not a valid boolean value."
            }
        }
        ^timestamp$ {
            if {! [qc::is timestamp $value] && ! [qc::castable timestamp $value]} {
                return "\"$value\"is not a valid timestamp."
            }
            
        }
        ^timestamptz$ {
            if {! [qc::is timestamptz $value] && ! [qc::castable timestamptz $value]} {
                return "\"$value\"is not a valid timestamptz."
            }           
        }
        {^(numeric|decimal)$} {
            if {! [qc::is decimal $value] && ! [qc::castable decimal $value]} {
                return "\"$value\"is not a valid decimal."
            }
        }
        ^text$ {
            return
        }
        ^safe_html$ {
            if {! [qc::is safe_html $value]} {
                return "\"[qc::trunc $value 50]...\" contains invalid or unsafe HTML."
            }
        }
        ^safe_markdown$ {
            if {! [qc::is safe_markdown $value]} {
                return "\"[qc::trunc $value 50]...\" contains invalid or unsafe HTML."
            }
        }
        default {
            # might be an enumeration or domain
            if {[qc::memoize qc::db_enum_exists $data_type]} {
                if {! [qc::castable enumeration $data_type $value]} {
                    return "\"$value\" is not a valid value for enum \"$data_type\"."
                }
            } elseif {[qc::memoize qc::db_domain_exists $data_type]} {
                set base_type [qc::memoize qc::db_domain_base_type $data_type]
                set constraints [qc::memoize qc::db_domain_constraints $data_type]
                set is_base_type [qc::is $base_type $value]
                set failed_constraints [list]
                dict for {constraint_name check_clause} $constraints {
                    if { ! [qc::db_eval_domain_constraint $value $base_type $check_clause] } {
                        lappend failed_constraints $constraint_name
                    }
                }
                if { ! $is_base_type && [llength $failed_constraints] > 0 } {
                    return "[data_type_error_check $base_type $value] and failed to meet the constraint(s) [join $failed_constraints ", "]"
                } elseif { ! $is_base_type } {
                    return [qc::data_type_error_check $base_type $value]
                } elseif { [llength $failed_constraints] > 0 } {
                    return "\"[qc::trunc $value 100]...\" failed to meet the constraint(s) [join $failed_constraints ", "]"
                }
            } else {
                return -code error "Unrecognised data type \"$data_type\""
            }
        }
    }
    return
}


namespace eval qc::cast {
    
    namespace export integer bigint smallint decimal boolean timestamp timestamptz char varchar text enumeration domain safe_html safe_markdown date postcode creditcard period epoch url url_relative url_path s3_uri time interval next_url
    namespace ensemble create -unknown {
        data_type_parser
    }

    proc integer {string} {
        #| Try to cast given string into an integer
        set result [integer_convert $string]
        if { [qc::is integer $result] } {
            return $result
        } else {
            return -code error -errorcode CAST "Could not cast $string to integer."
        }
    }

    proc bigint {string} {
        #| Try to cast the given string into an integer checking if it falls into big int range.
        set result [integer_convert $string]
        if { [qc::is bigint $result] } {
            return $result
        } else {
            return -code error -errorcode CAST "Could not cast $string to bigint."
        }
    }

    proc smallint {string} {
        #| Try to cast the given string into an integer checking if it falls into small int range.
        set result [integer_convert $string]
        if { [qc::is smallint $result] } {
            return $result
        } else {
            return -code error -errorcode CAST "Could not cast $string to smallint."
        }
    }

    proc integer_convert {string} {
        #| Tries to form an integer from the given string.
        set original $string
        # Convert e notation
        if { [string first e $string]!=-1 || [string first E $string]!=-1 } {
            set string [qc::exp2string $string]
        }
        set string [string map {, {} % {}} $string]
        # Strip leading zeros if followed by digit
        # This copes with 0 and 00
        regsub {^(-?)0+([0-9]+)$} $string {\1\2} string
        # Convert decimals
        if { [string first . $string]!=-1 } {
            set string [qc::round $string 0]
        }
        return $string
    }

    proc decimal {args} {
        #| Try to cast given string into a decimal value with the given precision and/or scale if present.
        qc::args $args -precision ? -scale ? -- string
        set original $string
        # Strip leading zeros if followed by digit
        # This copes with 0 and 00
        regsub {^(-?)0+([0-9].*)$} $string {\1\2} string
        set string [string map {, {} % {}} $string]
        if { [string is double -strict $string] } {
            if { ! [info exists scale] && ! [info exists precision] } {
                # Scale and precision not given.
                return [qc::exp2string $string]
            } elseif { [info exists scale] && ! [info exists precision] } {
                # Scale given but not precision.
                return -code error -errorcode CAST "Precision must be provided with scale."
            } elseif { [info exists precision] } {
                # Precision given.
                if { $precision <= 0 || ! [qc::is integer $precision] } {
                    return -code error -errorcode CAST "Precision must be a positive integer."
                }

                if { ! [info exists scale] } {
                    # Scale wasn't given so set it to 0.
                    set scale 0
                }

                if { $scale < 0 || ! [qc::is integer $scale]} {
                    return -code error "Scale must be a non-negative integer."
                }
                
                # Expand the number and round it to the scale and check the result.
                set rounded [qc::round [qc::exp2string $string] $scale]
                set parts [split $rounded .]
                set left_count [string length [lindex $parts 0]]
                set right_count [string length [lindex $parts 1]]
                if { $left_count == 0 } {
                    # The leading zero wasn't given.
                    return -code error -errorcode CAST "Incomplete numeric string."
                }
                
                set left_digit_max [expr {$precision - $scale}]
                if { $left_count > $left_digit_max || $right_count > $scale} {
                    return -code error -errorcode CAST "The resulting number ($rounded) is too large for the given precision ($precision) and scale ($scale)."
                } else {
                    return $rounded
                }
            }
        } else {
            return -code error -errorcode CAST "Could not cast $original to decimal."
        }
    }

    proc boolean { string {true t} {false f} } {
        #| Cast a string as a boolean.
        if { [string toupper $string] in {Y YES TRUE T 1} } {
            return $true
        } elseif {[string toupper $string] in {N NO FALSE F 0} } {
            return $false
        } else {
            return -code error -errorcode CAST "Can't cast \"$string\" to boolean data type."
        }
    }

    proc timestamp {string} {
        #| Try to convert the given string into an ISO datetime.
        return [clock format [epoch $string] -format "%Y-%m-%d %H:%M:%S"]
    }

    proc timestamptz {string} {
        #| Try to convert the given string into an ISO datetime with timezone.
        return [clock format [epoch $string] -format "%Y-%m-%d %H:%M:%S %z"]
    }

    proc char {length string} {
        #| Cast to char.
        if { [string length $string] == $length } {
            return $string
        } else {
            return -code error -errorcode CAST "Can't cast \"$string\" to char($length) data type."
        }
    }

    proc varchar {length string} {
        #| Cast to varchar.
        if { [string length $string] <= $length } {
            return $string
        } else {
            return -code error -errorcode CAST "Can't cast \"$string\" to varchar($length). String is too long."
        }
    }

    proc text {string} {
        #| Cast to text.
        return $string
    }
    
    proc enumeration {name value} {
        #| Cast $value to enumeration of $name.
        set value [string toupper $value]
        if {$value in [qc::memoize qc::db_enum_values $name]} {
            return $value
        } else {
            return -code error -errorcode CAST "Can't cast \"$value\": not a valid value for enumeration \"$name\"."
        }
    }

    proc domain {domain_name value} {
        #| Cast value to the domain $domain_name.
        set base_type [qc::memoize qc::db_domain_base_type $domain_name]
        if { [qc::castable $base_type $value] } {
            set value [qc::cast $base_type $value]
        } else {
            return -code error -errorcode CAST "Can't cast \"[qc::trunc $value 100]...\": not a valid value for base type \"$base_type\" while checking \"$domain_name\" type."
        }

        if { [qc::is $domain_name $value] } {
            return $value
        } else {
            return -code error -errorcode CAST "Can't cast \"[qc::trunc $value 100]...\": not a valid value for \"$domain_name\"."
        }
    }

    proc safe_html {text} {
        #| Cast text to safe html.
        if { [qc::is safe_html $text] } {
            return $text
        } else {
            return -code error -errorcode CAST "Can't cast \"[qc::trunc $text 100]...\": not safe html."
        }
    }

    proc safe_markdown {text} {
        #| Cast text to safe markdown.
        if {[qc::is safe_markdown $text]} {
            return $text
        } else {
            return -code error -errorcode CAST "Can't cast \"[qc::trunc $text 100]...\": not safe markdown."
        }
    }

    proc date {string} {
        #| Try to convert the given string into an ISO date.
        return [clock format [epoch $string] -format "%Y-%m-%d"]
    }

    proc time {time} {
        #| Try to convert the given string into a time,
        # in hh:mm:ss or hh:mm:ss.xxxxxx format
        if { [regexp {^((?:[0-1])?[0-9]|2[0-3]):((?:[0-5])?[0-9]):([0-5][0-9])\.(\d+)$} \
                  $time -> hours minutes seconds subseconds] } {
        } elseif {[regexp {^((?:[0-1])?[0-9]|2[0-3]):((?:[0-5])?[0-9]):([0-5][0-9])$} \
                  $time -> hours minutes seconds] } {
            set subseconds 0
        } elseif {[regexp {^((?:[0-1])?[0-9]|2[0-3]):((?:[0-5])?[0-9])$} \
                       $time -> hours minutes] } {
            set seconds 00
            set subseconds 0
        } elseif { [regexp {^24:00(:00(\.0+)?)?$} $time] } {
            set hours 24
            set minutes 00
            set seconds 00
            set subseconds 0
        } else {
            return -code error -errorcode CAST "Can't cast \"[qc::trunc $time 100]...\": not recognised time."            
        }
        set hours [format %02s $hours]
        set minutes [format %02s $minutes]
        set seconds [format %02s $seconds]
        set subseconds [qc::cast decimal -precision 7 -scale 6 "0.$subseconds"]

        if { $subseconds > 0.0 } {
            set subseconds [string range [string trimright $subseconds 0] 2 end]
            return "${hours}:${minutes}:${seconds}.$subseconds"
            
        } else {
            return "${hours}:${minutes}:${seconds}"
        }
    }

    proc epoch {string} {
        #| Try to convert the given string into an epoch.
        set string [string map [list "&#8209;" -] $string]

        #### EXACT MATCHES ####
        if { [string equal $string ""] } {
            return -code error -errorcode CAST "Can't cast an empty string to epoch"
        }

        # Looks like an eight-digit number, try to interpret as yyyymmdd or ddmmyyyy
        if { [regexp {^\d{8}$} $string] } {

            # Try yyyymmdd
            set year1 [string range $string 0 3]
            set month1 [string range $string 4 5]
            set day1 [string range $string 6 7]

            if { 0 < $month1 && $month1 <= 12
                 &&
                 0 < $day1 && $day1 <= [lindex {
                     - 31 29 31 30 31 30 31 31 30 31 30 31
                 } [string trimleft $month1 0]]
                 &&
                 1990 <= $year1 && $year1 < 2100
             } {
                set date1_valid true
            } else {
                set date1_valid false
            }

            # Try ddmmyyyy
            set year2 [string range $string 4 7]
            set month2 [string range $string 2 3]
            set day2 [string range $string 0 1]

            if { 0 < $month2 && $month2 <= 12
                 &&
                 0 < $day2 && $day2 <= [lindex {
                     - 31 29 31 30 31 30 31 31 30 31 30 31
                 } [string trimleft $month2 0]]
                 &&
                 1990 <= $year2 && $year2 < 2100
             } {
                set date2_valid true
            } else {
                set date2_valid false
            }

            # If one result is valid, return it. If both are, error.
            if { $date1_valid && $date2_valid } {
                return -code error -errorcode CAST "Ambiguous date"

            } elseif { $date1_valid } {
                return [clock scan "$year1-$month1-$day1"]

            } elseif { $date2_valid } {
                return [clock scan "$year2-$month2-$day2"]

            }
        }

        # Already an epoch
        if { [string is wideinteger -strict $string] } {
            if { $string > [clock scan "5000-01-01"] } {
                # Interpret as milliseconds past epoch
                return [expr {${string}/1000}]
            } elseif { $string>31 } {
                return $string
            }
        }

        # Exact ISO date
        if { [regexp {^(\d{4}|\d{2}|\d)-(\d{1,2})-(\d{1,2})$} $string -> year month day] } {
            return [clock scan "$year-$month-$day"]
        }

        # Exact ISO datetime
        if { [regexp {^(\d{4}|\d{2}|\d)-(\d{1,2})-(\d{1,2})( |T)(\d{1,2}:\d{1,2}(:\d{1,2})?)$} $string -> year month day . time] } {
            return [clock scan "$year-$month-$day $time"]
        }

        # Exact ISO datetime with offset timezone e.g. "2012-08-13 10:21:23.7777 -06:00"
        # Accepts offsets in formats -hh, -hhmm, or -hh:mm
        if { [regexp {^(\d{4}|\d{2}|\d)-(\d{1,2})-(\d{1,2})(?: |T)(\d{1,2}:\d{1,2})(?::(\d{1,2})(?:\.\d+)?)?\s?(Z|[-+]\d\d(:?\d\d)?)$} $string -> year month day time sec timezone] } {
            if { $timezone eq "Z" } {
                set timezone "+00"
            }
            if { $sec ne "" } {
                set time "$time:$sec"
            }
            return [clock scan "$year-$month-$day $time" -timezone "$timezone"]
        }

        # rfc-822 style dates used in emails.
        # "Fri, 17 Aug 2012 12:51:36 +0100"
        if { [regexp {(\d{1,2}) +(Jan|January|Feb|February|Mar|March|Apr|April|May|Jun|June|Jul|July|Aug|August|Sep|September|Oct|October|Nov|November|Dec|December) +(\d{4}) +(\d{1,2}:\d{1,2})(?::(\d{1,2})(?:\.\d+)?)?\s?(Z|[-+]\d\d(:?\d\d)?)} $string -> day mon year time sec timezone] } {
            if { $timezone eq "Z" } {
                set timezone "+00"
            }
            if { $sec ne "" } {
                set time "$time:$sec"
            }
            set month [expr {[lsearch {January February March April May June July August September October November December} "${mon}*"]+1}]
            return [clock scan "$year-$month-$day $time" -timezone "$timezone"]
        }

        # ISO datetime Don't match the end of line for e.g. "2009-04-06 12:25:18.343"
        if { [regexp {^(19\d\d|20\d\d)(\d\d)(\d\d)( |T)(\d{1,2}:\d{1,2}(:\d{1,2})?)} $string -> year month day . time] } {
            return [clock scan "$year-$month-$day $time"]
        }

        # dd/mm/yyYY
        if { [regexp {^(\d{1,2})[-/\.](\d{1,2})[-/\.](\d{4}|\d{2}|\d)$} $string -> day month year] } {
            # Assume UK locale dd/mm/yy, dd.mm.yy, or dd-mm-yy
            return [clock scan "$year-$month-$day"]
        }

        #### RELATIVE ####
        if { [regexp -nocase -- {(\+-)? ?(this|last|next|[0-9]+) ?(year|month|week|day|hour|minute)s?( ago)?} $string] \
                 || [regexp -nocase -- {(this|last|next) (Mon|Monday|Tue|Tuesday|Wed|Wednesday|Thu|Thurs|Thursday|Fri|Friday|Sat|Saturday|Sun|Sunday)} $string] \
                 || [regexp -nocase -- {today|yesterday|tomorrow} $string] } {
            return [clock scan $string]
        }

        #### RELAXED ####

        # Look for time hh:mm or hh:mm:ss
        if { ![regexp {(^|\W)(\d{1,2}:\d{1,2}(:\d{1,2})?)(\W|$)} $string -> ~ time ~] } { set time "" }

        # ISO format
        if { [regexp {(^|[^0-9])(\d{4}|\d{2}|\d)-(\d{1,2})-(\d{1,2})([^0-9]|$)} $string -> start year month day end] } {
            return [clock scan "$year-$month-$day $time"]
        }

        if { [regexp {(^|[^0-9])(\d{1,2})[/\.](\d{1,2})[/\.](\d{4}|\d{2}|\d)([^0-9]|$)} $string -> start day month year end] } {
            # Assume UK locale dd/mm/yy or dd.mm.yy
            return [clock scan "$year-$month-$day $time"]
        }

        # dd/mm or dd.mm
        if { [regexp {(^|\W)(\d{1,2})[/\.](\d{1,2})(\W|$)} $string -> start day month end] } {
            set year [clock format [clock seconds] -format "%Y"]
            return [clock scan "$year-$month-$day $time"]
        }

        if { [regexp {(^|[^0-9])(\d{4})([^0-9]|$)} $string -> start year end] } {
            #### Matched YEAR ####
            # year && month && dom
            # 23 June 2006 or June 23rd 2006
            if { [regexp -nocase -- {(^|[^a-zA-Z])(Jan|January|Feb|February|Mar|March|Apr|April|May|Jun|June|Jul|July|Aug|August|Sep|September|Oct|October|Nov|November|Dec|December)([^a-zA-Z]|$)} $string -> start month_name end] \
                     && [regexp -nocase -- {(^|\W)(\d{1,2})(st|nd|rd|th)?(\W|$)} $string -> start dom suffix end] } {
                # month and dom
                return [clock scan "$dom $month_name $year $time"]
            }
        } else {
            #### NO YEAR ####
            # no year && month && dom
            # 23rd June or Sept 11th
            if { [regexp -nocase -- {(^|[^a-zA-Z])(Jan|January|Feb|February|Mar|March|Apr|April|May|Jun|June|Jul|July|Aug|August|Sep|September|Oct|October|Nov|November|Dec|December)([^a-zA-Z]|$)} $string -> start month_name end] \
                     && [regexp -nocase -- {(^|\W)(\d{1,2})(st|nd|rd|th)(\W|$)} $string -> start dom suffix end] } {	 
                set year [clock format [clock seconds] -format "%Y"]
                return [clock scan "$dom $month_name $year $time"]
            }
            # dd or dd+suffix eg 23rd or 2nd ONLY 
            if { [regexp -nocase -- {(^|\W)(\d{1,2})(st|nd|rd|th)?(\W|$)} $string -> start day suffix end] } {
                set year [clock format [clock seconds] -format "%Y"]
                set month [clock format [clock seconds] -format "%m"]
                return [clock scan "$year-$month-$day $time"]
            }	
            if { [regexp -nocase -- {(^|\W)(Mon|Monday|Tue|Tuesday|Wed|Wednesday|Thu|Thurs|Thursday|Fri|Friday|Sat|Saturday|Sun|Sunday)(\W|$)} $string -> start dow end] } {
                return [clock scan "$dow $time"]
            }
        }

        if { [regexp -nocase -- {(year|fortnight|month|week|day|hour|min|sec|tomorrow|yesterday|today|now|last|next|ago)} $string] } {
            return [clock scan $string]
        }

        return -code error -errorcode CAST "Could not cast string to epoch"
    }

    proc postcode {string} {
        #| Try to cast a string into UK Postcode form
        set saved $string
        set postcode [string toupper $string]
        # BFPO 
        if { [string range $postcode 0 3] eq "BFPO" } {
            return $postcode
        }
        # convert AB12CD -> AB1 2CD or AB123CD -> AB12 3CD
        if { [string first " " $postcode] == -1 } {
            set cut [expr {[string length $postcode]-3-1}]
            set postcode "[string range $postcode 0 $cut] [string range $postcode [expr {$cut+1}] end]"
        }
        if { [qc::is postcode $postcode] } {
            return $postcode
        }
        # Convert zero -> CAPITAL O e.g. "Y023 3CD" -> "YO23 3CD"
        regsub {^([A-Z])0([0-9]{1,2}) (.+)$} $postcode {\1O\2 \3} postcode

        if { [qc::is postcode $postcode] } {
            return $postcode
        } else {
            return -code error -errorcode CAST "Could not cast $saved to postcode."
        }
    }

    proc creditcard {string} {
        #| Cast the given string to a credit card number.
        regsub -all {[^0-9]} $string "" number
        if { [qc::is creditcard $number] } {
            return $number
        } else {
            return -code error -errorcode CAST "$number is not a valid credit card number"
        }
    }

    proc period {string} {
        #| Return a pair of dates defining the period.
        set month_names [list Jan January Feb February Mar March Apr April May Jun June Jul July Aug August Sep September Oct October Nov November Dec December]
        set regexp_map [list \$month_names_regexp [join $month_names |]]

        if { [regexp -nocase {^\s*(.*?)\s+to\s+(.*?)\s*$} $string -> period1 period2] } {
            # Period defined by two periods eg "Jan 2011 to March 2011"
            lassign [period $period1] from_date .
            lassign [period $period2] . to_date

        } elseif { [regexp {^\d{4}-\d{2}-\d{2}$} $string] } {
            # String is an iso date eg "2014-01-01"
            set from_date [qc::cast date $string]
            set to_date $from_date

        } elseif { [regexp {^([12]\d{3})$} $string -> year] } {
            # Exact match for year eg "2006"
            set from_date [qc::date_year_start $year-01-01]
            set to_date [qc::date_year_end $year-01-01]

        } elseif { [regexp -nocase -- [string map $regexp_map {^($month_names_regexp)\s+([12]\d{3})$}] $string -> month_name year] } {
            # Exact match in format "Jan 2006"
            set epoch [clock scan "01 $month_name $year"]
            set from_date [qc::date_month_start $epoch]
            set to_date [qc::date_month_end $epoch]

        } elseif { [regexp -nocase -- [string map $regexp_map {^($month_names_regexp)$}] $string -> month_name] } {
            # Exact match in format "Jan" (assume current year)
            set epoch [clock scan "01 $month_name [qc::date_year now]"]
            set from_date [qc::date_month_start $epoch]
            set to_date [qc::date_month_end $epoch]

        } elseif { [regexp -nocase -- [string map $regexp_map {^([0-9]{1,2})(?:st|th|nd|rd)?\s+($month_names_regexp)\s+([12]\d{3})$}] $string -> dom month_name year] } {
            # Exact match for castable date in format "1st Jan 2014"
            set epoch [clock scan "$dom $month_name $year"]
            set from_date [date $epoch]
            set to_date $from_date

        } elseif { [regexp -nocase -- [string map $regexp_map {^([0-9]{1,2})(?:st|th|nd|rd)?\s+($month_names_regexp)$}] $string -> dom month_name] } {
            # Exact match for castable date in format "1st Jan" (assume current year)
            set epoch [clock scan "$dom $month_name [qc::date_year now]"]
            set from_date [date $epoch]
            set to_date $from_date

        } elseif { [regexp -nocase -- [string map $regexp_map {^($month_names_regexp)\s+([0-9]{1,2})(?:st|th|nd|rd)?\s+([12]\d{3})$}] $string -> month_name dom year] } {
            # Exact match for castable date in format "Jan 1st 2014"
            set epoch [clock scan "$dom $month_name $year"]
            set from_date [date $epoch]
            set to_date $from_date

        } elseif { [regexp -nocase -- [string map $regexp_map {^($month_names_regexp)\s+([0-9]{1,2})(?:st|th|nd|rd)?$}] $string -> month_name dom] } {
            # Exact match for castable date in format "Jan 1st" (assume current year)
            set epoch [clock scan "$dom $month_name [qc::date_year now]"]
            set from_date [date $epoch]
            set to_date $from_date

        } else {
            # error - could not parse string
            return -code error -errorcode CAST "Could not parse string \"$string\" into dates that define a period."
        }
        
        return [list $from_date $to_date]
    }

    proc url {string} {
        #| Cast the given string to an url
        #| (See also qc::is url)
        set lower [string tolower $string]
        if { [qc::is url $lower] } {
            return $lower
        }
        if { [qc::is url "http://${lower}"] } {
            return "http://${lower}"
        }
        return -code error -errorcode CAST "Could not cast $string to an url."
    }

    proc relative_url {string} {
        #| Cast the given string to a relative url
        #| (See also qc::is url)
        set lower [string tolower $string]
        if { [qc::is url -relative $lower] } {
            return $lower
        }  
        return -code error -errorcode CAST "Could not cast $string to an url."      
    }
    
    proc url_path {string} {
	#| Cast the given string to an url_path
	#| (See also qc::is url_path)
	set lower [string tolower $string]
	if { [qc::is url_path $lower] } {
	    return $lower
	}
	if { [qc::is url_path "/${lower}"] } {
	    return "/${lower}"
	}
	return -code error -errorcode CAST "Could not cast $string to an url path"
    }

    proc s3_uri {string} {
        #| Cast the given string to an s3 uri
        #| (See also qc::is s3_uri)

        # Split the string into bucket and object key
        lassign [qc::s3 uri_bucket_object_key $string] bucket object_key
        if { ![qc::is s3_bucket $bucket] } {
            return -code error -errorcode CAST "Could not cast \"$string\" to an s3_uri"
        }
        if { $object_key ne "" && ![qc::is s3_object_key $object_key] } {
            return -code error -errorcode CAST "Could not cast \"$string\" to an s3_uri"
        }
        
        return "s3://${bucket}/${object_key}"
    }

    proc interval {string} {
        #| Cast string to interval
        if { [qc::is interval $string] } {
            return [string tolower $string]
        }
	return -code error -errorcode CAST "Could not cast $string to an interval"
    }

    proc next_url {string} {
        if { ![regexp {^https?://} $string] } {
            # Relative url            
            set next_url [string trimleft $string /]
            set next_url "[qc::conn_location]/$next_url"           
        } else {
            set next_url $string
        }

        # check for malicious mal-formed url
        if { ![qc::is url $next_url] } {
            return -code error -errorcode CAST "Could not cast \"[html_escape $next_url]\" to valid next_url. Invalid URL."
        }
        
        # check that redirection is to the same domain
        set conn_host [qc::conn_host]
        if { ![regexp "^https?://${conn_host}(:\[0-9\]+)?(\\?|#|/|\$)" $next_url] } {
            return -code error -errorcode CAST "Could not cast \"[html_escape $next_url]\" to valid next_url. Domain must be \"[html_escape $conn_host]\"."
        }       

        return $next_url
    }
}

namespace eval qc {
    namespace export perl_quote list2perl_array dict2perl_hash llist2perl_aarray ddict2perl_hhash
}

proc qc::perl_quote {value} {
    #| Quote a value for use as a perl scalar
    return '[string map [list ' \\'] $value]'
}

proc qc::list2perl_array {list} {
    #| Convert a list into a perl array
    set quoted_values {}
    foreach value $list {
        lappend quoted_values [perl_quote $value]
    }
    return \[[join $quoted_values ", "]\]
}

proc qc::dict2perl_hash {dict} {
    #| Convert a dict into a perl hash
    set pairs {}
    dict for {name value} $dict {
        lappend pairs "$name => [perl_quote $value]"
    }
    return \{[join $pairs ", "]\}
}

proc qc::llist2perl_aarray {llist} {
    #| Convert a list of lists into a perl array of arrays
    set array_list {}
    foreach list $llist {
        lappend array_list [list2perl_array $list]
    }
    return \[[join $array_list ", "]\]
}

proc qc::ddict2perl_hhash {ddict} {
    #| Convert a nested dict into a perl hash of hashes
    set pairs {}
    dict for {key dict} $ddict {
        lappend pairs "$key => [dict2perl_hash $dict]"
    }
    return \{[join $pairs ", "]\}
}

proc qc::ldict2perl_ahash {ldict} {
    #| Convert a list of dictionaries into a perl array of hashes
    set hash_list {}
    foreach dict $ldict {
        lappend hash_list [dict2perl_hash $dict]
    }
    return \[[join $hash_list ", "]\]
}
namespace eval qc {
    namespace export db_*
}

proc qc::db_qualify_column {args} {
    #| Attempt to resolve 1-3 args as schema, table, column
    # Assume args are one of:
    # column
    # table column
    # schema table column
    switch [llength $args] {
        1 {
            lassign $args column_name

            # Select whichever table(s) in the current search path contain
            # column, and are earliest in schema search path
            set data [list]
            db_cache_foreach -ttl 86400 {
                select *
                from (
                      select
                      table_schema,
                      index,
                      table_name,
                      min(index) over () as min_index

                      from information_schema.columns
                      join (
                            select distinct on(table_name)
                            table_name,
                            table_schema,
                            index

                            from information_schema.tables
                            cross join
                            generate_series(1,
                                            array_length(current_schemas(true),1)
                                            ) as index

                            where table_schema = (current_schemas(true))\[index\]

                            order by
                            table_name,
                            index
                            ) i using(table_schema,table_name)
                      
                      where column_name = :column_name
                      ) t
                where index=min_index
            } {
                lappend data [dict_from {*}{
                    table_schema
                    table_name
                }]
            }

            # No tables in search path, check for all tables containing column
            if { [llength $data] == 0 } {
                db_cache_foreach -ttl 86400 {
                    select
                    table_schema,
                    table_name
                    
                    from information_schema.columns

                    where column_name = :column_name
                } {
                    lappend data [dict_from {*}{
                        table_schema
                        table_name
                    }]
                }
            }

            # Still no tables found
            if { [llength $data] == 0 } {
                error "No table containing column \"$column_name\""
            }

            # Exactly one table found
            if { [llength $data] == 1 } {
                dict2vars [lindex $data 0] table_schema table_name
                return [list $table_schema $table_name $column_name]
            }
            
            # Filter selected tables for one where column is a primary key
            set sql_value_rows [list]
            foreach row $data {
                dict2vars $row table_schema table_name
                lappend sql_value_rows \
                    "([db_quote $table_schema], [db_quote $table_name])"
            }
            db_cache_0or1row -ttl 86400 {
                select
                table_schema,
                table_name

                from (VALUES [join $sql_value_rows ,])
                as t (table_schema, table_name)
                
                join information_schema.table_constraints tc
                using (table_schema, table_name)

                join information_schema.constraint_column_usage ccu
                using (table_schema, table_name, constraint_name)

                where column_name=:column_name
                and constraint_type='PRIMARY KEY'

                limit 1
            } {
                # No tables with column as a primary key;
                # just return the first table found
                dict2vars [lindex $data 0] table_schema table_name                
            }
            return [list $table_schema $table_name $column_name]
        }
        2 {
            lassign $args table_name column_name

            # Look for a table in the current search path matching
            # table name
            db_cache_0or1row -ttl 86400 {
                select table_schema
                from information_schema.tables
                cross join generate_series(1,
                                           array_length(current_schemas(true),1)
                                           ) as index
                where table_schema = (current_schemas(true))\[index\]
                and table_name = :table_name
                order by index
                limit 1
            } {
                # No table found in search path;
                # fall back to any table containing column
                db_cache_1row -ttl 86400 {
                    select table_schema
                    from information_schema.columns
                    where table_name=:table_name
                    and column_name=:column_name
                    limit 1
                }
            } {
                # Check that selected table contains column
                db_cache_0or1row -ttl 86400 {
                    select true as exists
                    from information_schema.columns
                    where table_name=:table_name
                    and table_schema=:table_schema
                    and column_name=:column_name
                } {
                    error "Column \"$column_name\" not found in table \"$table_name\""
                }
            }
        }
        3 { lassign $args table_schema table_name column_name }
        default {
            error "Unable to resolve args to db_qualify_column"
        }
    }
    return [list $table_schema $table_name $column_name]
}

proc qc::db_resolve_field_name {name} {
    #| Resolve a field name to schema, table, column
    set parts [split $name "."]
    return [qc::db_qualify_column {*}$parts]
}

proc qc::db_qualify_table {args} {
    #| Resolve args to schema and table
    switch [llength $args] {
        1 {
            lassign $args table_name

            # Look for first table in search path matching table_name
            db_cache_0or1row -ttl 86400 {
                select table_schema

                from information_schema.tables
                cross join generate_series(1,
                                           array_length(current_schemas(true),1)
                                           ) as index

                where table_schema = (current_schemas(true))\[index\]
                and table_name = :table_name

                order by index
                
                limit 1
            } {
                # Fall back to any table matching table_name
                db_cache_1row -ttl 86400 {
                    select table_schema
                    from information_schema.tables
                    where table_name = :table_name
                    limit 1
                }
            }
        }
        2 {
            lassign $args table_schema table_name
        }
        default {
            error "Unable to resolve args to db_qualify_table"
        }
    }
    return [list $table_schema $table_name]
}

proc qc::db_col_varchar_length {args} {
    #| Returns the varchar length of a db table column
    lassign [qc::db_qualify_column {*}$args] {*}{
        schema
        table
        column
    }
    set qry {
        SELECT
        a.atttypmod-4 AS lengthvar,
        t.typname AS type

        FROM
        pg_namespace n,
        pg_attribute a,
        pg_class c,
        pg_type t

        WHERE
        n.nspname = :schema
        and c.relnamespace = n.oid
        and c.relname = :table
        and a.attnum > 0
        and a.attrelid = c.oid
        and a.atttypid = t.oid
        and a.attname = :column

        ORDER BY a.attnum
    }
    db_cache_0or1row -ttl 86400 $qry {
        error "No such column \"$column\" in table \"${schema}.${table}\""
    } 
    if { [eq $type varchar] } {
        return $lengthvar
    } else {
        error "Col \"$column\" is not type varchar it is type \"$type\""
    }
}

proc qc::db_column_exists {column} {
    #| Checks if the given column name exists in the database.
    db_cache_1row -ttl 86400 {
        SELECT count(*) as count 
        FROM information_schema.columns
        WHERE column_name=:column
    }
    if {$count > 0} {
        return true
    } else {
        return false
    }
}

proc qc::db_table_columns {args} {
    #| Returns a list of columns for the given table.
    lassign [qc::db_qualify_table {*}$args] {*}{
        table_schema
        table_name
    }
    set qry {
        SELECT column_name
        FROM information_schema.columns
        WHERE table_name=:table_name
        AND table_schema=:table_schema
        ORDER BY ordinal_position;
    }
    set columns {}
    qc::db_cache_foreach -ttl 86400 $qry {
        lappend columns $column_name
    }
    return $columns
}

proc qc::db_table_column_exists {args} {
    #| Checks if the column exists in the given table.
    lassign [qc::db_qualify_table {*}[lrange $args 0 end-1]] {*}{
        table_schema
        table_name
    }
    set column [lindex $args end]
    set qry {
        SELECT column_name
        FROM information_schema.columns
        WHERE column_name=:column
        AND table_name=:table_name
        AND table_schema=:table_schema;
    }
    qc::db_cache_0or1row -ttl 86400 $qry {
        return false
    } {
        return true
    }
}

proc qc::db_column_table {column} {
    #| Returns a list of tables that the given column name is part of.
    set qry {
        SELECT table_name
        FROM information_schema.columns
        WHERE column_name=:column
    }
    set tables {}
    qc::db_cache_foreach -ttl 86400 $qry {
        lappend tables $table_name
    }
    return $tables
}

proc qc::db_qualified_table_column {column} {
    #| Returns a fully qualified table column as a list of a pair: {table column}
    # No multi-schema support; use qc::db_qualify_column instead
    if {[qc::db_column_exists $column]} {
        set tables [qc::db_column_table $column]
        if {[llength $tables] > 1 && [qc::db_column_table_primary_exists $column]} {
            # try a column that's a also primary key
            set table [qc::db_column_table_primary $column]
        } else {
            # try the first table anyway
            set table [lindex $tables 0]
        }
        return [list $table $column]
    } else {
        return -code error "\"$column\" doesn't exist as a column in the database."
    }
}

proc qc::db_column_table_primary_exists {column} {
    #| Checks whether the given column exists as a primary key in the database.
    ::try {
        qc::db_column_table_primary $column
        return true
    } on error [list error_message options] {
        return false
    }
}

proc qc::db_column_table_primary {column} {
    #| Returns a table that the given column is a primary key in.
    set qry {
        SELECT tc.table_name
        FROM information_schema.table_constraints tc
        JOIN information_schema.constraint_column_usage ccu
        USING (table_name, constraint_name)
        WHERE column_name=:column
        AND constraint_type='PRIMARY KEY'
        LIMIT 1;
    }
    qc::db_cache_1row -ttl 86400 $qry
    return $table_name
}

proc qc::db_column_type {args} {
    #| Returns the data type of the column in the given table.
    qc::args $args -qualified -- args
    default qualified false
    lassign [qc::db_qualify_column {*}$args] {*}{
        table_schema
        table_name
        column_name
    }
    set qry {
        SELECT
        column_name,
        domain_name,
        domain_schema,
        udt_name,
        udt_schema,
        character_maximum_length,
        numeric_precision,
        numeric_scale

        FROM information_schema.columns

        WHERE table_schema=:table_schema
        and table_name=:table_name
        and column_name=:column_name
    }
    qc::db_cache_0or1row -ttl 86400 $qry {
        return -code error -errorcode DB_COLUMN "Column \"$column\" does not exist for table \"$table\"."
    } {
        if { $qualified } {
            if { $domain_name ne "" } {
                set data_type "${domain_schema}.${domain_name}"
            } elseif { $udt_schema ne "pg_catalog" } {
                set data_type "${udt_schema}.${udt_name}"
            } else {
                set data_type $udt_name
            }
        } else {
            if { $domain_name ne "" } {
                set data_type $domain_name
            } else {
                set data_type $udt_name
            }
        }
        return [qc::db_canonical_type $data_type $character_maximum_length $numeric_precision $numeric_scale]
    }
}

proc qc::db_table_column_types {args} {
    #| Returns a dict of all columns and their types in the given table.
    qc::args $args -qualified -- args
    default qualified false
    lassign [qc::db_qualify_table {*}$args] {*}{
        table_schema
        table_name
    }
    set qry {
        SELECT column_name,
        character_maximum_length,
        domain_name,
        domain_schema,
        udt_name,
        udt_schema,
        numeric_precision,
        numeric_scale

        FROM information_schema.columns

        WHERE table_schema=:table_schema
        and table_name=:table_name

        ORDER BY ordinal_position;
    }
    set column_types {}
    qc::db_cache_foreach -ttl 86400 $qry {
        if { $qualified } {
            if { $domain_name ne "" } {
                set data_type "${domain_schema}.${domain_name}"
            } elseif { $udt_schema ne "pg_catalog" } {
                set data_type "${udt_schema}.${udt_name}"
            } else {
                set data_type $udt_name
            }
        } else {
            if { $domain_name ne "" } {
                set data_type $domain_name
            } else {
                set data_type $udt_name
            }
        }
        lappend column_types $column_name [qc::db_canonical_type $data_type $character_maximum_length $numeric_precision $numeric_scale]
    }
    return $column_types
}

proc qc::db_column_nullable {args} {
    #| Checks if the column in the given table can hold null values.
    lassign [qc::db_qualify_column {*}$args] {*}{
        table_schema
        table_name
        column_name
    }
    set qry {
        SELECT is_nullable
        FROM information_schema.columns
        WHERE table_schema=:table_schema
        and table_name=:table_name
        and column_name=:column_name
    }
    qc::db_cache_1row -ttl 86400 $qry
    if {$is_nullable} {
        return true
    } else {
        return false
    }
}

proc qc::db_resolve_type_name {name} {
    #| Attempt to resolve args to schema and type_name
    set parts [split $name "."]
    switch [llength $parts] {
        2 { return $parts }
        1 {
            set type_name [lindex $parts 0]
            db_cache_0or1row -ttl 86400 {
                select
                n.nspname as schema

                from
                pg_namespace n,
                pg_type t,
                generate_series(1,
                                array_length(current_schemas(true),1)
                                ) as index

                where t.typnamespace = n.oid
                and t.typname = :type_name
                and n.nspname = (current_schemas(true))\[index\]

                order by index

                limit 1
            } {
                db_cache_1row -ttl 86400 {
                    select
                    n.nspname as schema

                    from
                    pg_namespace n,
                    pg_type t

                    where t.typnamespace = n.oid
                    and t.typname = :type_name

                    limit 1
                }
            }
            return [list $schema $type_name]
        }
        default {
            error "Unable to resolve type name \"$name\""
        }
    }
}

proc qc::db_enum_values {enum_name} {
    #| Returns a list of the values for the given enumeration
    lassign [qc::db_resolve_type_name $enum_name] {*}{
        schema
        enum_name
    }
    set qry {
        SELECT e.enumlabel as value
        FROM pg_enum e 
        JOIN pg_type t
        ON t.oid = e.enumtypid
        JOIN pg_namespace n
        ON t.typnamespace = n.oid
        WHERE t.typname=:enum_name
        AND n.nspname=:schema;
    }
    set values {}
    qc::db_cache_foreach -ttl 86400 $qry {
        lappend values $value
    }
    return $values
}

proc qc::db_enum_exists {args} {
    #| Checks if the given enum exists in the database.
    qc::args $args -no-cache -- enum_name
    set parts [split $enum_name "."]
    if { [llength $parts] == 2 } {
        lassign $parts schema enum_name
        set qry {
            SELECT e.enumtypid
            FROM pg_enum e
            JOIN pg_type t
            ON t.oid = e.enumtypid
            JOIN pg_namespace n
            ON t.typnamespace = n.oid
            WHERE t.typname=:enum_name
            AND n.nspname=:schema
            LIMIT 1;
        }
    } else {
        set qry {
            SELECT e.enumtypid
            FROM pg_enum e
            JOIN pg_type t
            ON t.oid = e.enumtypid
            WHERE t.typname=:enum_name
            LIMIT 1;
        }
    }
    if { [info exists no-cache] } {
	set ttl -1
    } else {
	set ttl 86400
    }
    qc::db_cache_0or1row -ttl $ttl $qry {
        return false
    } {
        return true
    }
}

proc qc::db_domain_exists {args} {
    #|Checks if the given domain exists in the database
    qc::args $args -no-cache -- domain_name
    set parts [split $domain_name "."]
    if { [llength $parts] == 2 } {
        lassign $parts domain_schema domain_name
        set qry {
            SELECT domain_name
            FROM information_schema.domains
            WHERE domain_name=:domain_name
            AND domain_schema=:domain_schema;
        }
    } else {
        set qry {
            SELECT domain_name
            FROM information_schema.domains
            WHERE domain_name=:domain_name
            LIMIT 1;
        }
    }
    if { [info exists no-cache] } {
	set ttl -1
    } else {
	set ttl 86400
    }
    qc::db_cache_0or1row -ttl $ttl $qry {
	return false
    } {
	return true
    }
}

proc qc::db_resolve_domain_name {name} {
    #| Resolve args to schema and domain
    set parts [split $name "."]
    switch [llength $parts] {
        2 { return $parts }
        1 {
            set domain_name [lindex $parts 0]
            db_cache_0or1row -ttl 86400 {
                select domain_schema

                from information_schema.domains,
                generate_series(1,
                                array_length(current_schemas(true),1)
                                ) as index

                where domain_schema = (current_schemas(true))\[index\]
                and domain_name=:domain_name

                order by index
                limit 1
            } {
                db_cache_1row -ttl 86400 {
                    select domain_schema

                    from information_schema.domains

                    where domain_name=:domain_name
                    
                    limit 1
                }
            }
            return [list $domain_schema $domain_name]
        }
        default {
            error "Unable to resolve domain name"
        }
    }
}

proc qc::db_domain_constraints {name} {
    #| Returns a dict of the constraint name with the check clause for the given domain in the database.
    lassign [qc::db_resolve_domain_name $name] {*}{
        domain_schema
        domain_name
    }
    set constraints [dict create]
    if {[qc::memoize qc::db_domain_exists $domain_name]} {
        set qry {
            SELECT cc.constraint_name, check_clause
            FROM information_schema.check_constraints cc
            JOIN information_schema.domain_constraints dc
            ON dc.constraint_name=cc.constraint_name
            AND dc.constraint_schema=cc.constraint_schema
            WHERE dc.domain_name=:domain_name
            AND dc.domain_schema=:domain_schema;
        }
        qc::db_cache_foreach -ttl 86400 $qry {
            dict set constraints $constraint_name $check_clause
        }
        return $constraints
    } else {
        return -code error -errorcode DB_DOMAIN "Domain \"$domain_name\" does not exist."
    }
}

proc qc::db_column_constraints {args} {
    #| Returns a dict of constraint name and check clause for the given column
    lassign [qc::db_qualify_column {*}$args] {*}{
        schema
        table
        column
    }
    if {[qc::db_table_column_exists $table $column]} {
        set qry {
            SELECT cc.constraint_name, check_clause
            FROM information_schema.check_constraints cc
            JOIN information_schema.constraint_column_usage ccu
            ON cc.constraint_name=ccu.constraint_name
            AND ccu.constraint_schema=cc.constraint_schema
            WHERE ccu.table_schema=:schema
            AND ccu.table_name=:table
            AND ccu.column_name=:column;
        }
        set constraints {}
        qc::db_cache_foreach -ttl 86400 $qry {
            lappend constraints $constraint_name $check_clause
        }
        return $constraints
    } else {
        return -code error -errorcode DB_COLUMN "Column \"$schema.$table.$column\" does not exist."
    }
}

proc qc::db_eval_constraint {args} {
    #| Check a db constraint expression by substituting in corresponding values from args
    # eg constraint_test {(col1 > col2)} col1 17 col2 16
    qc::args $args -schema ? -- table constraint args
    if { ! [info exists schema] } {
        lassign [qc::db_qualify_table $table] {*}{
            schema
            table
        }
    }
    set column_types [qc::memoize qc::db_table_column_types \
                          -qualified -- $schema $table]
    set columns [dict keys $column_types]
    set tq_columns [qc::map [list x "return $table.\$x"] $columns]
    set sq_columns [qc::map [list x "return $schema.$table.\$x"] $columns]
    set column_values [qc::dict_subset $args {*}$columns {*}$tq_columns {*}$sq_columns]
    set list {}
    foreach {name value} $column_values {
        lassign [qc::db_resolve_field_name $name] {*}{
            schema
            table
            column
        }            
        lappend list "[qc::db_quote $value]::[dict get $column_types $column] as $column"
    }    
    set sub_select "SELECT [join $list ,]"

    set qry {
        SELECT coalesce($constraint, true) AS result
        FROM ($sub_select) alias;
    }

    qc::db_cache_1row -ttl 86400 $qry
    return $result
}

proc qc::db_eval_domain_constraint {value base_type check_clause} {
    #| Evaluates the domain constraint against the given value.
    set qry {
        SELECT $check_clause AS result
        FROM (SELECT :value::$base_type as VALUE) alias;
    }
    ::try {
        qc::db_cache_1row -ttl 86400 $qry
        return $result
    } on error [list error_message options] {
        return f
    }
}

proc qc::db_eval_column_constraints {args} {
    #| Evaluates constraints on the given table.column with the given values.
    #| Returns a dict of the constraints and their results.
    set values [lindex $args end]
    lassign [qc::db_qualify_column {*}[lrange $args 0 end-1]] {*}{
        schema
        table
        column
    }
    set constraints \
        [qc::memoize qc::db_column_constraints $schema $table $column]
    set results {}
    foreach {constraint_name check_clause} $constraints {
        lappend results $constraint_name \
            [qc::db_eval_constraint \
                 -schema $schema -- $table $check_clause {*}$values]
    }
    return $results
}

proc qc::db_domain_base_type {domain} {
    #| Returns the base type of the given domain.
    if { [qc::memoize qc::db_domain_exists $domain] } {
        lassign [qc::db_resolve_domain_name $domain] {*}{
            schema
            domain_name
        }
        set qry {
            SELECT
            udt_name,
            character_maximum_length,
            numeric_precision,
            numeric_scale

            FROM information_schema.domains

            WHERE domain_name=:domain_name
            AND domain_schema=:schema;
        }
        qc::db_cache_1row -ttl 86400 $qry
        return [qc::db_canonical_type $udt_name $character_maximum_length $numeric_precision $numeric_scale]
    } else {
        return -code error -errorcode DB_DOMAIN "Domain \"$domain_name\" does not exist."
    }
}

proc qc::db_canonical_type {udt_name {character_maximum_length ""} {numeric_precision ""} {numeric_scale ""}} {
    #| Returns the canonical type name for the given type name.
    switch -glob -- $udt_name {
        varchar {
            return [expr {$character_maximum_length ne "" ? "varchar($character_maximum_length)": "text"}]
        }
        numeric {
            return [expr {$numeric_precision ne "" ? "decimal($numeric_precision,$numeric_scale)": "decimal"}]
        }
        bpchar {
            return "char($character_maximum_length)"
        }
        bit {
            return "bit($character_maximum_length)"
        }
        default {
            return $udt_name
        }
    }
}

proc qc::db_validation_message {table column} {
    #| Returns the validation message associated with the given table.column
    set qry {
        SELECT message
        FROM validation_messages
        WHERE
        column_name=:column
        order by table_name=:table DESC
        LIMIT 1
    }
    qc::db_cache_0or1row -ttl 86400 $qry {
        set message "Invalid $column."
    }
    return $message
}

proc qc::db_sequence_exists {args} {
    #| Checks if the given sequence exists in the database
    qc::args $args -no-cache -- sequence_name
    set qry {
	SELECT sequence_name 
	FROM information_schema.sequences
	WHERE sequence_name=:sequence_name;
    }
    if { [info exists no-cache] } {
	set ttl -1
    } else {
	set ttl 86400
    }
    qc::db_cache_0or1row -ttl $ttl $qry {
	return false
    } {
	return true
    }
}

proc qc::db_owner {database_name} {
    #| Gets the owner of the database
    set qry {
	SELECT u.usename as database_owner
	FROM pg_database d
	JOIN pg_user u ON (d.datdba=u.usesysid)
	WHERE d.datname=:database_name
    }
    db_0or1row $qry {
	error "Database \"$database_name\" does not exists"
    } {
	return $database_owner
    }
}

proc qc::db_database_name {{poolname DEFAULT}} {
    #| Gets the name of the database using poolname
    set qry {
	select current_database() as current_database
    }
    db_1row -db $poolname $qry
    return $current_database
}

proc qc::db_user {{poolname DEFAULT}} {
    #| Gets the user configured to connect to database using poolname
    db_1row {select current_user}
    return $current_user
}

proc qc::db_extension_exists {extension_name} {
    #| Checks if the extension exists in database
    set qry {
	SELECT extname
	FROM pg_extension
	WHERE extname=:extension_name
    }
    db_0or1row $qry {
	return false
    } {
	return true
    }
}

proc qc::db_user_is_superuser {username} {
    #| Returns true if username specified is a superuser
    set qry {
	SELECT usesuper
	FROM pg_user
	WHERE usename=:username;
    } 
    db_0or1row $qry {  
	# user does not exist
	error "Database user \"$username\" does not exist." 
    } {
	return $usesuper
    }
}

proc qc::db_user_is_member {username rolename} {
    #| Returns true if specified username is a member of $rolename

    set qry {
        select
        r.rolname,
        u.rolname
        FROM pg_catalog.pg_auth_members m
        JOIN pg_catalog.pg_roles r ON (m.roleid = r.oid)
        JOIN pg_catalog.pg_roles u ON (m.member=u.oid)
        where u.rolname=:username and r.rolname=:rolename;
    }
    db_0or1row $qry {
        return false
    } {
	return true
    }
}

proc qc::db_is {data_type value} {
    #| Determines if the value is a database data type.
    set qry {
        SELECT :value::$data_type as VALUE;
    }
    
    ::try {
        qc::db_cache_1row -ttl 86400 $qry
        return t
    } on error [list error_message options] {
        return f
    }
}

proc qc::db_table_check_constraints {schema table} {
    #| Get all check constraints for the given table.
    #| Return value is a list of constraints where each constraint is a dict
    #| with the following keys:
    #|   constraint_name - The name of the check constraint.
    #|   check_clause    - The check constraint clause.
    #|   column_names    - The columns used in the check constraint clause.

    set qry {
        select
        cc.constraint_name,
        check_clause,
        array_agg(ccu.column_name) as column_names

        from
        information_schema.check_constraints cc
        join information_schema.constraint_column_usage ccu
            on cc.constraint_name=ccu.constraint_name
            and ccu.constraint_schema=cc.constraint_schema

        where
        ccu.table_schema=:schema
        and ccu.table_name=:table

        group by
        cc.constraint_name,
        check_clause
    }

    set constraints [list]

    qc::db_cache_foreach -ttl 86400 $qry {
        lappend constraints [dict create \
                                 constraint_name $constraint_name \
                                 check_clause $check_clause \
                                 column_names [qc::sql_array2list $column_names]]
    }

    return $constraints
}

proc qc::db_table_check_constraints_eval {
    schema
    table
    column_values
} {
    #| Evaluate check constraints in the given table that use columns specified in column_values.
    #| Returns a dict of passed and failed check constraints e.g.:
    #| {
    #|   passed {
    #|     customer_firstname_check {firstname Joe}
    #|     customer_surname_check   {surname Bloggs}
    #|   }
    #|   failed {
    #|     customer_invoice_email_check {
    #|       send_invoice_by EMAIL
    #|       invoice_email   {}
    #|     }
    #|   }
    #| }
    set constraints [qc::memoize qc::db_table_check_constraints $schema $table]
    set results [dict create \
                     passed [dict create] \
                     failed [dict create]]

    foreach constraint $constraints {

        qc::dict2vars $constraint {*}{
            constraint_name
            check_clause
            column_names
        }

        # Narrow down values to pass to db_eval_constraint.
        # Evaluating a check constraint is much quicker if only the necessary values are given.
        set check_column_values [dict create]

        dict for {column value} $column_values {
            if { $column in $column_names } {
                dict set check_column_values $column $value
            }
        }

        if { [dict size $check_column_values] == 0 } {
            continue
        }

        if { [qc::db_eval_constraint \
                 -schema $schema \
                 -- \
                 $table \
                 $check_clause \
                 {*}$check_column_values] } {
            dict set results passed $constraint_name $check_column_values
        } else {
            dict set results failed $constraint_name $check_column_values
        }
    }

    return $results
}

namespace eval qc {
    namespace export dicts_diff_any
}

proc qc::dicts_diff_any {left_dict right_dict} {
    #| Return the first diff found
    dict for {name value} $right_dict {
        if { ! [dict exists $left_dict $name] } {
            return "Key $name not found in left dict"
        }
        if { [dict get $left_dict $name] ne $value } {
            return "Left $name : \"[dict get $left_dict $name]\" Right $name : \"$value\""
        }
    }
    dict for {name value} $left_dict {
        if { ! [dict exists $right_dict $name] } {
            return "Key $name not found in right dict"
        }
        if { [dict get $left_dict $name] ne $value } {
            return "Left $name : \"[dict get $left_dict $name]\" Right $name : \"$value\""
        }
    }
}

namespace eval qc {
    namespace export sql_where_in sql_where_in_not
}

proc qc::sql_where_in {args } {
    #| Construct part of a SQL WHERE clause using the IN construct.
    # SQL will test column against list of values.
    qc::args $args -type "" -- column list {default false}

    foreach item $list {
	lappend lquoted [db_quote $item $type]
    }
    if {[llength $list]==0} {
	return $default
    } elseif {[llength $list]==1} {
	return "$column=[lindex $lquoted 0]"
    } else {
	return "$column in ([join $lquoted ,])"
    }
}

proc qc::sql_where_in_not { args } {
    #| Construct part of a SQL WHERE clause using the NOT IN construct.
    # SQL will test column against list of values.
    qc::args $args -type "" -- column list {default true}

    foreach item $list {
	lappend lquoted [db_quote $item $type]
    }
    if {[llength $list]==0} {
	return $default
    } elseif {[llength $list]==1} {
	return "$column<>[lindex $lquoted 0]"
    } else {
	return "$column not in ([join $lquoted ,])"
    }
}


namespace eval qc {
    namespace export pkcs_padding_append pkcs_padding_strip encrypt_bf_tcl encrypt_bf_db encrypt_bf decrypt_bf_tcl decrypt_bf_db decrypt_bf
}
package require base64 
package require blowfish

proc qc::pkcs_padding_append {string} {
    #| Pads out the string to be multiple of 8 bytes in length.
    #| Padding character as per PKCS (RFC 2898).
    set padding_length [expr {8 - ([string length $string]%8)}]
    set padding_char [format %c $padding_length]
    set padding [string repeat $padding_char $padding_length]

    return "${string}${padding}"
}

proc qc::pkcs_padding_strip {string} {
    #| Trim PKCS padding from end of string.
    set padding_char [string index $string end] 
    set padding_length [scan $padding_char %c]
    set padding [string repeat $padding_char $padding_length]

    if { [string range $string end-[expr {$padding_length-1}] end] eq $padding } {
        return [string replace $string end-[expr {$padding_length-1}] end]
    } else {
        return $string
    }
}

proc qc::encrypt_bf_tcl {key plaintext} {
    #| Encrypt plaintext using TCLLib blowfish package. Return base64 encoded ciphertext.
    #| Ciphertext can be decrypted by qc::decrypt_bf_tcl and qc::decrypt_bf_db.
    if { $plaintext eq "" } {
        return $plaintext 
    } else {
        set padded_plaintext [qc::pkcs_padding_append [encoding convertto utf-8 $plaintext]]
        return [base64::encode [blowfish::blowfish -mode cbc -dir encrypt -iv [string repeat \0 8] -key $key $padded_plaintext]]
    }
}

proc qc::encrypt_bf_db {key plaintext} {
    #| Encrypt plaintext using Postgresql's pg_crypto blowfish functions. Return base64 encoded ciphertext.
    #| Ciphertext can be decrypted by qc::decrypt_bf_db and qc::decrypt_bf_tcl.
    set key_base64 [base64::encode $key]
    set plaintext_base64 [base64::encode [encoding convertto utf-8 $plaintext]]

    # Try clause to prevent disclosure of encryption key by error Handler
    db_1row {
        select 
        encode(
               encrypt_iv(
                          decode(:plaintext_base64,'base64'),
                          decode(:key_base64,'base64'),
                          repeat(E'\\000',8)::bytea,
                          'bf-cbc/pad:pkcs'
                          )
               ,'base64'
               ) as ciphertext
    }

    return $ciphertext
}

proc qc::encrypt_bf {key plaintext} {
    return [encrypt_db_tcl $key $plaintext]
}

proc qc::decrypt_bf_tcl {key ciphertext} {
    #| Decrypt base64 encoded ciphertext using TCLLib blowfish package. Return plaintext.   
    #| This proc can decrypt ciphertext generated by qc::decrypt_bf_tcl and qc::decrypt_bf_db.
    if { $ciphertext eq "" } {
        return $ciphertext 
    } else {
        set padded_plaintext [blowfish::blowfish -mode cbc -dir decrypt -iv [string repeat \0 8] -key $key [::base64::decode $ciphertext]]
        return [encoding convertfrom utf-8 [qc::pkcs_padding_strip $padded_plaintext]]
    }
}

proc qc::decrypt_bf_db {key ciphertext} {
    #| Decrypt base64 encoded ciphertext using Postgresql's pg_crypto blowfish functions. Return plaintext.   
    #| This proc can decrypt ciphertext generated by qc::decrypt_bf_db and qc::decrypt_bf_tcl.
    set key_base64 [base64::encode $key]

    # Try clause to prevent disclosure of encryption key by error Handler
    ::try {
        db_1row {
            select 
            encode(
                   decrypt_iv(
                              decode(:ciphertext::text,'base64'),
                              decode(:key_base64::text,'base64'),
                              repeat(E'\\000',8)::bytea,
                              'bf-cbc/pad:pkcs'
                              )
                   ,'base64'
                   ) as plaintext_base64
        }
    } on error {} {
        error "Unable to decrypt \"$ciphertext\""
    }
    return [encoding convertfrom utf-8 [::base64::decode $plaintext_base64]]
}

proc qc::decrypt_bf {key ciphertext} {
    return [decrypt_db_tcl $key $ciphertext]
}


namespace eval qc {
    namespace export is_*
}

namespace eval qc::is {
    
    namespace export *

    namespace ensemble create -unknown {
        data_type_parser
    }
}

namespace eval qc {
    namespace export perm_set perm_test_user perm_test perm perms perm_if
}

proc qc::perm_set {args} {
    #| Configure user permissions
    #| Usage: perm_set user_id perm_name ?method? ?method?
    qc::args $args -schema "" -- user_id perm_name args
    set methods [string toupper $args]
    if { $schema eq "" } {
        lassign [qc::db_qualify_table "perm"] schema
    }
    db_dml {
        -- Revoke any existing permissions on perm_name
        delete from [db_quote_identifier $schema].user_perm
        where user_id=:user_id 
        and perm_id in
        (
         select
         perm_id
         from [db_quote_identifier $schema].perm
         join [db_quote_identifier $schema].perm_class using (perm_class_id)
         where perm_name=:perm_name
         );
        
        -- Grant the specified method permissions on perm_name
        insert into [db_quote_identifier $schema].user_perm (user_id, perm_id)
        select :user_id, perm_id
        from [db_quote_identifier $schema].perm 
        join [db_quote_identifier $schema].perm_class using(perm_class_id)
        where perm_name=:perm_name
        and [qc::sql_where_in method $methods false];
    }
}

proc qc::perm_test_user { user_id perm_name method } {
    #| Test whether the user can perform $method on $perm_name
    #| Returns boolean
    set method [upper $method]
    db_0or1row {
        select 
        perm_id
        from user_perm
        join perm using(perm_id)
        join perm_class using(perm_class_id)
        where 
        user_id=:user_id
        and perm_name=:perm_name
        and method=:method        
    } {
        return false
    } {
        return true
    }    
}

proc qc::perm_test { perm_name method } {
    #| Test whether the current user can perform $method on $perm_name
    #| Returns boolean
    return [qc::perm_test_user [qc::auth] $perm_name $method]
}

proc qc::perm { perm_name method } {
    #| Test whether the current user can perform $method on $perm_name
    #| Throws an error and sets a global ldict errorList on failure.
    if { ! [perm_test $perm_name $method] } {
        global errorList
        set errorList [list [dict create perm_name $perm_name method [upper $method]]]
	error "You do not have $method permission on $perm_name." {} PERM
    }
}

proc qc::perms { body } {
    #| Test each line of permissions
    #| Throws an error after all tests if any fail, and sets a global ldict errorList
    global errorList
    set errorList {}
    set error_messages {}
    set lines [split $body \n]
    foreach line $lines {
        set line [string trim $line]
        if {$line ne ""} {
            set perm_name [lindex [split $line " "] 0]
            set method [lindex [split $line " "] 1]
            if { ! [perm_test $perm_name $method] } {
                lappend errorList [dict create perm_name $perm_name method [upper $method]]
                lappend error_messages "You do not have $method permission on $perm_name."
            }
        }
    }
    if { $errorList ne {} } {
        error [html_list $error_messages] {} PERM
    }
}

proc qc::perm_if {perm_name method if_code {. else} {else_code ""} } {
    #| Evaluate if_code if current user has permission else else_code
    set code ""
    if { [perm_test $perm_name $method] } {
	set code $if_code
    } elseif { $else_code ne "" } {
	set code $else_code
    }

    if { $code ne "" } {
        set return_code [catch {uplevel 1 $code} result options]
        if { $return_code != 0 } {
            # error, return, break, continue or custom
            # Preserve TCL_RETURN
            if { $return_code == 2 && [dict get $options -code] == 0 } {
                dict set options -code return
            } else {
                # Return in parent stack frame instead of here
                dict incr options -level
            }
            return -options $options $result
        }
    }
}

proc qc::perm_category_add {category} {
    #| Helper proc to create new permission category
    db_trans {
        set perm_category_id [db_seq perm_category_id_seq]
        db_dml {
            insert into perm_category
            (perm_category_id, description)
            values
            (:perm_category_id, :category)
        }
    }
}

proc qc::perm_add {perm_name description category args} {
    #| Helper proc to create new permissions
    db_trans {
        db_dml {lock table perm_category in exclusive mode}
        db_dml {lock table perm_class in exclusive mode}
        db_dml {lock table perm in exclusive mode}

        # Perm Category
        db_1row {
            select 
            perm_category_id
            from perm_category
            where description=:category
        }

        # Perm Class
        db_0or1row {
            select 
            perm_class_id
            from perm_class 
            where perm_name=:perm_name
        } {
            set perm_class_id [db_seq perm_class_id_seq]
            db_dml {
                insert into perm_class
                (perm_class_id, perm_name, description, perm_category_id)
                values
                (:perm_class_id, :perm_name, :description, :perm_category_id)
            }
        }

        # Perm Methods
        foreach perm_method $args {
            set perm_id [db_seq perm_id_seq]
            db_dml {
                insert into perm
                (perm_id, perm_class_id, method)
                values
                (:perm_id, :perm_class_id, :perm_method::perm_method)
            }
        }

    }    
}

namespace eval qc {
    namespace export tabular_text_parse
}

proc qc::tabular_text_parse  {args} {
    #| Parse text tabular data and return TCL table data structure
    # Example:
    #     % set lines {}
    #     % lappend lines "Column 1 Column 2"
    #     % lappend lines " a        b      "
    #     % set text [join $lines \n]
    #
    #     % set conf {}
    #     % lappend conf [list label "Column 1" var_name "col1" trim false]
    #     % lappend conf [list label "Column 2" var_name "col2"]
    #
    #     % return [tabular_text_parse $text $conf]
    #     {col1 col2} {{ a      } b}
    qc::args $args -ignore_empty_rows -- text columns_conf
    
    # convert text to list of lines of required min width (determined by line for table header)
    set temp {}
    foreach line [split [string map {\r ""} $text] \n] {
        lappend temp $line
    }
    # check table header line contains all column_headings and replace whitespace in headings with underscores
    set table_header_line [lindex $temp 0]
    foreach conf $columns_conf {
        dict2vars $conf label
        # replace whitespace with underscore to prevent column headers being split into multiple columns later on
        set new_label [regsub -all {\s} $label "_"]
        if { ! [regsub "(^|\\s)[regexp_escape $label](\\s|$)" $table_header_line "\\1$new_label\\2" table_header_line] } {
            error "Unable to locate column heading \"$label\""
        }
    }
    set lines [list $table_header_line]
    # padd all other lines with spaces to have at least same number of characters as the header
    set min_width [string length $table_header_line]    
    foreach line [lrange $temp 1 end] {
        if { [string length $line] < $min_width } {
            lappend lines [qc::format_left $line $min_width]
        } else {
            lappend lines $line
        }
    }
  
    # Generate list of indices for vertical columns of whitespace characters
    set vertical_whitespaces {}
    set i 0 
    foreach line $lines {
        incr i
        
        if { $i == 1 } {
            # first line - get indices of all whitespace characters
            set vertical_whitespaces [regexp -all -inline -indices {\s} $line]                 
        } else {
            # remaining lines - perform intersect to find indices for vertical columns of whitespace characters
            set vertical_whitespaces [qc::lintersect $vertical_whitespaces [regexp -all -inline -indices {\s} $line]]
        }
    }

    # Combine consecutive vertical whitespace indices 
    set column_separators {}
    set i 0
    set group_start 0
    set group_end 0
    foreach vertical_whitespace $vertical_whitespaces {
        incr i
        lassign $vertical_whitespace start end
        
        if { $i == 1 } {
            # first vertical whitespace - start new group
            set group_start $start
            set group_end $end
        }

        if { $start - $group_end > 1 } {
            # non-consequetive vertical whitespace - add column separator and start new group
            lappend column_separators [list $group_start $group_end]
            set group_start $start
            set group_end $end
        } else {
            # consecutive vertical whitespace - extend current group
            set group_end $end
        }

        if { $i == [llength $vertical_whitespaces] } {
            # last vertical whitespace - close current group then add as column separator
            set group_end $end
            lappend column_separators [list $group_start $group_end]
        }
    }
    
    # ignore column separators before first and after last table headings
    set result [regexp -inline -indices {^\s*\S} $table_header_line]
    set first_heading_start [lindex $result 0 1]
    set result [regexp -inline -indices {\S\s*$} $table_header_line]
    set last_heading_end [lindex $result 0 0]
    foreach column_separator $column_separators {
        lassign $column_separator column_separator_start column_separator_end
        if { $column_separator_end <= $first_heading_start || $column_separator_start >= $last_heading_end } {
            # discard column separators
            set column_separators [lexclude $column_separators $column_separator]
        }
    }
    # ignore multiple column separators between table headings (just keep the last one)
    set result [regexp -all -inline -indices {\s+} $table_header_line]
    foreach header_whitespace $result {
        lassign $header_whitespace header_whitespace_start header_whitespace_end
        set i 0 
        foreach column_separator [lreverse $column_separators] {
            lassign $column_separator column_separator_start column_separator_end
            if { $column_separator_start >= $header_whitespace_start && $column_separator_end <= $header_whitespace_end && $i == 0 } {
                # keep last column separator
                incr i
            } elseif { $column_separator_start >= $header_whitespace_start && $column_separator_end <= $header_whitespace_end } {
                # discard other column separators
                set column_separators [lexclude $column_separators $column_separator]
            }
        }
    }
    
    # generate TCL table data structure
    set table {}
    set i 0
    foreach line $lines {
        incr i
        if { [info exists ignore_empty_rows] && [trim $line] eq "" } {
            # ignore blank rows
            continue
        }
        
        # split lines on column separators and optionally trim whitespace from values
        foreach separator [lreverse $column_separators] {
            lassign $separator start end
            set line [string replace $line $start $end \0]
        }
        if { $i == 1 } {
            # table heading line - replace headings with desired variable names 
            foreach conf $columns_conf {
                dict2vars $conf label var_name
                set new_label [regsub -all {\s} $label "_"]
                if { ! [regsub "(^|\\0|\\s)[regexp_escape $new_label](\\0|\\s|$)" $line "\\1${var_name}\\2" line] } {
                    error "Unable to locate column heading \"$label\""
                }
            }
        }
        # raise error if we have too many/little columns
        if { [llength [split $line \0]] != [llength $columns_conf] } {
            set errored_line [string map [dict create "\0" "\\0"] $line]
            error "Found [llength [split $line \0]] columns, expected [llength $columns_conf] for line \"$errored_line\""
        }        
        set row {}        
        foreach conf $columns_conf value [split $line \0] {
            dict2vars $conf trim type
            default trim true

            if { $i == 1 || $trim } {
                set value [trim $value]
            }
            
            if { $i != 1 && [info exists type] && [qc::castable $type $value] } {
                set value [qc::cast $type $value]
            }
            
            lappend row $value
        }
        
        lappend table $row
    }
    return $table
}

namespace eval qc {
    namespace export html_options_*
}

proc qc::html_options_db { qry } {
    #| Expects a qry use columns named "name" and "value"
    #| Use aliases where required.
    #| E.g select foo_id as value,description as name from foo
    set options {}
    db_cache_foreach $qry {
	lappend options $name $value
    }
    return $options
}

proc qc::html_options_db_cache { qry {ttl 86400}} {
    #| Expects a qry use columns named "name" and "value"
    #| Use aliases where required.
    #| E.g select foo_id as value,description as name from foo
    #| Query results are cached 
    set options {}
    db_cache_foreach -ttl $ttl $qry {
	lappend options $name $value
    }
    return $options
}

proc qc::html_options_simple { args } {
    #| Use list items as both name and value
    #| Eg Converts one two three -> one one two two three three
    set options {}
    foreach item $args {
        lappend options $item $item
    }
    return $options
}


##################################################
#
# qc::is, qc::cast, qc::castable ensemble handler
#
###################################################
proc data_type_parser {args} {
    #| Try to map args to data type
    set namespace [lindex $args 0]
    set data_type [lindex $args 1]
    
    switch -regexp -matchvar numbers -- $data_type {
        {^varchar\(([0-9]+)\)$} {
            return [list $namespace varchar [lindex $numbers 1]]
        }
        {^char\(([0-9]+)\)$} {
            return [list $namespace char [lindex $numbers 1]]
        }
        {^(decimal|numeric)\(([0-9]+,[0-9]+)\)$} {
            set values [split [lindex $numbers 2] ,]
            return [list $namespace decimal -precision [lindex $values 0] -scale [lindex $values 1]]
        }
        {^(decimal|numeric)\(([0-9]+)\)$} {
            return [list $namespace decimal -precision [lindex $numbers 2]]
        }
        ^numeric$ {
            return [list $namespace decimal]
        }
        ^bool$ {
            return [list $namespace boolean]
        }
        {^(int|int4)$} {
            return [list $namespace integer]
        }
        ^int2$ {
            return [list $namespace smallint]
        }
        ^int8$ {
            return [list $namespace bigint]
        }
        ^float8$ {
            return [list $namespace decimal]
        }
        default {
            if { [regexp {\.([^\.]+)$} $data_type -> qc_data_type] && [info commands "${namespace}::${qc_data_type}"] ne "" } {
                # qcode-tcl data type prefixed by db schema - cast/validate using qcode-tcl 
                return [list $namespace $qc_data_type]
                
            } elseif { [qc::memoize qc::db_domain_exists $data_type] } {
                # db domain
                return [list $namespace domain $data_type]
                
            } elseif { [qc::memoize qc::db_enum_exists $data_type] } {
                # db enumerated type
                return [list $namespace enumeration $data_type]
                
            }
        }
    }
}

##################################################
#
# qc::response ensemble handler
#
##################################################
proc response_subcommand_map {ensemble subcommand args} {
    #| Map unknown subcommands to response extend
    return [list $ensemble extend $subcommand]
}

namespace eval qc {
    namespace export {*}{
        file_is_valid_image
        image_file_info
        image_resize
        image_file_autocrop
        image_file_resize
        image_file_convert
    }
}

proc qc::image_file_meta_strip {file} {
    #| Strip metadata from a file
    exec_proxy \
        -timeout 20000 \
        -ignorestderr \
        convert \
        $file \
        -quiet \
        -strip \
        -density 0 \
        $file
    
    return $file
}

proc qc::image_file_autocrop {old_file {crop_colour white}} {
    #| Autocrop image by trimming white border from jpgs,
    #| and transparent and white borders from pngs.
    #| Create a new file with the results and return the file path
    set file /tmp/[qc::uuid]
    set exec_proxy_flags {
        -timeout 20000
    }
    set exec_flags {
        -ignorestderr
    }
    set convert_options {
        -quiet
    }
    if { $crop_colour ne "" } {
        lappend convert_options \
            -bordercolor $crop_colour \
            -border 1x1
    }
    lappend convert_options {*}{
        -trim
        -format %wx%h%O info:
    }
    set crop_info \
        [exec_proxy \
             {*}$exec_proxy_flags \
             {*}$exec_flags \
             convert \
             $old_file \
             {*}$convert_options
             ]

    exec_proxy \
        {*}$exec_proxy_flags \
        {*}$exec_flags \
        convert \
        $old_file \
        -quiet \
        -crop $crop_info \
        +repage \
        $file

    return $file
}

proc qc::image_file_resize {old_file max_width max_height} {
    #| Resize an image to fit within max_width/height constraint.
    #| Create a new file with the results and return the file path
    set file /tmp/[qc::uuid]
    set exec_proxy_flags {
        -timeout 30000
    }
    set exec_flags {
        -ignorestderr
    }
    set convert_flags [subst {
        -quiet
        -thumbnail ${max_width}x${max_height}
        -strip
        -quality 75%
    }]
    exec_proxy \
        {*}$exec_proxy_flags \
        {*}$exec_flags \
        convert \
        {*}$convert_flags \
        $old_file \
        $file

    return $file
}

proc qc::image_resize {args} {
    #| Export an image from the database, constrained to max width/height,
    #| optionally auto-cropped,
    #| return dict of file, width, and height
    # Deprecated
    qc::args $args -autocrop -- file_id max_width max_height
    default autocrop false

    set file [qc::db_file_export $file_id]
    
    if { $autocrop } {
        set tmp_file $file
        ::try {
            set file [qc::image_file_autocrop $tmp_file]
        } finally {
            file delete $tmp_file
        }
    }

    dict2vars [qc::image_file_info $file] width height

    # Check whether image needs resizing
    if { $width > $max_width
         || $height > $max_height
     } {
        set tmp_file $file
        ::try {
            set file [qc::image_file_resize $tmp_file $max_width $max_height]
        } finally {
            file delete $tmp_file
        }
        dict2vars [qc::image_file_info $file] width height
    }

    return [qc::dict_from file width height]
}

proc qc::file_is_valid_image {file} {
    #| file (on local file system) is of an image type that we can stat.
    package require jpeg
    package require png

    return [expr {
                  [jpeg::isJPEG $file]
                  || [png::isPNG $file]
                  || [qc::is_gif $file]
                  || [qc::is_webp $file]
                  || [qc::is_svg $file]}]
}

proc qc::image_file_info {file} {
    #| dict of width, height, and mime_type of file (from local filesystem)
    package require jpeg
    package require png

    if { [jpeg::isJPEG $file] } {
        lassign [jpeg::dimensions $file] width height
        set mime_type "image/jpeg"
    } elseif { [png::isPNG $file] } {
        qc::dict2vars [png::imageInfo $file] width height
        set mime_type "image/png"
    } elseif { [qc::is_gif $file] } {
        lassign [qc::gif_dimensions $file] width height
        set mime_type "image/gif"
    } elseif { [qc::is_webp $file] } {
        lassign [qc::webpsize $file] width height
        set mime_type "image/webp"
    } elseif { [qc::is_svg $file] } {
        lassign [qc::svg_dimensions $file] width height
        set mime_type "image/svg+xml"
    } else {
        error "Unrecognised image type"
    }
    return [qc::dict_from width height mime_type]
}

proc qc::is_gif {file} {
    set f [open $file r]
    fconfigure $f -translation binary
    # read GIF signature -- check that this is
    # either GIF87a or GIF89a
    set sig [read $f 6]
    switch $sig {
        "GIF87a" -
        "GIF89a" {
            close $f
            return true
        }
        default {
            close $f
            return false
        }
    }
 }

proc qc::gif_dimensions {name} {
    set f [open $name r]
    fconfigure $f -translation binary
    # read GIF signature -- check that this is
    # either GIF87a or GIF89a
    set sig [read $f 6]
    switch $sig {
        "GIF87a" -
        "GIF89a" {
            # do nothing
        }
        default {
            close $f
            error "$f is not a GIF file"
        }
    }
    # read "logical screen size", this is USUALLY the image size too.
    # interpreting the rest of the GIF specification is left as an exercise
    binary scan [read $f 2] s wid
    binary scan [read $f 2] s hgt
    close $f

    return [list $wid $hgt]
}

proc qc::image_handler {
    cache_dir
    {error_handler "qc::error_handler"}
    {image_redirect_handler "qc::image_redirect_handler"}
    {allowed_max_width 2560}
    {allowed_max_height 2560}
} {
    #| URL handler to serve images that can not be served by fastpath.
    # Create image cache for canonical URL if it doesn't already exist.
    # If canonical URL was requested return file to client,
    # and register URL to be servered by fastpath for future requests.
    # Otherwise default redirect handler will redirect client,
    # to correct image dimensions or the canonical URL.
    # By default enforce a max width and height of 2560x2560
    # Uses default qc::error_handler
    setif error_handler "" "qc::error_handler"
    ::try {
        set request_path [qc::conn_path]
        setif image_redirect_handler "" "qc::image_redirect_handler"
        
        if { [regexp {^/image/([0-9]+)-([0-9]+)x([0-9]+)(?:/(.*)|$)} \
                  $request_path -> file_id max_width max_height filename]
         } {
            set autocrop false
        } elseif { [regexp {^/image/([0-9]+)/autocrop/([0-9]+)x([0-9]+)(?:/(.*)|$)} \
                        $request_path -> file_id max_width max_height filename]
               } {
            set autocrop true
        } elseif { [regexp {^/image/([0-9]+)\.svg$} \
                        $request_path -> file_id] } {
            # SVG images
            dict2vars [qc::image_data \
                           -mime_type "image/svg+xml" \
                           $cache_dir \
                           $file_id \
                           $allowed_max_width \
                           $allowed_max_height] url
            set file [ns_pagepath]$url
            ns_register_fastpath GET $url
            ns_register_fastpath HEAD $url
            return [ns_returnfile 200 "image/svg+xml" $file]
        } else {
            # Invalid image url
            return [ns_returnnotfound]
        }
        
        # Check requested max width and height does not exceed allowed maximums
        if { ($allowed_max_width ne "" && $max_width > $allowed_max_width)
             || ($allowed_max_height ne "" && $max_height > $allowed_max_height)
         } {
            return [ns_returnbadrequest "The requested image (${max_width}x${max_height}) exceeds the maximum width and height permitted (${allowed_max_width}x${allowed_max_height})."]
        }

        set mime_type [qc::mime_type_guess $filename]

        set cache_args [list]
        if { $autocrop } {
            lappend cache_args -autocrop
        }
        lappend cache_args \
            -mime_type $mime_type \
            -- $cache_dir $file_id $max_width $max_height
        
        dict2vars [qc::image_data {*}$cache_args] width height url
        set canonical_url $url
        set canonical_file [ns_pagepath]$url

        if { $request_path eq $canonical_url } {
            # Canonical URL was requested - return file
            ns_register_fastpath GET $canonical_url
            ns_register_fastpath HEAD $canonical_url
            return [ns_returnfile 200 [mime_type_guess $canonical_file] $canonical_file]
        } 

        # Redirect handler
        return [$image_redirect_handler $cache_dir]
    } on error {error_message options} {
        # Error handler
        return [$error_handler $error_message $options]
    }
}

proc qc::image_redirect_handler {cache_dir}  {
    #| Redirect client to correct image dimesions or the canonical URL.
    set request_path [qc::conn_path]

    if { [regexp {^/image/([0-9]+)-([0-9]+)x([0-9]+)(?:/(.*)|$)} \
              $request_path -> file_id max_width max_height filename]
     } {
        set autocrop false
    } else {
        regexp {^/image/([0-9]+)/autocrop/([0-9]+)x([0-9]+)(?:/(.*)|$)} \
            $request_path -> file_id max_width max_height filename
        set autocrop true
    }

    set mime_type [qc::mime_type_guess $filename]

    if { $autocrop } {
        dict2vars \
            [qc::image_cache_data \
                 -autocrop \
                 -mime_type $mime_type \
                 -- $cache_dir $file_id $max_width $max_height] \
            width height url
    } else {
        dict2vars \
            [qc::image_cache_data \
                 -mime_type $mime_type \
                 -- $cache_dir $file_id $max_width $max_height] \
            width height url
    }
    set canonical_url $url
    set canonical_file [ns_pagepath]$canonical_url
    
    # Check requested image dimensions
    if { $width != $max_width || $height != $max_height } {            
        # Redirect to url with correct image dimensions 
        return [ns_returnredirect \
                    "[conn_location][file dirname $canonical_url]/$filename"]
    }
    
    # Catch All - redirect to Canonical URL
    return [ns_returnredirect $canonical_url]      
}

proc qc::image_file_convert {old_file mime_type max_width max_height autocrop} {
    #| Convert an image, return file path of new image
    set ext [qc::mime_file_extension $mime_type]
    set file /tmp/[qc::uuid]${ext}

    if { $autocrop } {
        set old_file [qc::image_file_autocrop $old_file]
    }
    
    set exec_proxy_flags {
        -timeout 30000
    }
    set exec_flags {
        -ignorestderr
    }
    set convert_flags {
        -quiet
        -strip
    }
    if { $mime_type eq "image/webp" } {
        lappend convert_flags \
            -quality 100%            
    } else {
        lappend convert_flags \
            -quality 75%
    }
    lappend convert_flags \
        -thumbnail ${max_width}x${max_height}
    
    exec_proxy \
        {*}$exec_proxy_flags \
        {*}$exec_flags \
        convert \
        $old_file \
        {*}$convert_flags \
        $file
    
    if { $autocrop } {
        file delete $old_file
    }

    return $file
}

proc qc::is_svg {file} {
    #| Test if file is an SVG file
    set f [open $file r]
    set contents [read $f]
    close $f

    if { [regexp {^(<\?xml\s+[^>]*\?>).*} $contents -> declaration]
         && [qc::xml_declaration_valid $declaration]
         && [regexp {<svg\s+} $contents] } {
        return true
    } elseif { [regexp {^\s*<svg\s+} $contents] } {
        return true
    } else {
        return false
    }
}

proc qc::svg_dimensions {file} {
    #| Get the preferred dimensions of an SVG file
    package require tdom
    set f [open $file r]
    set svg [read $f]
    dom parse $svg doc
    set root [$doc documentElement]

    # Check for width and/or height attributes
    if { [$root hasAttribute "width"] } {
        set width [$root getAttribute "width"]
        set width [qc::svg_units2pixels $width]
    }
    
    if { [$root hasAttribute "height"] } {
        set height [$root getAttribute "height"]
        set height [qc::svg_units2pixels $height]
    }
    
    if { [info exists width] && [info exists height] } {
        return [list $width $height]
    }

    # Check for viewbox attribute
    if { [$root hasAttribute "viewBox"] } {
        set viewBox [$root getAttribute "viewBox"]
        lassign [split $viewBox] min_x min_y box_width box_height
        if { [info exists width] } {
            set ratio [expr {1.0 * $box_height / $box_width}]
            set height [qc::round [expr {$width * $ratio}] 0]
            return [list $width $height]
        }
        if { [info exists height] } {
            set ratio [expr {1.0 * $box_width / $box_height}]
            set width [qc::round [expr {$height * $ratio}] 0]
            return [list $width $height]
        }
        return [list \
                    [qc::round $box_width 0] \
                    [qc::round $box_height 0]]
    }

    # Most browsers now use 300x150 as the default
    return [list 300 150]
}

proc qc::svg_units2pixels {length} {
    #| Convert supported svg distance units to pixels
    if { ! [regexp \
                {^([0-9]+(?:\.[0-9]+))([a-zA-Z]*)$} \
                $length \
                -- number units] } {
        error "Unable to parse length \"$length\""
    }
    switch [string tolower $units] {
        "" -
        "px" {
            set pixels $number
        }
        "in" {
            set pixels [expr {$number * 96}]
        }
        "cm" {
            set pixels [expr {$number * 37.795}]
        }
        "mm" {
            set pixels [expr {$number * 3.7795}]
        }
        "pt" {
            set pixels [expr {$number * 1.3333}]
        }
        "pc" {
            set pixels [expr {$number * 16}]
        }
        default {
            error "Unrecognized units \"$units\""
        }
    }
    return [qc::round $pixels 0]
}

namespace eval qc {
    namespace export html_table_doc
}

proc qc::html_table_doc {args} {
    #| Create a 2 row table with labels on the first row and values on the 2nd row.
    # Requires a cols object
    set varNames [qc::args2vars $args]
    default tbody {}
    # data is pulled from caller's namespace
    set row {}
    foreach col $cols {
	qc::upcopy 1 [dict get $col name] value
	default value ""
	lappend row $value
    }
    lappend tbody $row
    lappend varNames cols tbody
    return [qc::html_table [dict_from {*}$varNames]]
}


proc qc::is_pnz_decimal { price } {
    #| Deprecated
    # positive non-zero double
    if { [string is double -strict $price] && $price>0 } {
	return 1
    } else {
	return 0
    }
}

proc qc::is_ipv4 {string} {
    #| Deprecated - see qc::is ipv4
    # TODO checks structure only, will allow 9999.9999.9999.9999
    return [qc::is ipv4 $string]
}

proc qc::is_integer {int} {
    #| Deprecated - see qc::is integer
    return [qc::is integer $int]
}

proc qc::is_timestamp_http { date } {
    #| Deprecated - see qc::is timestamp_http
    #| Returns true if date is an acceptable HTTP timestamp.
    #| Note although all three should be accepted, only RFC 1123 format should be
    #| generated.
    # RFC 1123 - Sun, 06 Nov 1994 08:49:37 GMT
    return [qc::is timestamp_http $date]
}

proc qc::is_timestamp { date } {
    #| Deprecated - see qc::is timestamp
    # timestamps are expected to be in iso format 
    return [qc::is timestamp $date]
}

proc qc::is_timestamp_castable {string} {
    #| Deprecated - see qc::castable timestamp
    #| Can string be cast into timestamp format?
    return [qc::castable timestamp $string]
}

proc qc::is_base64 {string} {
    #| Deprecated - see qc::is base64
    #| Checks input has only allowable base64 characters and is of the correct format
    return [qc::is base64 $string]
}

proc qc::is_url {args} {
    #| Deprecated - see qc::is url
    #| This is a more restrictive subset of all legal uri's defined by RFC 3986
    #| Relax as needed
    return [qc::is url {*}$args]
}

proc qc::is_mobile_number {string} {
    #| Deprecated - see qc::is mobile_number
    # uk mobile telephone number
    return [qc::is mobile_number $string]
}

proc qc::is_cidrnetv4 {string} {
    #| Deprecated - see qc::is cidrnetv4
    return [qc::is cidrnetv4 $string]
}

proc qc::is_uri_valid {uri} {
    #| Deprecated - see qc::is uri
    #| Test if the given uri is valid according to
    #| rfc3986 (https://tools.ietf.org/html/rfc3986)
    return [qc::is uri $uri]
}

proc qc::is_decimal { number } {
    #| Deprecated - see qc::is decimal
    return [qc::is decimal $number]
}

proc qc::is_decimal_castable {string} {
    #| Deprecated - see qc::castable decimal
    return [qc::castable decimal $string]
}

namespace eval qc::aws {
    namespace export s3
    namespace ensemble create
}

namespace eval qc::aws::s3 {
    namespace export sigv4_headers
    namespace ensemble create

    proc sigv4_headers {
        credentials
        region
        request
    } {
        #| SigV4 headers for AWS requests.
        #| See: https://docs.aws.amazon.com/AmazonS3/latest/API/sigv4-auth-using-authorization-header.html

        package require sha256

        qc::dict2vars $credentials {*}{
            access_key_id
            secret_access_key
        }

        qc::dict2vars $request {*}{
            headers
            http_verb
            url
            payload
        }

        set payload_hash [::sha2::sha256 -hex $payload]
        lappend headers "x-amz-content-sha256" $payload_hash

        # Date formats for the signature.
        if { [dict exists $headers "X-Amz-Date"] } {
            set date_string [dict get $headers "X-Amz-Date"]
        } elseif { [dict exists $headers "Date"] } {
            set date_string [dict get $headers "Date"]
        } else {
            error "A \"Date\" or \"X-Amz-Date\" header is required in the request."
        }

        set date_epoch [clock scan $date_string]
        set date [clock format $date_epoch \
            -format "%Y%m%d" \
            -timezone "UTC"]
        set datetime_iso8601 [clock format $date_epoch \
            -format "%Y%m%dT%H%M%SZ"\
            -timezone "UTC"]

        set credential_scope "${date}/${region}/s3/aws4_request"
        set signed_headers [_header_names_signed [dict keys $headers]]

        # Canonical request.
        set url_parts [qc::url_parts $url]
        set path [dict get $url_parts path]
        set params [dict get $url_parts params]
        set lcanonical_request [list $http_verb]
        lappend lcanonical_request [_uri_path_canonicalize $path]
        lappend lcanonical_request [_query_string_canonicalize $params]
        lappend lcanonical_request [_headers_canonicalize $headers]
        lappend lcanonical_request $signed_headers
        lappend lcanonical_request $payload_hash
        set canonical_request [join $lcanonical_request "\n"]

        # String to be signed.
        set parts [list \
            "AWS4-HMAC-SHA256" \
            $datetime_iso8601 \
            $credential_scope \
            [::sha2::sha256 -hex $canonical_request]]
        set string_to_sign [join $parts "\n"]

        # Signature.
        set date_key [::sha2::hmac -bin -key "AWS4$secret_access_key" $date]
        set date_region_key [::sha2::hmac -bin -key $date_key $region]
        set date_region_service_key [::sha2::hmac -bin -key $date_region_key s3]
        set signing_key [::sha2::hmac -bin -key $date_region_service_key "aws4_request"]
        set signature [::sha2::hmac -bin -key $signing_key $string_to_sign]
        set signature [binary encode hex $signature]

        # Authorization header.
        set auth_header "AWS4-HMAC-SHA256 "
        append auth_header "Credential=${access_key_id}/${credential_scope},"
        append auth_header "SignedHeaders=${signed_headers},"
        append auth_header "Signature=${signature}"

        set sigv4_headers [dict create \
            "Authorization" $auth_header \
            "x-amz-content-sha256" $payload_hash]

        return $sigv4_headers
    }

    proc _uri_path_canonicalize {path} {
        #| Canonicalize the URI path for an AWS signed request.

        set parts [split [string trim $path "/"] "/"]
        set decoded [qc::lapply qc::url_decode $parts]
        set encoded [qc::lapply qc::aws::s3::_uri_encode $decoded]

        return "/[join $encoded "/"]"
    }

    proc _query_string_canonicalize {params} {
        #| Canonicalize the query params for an AWS signed request.

        set encoded_pairs [list]

        foreach {name value} $params {
            lappend encoded_pairs \
                [qc::aws::s3::_uri_encode [qc::url_decode $name]] \
                [qc::aws::s3::_uri_encode [qc::url_decode $value]]
        }

        set sorted_pairs [lsort -stride 2 $encoded_pairs]
        set queries [list]

        foreach {name value} $sorted_pairs {
            lappend queries "${name}=${value}"
        }

        return [join $queries "&"]
    }

    proc _headers_canonicalize {headers} {
        #| Canonicalize the headers for an AWS signed request.

        set dict [dict create]

        foreach {name value} $headers {
            set lower_name [string tolower $name]
            regsub -all {\s+} [string trim $value] " " new_value

            if { [dict exists $dict $lower_name] } {
                set existing_value [dict get $dict $lower_name]
                dict set dict $lower_name "${existing_value},${new_value}"
            } else {
                dict set dict $lower_name $new_value
            }
        }

        set sorted [lsort -stride 2 $dict]
        set canonical_headers [list]

        foreach {name value} $sorted {
            lappend canonical_headers "${name}:${value}"
        }

        return "[join $canonical_headers "\n"]\n"
    }

    proc _header_names_signed {header_names} {
        #| Returns signed header names for an AWS signed request.

        set lower_names [lmap name [qc::lunique $header_names] \
            {string tolower $name}]
        return [join [lsort $lower_names] ";"]
    }

    proc _uri_encode {string} {
        #| URI encode a string according to AWS signed request requirements.

        variable ::qc::aws::s3::_uri_encode_map

        if { ! [info exists ::qc::aws::s3::_uri_encode_map] } { 
            qc::aws::s3::_uri_encode_map_init
        }

        return [string map \
            $::qc::aws::s3::_uri_encode_map \
            [encoding convertto "utf-8" $string]]
    }

    proc _uri_encode_map_init {} {
        #| Initialize the URI encode map for AWS URIs.

        variable ::qc::aws::s3::_uri_encode_map [list]

        for {set i 0} {$i < 256} {incr i} {
            set char [format %c $i]
            set hex %[format %02X $i]

            if { ! [string match {[-a-zA-Z0-9.~_]} $char] } {
                lappend ::qc::aws::s3::_uri_encode_map $char $hex
            }
        }
    }
}
namespace eval qc::aws {
    namespace export s3
    namespace ensemble create
}

namespace eval qc::aws::s3 {
    namespace export rest_api
    namespace ensemble create
}

namespace eval qc::aws::s3::rest_api {
    namespace export http_get
    namespace ensemble create

    proc http_get { args } {
        #| Makes http GET request (including auth headers) to 
        #| S3 API endpoint and returns result.
        qc::args $args \
            -timeout 60 \
            -region "" \
            -- \
            s3_uri \
            query_params

        if { $region eq "" } {
            if { [info exists ::env(AWS_DEFAULT_REGION)] } {
                set region $::env(AWS_DEFAULT_REGION)
            } else {
                error "No region specified."
            }
        }

        set headers [_http_headers \
                        GET \
                        $region \
                        $s3_uri \
                        $query_params \
                        "" \
                        ]
        set url [_endpoint $s3_uri $query_params]
        set result [qc::http_get \
                        -valid_response_codes {200} \
                        -timeout $timeout \
                        -headers $headers \
                        $url \
                    ]
        return $result
    }

    proc _credentials_check { } {
        #| Check for required AWS credentials in env vars
        if {
            (![info exists ::env(AWS_ACCESS_KEY_ID)] 
        ||   ![info exists ::env(AWS_SECRET_ACCESS_KEY)]) 
        } {
            error "No AWS credentials set."
        }
    }

    proc _http_headers { args} {
        #| Returns HTTP headers required for REST API auth.
        qc::args $args \
            -- \
            http_verb \
            region \
            s3_uri \
            query_params \
            payload

         _credentials_check

        set headers [dict create \
            Host [_endpoint_domain $s3_uri] \
            X-Amz-Date [clock format [clock scan now] \
                -format "%Y%m%dT%H%M%SZ" \
                -timezone "UTC"]]

        if { [info exists ::env(AWS_SESSION_TOKEN)]
             && $::env(AWS_SESSION_TOKEN) ne "" } {
            # We're using temporary security credentials - add token header
            dict set headers "x-amz-security-token" $::env(AWS_SESSION_TOKEN)
        }

        set credentials [dict create \
            access_key_id $::env(AWS_ACCESS_KEY_ID) \
            secret_access_key $::env(AWS_SECRET_ACCESS_KEY)]
        set request [dict create \
            http_verb $http_verb \
            url [_endpoint $s3_uri $query_params] \
            headers $headers \
            payload $payload]

        set sigv4_headers [qc::aws s3 sigv4_headers \
            $credentials \
            $region \
            $request]

        return [dict merge $headers $sigv4_headers]
    }

    proc _endpoint { s3_uri query_params } {
        #| Returns REST API endpoint.
        #| https://mybucket.s3.eu-west-1.amazonaws.com/object_key/?prefix=folder1
        dict2vars [qc::aws s3 s3_uri_parse $s3_uri] . object_key

        set url [list "https://[_endpoint_domain $s3_uri]"]

        # If object_key contains "/" we don't want that url encoded
        set segments [split $object_key "/"]
        set index 0
        set segment_values [list]
        foreach segment $segments {
            lappend segment_values segment_${index} $segment
            set segment_${index} $segment
            lappend url ":segment_${index}"
            incr index
        }

        return [qc::url \
            [join $url "/"] \
            {*}$segment_values \
            {*}$query_params \
            ]
    }

     proc _endpoint_domain { s3_uri } {
        #| Returns full REST API domain.
        #| mybucket.s3.eu-west-1.amazonaws.com
        dict2vars [qc::aws s3 s3_uri_parse $s3_uri] bucket

        set temp [list]
        if { $bucket ne "" } {
            lappend temp $bucket
        }
        # Use regional endpoint if default region is set otherwise default to
        # apex domain.
        if { [info exists ::env(AWS_DEFAULT_REGION)] } {
            lappend temp "s3.${::env(AWS_DEFAULT_REGION)}.amazonaws.com"
        } else {
            lappend temp "s3.amazonaws.com"
        }
        set domain [join $temp "."]
        return $domain
    }
}


proc qc::is_positive_decimal { number } {
    #| Deprecated
    if { [string is double -strict $number] && $number>=0 } {
	return 1
    } else {
	return 0
    }
}

proc qc::is_non_zero {number} {
    #| Deprecated
    if { [is_decimal $number] && $number!=0 } {
	return 1
    } else {
	return 0
    }
}

proc qc::is_varchar {string length} {
    #| Deprecated - see qc::is varchar
    #| Checks string would fit in a varchar of length $length
    return [qc::is varchar $length $string]
}

proc qc::is_pos {number} {
    #| Deprecated
    if { [is_decimal $number] && $number>=0 } {
	return 1
    } else {
	return 0
    }
}

proc qc::is::timestamp { date } {
    #| Checks if the given date is a timestamp (in iso format).
    return [regexp {^\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}$} $date]   
}

proc qc::is::timestamptz {date} {
    #| Checks if the given date is a timestamp with a time zone (in iso format).
    return [regexp {^\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}(\+|-)([0-9][0-9])$} $date] 
}

proc qc::is::timestamp_http {date} {
    #| Checks if the given date is an acceptable HTTP timestamp.
    #| Note although all three should be accepted, only RFC 1123 format should
    #| be generated.

    set short_day {[(Mon)|(Tue)|(Wed)|(Thu)|(Fri)|(Sat)|(Sun)]}
    set long_day {[(Monday)|(Tuesday)|(Wednesday)|(Thursday)|(Friday)|}
        append long_day {(Saturday)|(Sunday)]}
    set short_month {(Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)}
    set time {[0-2]\d(\:)[0-5]\d(\:)[0-5]\d}

    # RFC 1123 - Sun, 06 Nov 1994 08:49:37 GMT
    set pattern {(?x)
        $short_day[,]
        \s\d{2}\s$short_month\s\d{4}
        \s$time\s(GMT)
    }
    set pattern [subst -nobackslashes -nocommands $pattern]

    if { [regexp $pattern $date] } {
        return 1
    }

    # RFC 850 - Sunday, 06-Nov-94 08:49:37 GMT
    set pattern {(?x)
        $long_day[,]
        \s\d{2}-$short_month-\d{2}
        \s$time\s(GMT)
    }
    set pattern [subst -nobackslashes -nocommands $pattern]

    if { [regexp $pattern $date] } {
        return 1
    }

    # ANCI C - Sun Nov  6 08:49:37 1994
    set pattern {(?x)
        $short_day
        \s$short_month\s(\s|\d)\d
        \s$time\s\d{4}
    }
    set pattern [subst -nobackslashes -nocommands $pattern]

    if { [regexp $pattern $date] } {
        return 1
    }

    return 0
}

proc qc::is::ipv4 {string} {
    #| Checks if the given string follows the IPv4 format.
    # TODO checks structure only, will allow 9999.9999.9999.9999
    if { [regexp {^[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+$} $string] } {
        return 1
    } else {
        return 0
    }
}

proc qc::is::integer {int} {
    #| Checks if the given number is a 32-bit signed integer.
    if {[string is integer -strict $int]
        && $int >= -2147483648 && $int <= 2147483647} {
        return 1
    } else {
        return 0
    }
}

proc qc::is::text {string} {
    #| Check if the given string is text
    return 1
}

proc qc::is::safe_markdown {markdown} {
    #| Checks if the given markdown text contains HTML elements that are deemed safe.
    try {
        qc::commonmark2html $markdown
        return 1
    } on error [list error_message options] {
        return 0
    }
}

proc qc::is::cidrnetv4 {string} {
    #| Checks if the given string follows the CIDR NETv4 format.
    if { [regexp {^[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+/[0-9]+$} $string] } {
        return 1
    } else {
        return 0
    }
}

proc qc::is::base64 {string} {
    #| Checks if the given string has only allowable base64 characters and is of the
    #| correct format.
    if { [regexp {^[A-Za-z0-9/+\r\n]+=*$} $string] \
             && ([string length $string]-[regexp -all -- \r?\n $string])*6%8==0 } {
        return 1
    } else {
        return 0
    }
}

proc qc::is::uri {uri} {
    #| Test if the given uri is valid according to
    #| rfc3986 (https://tools.ietf.org/html/rfc3986)
    
    set unreserved {
        (?:[a-zA-Z0-9\-._~])
    }
    
    set sub_delims {
        (?:[!$&'()*+,;=])
    }
    
    set pct_encoded {
        (?:%[0-9a-fA-F]{2})
    }

    set pchar [subst -nocommands -nobackslashes {
        (?:${unreserved}|${pct_encoded}|${sub_delims}|[:@])
    }]
    
    set segment [subst -nocommands -nobackslashes {
        (?:${pchar}*)
    }]

    set segment_nz [subst -nocommands -nobackslashes {
        (?:${pchar}+)
    }]
    
    set segment_nz_nc [subst -nocommands -nobackslashes {
        (?:(?:${unreserved}|${pct_encoded}|${sub_delims}|[@])+)
    }]    

    set scheme {
        (?:[a-zA-Z][a-zA-Z+\-.]*)
    }

    set dec_octet {
        (?:[0-9]|[1-9][0-9]|1[0-9][0-9]|2[0-4][0-9]|25[0-5])
    }
    
    set ipv4_address [subst -nocommands -nobackslashes {
        (?:${dec_octet}\.${dec_octet}\.${dec_octet}\.${dec_octet})
    }]

    set h16 {
        (?:[0-9a-fA-F]{1,4})
    }

    set ls32 [subst -nocommands -nobackslashes {
        (?:
         ${h16}:${h16}
         |
         ${ipv4_address}
         )
    }]

    set ipv6_address [subst -nocommands -nobackslashes {
        (?:
         (?:${h16}:){6}${ls32}
         |
         ::(?:${h16}:){5}${ls32}
         |
         (?:${h16})?::(?:${h16}:){4}${ls32}
         |
         (?:(?:${h16}:)?${h16})?::(?:${h16}:){3}${ls32}
         |
         (?:(?:${h16}:){0,2}${h16})?::(?:${h16}:){2}${ls32}
         |
         (?:(?:${h16}:){0,3}${h16})?::${h16}:${ls32}
         |
         (?:(?:${h16}:){0,4}${h16})?::${ls32}
         |
         (?:(?:${h16}:){0,5}${h16})?::${h16}
         |
         (?:(?:${h16}:){0,6}${h16})?::
         )
    }]

    set ipvfuture [subst -nocommands -nobackslashes {
        (?:v[0-9a-fA-F]+\.(?:${unreserved}|${sub_delims}|:)+)
    }]
    
    set ip_literal [subst -nocommands -nobackslashes {
        (?:\[(?:${ipv6_address}|${ipvfuture})\])
    }]
    
    set host [subst -nocommands -nobackslashes {
        (?:${ip_literal}
         |${ipv4_address}
         |(?:${unreserved}
           |${pct_encoded}
           |${sub_delims}
           )*
         )
    }]

    set user_info [subst -nocommands -nobackslashes {
        (?:(?:${unreserved}|${pct_encoded}|${sub_delims}|:)*)
    }]

    set port [subst -nocommands -nobackslashes {
        (?:[0-9]*)
    }]

    set authority [subst -nocommands -nobackslashes {
        (?:${user_info}@)?${host}(?::${port})?
    }]           

    set path_abempty [subst -nocommands -nobackslashes {
        (?:(?:/${segment})*)
    }]
    
    set path_absolute [subst -nocommands -nobackslashes {
        (?:/(?:${segment_nz}(?:/${segment})*)?)
    }]
    
    set path_noscheme [subst -nocommands -nobackslashes {
        (?:${segment_nz_nc}(?:/${segment})*)
    }]

    set path_rootless [subst -nocommands -nobackslashes {
        (?:${segment_nz}(?:/${segment})*)
    }]

    set path_empty [subst -nocommands -nobackslashes {
        (?:${pchar}{0})
    }]

    set fragment_char [subst -nobackslashes {
        (?:${pchar}|/|\?)
    }]

    set query_char [subst -nobackslashes {
        (?:${pchar}|/|\?)
    }]   

    set absolute_uri_re [subst -nocommands -nobackslashes {
        ^
        (?:
         ${scheme}:
         (?:(?://${authority}${path_abempty})
          |${path_absolute}
          |${path_rootless}
          |${path_empty}
          )
         (?:\?${query_char}*)?
         (?:\#${fragment_char}*)?
         )
        $
    }]
    
    set relative_uri_re1 [subst -nocommands -nobackslashes {
        ^
        (?:
         ${path_absolute}
         (?:\?${query_char}*)?
         (\#${fragment_char}*)?
         )
        $
    }]

    set relative_uri_re2 [subst -nocommands -nobackslashes {
        ^
        (?:
         ${path_noscheme}
         (?:\?${query_char}*)?
         (\#${fragment_char}*)?
         )
        $
    }]

    set relative_uri_re3 [subst -nocommands -nobackslashes {
        ^
        (?:
         ${path_empty}
         (?:\?${query_char}*)?
         (\#${fragment_char}*)?
         )
        $
    }]

    set relative_uri_re4 [subst -nocommands -nobackslashes {
        ^
        (?:
         //${authority}${path_abempty}
         (?:\?${query_char}*)?
         (\#${fragment_char}*)?
         )
        $
    }]      

    if {
        [regexp -expanded $absolute_uri_re $uri] 
        || [regexp -expanded $relative_uri_re1 $uri] 
        || [regexp -expanded $relative_uri_re2 $uri]
        || [regexp -expanded $relative_uri_re3 $uri]
        || [regexp -expanded $relative_uri_re4 $uri] 
    } {
        return 1
    } else {
        return 0
    }
}

proc qc::is::next_url {next_url} {
    #| Absolute next url
    # check that redirection is to the same domain
    set conn_host [qc::conn_host]
    if { ![regexp "^https?://${conn_host}(:\[0-9\]+)?(\\?|#|/|\$)" $next_url] } {
        return 0
    }
    # check for malicious mal-formed url
    if { ![qc::is url $next_url] } {
        return 0 
    }
    
    return 1
}

proc qc::is::enumeration {enum_name value} {
    #| Checks if the given value is a value in enumeration enum_name.
    if {[qc::memoize qc::db_enum_exists $enum_name]
        && $value in [qc::memoize qc::db_enum_values $enum_name]} {
        return 1
    } else {
        return 0
    }
}

proc qc::is::decimal {args} {
    #| Checks if the given number is a decimal.
    qc::args $args -precision ? -scale ? -- string
    if {[string is double -strict $string]} {
        if { ! [info exists precision] && ! [info exists scale] } {
            # Scale and precision not given.
            return 1
        } elseif { ! [info exists precision] && [info exists scale] } {
            # Scale given but not precision.
            return -code error "Precision must be provided with scale."
        } elseif { [info exists precision] } {
            # Precision given.
            if { $precision <= 0 || ! [qc::is integer $precision]} {
                return -code error "Precision must be a positive integer."
            }
            # Count the number of digits before and after the decimal point.
            set number [qc::exp2string $string]
            set parts [split $number .]
            set left_count [string length [lindex $parts 0]]
            set right_count [string length [lindex $parts 1]]
            if { $left_count == 0 } {
                # The leading zero wasn't given.
                return -code error "Incomplete numeric string."
            }

            set total_count [expr {$left_count + $right_count}]
            if { $precision < $total_count } {
                # Precision is less than the total number of digits in the given
                # decimal.
                return 0
            }
            
            if { ! [info exists scale] } {
                # Scale wasn't given so set it to 0.
                set scale 0
            }

            if { $scale < 0 || ! [qc::is integer $scale]} {
                return -code error "Scale must be a non-negative integer."
            }

            set left_digit_max [expr {$precision - $scale}]
            if { $left_count > $left_digit_max || $right_count > $scale } {
                return 0
            } else {
                return 1
            }
            
        }
    } else {
        return 0
    }
}

proc qc::is::email {email} {
    #| Checks if the given string follows the form of an email address.
    set pattern {^[-a-zA-Z0-9!$&*=^`|~#%'+/?_{}]+(\.[-a-zA-Z0-9!$&*=^`|~#%'+/?_{}]+)*}
    append pattern {@[a-zA-Z0-9\-]+(\.[a-zA-Z0-9\-]+)+$}
    return [regexp $pattern $email]
}

proc qc::is::period {string} {
    #| Check if the given string is a period.

    if { [regexp -nocase {^\s*(.*?)\s+to\s+(.*?)\s*$} $string -> period1 period2] } {
        # Period defined by two periods eg "Jan 2011 to March 2011"
        if { [qc::is period $period1] && [qc::is period $period2] } {
            return 1
        } else {
            return 0
        }
    }

    if { [qc::is date $string] } {
        # String is an iso date eg "2014-01-01"
        return 1
    }

    if { [regexp {^([12]\d{3})$} $string -> year] } {
        # Exact match for year eg "2006"
        return 1
    }

    set month_names {Jan|January|Feb|February|Mar|March|Apr|April|May|Jun|June}
    append month_names {|Jul|July|Aug|August|Sep|September|Oct|October|Nov|November}
    append month_names {|Dec|December}

    set pattern [subst -nocommands -nobackslashes {($month_names)\s+([12]\d{3})$}]

    if { [regexp -nocase -- $pattern $string -> month_name year] } {
        # Exact match in format "Jan 2006"
        return 1
    }

    set pattern [subst {^($month_names)$}]

    if { [regexp -nocase -- $pattern $string -> month_name] } {
        # Exact match in format "Jan" (assume current year)
        return 1
    }

    set pattern {^([0-9]{1,2})(?:st|th|nd|rd)?\s+($month_names)\s+([12]\d{3})$}
    set pattern [subst -nocommands -nobackslashes $pattern]

    if { [regexp -nocase -- $pattern $string -> dom month_name year] } {
        # Exact match for castable date in format "1st Jan 2014"
        return 1
    }

    set pattern {^([0-9]{1,2})(?:st|th|nd|rd)?\s+($month_names)$}
    set pattern [subst -nocommands -nobackslashes $pattern]

    if { [regexp -nocase -- $pattern $string -> dom month_name] } {
        # Exact match for castable date in format "1st Jan" (assume current year)
        return 1       

    }

    set pattern {^($month_names)\s+([0-9]{1,2})(?:st|th|nd|rd)?\s+([12]\d{3})$}
    set pattern [subst -nocommands -nobackslashes $pattern]

    if { [regexp -nocase -- $pattern $string -> month_name dom year] } {
        # Exact match for castable date in format "Jan 1st 2014"
        return 1
    }

    set pattern {^($month_names)\s+([0-9]{1,2})(?:st|th|nd|rd)?$}
    set pattern [subst -nocommands -nobackslashes $pattern]

    if { [regexp -nocase -- $pattern $string -> month_name dom] } {
        # Exact match for castable date in format "Jan 1st" (assume current year)
        return 1
    }

    # could not parse string
    return 0
}

proc qc::is::domain {domain_name value} {
    #| Checks if the given value falls under the domain domain_name.
    if {[qc::memoize qc::db_domain_exists $domain_name]} {
        set base_type [qc::memoize qc::db_domain_base_type $domain_name]
        if { ! [qc::is $base_type $value] } {
            return 0
        }
        set constraints [qc::memoize qc::db_domain_constraints $domain_name]
        dict for {constraint_name check_clause} $constraints {
            if { ! [qc::db_eval_domain_constraint $value $base_type $check_clause] } {
                return 0
            }
        }
        return 1
    }
    return 0
}

proc qc::is::varchar {length string} {
    #| Checks if string would fit in a varchar of the given length.
    if { $length eq "" } {
        # PostgreSQL specification - missing length means any size
        return 1
    } elseif { [string length $string] <= $length } {
        return 1
    } else {
        return 0
    }
}

proc qc::is::html {text} {
    #| Checks if the given text contains valid html.
    try {
        # wrap the text up in <root> to preserve text outwith the html
        set text [qc::h root $text]
        set doc [dom parse -html $text]
        set root [$doc documentElement]
        $doc delete
        return 1
    } on error [list error_message options] {
        return 0
    }
}

proc qc::is::date {string} {
    #| Checks if the given string is a date.
    #| Dates are expected to be in ISO format.
    if { ! [regexp {^(\d{4})-(\d{2})-(\d{2})$} $string -- yyyy mm dd] } {
        return false
    }
    set yyyy [qc::cast int $yyyy]
    set mm [qc::cast int $mm]
    set dd [qc::cast int $dd]
    if { $mm > 12 || $mm == 0 } {
        return false
    }
    if { $dd == 0 } {
        return false
    }
    if { $mm in {1 3 5 7 8 10 12} } {
        if { $dd > 31 } {
            return false
        }
        return true
    }
    if { $mm in {4 6 9 11} } {
        if { $dd > 30 } {
            return false
        }
        return true
    }
    if { [qc::date_year_is_leap $yyyy] } {
        if { $dd > 29 } {
            return false
        }
        return true
    }
    if { $dd > 28 } {
        return false
    }
    return true
}

proc qc::is::safe_html {text} {
    #| Checks if the given text contains only safe html.
    try {
        # wrap the text up in <root> to preserve text outwith the html
        set text [qc::h root $text]
        set doc [dom parse -html $text]
        set root [$doc documentElement]
        
        if {$root eq ""} {
            $doc delete
            return 1
        } else {
            set safe [expr {[qc::safe_elements_check $root]
                            && [qc::safe_attributes_check $root]}]
            $doc delete
            return $safe
        }
    } on error [list error_message options] {
        return 0
    }
}

proc qc::is::boolean {bool} {
    #| Checks if the given number is a boolean.
    return [expr {[string toupper $bool] in {Y N YES NO TRUE FALSE T F 0 1}}]
}

proc qc::is::hex {string} {
    #| Checks if the given string is a hex number.
    return [regexp -nocase {^[0-9a-f]*$} $string]
}

proc qc::is::creditcard {number} {
    #| Checks if the given string is an allowable credit card number.
    #| Checks, number of digits are >13 & <19, all characters are integers,
    #| luhn 10 check
    regsub -all {[ -]} $number "" number
    set mult 1
    set sum 0
    if { [string length $number]<13 || [string length $number]>19 } {
        return 0
    }
    foreach digit [lreverse [split $number ""]] {
        if { ![qc::is integer $digit] } {
            return 0
        }
        set t [expr {$digit*$mult}]
        if { $t >= 10 } {
            set sum [expr {$sum + $t%10 +1}]
        } else {
            set sum [expr {$sum + $t}]
        }
        if { $mult == 1 } { set mult 2 } else { set mult 1 }
    }
    if { $sum%10 == 0 } {
        return 1
    } else {
        return 0
    }
}

proc qc::is::creditcard_masked {string} {
    #| Check the credit card number is masked to PCI requirements.
    regsub -all {[^0-9\*]} $string "" number

    # 13-19 chars masked with < 6 prefix and < 4 suffix digits
    return [expr {[regexp {[0-9\*]{13,19}} $number]
                  && [regexp {^[3-6\*][0-9]{0,5}\*+[0-9]{0,4}$} $number]}]
}

proc qc::is::url {args} {
    #| Checks if the given string is a URL.
    #| This is a more restrictive subset of all legal uri's defined by RFC 3986
    #| Relax as needed
    qc::args $args -relative -- url
    qc::default relative false
    if { $relative } {
        return [regexp -expanded {
            # path
            ^([a-zA-Z0-9_\-\.~+/%&;]+)?
            # query
            (\?[a-zA-Z0-9_\-\.~+/%=&:@;]+)?
            # anchor
            (\#[a-zA-Z0-9_\-\.~+/%;]+)?
            $
        } $url]
    } else {
        return [regexp -expanded {
            # protocol
            ^https?://
            # domain
            [a-z0-9\-\.;]+
            # port
            (:[0-9]+)?
            # path
            ([a-zA-Z0-9_\-\.~+/%&;]+)?
            # query
            (\?[a-zA-Z0-9_\-\.~+/%=&:@;]+)?
            # anchor
            (\#[a-zA-Z0-9_\-\.~+/%;]+)?
            $
        } $url]
    }
}

proc qc::is::url_path {string} {
    #| Checks if the given string is an url path.
    return [regexp {/([a-zA-Z0-9\-._~]|%[0-9a-fA-F]{2}|[!$&'()*+,;=:@]|/)*$} $string]
}

proc qc::is::s3_bucket {s3_bucket} {
    #| Checks if the given string is a valid s3 bucket
    #| Rules from:
    #| https://docs.aws.amazon.com/AmazonS3/latest/dev/BucketRestrictions.html
    
    # Bucket names must be between 3 and 63 characters long
    if { [string length $s3_bucket] < 3 ||
         [string length $s3_bucket] > 63 } {
        return 0
    }

    # Bucket names can consist only of lowercase letters, numbers, dots (.),
    # and hyphens (-)
    if { [regexp {[^a-z0-9.-]} $s3_bucket] } {
        return 0
    }

    # Bucket names must begin and end with a letter or number.
    if { [regexp {[^a-z0-9]} [string index $s3_bucket 0]] ||
         [regexp {[^a-z0-9]} [string index $s3_bucket end]] } {
        return 0
    }

    # Bucket names must not be formatted as an IP address (for example, 192.168.5.4).
    if { [qc::is ip $s3_bucket] } {
        return 0
    }

    return 1
}

proc qc::is::s3_object_key {s3_object_key} {
    #| Checks if the given string is a valid s3 object key
    #| Limits characters to those defined as safe:
    #| https://docs.aws.amazon.com/AmazonS3/latest/dev/UsingMetadata.html#object-keys

    # Restrict to safe characters (excluding the starting "/")
    if { [regexp {[^-a-zA-Z0-9/!_.*'()]} $s3_object_key] } {
        return 0
    }

    # Object keys ending in "." can cause issues
    if { [string index $s3_object_key end] eq "." } {
        return 0
    }

    return 1
}

proc qc::is::s3_uri {s3_uri} {
    #| Checks if the given string is a valid s3 URI
    # Valid Examples:
    # s3://bucket/location.png
    # S3://test/location/foo/bar
    # s3://bucket

    # Check for protocol 
    if { ![regexp {^[sS]3://} $s3_uri]} {
        return 0
    }

    # Check the bucket and key are valid
    lassign [qc::s3 uri_bucket_object_key $s3_uri] bucket object_key
    if { ![qc::is s3_bucket $bucket] } {
        return 0
    }
    if { $object_key ne "" && ![qc::is s3_object_key $object_key] } {
        return 0
    }
    
    return 1
}

proc qc::is::smallint {int} {
    #| Checks if the given number is an 8-bit signed integer.
    if {[string is integer -strict $int] && $int >= -32768 && $int <= 32767} {
        return 1
    } else {
        return 0
    }
}

proc qc::is::interval {text} {
    #| Checks if given text is an interval
    # (more restrictive than postgres, relax as needed)
    return \
        [regexp \
             {((^| +)(\+|-)?[0-9]+ +(year|month|week|day|hour|minute|second)s?)+$} \
             [string tolower $text]]
}

proc qc::is::time {time} {
    #| Check if the given date is a time
    #| in the form 23:59:59 or 23:59:59.01
    set pattern {^(([0-1][0-9]|2[0-3]):[0-5][0-9]:[0-5][0-9]|24:00:00)(\.\d{1,6})?$}
    return [regexp $pattern $time]
}

proc qc::is::bigint {int} {
    #| Checks if the given number is a 64-bit signed integer.
    if {[string is wideinteger -strict $int]
        && $int >= -9223372036854775808
        && $int <= 9223372036854775807} {
        return 1
    } else {
        return 0
    }
}

proc qc::is::postcode {postcode} {
    #| Checks if the given string is a UK postcode.
    return [expr {[regexp \
                       {^[A-Z]{1,2}[0-9R][0-9A-Z]? [0-9][ABD-HJLNP-UW-Z]{2}$} \
                       $postcode]
                  || [regexp {^BFPO ?[0-9]+$} $postcode]}]
}

proc qc::is::char {length string} {
    #| Checks if string would fit exactly into a character string of the given length.
    if { [string length $string] == $length } {
        return 1
    } else {
        return 0
    }
}

proc qc::is::webp {file} {
    #| Check if the file is a webp image.
    set fin [open $file rb]
    set chunk0 [read $fin 12]
    close $fin

    binary scan $chunk0 "a4ia4" riff size id

    return [expr {
                  [info exists riff]
                  && [info exists id]
                  && $riff eq "RIFF"
                  && $id eq "WEBP"
              }]
}

proc qc::is::mobile_number {string} {
    #| Checks if the given string is of the form of a UK mobile telephone number.
    regsub -all {[^0-9]} $string {} tel_no
    if {  [regexp {^(07[1-57-9][0-9]{8}|07624[0-9]{6})$} $tel_no] } {
        return 1
    } else {
        return 0
    }
}

proc qc::is_non_zero_integer {int} {
    #| Deprecated
    if { [is_integer $int] && $int!=0 } {
	return 1
    } else {
	return 0
    }
}

proc qc::is_email { email } {
    #| Deprecated - see qc::is email
    return [qc::is email $email]
}

proc qc::is_boolean {bool} {
    #| Deprecated - see qc::is boolean
    return [qc::is boolean $bool]
}

proc qc::is_webp {file} {
    #| Deprecated - see qc::is webp
    #| Check if the file is a webp image.

    return [qc::is webp $file]
}

proc qc::is_pnz {number} {
    #| Deprecated
    if { [is_decimal $number] && $number>0 } {
	return 1
    } else {
	return 0
    }
}

proc qc::is_pnz_int { int } {
    #| Deprecated
    # positive non zero integer
    if { [is_integer $int] && $int > 0 } {
	return 1
    } else {
	return 0
    }
}

proc qc::is_int_castable {string} {
    #| Deprecated - see qc::castable integer
    #| Can input be cast to an integer?
    return [qc::castable integer $string]
}

proc qc::is_non_zero_decimal { number } {
    #| Deprecated
    if { [string is double -strict $number] && $number!=0 } {
	return 1
    } else {
	return 0
    }
}

proc qc::is_postcode { postcode } {
    #| Deprecated - see qc::is postcode
    # uk postcode
    return [qc::is postcode $postcode]
}

proc qc::is_creditcard { no } {
    #| Deprecated - see qc::is creditcard
    #| Checks if no is an allowable credit card number
    #| Checks, number of digits are >13 & <19, all characters are integers, luhn 10 check
    return [qc::is creditcard $no]
}

proc qc::is_creditcard_masked { no } {
    #| Deprecated - see qc::is creditcard_masked
    #| Check the credit card number is masked to PCI requirements
    return [qc::is creditcard_masked $no]
}

namespace eval qc {
    namespace export contains_creditcard
}

proc qc::contains_creditcard {string} {
    #| Checks string for occurrences of credit card numbers
    set re {
	(?:^|[^0-9])
	(
	 # amex 15 digits long
	 (?:3\d{3}[ \-\.]*\d{6}[ \-\.]*\d{5})
	 |
	 # other cards 16 digits long
	 (?:[4|5|6]\d{3}[ \-\.]*\d{4}[ \-\.]*\d{4}[ \-\.]*\d{4})
	 )
	(?:[^0-9]|$)
    }
    foreach {match cc_no} [regexp -inline -all -expanded $re $string] {
	if { [is_creditcard [string map {" " "" - "" . ""} $cc_no]] } {
	    return true
	}
    }
    return false
}

namespace eval qc {
    namespace export webpsize
}

proc qc::webpsize {file} {
    #| Get the width and height of a webp image file
    set info [exec_proxy \
                  -timeout 1000 \
                  -ignorestderr \
                  identify \
                  $file]
    lassign [split $info " "] {*}{
        filename
        format
        dimensions
        page_dimensions
        colorspace
        usertime
        elaspedtime
    }
    return [split $dimensions "x"]
}

proc qc::is_date { date } {
    #| Deprecated - see qc::is date
    # dates are expected to be in iso format 
    return [qc::is date $date] 
}

proc qc::is_date_castable {string} {
    #| Deprecated - see qc::castable date
    #| Can string be cast into date format?
    return [qc::castable date $string]
}

proc qc::is_hex {string} {
    #| Deprecated - see qc::is hex
    #| Does the input look like a hex number?
    return [qc::is hex $string]
}

namespace eval qc {
    namespace export ofc_*
}

### chart.tcl ###

# New Charts by Daniel Clark 
proc qc::ofc_piechart {args} {

    args $args -id "" -title {} -animate false -width 100% -height 50% -- data
    
    # if id not supplied use global counter to ensure id is unique to allow multiple charts on a html page. 
    if { [eq $id ""] } {
	global ofc_id
	set id graph[incr0 ofc_id 1]
    }

    # convert args into tson and then json for ofc
    #
    # values
    set values {}
    foreach datum $data {
	dict2vars $datum label value tooltip
	set list [list label [list string $label] value $value]
	if { [info exists tooltip] } {
	    lappend list tip [list string $tooltip]
	}
	lappend values [list object {*}$list]
    }	
    set values [list array {*}$values]

    # title
    dict2vars $title label font-size
    if { [info exists label] } {
	set title_style [style_set "" font-family "Arial,Helvetica,sans-serif" text-align center font-size 20px]
	if { [info exists font-size] } {
	    set title_style [style_set $title_style font-size ${font-size}]
	}
	set title [list object text [list string $label] style $title_style]
    }
    
    # default colours
    set no_of_elements [llength $data]
    set colors [list array {*}[qc::ofc_colors $no_of_elements]]

    # construct tson that will be used to generate json loaded by ofc. 
    set tson [list object \
		  elements [list array \
				[list object \
				     type pie \
				     animate $animate \
				     values $values \
				     colours $colors \
				     font-size 14 \
				     tip "#label#:<br>#val# of #total#<br>#percent#" \
				     alpha 1 \
				     start-angle 35 \
				     gradient-fill true \
				     label-colour 0 \
				     border 2 ]] \
		  title $title \
		  bg_colour #FAFAFA]

    set html [ofc_html $id [tson2json $tson] $width $height]
}

proc /doc/ofc_pie_examples.html {} {
    sset html {
	<script type="text/javascript" src="/JavaScript/swfobject.js"></script> 
	<script type="text/javascript" src="/JavaScript/json2.js"></script> 
    }

    append html [html h2 "Piechart Examples"]

    append html [html h4 {Minimum Argument Usage: "ofc_piechart {$data}"}]
    set data [list {label "Label 1" value 12} {label "Label 2" value 14}]
    append html [ofc_piechart $data]
    
    append html [html h4 {Full Argument Usage: "ofc_piechart -id $id -title $title -animate $animate -width $width -height $height -- $data"}]
    set id chart2
    set title {label "My Wizzzy Pie Chart" font-size 25px}  
    set animate true
    set width 50%
    set height 50%  
    set data [list {label "Label 1" value 12 tooltip "tip 1"} {label "Label 2" value 14 tooltip "tip 2"}]
    append html [ofc_piechart -id $id -title $title -animate $animate -width $width -height $height -- $data]

    qc::return_html $html
}

proc qc::ofc_linechart {args} {

    args $args -id "" -title {} -x_axis {} -y_axis {} -width 100% -height 50% -- lines
    
    if { [eq $id ""] } {
	# if id not supplied use global counter to ensure id is unique to allow multiple charts on a html page. 
	global ofc_id
	set id graph[incr0 ofc_id 1]
    }

    set colors [qc::ofc_colors]

    # find min and max over all lines
    lappend min_values 0
    lappend max_values 0
    set elements {}
    set x_labels {}
    foreach line $lines {
 	dict2vars $line label color data 
	set number_of_points [llength $data]
	default color [qc::lshift colors]
	
	lassign [ofc_line_element $label $color $data] tson x_labels min max 
	lappend elements $tson
	lappend max_values $max
	lappend min_values $min
    }

    set x_labels [list array {*}$x_labels]
    set max_value [max {*}$max_values]  
    set min_value [min {*}$min_values]  
    
    # plot y=0 line if negative values exist.
    if { $min_value < 0 } { 
	for { set i 0 } { $i < $number_of_points } { incr i } {
	    lappend y_values 0
	} 
	lappend elements [list object \
			      type line \
			      line-style [list object style dash on 5 off 5] \
			      colour #999999 \
			      size 20 \
			      values [list array {*}$y_values]]
    }
    
    # title
    dict2vars $title label font-size
    if { [info exists label] } {
	set title_style [style_set "" font-family "Arial,Helvetica,sans-serif" text-align center font-size 20px]
	if { [info exists font-size] } {
	    set title_style [style_set $title_style font-size ${font-size}]
	}
	set title [list object text [list string $label] style $title_style]
    }

    # x legend
    dict2vars $x_axis label grid_step label_step offset
    default offset false
    default label "" 
    default grid_step 1
    default label_step $grid_step
    set style [style_set "" font-family "Arial,Helvetica,sans-serif" font-size 18px]
    set x_legend [list object style $style text [list string $label]]

    # y legend
    # Default y-axis with 10 steps.
    dict2vars $y_axis label min max step 
    default label "" 
    set style [style_set "" font-family "Arial,Helvetica,sans-serif" font-size 18px]
    set y_legend [list object style $style text [list string $label]]
    default min $min_value
    default max $max_value  
    default step [ofc_step $min $max] 
    set min [expr {floor(double($min)/$step) * $step}]
    set max [expr {$min + (ceil(double($max-$min)/$step) * $step)}]

    # Construct tson that will be used to generate json loaded by ofc. 
    set tson [list object \
		  elements [list array {*}$elements] \
		  x_legend $x_legend \
		  x_axis [list object \
			      steps $grid_step \
			      offset $offset \
			      stroke 1 \
			      colour #000000 \
			      width 6 \
			      grid-colour #dddddd \
			      labels [list object \
					  visible-steps $label_step \
					  size 13 \
					  labels $x_labels]] \
		  y_legend $y_legend \
		  y_axis [list object \
			      min $min \
			      max $max \
			      steps $step \
			      stroke 1 \
			      colour #000000 \
			      grid-colour #dddddd \
			      labels [list object \
					  size 13]] \
		  title $title \
		  bg_colour #FAFAFA]
    
    set html [ofc_html $id [tson2json $tson] $width $height]
}

proc /doc/ofc_line_examples.html {} { 

    sset html {
	<script type="text/javascript" src="/JavaScript/swfobject.js"></script> 
	<script type="text/javascript" src="/JavaScript/json2.js"></script> 
    }
    
    set data1 [list \
		   [list x Jan y 500] \
		   [list x Feb y 550] \
		   [list x Mar y 700] \
		   [list x Apr y 670] \
		   [list x May y 730]]
    set data2 [list \
		   [list x Jan y 50 tooltip "Adwords (Jan)<br>50 Bananas"] \
		   [list x Feb y 65 tooltip "Adwords (Feb)<br>65 Bananas"] \
		   [list x Mar y -3 tooltip "Adwords (Mar)<br>-3 Bananas"] \
		   [list x Apr y 67 tooltip "Adwords (Apr)<br>67 Bananas"] \
		   [list x May y 73 tooltip "Adwords (May)<br>73 Bananas"]]
    set data3 [list \
		   [list x Jan y 500 tooltip "Froogle (Jan)<br>500 Apples"] \
		   [list x Feb y 605 tooltip "Froogle (Feb)<br>605 Apples"] \
		   [list x Mar y 700 tooltip "Froogle (Mar)<br>700 Apples"] \
		   [list x Apr y 607 tooltip "Froogle (Apr)<br>607 Apples"] \
		   [list x May y 703 tooltip "Froogle (May)<br>703 Apples"]]

    append html [html h2 "Linechart Examples"]

    append html [html h4 {Minimum Arguments Usage: "ofc_linechart $lines"}]
    set lines [list \
		   [list label "Direct" data $data1] \
		   [list label "Adwords" data $data2] \
		   [list label "Froogle" data $data3]]
    append html [ofc_linechart $lines]

    append html [html h4 {Full Arguments Usage: "ofc_linechart -id $id -title $title -x_axis $x_axis -y_axis $y_axis -width $width -height $height --  $lines"}]
    set id chart2
    set title {label "My Wizzy Line Chart" font-size 20px}
    set x_axis {label "Months" grid_step 1 label_step 1}
    set y_axis {label "Orders" min -100 max 800 step 100} 
    set width 50%
    set height 25%  
    set lines [list \
		   [list label "Direct" color #00FF33 data $data1] \
		   [list label "Adwords" color #FF0033 data $data2] \
		   [list label "Froogle" color #3333CC data $data3]]
    append html [ofc_linechart -id $id -title $title -x_axis $x_axis -y_axis $y_axis -width $width -height $height --  $lines]

    qc::return_html $html
}

proc qc::ofc_line_element {label color data} {
    # return tson for the ofc element
    # each ofc element describes one line

    set x_labels {}
    set y_values {}
    foreach datum $data {
	dict2vars $datum x y tooltip
	default tooltip "$label<br>#val#"
	lappend x_labels [list string $x]
	lappend y_values [list object value $y tip [list string $tooltip]]
	lappend values $y
    }

    set tson [list object \
		  type line \
		  text [list string $label] \
		  colour $color \
		  size 20 \
		  values [list array {*}$y_values]]
    
    return [list $tson $x_labels [min {*}$values] [max {*}$values]]
}

proc qc::ofc_barchart {args} {

    args $args -id "" -title {} -x_axis {} -y_axis {} -width 100% -height 50% -- bars
    
    if { [eq $id ""] } {
	# if id not supplied use global counter to ensure id is unique to allow multiple charts on a html page. 
	global ofc_id
	set id graph[incr0 ofc_id 1]
    }

    set colors [qc::ofc_colors]
    
    # find min and max over all bars
    lappend max_values 0
    lappend min_values 0
    set elements {}
    set x_labels {}
    set values {}
    set tson_keys {}
    foreach bar $bars {
 	dict2vars $bar x data 
	lassign [ofc_bar2tson $colors $data] tson_values tson_keys pos_total neg_total 

	lappend x_labels [list string $x]
	lappend values $tson_values
	lappend max_values $pos_total
	lappend min_values $neg_total
    }
    set max_value [max {*}$max_values]
    set min_value [min {*}$min_values]
    
    # elements
    lappend elements [list object \
			  type bar_stack \
			  alpha 0.80 \
			  colours [list array {*}$colors] \
			  values [list array {*}$values] \
			  keys [list array {*}$tson_keys]] 
    
    # plot y=0 line if negative values exist.
    set y0 {}
    if { $min_value < 0 } { 
	for { set i 0 } { $i < [llength $bars] } { incr i } {
	    lappend y_values 0
	} 
	lappend elements [list object \
			      type line \
			      line-style [list object style dash on 5 off 5] \
			      colour #999999 \
			      size 20 \
			      values [list array {*}$y_values]]
    }
    
    # title
    dict2vars $title label font-size
    if { [info exists label] } {
	set title_style [style_set "" font-family "Arial,Helvetica,sans-serif" text-align center font-size 20px]
	if { [info exists font-size] } {
	    set title_style [style_set $title_style font-size ${font-size}]
	}
	set title [list object text [list string $label] style $title_style]
    }
    
    # x legend 
    dict2vars $x_axis label grid_step label_step 
    default label "" 
    default grid_step 1
    default label_step $grid_step
    set style [style_set "" font-family "Arial,Helvetica,sans-serif" font-size 18px]
    set x_legend [list object style $style text [list string $label]]
    
    # y legend
    # Default y-axis with 10 steps.
    dict2vars $y_axis label min max step 
    default label "" 
    set style [style_set "" font-family "Arial,Helvetica,sans-serif" font-size 18px]
    set y_legend [list object style $style text [list string $label]]
    default min $min_value
    default max $max_value  
    default step [ofc_step $min $max] 
    set min [expr {floor(double($min)/$step) * $step}]
    set max [expr {$min + (ceil(double($max-$min)/$step) * $step)}]
    
    # Construct tson that will be used to generate json loaded by ofc. 
    set tson [list object \
		  elements [list array {*}$elements] \
		  x_legend $x_legend \
		  x_axis [list object \
			      steps $grid_step \
			      colour #000000 \
			      grid-colour #dddddd \
			      labels [list object \
					  steps $label_step \
					  size 13 \
					  labels [list array {*}$x_labels]]] \
		  y_legend $y_legend \
		  y_axis [list object \
			      min $min \
			      max $max \
			      steps $step \
			      stroke 1 \
			      colour #000000 \
			      grid-colour #dddddd \
			      labels [list object \
					  size 13]] \
		  title $title \
		  tooltip [list object mouse 2] \
		  bg_colour #FAFAFA]
    
    set html [ofc_html $id [tson2json $tson] $width $height]
}

proc /doc/ofc_bar_examples.html {} { 

    sset html {
	<script type="text/javascript" src="/JavaScript/swfobject.js"></script> 
	<script type="text/javascript" src="/JavaScript/json2.js"></script> 
    }
    
    set data1 [list \
		   [list label Direct y 500] \
		   [list label Adwords y 750] \
		   [list label Froogle y 70]]
    set data2 [list \
		   [list label Direct y 560 tooltip "Direct (Jan)<br>560 of 1005 Orders"] \
		   [list label Adwords y 395 tooltip "Adwords (Feb)<br>395 of 1005 Orders"] \
		   [list label Froogle y 50 tooltip "Froogle (Mar)<br>50 of 1005 Orders"]]
    set data3 [list \
		   [list label Direct y 600 tooltip "Direct (Jan)<br>600 of 1410 Orders"] \
		   [list label Adwords y 360 tooltip "Adwords (Feb)<br>360 of 1410 Orders"] \
		   [list label Froogle y 450 tooltip "Froogle (Mar)<br>450 of 1410 Orders"]]    
    set bars [list \
		  [list x Jan data $data1] \
		  [list x Feb data $data2] \
		  [list x March data $data3]]

    append html [html h2 "Barchart Examples"]

    append html [html h4 {Minimum Arguments Usage: "ofc_barchart $bars"}]   
    append html [ofc_barchart $bars]

    append html [html h4 {Full Arguments Usage: "ofc_barchart -id $id -title $title -x_axis $x_axis -y_axis $y_axis -width $width -height $height --  $bars"}]
    set id chart2
    set title {label "My Wizzy Bar Chart" font-size 20px}
    set x_axis {label "Months" grid_step 1 label_step 1}
    set y_axis {label "Orders" min 0 max 1500 step 250} 
    set width 50%
    set height 25%     
    append html [ofc_barchart -id $id -title $title -x_axis $x_axis -y_axis $y_axis -width $width -height $height --  $bars]

    qc::return_html $html
}

proc qc::ofc_bar2tson {colors data} {
    # return tson for the ofc element
    # each ofc values describes one bar
    
    set tson_keys {}
    set y_values {}
    set pos_total 0
    set neg_total 0
    foreach datum $data {
	dict2vars $datum label y tooltip
	default tooltip "$label<br>#val# of #total#"
	lappend tson_keys [list object text [list string $label] colour [qc::lshift colors] font-size 13]
	lappend y_values [list object val $y tip [list string $tooltip]]
	if { $y > 0 } {
	    set pos_total [add $pos_total $y]
	} else {
	    set neg_total [add $neg_total $y]
	}
    }
    set tson_values [list array {*}$y_values]
    
    return [list $tson_values $tson_keys $pos_total $neg_total]
}

proc qc::ofc_step {min max} {
    # Determine a step value to be used to display a chart with 10 divisions. 
    # Step size should be rounded up to only one sig fig 
    # in order to have nicer increments 
    # eg. for max 960 min 0, step would be 100 instead of 96. 
    if { $min == $max } { 
	return 1 
    } else {
	set step [expr {double(abs($max - $min)) /10}]
	return [qc::sigfigs_ceil $step 1] 
    }
}

proc qc::ofc_colors {{no_of_elements 1}} {
    # 10 preset web safe colours to colour chart elements, if number of elements is odd use 11 instead.
    # #336699/*dark blue*/ #666600/*dark green*/ #CC9933/*dark orange*/ #993366/*dark red*/ 
    # #CCCCCC/*light gray*/ #6699CC/*light blue*/ #999900/*light green*/ #FFCC66/*light orange*/ 
    # #CC9999/*light red*/ #000000/*black*/ #669999/*aqua*/

    set colors [list #336699 #666600 #CC9933 #993366 #CCCCCC #6699CC #999900 #FFCC66 #CC9999 #000000]

    if { [expr $no_of_elements%2] ==  1 } {
	lappend colors #669999
    } 

    return $colors
}

proc qc::ofc_html {id json width height} {   
    # html to construct ofc object.
    sset html {
	<script type="text/javascript">
	var data_$id = ${json};
	function get_data_${id}() {
	    return JSON.stringify(data_$id);
	}
	jQuery(function(){
	    swfobject.embedSWF("/JavaScript/open-flash-chart.swf?"+Math.floor(Math.random()*1000), "$id", "$width", "$height", "9.0.0", false, {"get-data":"get_data_$id"}, {"wmode":"opaque"} );
	});
	</script>
    }
    # add div element for ofc to be written into. 
    append html [html div "" id $id style "width:$width;height:$height;"]

    return $html
}

namespace eval qc {
    namespace export lshift lunshift lintersect lexclude lexclude2 lunion ldelete lmove lunique lequal lsort_using in lpage ltotitle list2csv lconcat lsum laverage lreplace_values lapply
}

proc qc::lshift { stack } {
    #| Return leftmost value from list and remove it
    upvar 1 $stack list
    set value [lindex $list 0]
    set list [lrange $list 1 end]
    return $value 
}

proc qc::lunshift { stack value } {
    #| Adds $value as leftmost item in the list
    upvar 1 $stack list
    if { ![ info exists list ] } {
	set list {}
    }
    set list [linsert $list 0 $value]
}

global tcl_version
if { $tcl_version<8.5 } {
    proc lreverse { list } {
        #| List reverse for Tcl <8.5 
	set rlist {}
	foreach value $list {
	    set rlist [linsert $rlist 0 $value]
	}
	return $rlist
    }
    proc lassign {list args} {
        #| List assign for Tcl <8.5 
	foreach value $list variableName $args {
	    upset 1 $variableName $value
	}
    }
}

proc qc::lintersect { a b } {
    #| Returns the intersection of 2 lists
    set a [lsort $a]
    set result {}
    foreach element $b {
	if { [lsearch -sorted -increasing $a $element] != -1 } {
	    lappend result $element
	}
    }
    return $result
}

proc qc::lexclude { list args } {
    #| Return $list with and values listed in $args excluded
    set result {}
    foreach elem $list {
	if { ![in $args $elem] } {
	    lappend result $elem
	}
    }
    return $result
}

proc qc::lexclude2 { list args } { 
    #| Alternative implementation of qc::lexclude
    foreach item $args {
	if { [set index [lsearch $list $item]]!=-1 } {
	    ldelete list $index
	}
    }
    return $list
}

proc qc::lunion { a b } {
    #| Return union of 2 lists
    set result [concat $a $b]
    return [lsort -unique $result]
}

proc qc::ldelete {listVar index} {
    #| Deletes item at $index of list
    upvar 1 $listVar list

    if { $index < [llength $list] } {
        # Replace a deletion with null, much faster
        set list [lreplace [K $list [set list {}]] $index $index]
    }

    return $list
}

proc qc::lmove {list from to} {
    #| Move an element in a list from one place to another
    return [linsert [lreplace $list $from $from] $to [lindex $list $from]]
}

proc qc::lunique {list} {
    #| Returns a list of distinct list values
    set result {}
    foreach item $list {
	if {![info exists u($item)]} {
	    set u($item) {}
	    lappend result $item
	}
    }
    return $result
}

proc qc::lequal {a b} {
    #| Compare 2 lists
    # Author RS Tcl wiki
    
    # a or b not a list
    if { ! [string is list $a]
         || ! [string is list $b]
     } {
        return [string equal $a $b]
    }

    # a and b different lengths
    if { [llength $a] != [llength $b] } {
        return 0
    }

    # both are length 1
    if { [lindex $a 0] == $a } {
        return [string equal $a $b]
    }

    # compare each element
    foreach i $a j $b {
        if { ! [lequal $i $j] } {
            return 0
        }
    }
    return 1
} 

proc qc::lsort_using { list order } {
    #| Apply an arbitrary sort order to a list
    foreach item [lreverse $order] {
	foreach index [lsearch -all -exact $list $item] {
	    set list [lmove $list $index 0]
	}
    }
    return $list
}

# ::tcl::mathop::in has different argument sequence
#namespace import ::tcl::mathop::in
#namespace import ::tcl::mathop::ni

proc qc::in { list item } {
    #| Return 1 if $item appears in $list
    set list [lsort $list]
    if {[lsearch -sorted -increasing $list $item] == -1 } {
	return 0
    } else {
	return 1
    }
}

proc qc::lpage { list page_length } {
    #| Split list into sublists of length $page_length
    set lpage {}
    while { [llength $list]%$page_length != 0 || [llength $list]<$page_length} {
	lappend list ""
    }
    for { set line 0 } { $line < [llength $list] } { incr line $page_length } {
	lappend lpage [lrange $list $line [expr {$line + $page_length -1}]]
    }
    return $lpage
}

proc qc::ltotitle {list} {
    #| Make each word totitle excepting some industry specific acronyms
    if { [regexp {\"} $list] } {
	return $list
    } else {
	set new_list {}
	set upper {TV FM AM PVC BC DP ES CCTV TP&N GU10 SP&N MCB RCD AC DC}
	foreach word [split $list] {
	    if { [in $upper [upper $word]] } {
		lappend new_list [upper $word]
	    } else {
		lappend new_list [string totitle $word]
	    }
	}
	return [join $new_list]
    }
}

proc qc::list2csv {list {comma ,}} {
    #| Convert list to CSV (use of comma can be overridden)
    set out ""
    set separator {}
    foreach field $list {
	if { [string first $comma $field]==-1
             && [string first \" $field]==-1
             && [string first \n $field]==-1
         } {
	    append out $separator$field
	} else {
	    append out $separator\"[string map [list \" \"\"] $field]\"
	}
	set separator $comma
    }
    return $out
}

proc qc::lconcat {listVar list} {
    #| Concatenate list onto supplied listVar
    upvar $listVar var
    set var [concat $var $list]
}

proc qc::lsum {list} {
    #| Return sum of all list items
    set total 0
    foreach number $list {
	set total [expr {$total+$number}]
    }
    return $total
}

proc qc::laverage {list} {
    #| Returns average of all list elements
    return [expr {double([lsum $list])/[llength $list]}]
}

proc qc::lreplace_values {list find replace} {
    #| Replace any occurrence of $find in $list with $replace
    set index [lsearch -exact $list $find]
    while { $index!=-1 } {
	set list [lreplace $list $index $index $replace]
	set index [lsearch -exact $list $find]
    }
    return $list
}

proc qc::lapply { func list } {
    #| Apply the named procedure to all elements in the list and return a list of the results
    set result {}
    foreach item $list {
        lappend result [$func $item]
    }
    return $result
}

proc qc::map {lambda list} {
    #| Applies a function to each item in the list.
    set result {}
    foreach item $list {
        lappend result [apply $lambda $item]
    }
    return $result
}

namespace eval qc {
    namespace export ll_sum ll2csv
}

proc qc::ll_sum { llVar index } {
    #| Traverses a list of lists and returns the sum of values at $index in each list
    # TODO: Could be de-aolserverised
    set sum 0
    upvar 1 $llVar ll
    foreach list $ll {
	set value [lindex $list $index]
	set value [qc::strip_html $value]
	regsub -all {[, ]} $value {} value
	if { [string is double -strict $value] } {
	    set sum [expr {$sum + $value}]
	}
    }
    return $sum
}

proc qc::ll2csv {ll {comma ,}} {
    #| Convert a list of lists into a csv.
    #| Defaults to comma separated but allows the passing of alternative delimiters.
    set lines {}
    foreach list $ll {
	lappend lines [qc::list2csv $list $comma]
    }
    return [join $lines \r\n]
}

proc qc::ll2pg_copy {ll} {
    #| Return data in the format accepted by postgresql's copy statements
    set pg_copy_data ""
    foreach list $ll {
        append pg_copy_data "[qc::list2pg_copy $list]\n"
    }
    return $pg_copy_data
}

proc qc::list2pg_copy {list} {
    #| Return data in the format accepted by postgresql's copy statements
    set temp {}
    foreach value $list {
        if { $value eq "" } {
            # Encode empty string as NULL
            set value \\N
        } else {
            # Escape backslash, newline, carriage return, tab characters
            set value [string map {\\ \\\\ \n \\n \r \\r \t \\t \v \\v} $value]
        }
        lappend temp $value
    }
    return "[join $temp \t]"
}



namespace eval qc {
}

proc qc::_image_cache_original_exists {
    cache_dir file_id
} {
    #| Test whether a filesystem cache of the original image file exists
    set glob_pattern "${file_id}.*"
    set glob_options [list -nocomplain -types f -directory $cache_dir]
    if { [llength [glob {*}$glob_options {*}$glob_pattern]] == 1 } {
        return true
    } else {
        return false
    }
}

proc qc::_image_cache_original_file {
    cache_dir file_id
} {
    #| Test whether a filesystem cache of the original image file exists
    set glob_pattern "${file_id}.*"
    set glob_options [list -nocomplain -types f -directory $cache_dir]
    set matches [glob {*}$glob_options {*}$glob_pattern]
    return [lindex $matches 0]
}

proc qc::_image_cache_original_create {
    cache_dir file_id
} {
    #| Create a cache of an image in original form, in cache_dir
    #| Create symbolic link to cached image
    db_1row {
        select
        filename,
        width,
        height,
        mime_type
        
        from file
        join image using(file_id)
        
        where file_id=:file_id
    }
    # Only one canonical file required for svg
    if { $mime_type eq "image/svg+xml" } {
        set file [qc::db_file_export $file_id]
        file rename -force $file ${cache_dir}/${file_id}.svg
        return
    }
    
    set ext [file extension $filename]
    set cache_file_relative ${file_id}-${width}x${height}/${file_id}${ext}
    set cache_file ${cache_dir}/${cache_file_relative}
    if { ! [file exists $cache_file] } {
        set file [qc::db_file_export $file_id]
        qc::image_file_meta_strip $file
        
        file mkdir [file dirname $cache_file]
        file rename -force $file $cache_file
    }
    file link ${cache_dir}/${file_id}${ext} $cache_file_relative
}

namespace eval qc::castable {
    
    namespace export integer bigint smallint decimal boolean timestamp timestamptz char varchar enumeration text domain safe_html safe_markdown date postcode creditcard period url relative_url url_path s3_uri time interval next_url
    namespace ensemble create -unknown {
        data_type_parser
    }

    proc integer {string} {
        #| Test if the given string can be cast to an integer.
        try {
            qc::cast integer $string
            return true
        } on error [list error_message options] {
            return false
        }
    }

    proc bigint {string} {
        #| Test if the given string can be cast to a big integer.
        try {
            qc::cast bigint $string
            return true
        } on error [list error_message options] {
            return false
        }
    }

    proc smallint {string} {
        #| Test if the given string can be cast to a small integer.
        try {
            qc::cast smallint $string
            return true
        } on error [list error_message options] {
            return false
        }
    }

    proc decimal {args} {
        #| Test if the given value can be cast to a decimal.
        try {
            qc::cast decimal {*}$args
            return true
        } on error [list error_message options] {
            return false
        }
    }

    proc boolean { string {true t} {false f} } {
        #| Test if the given string can be cast to an boolean.
        try {
            qc::cast boolean $string
            return true
        } on error [list error_message options] {
            return false
        }        
    }

    proc timestamp {string} {
        #| Test if the given string can be cast to a timestamp without timezone.
        try {
            qc::cast timestamp $string
            return true
        } on error [list error_message options] {
            return false
        }        
    }

    proc timestamptz {string} {
        #| Test if the given string can be cast to a timestamp with timezone.
        try {
            qc::cast timestamptz $string
            return true
        } on error [list error_message options] {
            return false
        }        
    }

    proc char {length string} {
        #| Test if the given string can be cast to a fixed length string.
        try {
            qc::cast char $length $string
            return true
        } on error [list error_message options] {
            return false
        }        
    }

    proc varchar {length string} {
        #| Test if the given string can be cast to a varchar of given length.
        try {
            qc::cast varchar $length $string
            return true
        } on error [list error_message options] {
            return false
        }        
    }

    proc enumeration {name value} {
        #| Test if the given value can be cast to an enumeration value in $name.
        try {
            qc::cast enumeration $name $value
            return true
        } on error [list error_message options] {
            return false
        }
    }

    proc text {string} {
        #| Test if the given string can be cast to text.
        try {
            qc::cast text $string
            return true
        } on error [list error_message options] {
            return false
        }
    }

    proc domain {domain_name value} {
        #| Test if the given value can be cast to domain $domain_name.
        try {
            qc::cast domain $domain_name $value
            return true
        } on error [list error_message options] {
            return false
        }
    }

    proc safe_html {text} {
        #| Test if the given text can be cast to safe html.
        return [qc::is safe_html $text]
    }

    proc safe_markdown {text} {
        #| Test if the given text can be cast to safe markdown.
        return [qc::is safe_markdown $text]
    }

    proc date {string} {
        #| Test if the given string can be cast to a date.
        try {
            qc::cast date $string
            return true
        } on error [list error_message options] {
            return false
        }
    }

    proc time {string} {
        #| Test if the given string can be cast to a time.
        try {
            qc::cast time $string
            return true
        } on error [list error_message options] {
            return false
        }
    }

    proc postcode {string} {
        #| Test if the given string can be cast to a UK postcode.
        try {
            qc::cast postcode $string
            return true
        } on error [list error_message options] {
            return false
        }
    }

    proc creditcard {string} {
        #| Test if the given string can be cast to a credit card number.
        try {
            qc::cast creditcard $string
            return true
        } on error [list error_message options] {
            return false
        }
    }

    proc period {string} {
        #| Test if the given string can be cast to a period.
        try {
            qc::cast period $string
            return true
        } on error [list error_message options] {
            return false
        }
    }

    proc url {string} {
        #| Test if the given string can be cast to an url.
        #| (See qc::is url)
        try {
            qc::cast url $string
            return true
        } on error [list error_message options] {
            return false
        }
    }

    proc relative_url {string} {
        #| Test if the given string can be cast to a relative url.
        #| (See qc::is url)
        try {
            qc::cast relative_url $string
            return true
        } on error [list error_message options] {
            return false
        }
    }
    
    proc url_path {string} {
	#| Test if the given string can be cast to an url path
	#| (See qc::is url_path)
	try {
	    qc::cast url_path $string
	    return true
	} on error [list error_message options] {
	    return false
	}
    }

    proc s3_uri {string} {
        #| Test if the given string can be cast to an s3 uri
        #| (See also qc::is s3_uri)
        try {
	    qc::cast s3_uri $string
	    return true
	} on error [list error_message options] {
	    return false
	}
    }
    
    proc interval {string} {
        #| Test if the given string can be cast to an interval
	try {
	    qc::cast interval $string
	    return true
	} on error [list error_message options] {
	    return false
	}        
    }

    proc next_url {string} {
        try {
            qc::cast next_url $string
            return true
        } on error [list error_message options] {
            return false
        }
    }
}

namespace eval qc {
    namespace export round round_up rshift10 intplaces add sum subtr mult divide exp2string base frombase mantissa_exponent sigfigs sigfigs_ceil

    # Import Math functions
    namespace import ::tcl::mathfunc::max
    namespace import ::tcl::mathfunc::min
}
package require math::decimal

proc qc::round { value dec_places } {
    #| Perform rounding of $value to $dec_places places.
    #| Handles exponentials.
    #| Rounds up on 5
    #| e.g. 2.345 -> 2.35
    # TODO Susceptible to buffer overflow for large values
    
    if { [string first e $value]!=-1 || [string first E $value]!=-1 } {
	set value [exp2string $value]
    }
    set k [string first . $value]
    set l [string length $value]
    
    if { $k==-1 } {
	if { $dec_places == 0 } {
	    return $value
	} else {
	    return "$value.[string repeat 0 $dec_places]"
	}
    } 
    if { $l - $k -1 < $dec_places } {
	return "$value[string repeat 0 [expr {$dec_places-$l+$k+1}]]"
    }
    if { $l - $k -1 == $dec_places } {
	return $value
    }

    if { $value < 0 } {
	set int "[string range $value 1 [expr {$k-1}]][string range $value [expr {$k+1}] [expr {$k+1+$dec_places}]]"
    } else {
	set int "[string range $value 0 [expr {$k-1}]][string range $value [expr {$k+1}] [expr {$k+1+$dec_places}]]"
    }

    if { $int != 0 } {
	set int [string trimleft $int 0]
    }

    if { wide($int)%wide(10) >= wide(5) } {
	# round up
	set result [expr {wide($int)/wide(10) + wide(1)}]
    } else {
	set result [expr {wide($int)/wide(10)}]
    }
    if { $value < 0 } {
	return "-[rshift10 $result $dec_places]"
    } else {
	return [rshift10 $result $dec_places]
    }
}

proc qc::round_up { value trunc_places } {
    #| Round up to the nearest $trunc_places decimal places
    set value [expr {double(ceil($value*pow(10,$trunc_places)))/pow(10,$trunc_places)}]
    return [format "%.${trunc_places}f" $value]
}

proc qc::rshift10 {int dec_places} {
    #| Format right aligned padding left
    set s [string length $int]
    if { $int == 0 } {
	if { $dec_places > 0 } {
	    return "0.[string repeat 0 $dec_places]"
	} else {
	    return 0
	}
    }
    if { $dec_places == 0 } {
	return $int
    }
    if { $s < $dec_places } {
	return "0.[string repeat 0 [expr {$dec_places - $s}]]$int"
    }
    if { $s == $dec_places } {
	return "0.$int"
    }
    if { $s > $dec_places } {
	return "[string range $int 0 [expr {$s - $dec_places-1}]].[string range $int [expr {$s - $dec_places}] end]"
    }
}

proc qc::intplaces { number } {
    #| Shift the decimal point n places to the right until $number is an int. Return int and n.
    set k  [string first . $number]
    if { $k ==-1 } {
	set dec_places 0
    } else {
	set dec_places [expr {[string length $number] - $k - 1}]
	set number "[string range $number 0 [expr {$k-1}]][string range $number [expr {$k+1}] end]"
    }

    if { $number!=0 } {
	# Prevent octal interpretation
	set number [string trimleft $number 0]
    } else {
	set number 0
    }
    return [list $number $dec_places]
}

proc qc::add { n1 n2 } {
    #| Adds 2 numbers using decimal arithmetic (::math::decimal).
    set n1 [::math::decimal::fromstr $n1]
    set n2 [::math::decimal::fromstr $n2]
    return [::math::decimal::tostr [::math::decimal::add $n1 $n2]]
}

proc qc::sum { sum args } {
    #| Returns the sum of list of number using decimal arithmetic (::math::decimal).
    set sum [::math::decimal::fromstr $sum]
    foreach arg $args {
	set sum [::math::decimal::add $sum [::math::decimal::fromstr $arg]]
    }
    return [::math::decimal::tostr $sum]
}

proc qc::subtr { n1 n2 } {
    #| Subtracts 2 numbers using decimal arithmetic (::math::decimal).
    set n1 [::math::decimal::fromstr $n1]
    set n2 [::math::decimal::fromstr $n2]
    return [::math::decimal::tostr [::math::decimal::subtract $n1 $n2]]
}

proc qc::mult { n1 n2 } {
    #| Multiplies 2 numbers using decimal arithmetic (::math::decimal).
    set n1 [::math::decimal::fromstr $n1]
    set n2 [::math::decimal::fromstr $n2]
    return [::math::decimal::tostr [::math::decimal::multiply $n1 $n2]]
}

proc qc::divide { n1 n2 } {
    #| Divides 2 numbers using decimal arithmetic (::math::decimal).
    if { $n2 == 0 } {
        error "divide by zero"
    }
    set n1 [::math::decimal::fromstr $n1]
    set n2 [::math::decimal::fromstr $n2]
    return [::math::decimal::tostr [::math::decimal::divide $n1 $n2]]
}

proc qc::exp2string { number } {
    #| Convert floats including exponentials to strings
    set number [string tolower $number]
    set k  [string first e $number]
    if { $k == -1 } {
	return $number
    }
    if { [eq [string index $number 0] -] } {
	set mantissa [string range $number 1 [add $k -1]]
	set sign -
    } else {
	set mantissa [string range $number 0 [add $k -1]]
	set sign ""
    }
    set exp [qc::cast integer [string range $number [add $k 1] end]]
    lassign [intplaces $mantissa] int places
    set exp [expr {$exp - $places}]
    if { $exp > 0 } {
	return $sign[append int [string repeat 0 $exp]]
    } else {
	return $sign[rshift10 $int [mult -1 $exp]]
    }
}

proc qc::base {base number} {
    #| Convert supplied number to specified base 
    set negative [regexp ^-(.+) $number -> number] ;# (1)
    set digits {0 1 2 3 4 5 6 7 8 9 A B C D E F G H I J K L M N
        O P Q R S T U V W X Y Z a b c d e f g h i j k l m n o p
        q r s t u v w x y z}

    set res {}
    while {$number} {
        set digit [expr {$number % $base}]
        set res [lindex $digits $digit]$res
        set number [expr {$number / $base}]
    }
    if $negative {set res -$res}
    set res
 }

proc qc::frombase {base number} {
    #| Converts a number from the specified base to base 10
    set digits {0 1 2 3 4 5 6 7 8 9 A B C D E F G H I J K L M N
        O P Q R S T U V W X Y Z a b c d e f g h i j k l m n o p
        q r s t u v w x y z}
    set negative [regexp ^-(.+) $number -> number]
    set res 0
    foreach digit [split $number ""] {
        set decimalvalue [lsearch $digits $digit]
        if {$decimalvalue<0 || $decimalvalue >= $base} {
            error "bad digit $decimalvalue for base $base"
        }
        set res [expr {$res*$base + $decimalvalue}]
    }
    if $negative {set res -$res}
    set res
}

proc qc::mantissa_exponent {x} {
    #| Returns the mantissa and exponent which would be used to represent x
    if { $x==0 } { return [list 0 0] }
    set m $x;set e 0
    while { abs($m) >=10 } { incr e; set m [expr {double($m)/10}] }
    while { abs($m)<1.0 } { incr e -1; set m [expr {$m*10}] }
    return [list $m $e]
}

proc qc::sigfigs {x n} {
    #| Returns x to n significant figures
    if { $x==0 } { return 0 }
    lassign [qc::mantissa_exponent $x] m e
    set p [expr {pow(10,$e-$n+1)}]
    if { $n-$e-1>0 } {
	return [qc::round [expr {round(double($x)/$p)*$p}] [expr {$n-$e-1}]]
    } else {
	return [qc::round [expr {round(double($x)/$p)*$p}] 0]
    }
}

proc qc::sigfigs_ceil {x n} {
    #| Returns x to n significant figures rounding up
    if { $x==0 } { return 0 }
    lassign [qc::mantissa_exponent $x] m e
    set p [expr {pow(10,$e-$n+1)}]
    if { $n-$e-1>0 } {
	return [qc::round [expr {ceil(double($x)/$p)*$p}] [expr {$n-$e-1}]]
    } else {
	return [qc::round [expr {ceil(double($x)/$p)*$p}] 0]
    }
}


namespace eval qc {
    namespace export \
        image_resize_task_add
}

proc qc::image_resize_task_add {args} {
    #| Add a new resize task for an image in the database.

    qc::args $args {*}{
        -autocrop
        -mime_type "*/*"
        -cache_dir ""
        --
        file_id
        width
        height
    }

    default autocrop false

    if { $cache_dir eq "" } {
        set cache_dir [ns_pagepath]/image
    }

    if { $mime_type eq "*/*" } {
        db_1row {
            select
            mime_type

            from
            file

            where
            file_id=:file_id
        }
    }

    db_trans {
        qc::db_advisory_trans_lock image_resize_task add

        db_0or1row {
            select
            row_id

            from
            image_resize_task

            where
            file_id=:file_id
            and cache_dir=:cache_dir
            and autocrop=:autocrop
            and task_state = 'QUEUED'
            and mime_type=:mime_type
            and (
                 (
                  width=:width
                  and height<=:height
                 )
                 or (
                     width<=:width
                     and height=:height
                 )
            )
        } {
            # Image hasn't been queued at this height and width for the cache directory.
            set row_id [qc::db_seq "row_id_seq"]
            set task_state "QUEUED"
            set date_added [qc::cast timestamptz "now"]

            db_dml "
                insert into
                image_resize_task
                [qc::sql_insert {*}{
                    row_id
                    file_id
                    cache_dir
                    height
                    width
                    date_added
                    task_state
                    autocrop
                    mime_type
                }]
            "
        }
    }

    return $row_id
}

namespace eval qc {
    namespace export schema_update
}

proc qc::schema_update {version code} {
    #| Run the code if it applies to the current schema in order to bring the schema up to the next version.
    db_trans {
	db_1row {select version as current_version from schema for update}
	if { $current_version == $version } {
	    log "Updating schema version $version ..."
	    uplevel 1 $code
	    incr version
	    db_dml {update schema set version=:version}
	    log "Schema updated to version $version"
            # Clear db_introspection caches
            qc::db_cache_clear
            if { [info commands "ns_memoize_flush"] eq "ns_memoize_flush" } {
                ns_memoize_flush db_*
            }
	}
    }
}


namespace eval qc {
}

proc qc::_image_filesystem_cache_glob {args} {
    #| Return a list of files on the filesystem cache that match the given
    #| file_id, and match at least one size contraint exactly
    qc::args $args {*}{
        -autocrop
        -mime_type */*
        --
        cache_dir
        file_id
        max_width
        max_height
    }
    default autocrop false
    if { $mime_type eq "*/*" } {
        set extension ".*"
    } else {
        # Construct a case-insensitive glob pattern
        set extension ""
        foreach char [split [qc::mime_file_extension $mime_type] ""] {
            if { [string is alpha $char] } {
                append extension \[
                append extension [string tolower $char]
                append extension [string toupper $char]
                append extension \]
            } else {
                append extension $char
            }
        }
    }
    if { $autocrop } {
        set glob_patterns \
            [list \
                 "${file_id}/autocrop/${max_width}x*/${file_id}${extension}" \
                 "${file_id}/autocrop/*x${max_height}/${file_id}${extension}"]

    } else {
        set glob_patterns \
            [list \
                 "${file_id}-${max_width}x*/${file_id}${extension}" \
                 "${file_id}-*x${max_height}/${file_id}${extension}"]
    }
    
    set data [list]
    set glob_options [list -nocomplain -types f -directory $cache_dir]
    set return_data [list]
    return [glob {*}$glob_options {*}$glob_patterns]
}

proc qc::_image_filesystem_cache_glob2 {args} {
    #| Return a list of files on the filesystem cache that match the given
    #| file_id, mime_type and match at least one size constraint exactly
    qc::args $args {*}{
        -autocrop
        --
        cache_dir
        file_id
        mime_type
        max_width
        max_height
    }
    default autocrop false
    
    # Construct a case-insensitive glob pattern
    set extension ""
    foreach char [split [qc::mime_file_extension $mime_type] ""] {
        if { [string is alpha $char] } {
            append extension \[
            append extension [string tolower $char]
            append extension [string toupper $char]
            append extension \]
        } else {
            append extension $char
        }
    }
    
    if { $autocrop } {
        set glob_patterns \
            [list \
                 "${file_id}/autocrop/${max_width}x*/${file_id}${extension}" \
                 "${file_id}/autocrop/*x${max_height}/${file_id}${extension}"]

    } else {
        set glob_patterns \
            [list \
                 "${file_id}-${max_width}x*/${file_id}${extension}" \
                 "${file_id}-*x${max_height}/${file_id}${extension}"]
    }
    
    set data [list]
    set glob_options [list -nocomplain -types f -directory $cache_dir]
    set return_data [list]
    return [glob {*}$glob_options {*}$glob_patterns]
}

proc qc::_image_filesystem_cache_file2dimensions {file} {
    #| Extract the width and height from a filesystem cache path
    set expression {[^0-9]([0-9]+)x([0-9]+)[^0-9]}
    regexp $expression $file -> width height
    return [list $width $height]
}

namespace eval qc {
    namespace export html_table_list
}

proc qc::html_table_list {args} {
    # DEPRECATED - only used in mla. This file to be deleted. 
    set varNames [qc::args2vars $args]
    default height 120
    default id dbGridList

    append html [html_tag div class scroll style "height:${height}px;border-top:3px solid #ece9d8;border-left:3px solid #ece9d8;border-right:3px solid #ece9d8;"]
    set class "db-list grid"
    lappend varNames class id
    append html [qc::html_table [dict_from {*}$varNames]]
    append html "</div>\n"
    return $html
}

namespace eval qc {
    namespace export lock
}

proc qc::lock {lock_id timeout code} {
    #| Wait up to $timeout seconds to obtain a lock and then
    #| execute code 
    # TODO Aolserver only
    while { [nsv_incr lock $lock_id] != 1 } {
	if { $timeout>0 } {
	    ns_sleep 1
	    incr timeout -1
	} else {
	    error "Timout waiting for lock on $lock_id" {} TIMEOUT
	}
    }
    set return_code [ catch { uplevel 1 $code } result options ]
    nsv_unset lock $lock_id
    # Preserve TCL_RETURN
    if { $return_code == 2 && [dict get $options -code] == 0 } {
        dict set options -code return
    } else {
        dict incr options -level
    }
    return -options $options $result
}

namespace eval qc {
    namespace export ldicts_diff_any
}

proc qc::ldicts_diff_any {left_ldict right_ldict} {
    #| Return the first diff
    set index 0
    foreach left_dict $left_ldict right_dict $right_ldict {
        set diff [dicts_diff_any $left_dict $right_dict]
        if { $diff ne "" } {
            return "Index $index: $diff"
        }
        incr index
    }

    return ""
}

namespace eval qc {
    namespace export args2dict args2vars args_check_required args_definition_split args_split args
}

proc qc::args2dict {callers_args} {
    #| Parse callers args. Interpret as regular dict unless first item is ~ 
    #| in which case interpret as a list of variable names to pass-by-name.
    #| Return dict of resulting name value pairs.

    ### This line below has got to go but will break things. ####
    if { [llength $callers_args]==1 } {set callers_args [lindex $callers_args 0]}
    ##################
    if { [eq [lindex $callers_args 0] ~] } {
	# Pass by Name
	set dict {}
	foreach varName [lrange $callers_args 1 end] {
	    if { [uplevel 2 info exists $varName] } {
		lappend dict $varName [upset 2 $varName]
	    }
	}
	return $dict
    } else {
        if { [llength $callers_args] % 2 != 0 } {
            error "Missing name or value in args for qc::args2dict"
        }
        
        set dict [dict create {*}$callers_args]
        if { [llength $callers_args] ne [llength $dict] } {
            error "Duplicate key for qc::args2dict"
        }
	return $dict
    }
}

proc qc::args2vars {callers_args args} {
    #| Parse callers args. Interpret as regular dict unless first item is ~ 
    #| in which case interpret as a list of variable names to pass-by-name.
    #| Dict - set all variables or just those specified that exists in the dict
    #| Pass-by-Value - set all variables or just those specified that exists in caller's namespace.

    if { [llength $callers_args]==1 } {set callers_args [lindex $callers_args 0]}
    set varNames {}

    if { [eq [lindex $callers_args 0] ~] } {
	# Pass by Name
	foreach varName [lrange $callers_args 1 end] {
	    if { [uplevel 2 info exists $varName] && ([llength $args]==0 || [in $args $varName]) } {
		upset 1 $varName [upset 2 $varName]
		lappend varNames $varName
	    }
	}
    } else {
	foreach {varName varValue} $callers_args {
            if { [llength $args]==0 || [in $args $varName] } {
		upset 1 $varName $varValue
		lappend varNames $varName
	    }
	}
    }
    return $varNames
}

proc qc::args_check_required {callers_args args} {
    #| Assume callers_args is a dict of name value pairs
    #| check that all the keys given exist.

    foreach arg $args {
	if { ![dict exists $callers_args $arg] } {
	    error "Missing value for arg \"$arg\" when calling \"[info level -1]\""
	}
    }
}

proc qc::args_definition_split {def_args} {
    #| Return three lists for switches, option-default pairs and other args
    set switches {}
    set options {}
    set others {}
    set index 0
    while {$index<[llength $def_args]} {
        set arg [lindex $def_args $index]
        set next [lindex $def_args [expr {$index+1}]]
        if { [eq $arg "--"] } {
            # -- indicates the end of switches and options, all remaining args are value arguments
            incr index 1
            set others [lrange $def_args $index end]
            break
        } elseif { [eq [string index $arg 0] "-"] && [eq [string index $next 0] "-"] } {
            # switch
            lappend switches [string range $arg 1 end]
            incr index 1
        } elseif { [eq [string index $arg 0] "-"] } {
            # option
            lappend options [string range $arg 1 end] $next 
            incr index 2
        } else {
            # value argument
            lappend others $arg
            incr index 1
        }
    }
    return [list $switches $options $others]
}

proc qc::args_split {caller_args switch_names option_names} {
    #| Return three lists for switches, options pairs and other args
    set caller_switches {}
    set caller_options {}
    set caller_others {}

    # If this proc takes switches/options and the first arg might be a switch or option
    if { [eq [string index [lindex $caller_args 0] 0] "-"] && ([llength $switch_names] > 0 || [llength $option_names] > 0) } {
        set index 0
        while {$index<[llength $caller_args]} {
            set arg [lindex $caller_args $index]
            set next [lindex $caller_args [expr {$index+1}]]
            if { [eq $arg "--"] } {
                # -- indicates the end of switches and options - all that remain are regular args
                incr index 1
                set caller_others [lrange $caller_args $index end]
                break
            } elseif { [eq [string index $arg 0] "-"] && [in $switch_names [string range $arg 1 end]] } {
                # switch
                lappend caller_switches [string range $arg 1 end]
                incr index 1
            } elseif { [eq [string index $arg 0] "-"] && [in $option_names [string range $arg 1 end]] } {
                # option
                lappend caller_options [string range $arg 1 end] $next 
                incr index 2
            } else {
                # Once a value argument has been found, all that remain must be value args
                set caller_others [lrange $caller_args $index end]
                incr index 1
                break
            }
        }
    } else {
        set caller_others $caller_args
    }
    return [list $caller_switches $caller_options $caller_others]
}

proc qc::args {callers_args args} {
    #| Assign caller arguments to variables as specified
    lassign [args_definition_split $args] switches options others
    lassign [args_split $callers_args $switches [dict keys $options]] callers_switches callers_options callers_others

    if { [llength $args]==0 } {
	foreach switch $callers_switches {
	    upset 1 $switch true
	}
	foreach {name value} $callers_options {
	     upset 1 $name $value
	}
	upset 1  args $callers_others
    } else {
	# prescribed
	# Usage Switches
	foreach switch $callers_switches {
	    upset 1 $switch true
	}
	# Options
	foreach {name defaultValue} $options {
	    if { [dict exists $callers_options $name] } {
		upset 1 $name [dict get $callers_options $name]
	    } elseif { [eq $defaultValue *] } {
		# Error - missing required option
		error "Missing required option for \"$name\""
	    } elseif { [eq $defaultValue ?] } {
		# Optional switch with no default
	    } else {
		# set default option value
		upset 1 $name $defaultValue
	    }
	}
	
	# Listed Arguments
	if { ![in $others args] && [llength $callers_others]>[llength $others] } {
	    # Usage Error - too many values
	    error "Too many values; expected [llength $others] but got [llength $callers_others] in \"$callers_args\""
	}
	
	set index 0
	foreach name $others {
	    if { [eq $name args] } {
		break 
	    } elseif { $index>([llength $callers_others]-1)} {
		if { [llength $name]==2 } {
		    # default value
		    upset 1 [lindex $name 0] [lindex $name 1]
		} else {
		    # Usage Error - too few values
		    error "Too few values"
		}
	    } else {
		upset 1 [lindex $name 0] [lindex $callers_others $index]
	    }
	    incr index
	}
	upset 1 args [lrange $callers_others $index end]
    }
}


namespace eval qc {
    namespace export widget widget_*
}

proc qc::widget {args} {
    #| Look for a proc "widget_$type" to make the widget
    set type [dict get $args type]
    # Check the qc namespace for a widget proc to create the widget type.
    if { [eq [info procs "::qc::widget_$type"] "::qc::widget_$type"] } {
	return ["::qc::widget_$type" {*}$args]
    }

    # No widget proc found in the qc namespace for the type.
    # Check the global namespace for a proc to create the widget type.
    if { [eq [info procs "::widget_$type"] "::widget_$type"] } {
	return ["::widget_$type" {*}$args]
    }
    error "No widget proc defined for $type"
}

proc qc::widget_label { args } {
    #| Return an HTML form label element.
    args_check_required $args name label 
    array set this $args
    default this(id) $this(name)
    set this(for) $this(id)
    if { [info exists this(tooltip)] } {
        set this(title) $this(tooltip)
    }

    set attributes [dict_subset [array get this] for title]
    set html $this(label)
    if {[info exists this(required)] && [string is true $this(required)] } {
        dict set attributes class required
	append html [html span * style "color:#CC0000;"]
    } 

    return [html label $html {*}$attributes]
}

proc qc::widget_text { args } {
    #| Return an HTML form text input widget.
    args_check_required $args name value 

    array set this $args
    set this(type) text
    default this(id) $this(name)
    default this(width) 160
    set this(style) [qc::style_set [coalesce this(style) ""] width $this(width)]
    if { [info exists this(tooltip)] } {
        set this(title) $this(tooltip)
    }

    if { [info exists this(sticky)] && [true $this(sticky)] } {
        # sticky values
        if { [info exists this(sticky_url)] && [sticky_exists -url $this(sticky_url) $this(name)] } {
            set this(value) [sticky_get -url $this(sticky_url) $this(name)]
        } elseif { [sticky_exists $this(name)] } {
            set this(value) [sticky_get $this(name)]
        }
    }

    if { [info exists this(disabled)] && [string is true $this(disabled)] } {
        set this(type) hidden
        set html [html span $this(value) {*}[dict_subset [array get this] title]]
        append html [html_tag input {*}[dict_subset [array get this] type name value id]]
    } else {
	set html [html_tag input {*}[dict_exclude [array get this] required label width units tooltip]]
    }

    return $html
}

proc qc::widget_compare { args } {
    #| Return an HTML form widget with an operator drop down and input box.
    args_check_required $args name value 
    array set this $args
    default this(operator) =
    default this(options) [list "greater than" ">" equals = "less than" "<"]
    default this(sticky) no
    if { [info exists this(tooltip)] } {
        set this(title) $this(tooltip)
    }

    set html [qc::widget_select name $this(name)_op type select value $this(operator) {*}[dict_subset [array get this] options sticky tooltip title]]
    append html " "
    append html [qc::widget_text {*}[qc::dict_exclude [array get this] required operator options sticky]]

    return $html
}

proc qc::widget_combo { args } {
    #| Return an DHTML form widget with text input and dropdown for completion.
    array set this $args
    args_check_required $args name value searchURL
    default this(searchLimit) 10
    set this(class) "db-form-combo"
    set this(AUTOCOMPLETE) off
    set html [qc::widget_text {*}[array get this]]
    if {[info exists this(boundName)]} {
        append html [html_tag input type hidden name $this(boundName) value $this(boundValue)]
    }
    return $html
}

proc qc::widget_htmlarea { args } {
    #| 	Return an HTML form htmlarea widget made from an editable div tag.
    array set this $args
    if { [info exists this(sticky)] && [true $this(sticky)] } {
        # sticky values
        if { [info exists this(sticky_url)] && [sticky_exists -url $this(sticky_url) $this(name)] } {
            set this(value) [sticky_get -url $this(sticky_url) $this(name)]
        } elseif { [sticky_exists $this(name)] } {
            set this(value) [sticky_get $this(name)]
        }
    }
    args_check_required $args name value 
    default this(width) 160
    default this(height) 100
    set this(style) [qc::style_set [coalesce this(style) ""] width $this(width) height $this(height)]

    if { [info exists this(name)] } { 
	default this(id) $this(name)
    }
    set this(contentEditable) true
    set this(class) "db-form-html-area"
    if { [info exists this(tooltip)] } {
        set this(title) $this(tooltip)
    }
    return [html div $this(value) {*}[qc::dict_exclude [array get this] label required units width height tooltip]]
}

proc qc::widget_textarea { args } {
    #| Return an HTML form textarea element.
    array set this $args
    if { [info exists this(sticky)] && [true $this(sticky)] } {
        # sticky values
        if { [info exists this(sticky_url)] && [sticky_exists -url $this(sticky_url) $this(name)] } {
            set this(value) [sticky_get -url $this(sticky_url) $this(name)]
        } elseif { [sticky_exists $this(name)] } {
            set this(value) [sticky_get $this(name)]
        }
    }
    args_check_required $args name value 
    default this(width) 160
    default this(height) 100
    set this(style) [qc::style_set [coalesce this(style) ""] width $this(width) height $this(height)]
    if { [info exists this(name)] } { 
	default this(id) $this(name)
    }
    if { [info exists this(tooltip)] } {
        set this(title) $this(tooltip)
    }
    set html [html textarea $this(value) {*}[qc::dict_exclude [array get this] type value label required units width height tooltip]]
}

proc qc::widget_select { args } {
    #| 	Return an HTML form dropdown list.
    args_check_required $args name value options
    set name_selected "-"
    array set this $args
    if { [info exists this(sticky)] && [true $this(sticky)] } {
        # sticky values
        if { [info exists this(sticky_url)] && [sticky_exists -url $this(sticky_url) $this(name)] } {
            set this(value) [sticky_get -url $this(sticky_url) $this(name)]
        } elseif { [sticky_exists $this(name)] } {
            set this(value) [sticky_get $this(name)]
        }
    }
    default this(null_option) no
    if { [info exists this(name)] } { 
	default this(id) $this(name)
    }
    if { [info exists this(tooltip)] } {
        set this(title) $this(tooltip)
    }

    set options_html {}
    if { [string is true $this(null_option)] } {
	lappend options_html "<option value=\"\">- Select -</option>"
    }
    foreach {name value} $this(options) {
	if { [string equal $name $this(value)] || [string equal $value $this(value)] } {
	    set name_selected $name
	    lappend options_html [html option $name value $value selected true]
	} else {
	    lappend options_html [html option $name value $value] 
	}
    }

    if { [info exists this(disabled)] && [string is true $this(disabled)] } {
        set html [html span $name_selected {*}[dict_subset [array get this] title]]
        append html [html_tag input type hidden {*}[dict_subset [array get this] name value id]]
    } else {
        set html [html select [join $options_html \n] {*}[qc::dict_exclude [array get this] required label type options null_option value units tooltip]]
    }

    return $html
}

proc qc::widget_datalist { args } {
    #| 	Return an HTML datalist.
    args_check_required $args name value options
    set name_selected "-"
    array set this $args
    if { [info exists this(sticky)] && [true $this(sticky)] } {
        # sticky values
        if { [info exists this(sticky_url)] && [sticky_exists -url $this(sticky_url) $this(name)] } {
            set this(value) [sticky_get -url $this(sticky_url) $this(name)]
        } elseif { [sticky_exists $this(name)] } {
            set this(value) [sticky_get $this(name)]
        }
    }
    default this(null_option) no
    if { [info exists this(name)] } { 
	default this(id) $this(name)
    }
    if { [info exists this(tooltip)] } {
        set this(title) $this(tooltip)
    }

    default this(style) ""
    if { [info exists this(width)] } {
        set this(style) [qc::style_set $this(style) width $this(width)]
    }
    
    set options_html [list]
    foreach {name value} $this(options) {
	if { [string equal $name $this(value)] || [string equal $value $this(value)] } {
	    set name_selected $name
	    lappend options_html [html option $name value $value selected true]
	} else {
	    lappend options_html [html option $name value $value] 
	}
    }

    if { [info exists this(disabled)] && [string is true $this(disabled)] } {
        set html [html span $name_selected {*}[dict_subset [array get this] title]]
        append html [html_tag input type hidden {*}[dict_subset [array get this] name value id]]
    } else {
        set html [h input \
                      {*}[qc::dict_exclude [array get this] required label type options null_option value units tooltip width] \
                      list $this(id)_datalist \
                     ]
        append html [h datalist \
                         id $this(id)_datalist \
                         [join $options_html \n] \
                        ]
    }

    return $html
}

proc qc::widget_span { args } {
    #| Return a span element showing the value of the widget.
    args_check_required $args name value 
    array set this $args
    
    if { [info exists this(name)] } { 
	default this(id) $this(name)
    }
    # height and width
    if { [info exists this(width)] } {
	set this(style) [qc::style_set [coalesce this(style) ""] width $this(width)]
    }
    if { [info exists this(height)] } {
	set this(style) [qc::style_set [coalesce this(style) ""] height $this(height)]
    }
    if { [info exists this(tooltip)] } {
        set this(title) $this(tooltip)
    }
    return [html span $this(value) {*}[qc::dict_exclude [array get this] label type value width height tooltip]] 
}

proc qc::widget_password { args } {
    #| Return an HTML form, password input widget.
    args_check_required $args name value
    array set this $args
    if { [info exists this(sticky)] && [true $this(sticky)] } {
        # sticky values
        if { [info exists this(sticky_url)] && [sticky_exists -url $this(sticky_url) $this(name)] } {
            set this(value) [sticky_get -url $this(sticky_url) $this(name)]
        } elseif { [sticky_exists $this(name)] } {
            set this(value) [sticky_get $this(name)]
        }
    }
    set this(type) password
    if { [info exists this(name)] } { 
	default this(id) $this(name)
    }
    default this(width) 160
    set this(style) [qc::style_set [coalesce this(style) ""] width $this(width)]
    if { [info exists this(tooltip)] } {
        set this(title) $this(tooltip)
    }
    return [html_tag input {*}[qc::dict_exclude [array get this] label width height units tooltip]]
}

proc qc::widget_bool { args } {
    args_check_required $args name value
    array set this $args
    if { [info exists this(sticky)] && [true $this(sticky)] } {
        # sticky values
        if { [info exists this(sticky_url)] && [sticky_exists -url $this(sticky_url) $this(name)] } {
            set this(value) [sticky_get -url $this(sticky_url) $this(name)]
        } elseif { [sticky_exists $this(name)] } {
            set this(value) [sticky_get $this(name)]
        }
    }
    if { [info exists this(name)] } { 
	default this(id) $this(name)
    }
    set this(checked) [true $this(value)]
    # A boolean checkbox ALWAYS has a value of true
    set this(type) checkbox
    set this(value) true
    set this(boolean) true
    if { [info exists this(tooltip)] } {
        set this(title) $this(tooltip)
    }
    return [html_tag input {*}[qc::dict_exclude [array get this] label width height units tooltip]]
}

proc qc::widget_checkbox { args } {
    array set this $args
    if { [info exists this(name)] } { 
	default this(id) $this(name)
    }
    set this(type) checkbox
    if { [info exists this(tooltip)] } {
        set this(title) $this(tooltip)
    }
    return [html_tag input {*}[qc::dict_exclude [array get this] label width height units tooltip]]
}

proc qc::widget_button { args } {
    #| Return an HTML form, button
    args_check_required $args name value
    array set this $args
    set this(type) button
    if { [info exists this(name)] } { 
	default this(id) $this(name)
    }
    if { [info exists this(tooltip)] } {
        set this(title) $this(tooltip)
    }
    return [html_tag input {*}[qc::dict_exclude [array get this] label width height units tooltip]]
}

proc qc::widget_submit { args } {
    #| Return an HTML form, submit button
    args_check_required $args name value
    array set this $args
    if { [info exists this(name)] } { 
	default this(id) $this(name)
    }
    set this(type) submit
    if { [info exists this(tooltip)] } {
        set this(title) $this(tooltip)
    }
    return [html_tag input {*}[qc::dict_exclude [array get this] label width height units tooltip]] 
}

proc qc::widget_radio { args } {
    #| Return an HTML form, radio button
    args_check_required $args name value
    array set this $args
    set this(type) radio
    if { [info exists this(name)] } { 
	default this(id) $this(name)
    }
    default this(checked) false
    if { [info exists this(tooltip)] } {
        set this(title) $this(tooltip)
    }
    return [html_tag input {*}[qc::dict_exclude [array get this] label width height units tooltip]]
}

proc qc::widget_radiogroup { args } {
    args_check_required $args name value options
    array set this $args
    if { [info exists this(sticky)] && [true $this(sticky)] } {
        # sticky values
        if { [info exists this(sticky_url)] && [sticky_exists -url $this(sticky_url) $this(name)] } {
            set this(value) [sticky_get -url $this(sticky_url) $this(name)]
        } elseif { [sticky_exists $this(name)] } {
            set this(value) [sticky_get $this(name)]
        }
    }

    set group_name $this(name)
    set group_value $this(value)
    
    set buttons {}
    foreach {option_name option_value} $this(options) {
	set id ${group_name}${option_value}
	set widget [qc::widget_radio name $group_name value $option_value checked [eq $option_value $group_value] id $id {*}[dict_subset [array get this] title tooltip]]
	set label [qc::widget_label label $option_name name $id {*}[dict_subset [array get this] title tooltip]]
	lappend buttons \
            [h span \
                 style "display:inline-block" \
                 "$widget&nbsp;$label"]
    }
    return [html div [join $buttons "&nbsp; &nbsp;"] class "radio-group" name $group_name id $group_name]
}

proc qc::widget_image_combo { args } {
    #| A widget continaing an image and a combo-input for filenames
    array set this $args
    args_check_required $args name value searchURL imageURL
    default this(searchLimit) 10
    default this(class) "image-combo"
    default this(width) 150
    default this(height) 150
    default this(defaultImage) "/Graphics/noimage.png"
    set style [qc::style_set "" max-width $this(width)px max-height $this(height)px]
    if {$this(value) ne ""} {
        set html [html_tag img src [qc::url $this(imageURL) filename $this(value)] style $style]
    } else {
        set html [html_tag img src $this(defaultImage) style $style]
    }
    append html [qc::widget_text {*}[qc::dict_exclude [array get this] class height type]]
    return [html div $html class $this(class)]
}

namespace eval qc {
    namespace export return2client return_*
}

proc qc::return2client { args } {
    #| Return data to http client.
    # Usage return2client ?code code? ?content-type mime-type? ?html html? ?text text? ?xml xml? ?json json? ?csv csv? ?file file? ?filename filename? ?download boolean? ?filter_cc boolean? ?header header? .. 
    set arg_names [qc::args2vars $args]
    set headers [lexclude $arg_names html xml text json csv file filename download code content-type filter_cc]
    default code 200
    default filter_cc no
    default filename [string trimleft [qc::url_path [ns_conn url]] /]

    # Determine type of payload and configure defaults
    if { [info exists file] } {
        # File
        default download yes
        default content-type [qc::mime_type_guess $filename]
        set var file
    } elseif { [info exists html] } {
        # HTML payload
        default download no
	default content-type "text/html; charset=utf-8"
	set var html
    } elseif { [info exists xml] } {
        # XML payload
        default download no
	default content-type "text/xml; charset=utf-8"
	set var xml
    } elseif { [info exists text] } {
        # Text payload
        default download no
	default content-type "text/plain; charset=utf-8"
	set var text
    } elseif { [info exists csv] } {
        # CSV payload
        default download yes
	default content-type "text/csv; charset=utf-8"
	set var csv
    } elseif { [info exists json] } {
        # JSON payload
        default download no
	default content-type "application/json; charset=utf-8"
	set var json
    } else {
        # Unknown payload
	error "No payload given in html or xml or text or json or file" 
    }

    # Content-Disposition
    if { $download } {
        # Download payload as an attachment
        default content-disposition "attachment; filename=$filename"
    } else {
        # Display payload inline
        default content-disposition "inline; filename=$filename"
    }
    if { "content-disposition" ni $headers } {
        lappend headers content-disposition       
    }    

    # Mask plain text credit card numbers in payload
    if { $filter_cc && $var ne "file" } {
        set $var [qc::format_cc_masked_string [set $var]] 
    }
    
    # Update headers
    foreach name $headers {
        set conn_headers [ns_conn outputheaders]
        ns_set update $conn_headers $name [set $name]
    }
    
    # Return payload to client
    if { $var eq "file" } {
        ns_returnfile $code ${content-type} [set $var]
    } else {
        ns_return $code ${content-type} [set $var]
    }
}

proc qc::return_html { html } { 
    # Deprecated
    return2client html $html
}

proc qc::return_xml { xml } {
    # Deprecated
    return2client xml $xml
}

proc qc::return_csv { text } { 
    return2client content-type "text/csv; charset=utf-8" text $text content-disposition "attachment;"
}

proc qc::return_soap { string } { 
    ns_return 200 "application/soap+xml; charset=utf-8" $string
}

proc qc::return_headers {} {
    set list {}
    lappend list "HTTP/1.0 200 OK"
    lappend list "Date: [ns_httptime [clock seconds]]"
    lappend list "MIME-Version: 1.0"
    lappend list "Content-Type: text/html"
    set output [join $list \r\n]
    append output \r\n\r\n
    ns_write $output
}

proc qc::return_headers_chunked {} {
    set list {}
    lappend list "HTTP/1.1 200 OK"
    lappend list "Date: [ns_httptime [clock seconds]]"
    lappend list "MIME-Version: 1.0"
    lappend list "Content-Type: text/html"
    lappend list "Transfer-Encoding: chunked"
    set output [join $list \r\n]
    append output \r\n\r\n
    ns_write $output
}

proc qc::return_chunks {string} {
    regsub -all {\r\n} $string \n string
    foreach line [split $string \n] {
	qc::return_chunk $line
    }
}

proc qc::return_chunk {string} {
    ns_write [format %X [string bytelength $string]]\r\n$string\r\n
}

proc qc::return_next { next_url } {
    #| Redirect to an internal url
    ns_returnredirect [qc::cast next_url $next_url]
}

proc ns_returnmoved {url} {
    ns_set update [ns_conn outputheaders] Location $url
    ns_return 301 "text/html" [subst \
  {<!DOCTYPE HTML PUBLIC "-//IETF//DTD HTML 2.0//EN">
  <HTML>
  <HEAD>
  <TITLE>Moved</TITLE>
  </HEAD>
  <BODY>
  <H2>Moved</H2>
  <A HREF="$url">The requested URL has moved here.</A>
  <P ALIGN=RIGHT><SMALL><I>[ns_info name]/[ns_info patchlevel] on [ns_conn location]</I></SMALL></P>
  </BODY></HTML>}]
}

proc qc::return_response {args} {
    #| Returns the global data structure to the client.
    #| Content negotiates to try and find a suitable content type.
    qc::args $args -code 200 -response2html qc::response2html -- args
    
    set mime_types [list "text/html" "application/json" "application/xml" "text/xml"]
    set mime_type [qc::http_accept_header_best_mime_type $mime_types]
    set media_type [lindex [split $mime_type "/"] 1]
    if { $media_type eq "" } {
        # Couldn't negotiate an acceptable response type.
        return [qc::return2client code 406 text "Couldn't respond with an acceptable content type. Available content types: [join $mime_types ", "]"]
    }
    
    switch -nocase -- $media_type {
        xml {
            set response [qc::response2xml]
        }
        json {
            set response [qc::response2json]
        }
        * -
        html {
            set media_type html
            set response [$response2html]           
        }
    }

    qc::return2client code $code $media_type $response
}

namespace eval qc {
    namespace export db_trans db_trans_*
}

proc qc::db_trans {args} {
    #| Execute code within a transaction
    #| Rollback on database or tcl error
    #| Nested db_trans structures do not
    #| emulate postgres nested transactions but simply
    #| ensure code is executed in a transaction by
    #| maintaining a global db_trans_level
    args $args -db DEFAULT -- code {error_code ""}

    db_trans_start $db
    
    set return_code [ catch { uplevel 1 $code } result options ]
    
    switch $return_code {
	1 {
	    # Error
            db_trans_abort $db

	    uplevel 1 $error_code
            # Return in parent stack frame instead of here
            dict incr options -level
	}
	default {
	    # ok, return, break, continue
            db_trans_end $db

            # Preserve TCL_RETURN
            if { $return_code == 2 && [dict get $options -code] == 0 } {
                dict set options -code return
            } else {
                # Return in parent stack frame instead of here
                dict incr options -level
            }
	}
    }
    return -options $options $result
}

proc qc::db_trans_start {{db DEFAULT}} {
    #| Start a database transaction or add a savepoint
    global db_trans_level
    if { ![info exists db_trans_level] } {
	set db_trans_level($db) 0
    }
    incr db_trans_level($db)

    set savepoint "db_trans_level_$db_trans_level($db)"

    if { $db_trans_level($db) == 1 } {
	db_dml -db $db "BEGIN WORK"
    } else {
        db_dml -db $db "SAVEPOINT $savepoint"
    }
}

proc qc::db_trans_end {{db DEFAULT}} {
    #| Release a savepoint or commit current transaction
    global db_trans_level
    
    set savepoint "db_trans_level_$db_trans_level($db)"
    
    if { $db_trans_level($db) > 1 } {
        db_dml -db $db "RELEASE SAVEPOINT $savepoint"
    } else {
        db_dml -db $db "COMMIT WORK"
    }
    incr db_trans_level($db) -1
}

proc qc::db_trans_abort {{db DEFAULT}} {
    #| Rollback to latest savepoint or out of transaction
    global db_trans_level
    
    set savepoint "db_trans_level_$db_trans_level($db)"
    
    if { $db_trans_level($db) > 1 } {
        db_dml -db $db "ROLLBACK TO SAVEPOINT $savepoint"
    } else {
        db_dml -db $db "ROLLBACK WORK"
    }
    incr db_trans_level($db) -1
}

namespace eval qc {
    namespace export db_cache_*
}

proc qc::db_cache_1row { args } {
    # Cached equivalent of db_1row
    args $args -ttl ? -- qry 
    if { [info exists ttl] } {
	set table [db_cache_select_table -ttl $ttl $qry 1]
    } else {
	set table [db_cache_select_table $qry 1]
    }
    set db_nrows [expr {[llength $table]-1}]
    
    if { $db_nrows!=1 } {
	error "The qry <code>[db_qry_parse $qry 1]</code> returned $db_nrows rows"
    }
    foreach key [lindex $table 0] value [lindex $table 1] { upset 1 $key $value }
    return
}

proc qc::db_cache_0or1row { args } {
    # Cached equivalent of db_0or1row
    args $args -ttl ? -- qry {no_rows_code ""} {one_row_code ""}
    if { [info exists ttl] } {
	set table [db_cache_select_table -ttl $ttl $qry 1]
    } else {
	set table [db_cache_select_table $qry 1]
    }
    set db_nrows [expr {[llength $table]-1}]

    if {$db_nrows==0} {
	# no rows
	set return_code [ catch { uplevel 1 $no_rows_code } result options ]
        # Preserve TCL_RETURN
	if { $return_code == 2 && [dict get $options -code] == 0 } {
            dict set options -code return
        } else {
            # Return in parent stack frame instead of here
            dict incr options -level
        }
        return -options $options $result
    } elseif { $db_nrows==1 } { 
	# 1 row
	foreach key [lindex $table 0] value [lindex $table 1] { upset 1 $key $value }
	set return_code [ catch { uplevel 1 $one_row_code } result options ]
        # Preserve TCL_RETURN
        if { $return_code == 2 && [dict get $options -code] == 0 } {
            dict set options -code return
        } else {
            # Return in parent stack frame instead of here
            dict incr options -level
        }
        return -options $options $result
    } else {
	# more than 1 row
	error "The qry <code>[db_qry_parse $qry 1]</code> returned $db_nrows rows"
    }
}

proc qc::db_cache_foreach { args } {
    # Cached equivalent of db_foreach
    args $args -ttl ? -- qry foreach_code { no_rows_code ""}

    # save special db variables
    qc::upcopy 1 db_nrows      saved_db_nrows
    qc::upcopy 1 db_row_number saved_db_row_number

    if { [info exists ttl] } {
	set table [db_cache_select_table -ttl $ttl $qry 1]
    } else {
	set table [db_cache_select_table $qry 1]
    }
    set db_nrows [expr {[llength $table]-1}]
    set db_row_number 0

    if { $db_nrows == 0 } {
	upset 1 db_nrows 0
	upset 1 db_row_number 0
	set return_code [ catch { uplevel 1 $no_rows_code } result options ]
        switch $return_code {
            0 {
                # ok
            }
            default {
                # error, return

                # Preserve TCL_RETURN
                if { $return_code == 2 && [dict get $options -code] == 0 } {
                    dict set options -code return
                } else {
                    # Return in parent stack frame instead of here
                    dict incr options -level
                }
                return -options $options $result
            }
        }
    } else {
	set masterkey [lindex $table 0]
	foreach list [lrange $table 1 end] {
	    upset 1 db_nrows $db_nrows
	    upset 1 db_row_number [incr db_row_number]
	    foreach key $masterkey value $list {
		upset 1 $key $value
	    }
            set return_code [ catch { uplevel 1 $foreach_code } result options ]
            switch $return_code {
                0 {
                    # ok
                }
                3 -
                4 {
                    # break, continue
                    return -options $options $result
                }
                default {
                    # error, return

                    # Preserve TCL_RETURN
                    if { $return_code == 2 && [dict get $options -code] == 0 } {
                        dict set options -code return
                    } else {
                        # Return in parent stack frame instead of here
                        dict incr options -level
                    }
                    return -options $options $result
                }
            }

            # Clean up the result variable to prevent Tcl's Copy on Write
            # process from adversely affecting performance
            unset result
	}
    }
    # restore saved variables
    if { [info exists saved_db_nrows] } {
	upset 1 db_nrows      $saved_db_nrows
	upset 1 db_row_number $saved_db_row_number
    }
}

proc qc::db_cache_select_table { args } {
    #| Check if the results of the qry have already been saved.
    #| If never saved (or has expired due to ttl) then run the qry and place the results
    #| as a table in either db_thread_cache global array, or if ttl was specified,
    #| a time limited ns_cache cache.
    args $args -ttl ? -- qry {level 0}
    incr level
    set hash [qc::md5 [db_qry_parse $qry $level]]

    # Use global array or ns_cache with ttl?
    if { [info exists ttl] && [info procs ::ns_cache] ne "" } {
        # Use ns_cache
        # Create the cache if it doesn't exist yet
        if { ! [in [ns_cache_names] db] } {
            # DB cache size
            set param_name db_cache_size
            db_0or1row {select param_value from param where param_name=:param_name} {
                # DB param does not exist
                if { [set param_value [ns_config ns/server/[ns_info server] $param_name]] ne "" } {
                    # naviserver config param exists
                    set db_cache_size $param_value
                } else {
                    # use default of 10 MB
                    set db_cache_size [expr 1024*1024]
                }
            } {
                # DB param exists 
                set db_cache_size $param_value
            }

            # Initialise db cache
	    ns_cache create db -size $db_cache_size 
	}

        if { [ns_cache names db $hash] ne "" \
		 && [nsv_exists db $hash] \
		 && (([clock seconds]-[nsv_get db $hash])<=$ttl) 
	 } {
	    # age of cache value < ttl
	    return [ns_cache get db $hash]
        } else {
	    set table [db_select_table $qry $level]
	    ns_cache set db $hash $table
	    nsv_set db $hash [clock seconds]
	    return $table
        }
    } else {
        # No ttl specified - use global array.
        global db_thread_cache    
        if { [info exists db_thread_cache($hash)] } {
	    return $db_thread_cache($hash)
        } else {
	    set db_thread_cache($hash) [db_select_table $qry $level]
        }
    }
}

proc qc::db_cache_clear { {qry ""} } {
    #| Clear the cache for the qry or all
    set hash [qc::md5 [db_qry_parse $qry 1]]

    # ns_cache cache
    if { [info procs ::ns_cache] ne "" } {
        if { [in [ns_cache_names] db] } {
            foreach key [ns_cache names db] {
                if { $qry eq "" || $key eq $hash } {
                    ns_cache flush db $key
                }
            }
        }
    }

    # Thread cache
    global db_thread_cache
    if { [info exists db_thread_cache] } {
        if { $qry eq ""} {
            unset db_thread_cache
        } else {
            if { [info exists db_thread_cache($hash)] } {
                unset db_thread_cache($hash)
            }
        }
    }
}

proc qc::db_cache_ldict { qry } {
    #| Cached equivalent of db_select_ldict
    #| Select the results of qry into a ldict
    set table [db_cache_select_table $qry 1]
    return [qc::table2ldict $table]
}


namespace eval qc {
    namespace export section section_menu
}

proc qc::section {conf section default_url {section_var section}} {
    set html [qc::section_menu $conf $section $default_url $section_var]
    foreach dict $conf {
	if { [string equal [dict get $dict label] $section] } {
	    set cmd [dict get $dict content]
	    append html [eval $cmd]
	    return $html
	}
    }
}

proc qc::section_menu {conf section default_url section_var} {
    set lmenu {}
    foreach dict $conf {
	set label [dict get $dict label]

        if { [dict exists $dict class] } {
            set class [dict get $dict class]
        } else {
            set class ""
        }

	# AccessKey
	for {set i 0} {$i<[string length $label]} {incr i} {
	    set letter [string index $label $i]
	    if { ![info exist usedkeys($letter)] } {
		set accesskey [string tolower $letter]
		set usedkeys($letter) 1
		set link_label "[string range $label 0 [expr {$i-1}]]<u>$letter</u>[string range $label [expr {$i+1}] end]"
		break
	    } else {
		set accesskey .
		set link_label $label
	    }
	}

	# Spacer - TLC
	if { [dict exists $dict type] && [string equal [dict get $dict type] spacer] } {
            lappend class "spacer"
	    lappend lmenu [html span "" style "width:[dict get $dict width]px" class $class]
	    continue
	}

	# Use default url with section
	if {[string equal $label $section]} {
            lappend class "selected"
	    lappend lmenu [html span $label class $class]
	} else {
	    if { [dict exists $dict count] } {
		set count [dict get $dict count]
		lappend lmenu [html span "[html_a $link_label [url $default_url $section_var $label] accesskey $accesskey] ($count)" class $class]
	    } else {
		lappend lmenu [html span [html_a $link_label [url $default_url $section_var $label] accesskey $accesskey] class $class]
	    }
	}
    }
    return [html div [join $lmenu ""] class "section-menu"]
}

namespace eval qc {
    namespace export format_timestamp format_timestamp_* format_timestamp2hour
}

proc qc::format_timestamp_iso { args } {
    #| Format string as an ISO timestamp
    qc::args $args -text -html -- string
    if { [info exists text] } {
        return [clock format [cast epoch $string] -format "%Y-%m-%d %H:%M:%S"]
    } else {
        return [string map [list - "&#8209;"] [clock format [cast epoch $string] -format "%Y-%m-%d %H:%M:%S"]]
    }
}

proc qc::format_timestamptz { args } {
    #| Format string as an ISO timestamp with time zone
    qc::args $args -text -html -- string
    if { [info exists text] } {
        return [clock format [cast epoch $string] -format "%Y-%m-%d %H:%M:%S %z"]
    } else {
        return [string map [list - "&#8209;"] [clock format [cast epoch $string] -format "%Y-%m-%d %H:%M:%S %z"]]
    }
}

proc qc::format_timestamp_http { string } {
    #| Format string as http timestamp according to RFC 1123
    return [clock format [cast epoch $string] -timezone :GMT -format "%a, %d %b %Y %H:%M:%S %Z"]
}

proc qc::format_timestamp_rel { string } {
    #| Format relative to age with date and time
    return [qc::format_timestamp_relative now $string]
}

proc qc::format_timestamp_relative { timestamp_now timestamp } {
    #| Format relative to age with date and time
    set epoch [cast epoch $timestamp]
    set epoch_now [cast epoch $timestamp_now]
    # Today return time
    if { [string equal [cast date $epoch_now] [cast date $epoch]] } {
        return [clock format $epoch -format "%H:%M"]
    }
    # Same Week
    if { [string equal [clock format $epoch_now -format "%Y%U"] [clock format $epoch -format "%Y%U"]] } {
        return [clock format $epoch -format "%a %H:%M"]
    }
    # Same Year
    if { [string equal [clock format $epoch_now -format "%Y"] [clock format $epoch -format "%Y"]] } {
	return "[date_month_shortname $epoch] [format_ordinal [date_dom $epoch]]"
    }
    return [clock format $epoch -format "%Y-%m-%d"]
}

proc qc::format_timestamp2hour { args } {
    qc::args $args -text -html -- string
    if { [info exists text] } {
        return [clock format [cast epoch $string] -format "%Y-%m-%d %H:%M"]
    } else {
        return [string map [list - "&#8209;"] [clock format [cast epoch $string] -format "%Y-%m-%d %H:%M"]]
    }
}

proc qc::format_timestamp { args } {
    #| Format string as datetime for user.
    #| Will be customizable in future but at present chooses the ISO format.
    return [format_timestamp_iso {*}$args]
}

proc qc::format_timestamp_rel_age {args} {
    #| Return the approximate relative age of a timestamp
    return [qc::format_timestamp_relative_age {*}[linsert $args end-1 now]]
}

proc qc::format_timestamp_relative_age {args} {
    #| Return the approximate relative age of a timestamp
    qc::args $args -long -- timestamp_now timestamp
    
    set days [qc::date_days $timestamp $timestamp_now]
    if { $days == 0 } {
        return "today"
    }
  
    set years [expr {round ($days/365.0)}]
    # Return to the nearest year if 361 days or more have elapsed
    if { $years > 0 && $days > 360} {
        set rel_age "$years [iif {$years==1} year years]"
        if { [info exists long] } {
            append rel_age " ago"
        }
        return $rel_age
    }
  
    set months [expr {round($days/30.0)}]
    # Return to the nearest month if 30 days or more have elapsed
    if {$months > 0 && $days > 29} {
        set rel_age "$months [iif {$months==1} month months]"
        if { [info exists long] } {
            append rel_age " ago"
        }
        return $rel_age
    }

    set weeks [expr {round($days/7.0)}]
    # Return to the nearest week if 5 days or more have elapsed
    if {$weeks > 0 && $days > 5} {
        set rel_age "$weeks [iif {$weeks==1} week weeks]"
        if { [info exists long] } {
            append rel_age " ago"
        }
        return $rel_age
    }
    set rel_age "$days [iif {$days==1} day days]"
    if { [info exists long] } {
        append rel_age " ago"
    }
    return $rel_age
}


namespace eval qc::aws {
    namespace export s3
    namespace ensemble create
}

namespace eval qc::aws::s3 {
    namespace export s3_uri_parse
    namespace ensemble create

    proc s3_uri_parse { s3_uri } {
        #| Parse s3 uri to extract bucket and object key into a dict
       
        if { [qc::is s3_uri $s3_uri] } {
            regexp {^(?:[sS]3://|/)?([^/]+)/?(.*)} $s3_uri -> bucket object_key
        } {
            error "Unable to parse s3 uri, \"$s3_uri\""
        }
        return [qc::dict_from bucket object_key]
    }
}

namespace eval qc {
    namespace export email_* smtp_send smtp_recv sendmail email2multimap mime_type_guess qp_encode
}
package require mime
package require base64 

proc qc::email_send {args} {
    set argnames [qc::args2vars $args]
    # email_send from to subject text|html ?cc? ?bcc? ?reply-to? ?bounce-to? ?sender? ?attachment? ?attachments? ?filename? ?filenames?

    #| attachments is a list of dicts
    #| dict keys are encoding data filename ?cid?
    #| Example dict - {encoding base64 data aGVsbG8= cid 1312967973006309 filename attach1.pdf}
    #| Including cid in this dict is optional, if provided it must be world-unique
    #| cid can be used to reference an attachment within the email's html.
    #| eg. embed an image (<img src="cid:1312967973006309"/>).

    # Return-Path
    # The MTA will set the Return-Path based on the $mail_from value
    if { [info exists bounce-to] } {
        set mail_from [qc::email_address ${bounce-to}]
    } elseif { [info exists sender] } {
        set mail_from [qc::email_address $sender]
    } else {
        set mail_from [qc::email_address $from]
    }

    # Sender
    if { [info exists sender] } {
        lappend headers Sender $sender
    }

    # From
    lappend headers From $from
    # To
    set rcpts [list]
    foreach address [qc::email_addresses $to] {
	lappend rcpts $address
    }
    lappend headers To $to
    # CC
    if { [info exists cc] } {
	foreach address [qc::email_addresses $cc] {
	    lappend rcpts $address
	}
	lappend headers CC $cc
    }
    # BCC
    if { [info exists bcc] } {
	foreach address [qc::email_addresses $bcc] {
	    lappend rcpts $address
	}
    }
    # Subject
    lappend headers Subject $subject
    # Reply-To
    if { [info exists reply-to] } {
	lappend headers Reply-To ${reply-to}
    }
    # Date
    lappend headers Date [format_timestamp_http now] 
    # MIME
    lappend headers MIME-Version 1.0
    # Message-ID
    #lappend headers Message-ID "<[qc::uuid]>"

    default attachments [list]
    # Single attachment
    if { [info exists attachment] } {
	lappend attachments $attachment
    }
    # Single file 
    if { [info exists filename] } {
	set filenames [list $filename]
    }
    # Attachments from filenames
    if { [info exists filenames] } {
	foreach filename $filenames {
	    lappend attachments [qc::email_file2attachment $filename]
	}
    }

    # Split up mixed and related (cid that can be referenced in html) attachments 
    set mixed_attachments {}
    set related_attachments {}
    foreach attachment $attachments {
        if { [dict exists $attachment cid] } {
            lappend related_attachments $attachment
        } else {
            lappend mixed_attachments $attachment
        }
    }

    if { [info exists html] } {
        # HTML with text alternative
        set alternative_boundary [qc::uuid]
        if { [string first "data:image/" $html]!=-1 } {
            # Embedded image data
            lassign [qc::email_html_embedded_images2attachments $html] html generated_attachments
            lappend related_attachments {*}$generated_attachments
        }
        set mime_body [qc::email_mime_html_alternative $html $alternative_boundary]
        set mime_headers [list Content-Type "multipart/alternative; boundary=\"$alternative_boundary\""]
    } else {
        # Text Only
        set mime_body [qc::qp_encode $text]
        set mime_headers [list Content-Transfer-Encoding quoted-printable Content-Type "text/plain; charset=utf-8"]
    }
    
    if { [llength $related_attachments] } {
        # Related attachments have a cid that can be referenced in email's html
        set related_boundary [qc::uuid]
        set related_parts [list [list headers $mime_headers body $mime_body]]
        foreach attachment $related_attachments {
	    lappend related_parts [qc::email_mime_attachment $attachment]
	}

	set mime_body [qc::email_mime_join $related_parts $related_boundary]
        set mime_headers [list Content-Type "multipart/related; boundary=\"$related_boundary\""]
    }
    if { [llength $mixed_attachments] } {
        # Mixed attachments (standard attachments)
        set mixed_boundary [qc::uuid]
        set mixed_parts [list [list headers $mime_headers body $mime_body]]
        foreach attachment $mixed_attachments {
	    lappend mixed_parts [qc::email_mime_attachment $attachment]
	}

	set mime_body [qc::email_mime_join $mixed_parts $mixed_boundary]
        set mime_headers [list Content-Type "multipart/mixed; boundary=\"$mixed_boundary\""]
    }

    lappend headers {*}$mime_headers
    qc::sendmail $mail_from $rcpts $mime_body {*}$headers
}

proc qc::email_file2attachment {filename} {
    #| Return an attachment dict for this file
    set fhandle [open $filename r]
    fconfigure $fhandle -buffering line -translation binary -blocking 1
    set base64 [::base64::encode [read $fhandle]]
    close $fhandle
    set filename [file tail $filename]
    return [list encoding base64 data $base64 filename $filename]
}

proc qc::email_address {text} {
    #| Extract the email address from text
    return [lindex [qc::email_addresses $text] 0]
}

proc qc::email_addresses {text} {
    #| Return a list of email addresses in the text
    set list [list]
    foreach dict [mime::parseaddress $text] {
	lappend list [dict get $dict address]
    }
    return $list
}

proc qc::email_mime_html_alternative {html boundary} {
    #| Helper to return mime part for html part with plain text alternative
    set text [html2text $html]
    lappend parts [list headers [list Content-Type "text/plain;charset=\"utf-8\"" Content-Transfer-Encoding quoted-printable] body [qc::qp_encode [encoding convertto utf-8 $text]]]
    lappend parts [list headers [list Content-Type "text/html;charset=\"utf-8\"" Content-Transfer-Encoding quoted-printable] body [qc::qp_encode [encoding convertto utf-8 $html]]]
    return [qc::email_mime_join $parts $boundary]
}

proc qc::email_mime_attachment {dict} {
    #| Helper to return mime part for attachment
    # dict keys: data filename ?encoding? ?cid?
    dict2vars $dict encoding data cid filename
    set headers [list]
    set mimetype [qc::mime_type_guess [file tail $filename]]
 
    if { ![info exists encoding] } {
	# No encoding provided so assume binary data even if text
	set encoding base64
	set data [::base64::encode $data]
    }
    lappend headers Content-Type "$mimetype;name=\"$filename\""
    lappend headers Content-Transfer-Encoding $encoding
    if { [info exists cid] } {
        lappend headers Content-Disposition "inline;filename=\"$filename\""
	lappend headers Content-ID <$cid>
    } else {
	lappend headers Content-Disposition "attachment;filename=\"$filename\""
    }

    return [list headers $headers body $data]
}

proc qc::email_mime_join {parts boundary} {
    #| Helper to join mime parts
    lappend list "--${boundary}"
    foreach part $parts {
	lappend list [qc::email_mime_part $part]
	lappend list "--${boundary}"
    }
    lset list end "--${boundary}--"
    return [join $list \r\n]
}

proc qc::email_mime_part {part} {
    # Helper to construct mime part
    dict2vars $part headers body
    set list [list]
    foreach {name value} $headers {
	lappend list "$name: $value"
    }
    return [join $list \r\n]\r\n\r\n$body
}

proc qc::smtp_send {sock string timeout} {
    #| Write data to the smtp server via sock

    # All SMTP commands should be terminated by CRLF
    socket_puts -nonewline $sock "$string\r\n" $timeout
    flush $sock
}

proc qc::smtp_recv {sock check timeout} {
    #| Read data from the smtp server
    while (1) {
	set line [socket_gets $sock $timeout]
	set code [string range $line 0 2]
	if {$check ne $code} {
	    error "Expected a $check status line; got:\n$line"
	}
	if ![string match "-" [string range $line 3 3]] {
	    break;
	}
    }
}

proc qc::sendmail {mail_from rcpts body args} {
    #| Connect to the smtp host and send email message
    #| mail_from is a bare email address eg. root@localhost
    #| rcpts is a list of bare rcpt email addresses
    #| body is the plain text message usually in mime format.
    #| args is a name value pair list of mail headers  

    set smtphost localhost
    
    set smtpport 25
    set timeout 60
    set headers {}

    ## CONSTRUCT THE MESSAGE ##
    # Start with headers
    foreach {name value} $args {
	lappend headers [email_header_fold "$name: $value"]
    }
    set msg [join $headers \r\n]

    # Blank line between headers and body
    append msg \r\n\r\n

    # Add the body    
    # Convert all lines to \r\n
    regsub -all {([^\r])\n} $body "\\1\r\n" body
    regsub -all {\r([^\n])} $body "\r\n\\1" body

    # Escape lines starting with a dot .
    regsub -all {(^|\r\n)\.} $body "\\1.." body

    append msg $body

    # Termination
    append msg "\r\n."

    ## Open the connection ##
    set sock [socket_open $smtphost $smtpport $timeout]

    ## Perform the SMTP conversation
    if { [catch {
        qc::smtp_recv $sock 220 $timeout
        qc::smtp_send $sock "HELO [qc::my hostname]" $timeout
        qc::smtp_recv $sock 250 $timeout
        qc::smtp_send $sock "MAIL FROM:<$mail_from>" $timeout
        qc::smtp_recv $sock 250 $timeout
	
        foreach rcpt_to $rcpts {
            qc::smtp_send $sock "RCPT TO:<$rcpt_to>" $timeout
            qc::smtp_recv $sock 250 $timeout	
        }

        #qc::smtp_send $sock "SIZE=[string bytelength $msg]" $timeout
        #qc::smtp_recv $sock 250 $timeout
    
        qc::smtp_send $sock DATA $timeout
        qc::smtp_recv $sock 354 $timeout
        qc::smtp_send $sock $msg $timeout
        qc::smtp_recv $sock 250 $timeout
        qc::smtp_send $sock QUIT $timeout
        qc::smtp_recv $sock 221 $timeout
    } errMsg ] } {
        ## Error, close and report
        close $sock
        return -code error $errMsg
    }

    ## Close the connection
    close $sock
}

proc qc::email2multimap {text} {
    # Convert an email message into a multimap data structure
    # Header values are mapped as key value pairs
    # If the message is multipart then the "bodies" key contains a list of the email parts
    # else the key "body" contains the message body
    #
    # A tree of multipart messages goes bodies->bodies->body

    set email {}
    regsub -all {\r\n} $text \n text
    lassign [qc::email_header_split_pair $text "\n\n"] head body
    # remove line breaks from header values
    regsub -all {\n[ \t]+} $head { } head
    foreach line [qc::email_header_split $head \n] {
	lassign [split_pair $line :] key value
	# Check if value is encoded
	lappend email $key [qc::email_header_value_decode $value]
    }
    
    if { [multimap_exists $email Content-Type] } {
        # MIME message
	set bodies {}
	array set header [email_header_values Content-Type [multimap_get_first $email Content-Type]]
	if { [string match -nocase multipart/* $header(Content-Type)] } {
            # Multi-part MIME
	    foreach part [lrange [mcsplit $body "--$header(boundary)"] 1 end-1] {
                lappend bodies [qc::email2multimap $part]
            }
	    lappend email bodies $bodies
	} else {
	    # One part MIME
	    # Content-Transfer-Encoding
	    if { [multimap_exists $email Content-Transfer-Encoding] } {
                switch [multimap_get_first $email Content-Transfer-Encoding] {
		    7bit -
		    8bit -
		    binary {
		    }
		    quoted-printable {
			set body [::mime::qp_decode $body]
                        if { [info exists header(charset)] } {
                            set encoding [IANAEncoding2TclEncoding $header(charset)]
                            set body [encoding convertfrom $encoding $body]
                        } 
		    }
		    base64 {
			set body [::base64::decode $body]
                        if { [info exists header(charset)] } {
                            set encoding [IANAEncoding2TclEncoding $header(charset)]
                            set body [encoding convertfrom $encoding $body]
                        } 
		    }
		}
            }
	    lappend email body $body
	}
    } else {
        # Non-MIME message
        lappend email body $body
    }
    return $email
}

proc qc::email_header_value_decode {value} {
    #| Decode an encoded header value as per rfc-2047
    # eg. =?UTF-8?B?UWNvZGUgUm9ja3M=?=
    if { [regexp {^=\?[^ \?]+\?[QBqb]\?[^ ]+?=$} $value] } {
        lassign [mime::word_decode $value] charset method string
        return [encoding convertfrom $charset $string]
    } else {
        return $value
    }
}

proc qc::email_header_values {key value} {
    #| Convert header values to dict
    # Convert
    # Content-Type: multipart/report; report-type=delivery-status; boundary="=_ventus"
    # to dict
    # $key {Content-Type: multipart/report} report-type delivery-status boundary =_ventus
    # lower case parameter attribute names to allow case insensitive matching
    set dict {}
    set remainder [string trimright [string trimright $value] ";"]

    # Get first value
    set index [string first ";" $remainder]    
    if { $index == -1 } {
        dict set dict $key $remainder
        return $dict
    }
    dict set dict $key [string range $remainder 0 ${index}-1]
    set remainder [string range $remainder ${index}+1 end]
    set remainder [string trimleft $remainder]

    # Loop until entire string is parsed, limit to string length as sanity check
    foreach saftey [.. 0 [string length $remainder]] {

        # Remove key from start of remainder
        set index [string first "=" $remainder]
        if { $index == -1 } {
            error "Missing value for key \"$remainder\""
        }
        set key [string range $remainder 0 ${index}-1]
        set remainder [string range $remainder ${index}+1 end]

        set remainder [string trimleft $remainder]

        # Remove value from start of remainder
        set first_char [string index $remainder 0]
        switch $first_char {
            ' -
            \" {
                # Value begins with quote, search for next matching quote
                # (excluding escaped quotes)
                set index 0
                foreach saftey2 [.. 0 [string length $remainder]] {
                    set index [string first $first_char $remainder ${index}+1]
                    if { $index == -1 } {
                        error "Unbalanced quotes in \"$remainder\""
                    }
                    set preceding_backslashes 0
                    foreach index2 [.. $index 1 -1] {
                        if { [string index $remainder ${index2}-1] eq "\\" } {
                            incr preceding_backslashes
                        } else {
                            break
                        }
                    }
                    if { $preceding_backslashes % 2 == 0 } {
                        break
                    }
                }
                set value [string range $remainder 1 ${index}-1]

                # Un-escape escaped characters
                set value [regsub -all {\\(.)} $value {\1}]
                
                set remainder [string range $remainder ${index}+1 end]
                
                set remainder [string trimleft $remainder]

                if { [string length $remainder] > 0
                     &&
                     [string index $remainder 0] ne ";"
                 } {
                    error "Invalid characters after close-quote ($remainder)"
                }

                set remainder [string range $remainder 1 end]
            }
            default {
                # Value is all remaining string up to the first ;
                set index [string first ";" $remainder]
                if { $index == -1 } {
                    set value $remainder
                    set remainder ""
                } else {
                    set value [string range $remainder 0 ${index}-1]
                    set remainder [string range $remainder ${index}+1 end]
                }
            }
        }
        set remainder [string trimleft $remainder]
        
        set value [qc::email_header_value_decode $value]
	lappend dict [string tolower $key] $value
        if { $remainder eq "" } {
            break
        }
    }
    return $dict
}

proc qc::email_header_fold {string} {
    #| Fold header into lines starting with a space as per rfc2822
    set width 78

    # Convert to unix newlines for processing
    regsub -all {\r\n} $string \n string

    set start 0
    set list {}
    while { $start<[string length $string] && [regexp -indices -start $start -- {(\"[^\"]+\")|(\([^\)]+\))|([^ ]+)} $string match] } {
	set atom [string range $string [lindex $match 0] [lindex $match 1]]
	# If characters above 127 then encode
	if { [regexp {([\u007F-\u00FF])} $atom] } {
	    set list [concat $list [split [mime::word_encode utf-8 quoted-printable $atom] \n]]
	} else {
	    lappend list $atom
	}
	set start [expr {[lindex $match 1]+1}]
    }
    set result {}
    set line [lindex $list 0]
    foreach string [lrange $list 1 end] {
	if { [string length "$line $string"]<=78 } {
	    append line " $string"
	} else {
	    lappend result $line
	    set line $string
	}
    }
    lappend result $line
    return [string map {\n "\r\n "} [join $result \r\n]]
}

proc qc::email_token2dict {token} {
    # Helper to convert a mime token to dict
    set dict {}
    set mime_type [mime::getproperty $token content]
    switch -glob -- $mime_type {
	multipart/* {
	    foreach part [mime::getproperty $token parts] {
		lappend dict $mime_type [qc::email_token2dict $part]
	    }
	}
	default {
	    lappend dict $mime_type [mime::getbody $token]
	}
    }
    return $dict
}

proc qc::email_support { args } {
    #| Send email to support.
    # Masking any card numbers before sending.
    # Usage: email_support subject $subject ?html $html? ?text $text?

    qc::args2vars $args subject html text
    if { [qc::param_exists hostname] } {
        set hostname [qc::param_get hostname]
    } else {
        set hostname [ns_info hostname]
    }
    set email_args [list from "nsd@$hostname"]
    lappend email_args to [qc::param_get email_support]
    lappend email_args subject [qc::format_cc_masked_string $subject]
    if { [info exists html] } {
	lappend email_args html [qc::format_cc_masked_string $html]
    } else {
	lappend email_args text [qc::format_cc_masked_string $text]
    }
    qc::email_send {*}$email_args
}

proc qc::qp_encode {string {encoded_word 0} {no_softbreak 0}} {
    # Based on ::mime::qp_encode, but uses \r\n instead of only \n

    regsub -all -- \
	    {[\x00-\x08\x0B-\x1E\x21-\x24\x3D\x40\x5B-\x5E\x60\x7B-\xFF]} \
	    $string {[format =%02X [scan "\\&" %c]]} string

    # Replace the format commands with their result

    set string [subst -novariable $string]

    # soft/hard newlines and other
    # Funky cases for SMTP compatibility
    set mapChars [list " \n" "=20\n" "\t\n" "=09\n" \
	    "\n\.\n" "\n=2E\n" "\nFrom " "\n=46rom "]
    if {$encoded_word} {
	# Special processing for encoded words (RFC 2047)
	lappend mapChars " " "_"
    }
    set string [string map $mapChars $string]

    # Break long lines - ugh

    # Implementation of FR #503336
    if {$no_softbreak} {
	set result $string
    } else {
	set result ""
	foreach line [split $string \n] {
	    while {[string length $line] > 72} {
		set chunk [string range $line 0 72]
		if {[regexp -- (=|=.)$ $chunk dummy end]} {

		    # Don't break in the middle of a code

		    set len [expr {72 - [string length $end]}]
		    set chunk [string range $line 0 $len]
		    incr len
		    set line [string range $line $len end]
		} else {
		    set line [string range $line 73 end]
		}
		append result $chunk=\r\n
	    }
	    append result $line\r\n
	}
    
	# Trim off last \r\n, since the above code has the side-effect
	# of adding an extra \r\n to the encoded string and return the
	# result.
	set result [string range $result 0 end-2]
    }

    # If the string ends in space or tab, replace with =xx

    set lastChar [string index $result end]
    if {$lastChar==" "} {
	set result [string replace $result end end "=20"]
    } elseif {$lastChar=="\t"} {
	set result [string replace $result end end "=09"]
    }

    return $result
}

proc qc::email_html_embedded_images2attachments {html} {
    #| Converts embedded images in the given html string to attachments
    # Returns the html with image src cids, and an ldict of attachments
    set attachments {}
    dom parse -html $html doc
    set nodes [$doc selectNodes "//img\[starts-with(@src,'data:image/')\]"]
    
    foreach node $nodes {
        set src [$node getAttribute src]
        regexp {^data:image/([a-z]+);base64,(.*)$} $src -> file_type base64
        set filename [qc::uuid].${file_type}
        set cid ${filename}@[clock seconds]
        lappend attachments [dict create encoding base64 data $base64 filename $filename cid $cid]
        $node setAttribute src "cid:${cid}"
    }
    set html [$doc asHTML]
    $doc delete
    return [list $html $attachments]
}

proc qc::email_header_split {args} {
    #| Split a string into a list, ignoring quoted or escaped delimiters
    qc::args $args -escape \\ -quote \" -- string delimiter
    set length [string length $string]
    set list {}
    
    # Loop with safety limit
    for {set i $length} {$i>0} {incr i -1} {
        set pair [qc::email_header_split_pair \
                      -nocomplain \
                      -escape $escape \
                      -quote $quote \
                      $string $delimiter]
        if { [llength $pair] == 1 } {
            lappend list [lindex $pair 0]
            break
        } else {
            lappend list [lindex $pair 0]
            set string [lindex $pair 1]
        }
    }
    return $list
}

proc qc::email_header_split_pair {args} {
    #| Split a string into 2 parts at the first occurence of the delimiter,
    #| ignoring quoted or escaped delimiters
    qc::args $args -nocomplain -escape \\ -quote \" -- string delimiter
    default nocomplain false
    set delimiter_index -1
    set quote_index -1
    set length [string length $string]

    # Loop with safety limit
    for {set i $length} {$i>0} {incr i -1} {

        # Find first instance of delimiter
        set delimiter_index [string first \
                                 $delimiter $string ${delimiter_index}+1]        
        if { $delimiter_index == -1 } {
            if { $nocomplain } {
                return [list $string]
            } else {
                error "Delimiter \"$delimiter\" was not found in the string \"$string\""
            }
        }
        if { [qc::string_is_escaped $string $delimiter_index $escape] } {
            # Proceed to next iteration, to find next instance of delimiter
            continue
        }
        
        set in_quotes false
        
        # Loop with safety limit
        for {set j $length} {$j>0} {incr j -1} {
            
            # Find opening quote
            set quote_index [string first $quote $string ${quote_index}+1]

            # Delimiter is not inside quotes
            if { $quote_index > $delimiter_index
                 || $quote_index == -1 } {
                break
            }

            if { [qc::string_is_escaped $string $quote_index $escape] } {
                # Find next opening quote
                continue
            }

            # Loop with safety limit
            for {set k $length} {$k>0} {incr k -1} {

                # Find closing quote
                set quote_index [string first $quote $string ${quote_index}+1]
                if { $quote_index == -1 } {
                    error "Mismatched parenthesis in \"$string\""
                }
                if { ! [qc::string_is_escaped $string $quote_index $escape] } {
                    break
                }
            }
            if { $quote_index > $delimiter_index } {
                set in_quotes true
                break
            }
        }

        if { ! $in_quotes } {
            break
        }
    }

    # Delimiter is found, split string on delimiter
    set list {}
    lappend list \
        [string trim \
             [string range \
                  $string \
                  0 \
                  [expr {$delimiter_index-1}]]]
    lappend list \
        [string trim \
             [string range \
                  $string \
                  [expr {$delimiter_index+[string length $delimiter]}] \
                  end]]
    
    return $list
}

namespace eval qc {
    namespace export html_table_scroll
}

proc qc::html_table_scroll {args} {
    # varNames can be one of
    # cols tbody tfoot height
    set varNames [qc::args2vars $args]

    default height 600

    set div_style ""
    set div_class [list "scroll"]
    if { [lower $height] eq "max" } {
	lappend div_class "maximize-height"
    } else {
	set div_style [style_set $div_style height ${height}px]
    }
    append html [html_tag div class $div_class style $div_style]
    append html [qc::html_table [dict_from {*}$varNames]]
    append html "</div>\n"

    return $html
}

namespace eval qc {
    namespace export param_exists param_set param_get
}
namespace eval qc::param {}

proc qc::param_get { args } {
    #| Return param value.
    qc::args $args -ttl [expr {60*60*24}] -- param_name args
    
    if { [llength $args] > 0 } {
        return [dict get [qc::param_get $param_name] {*}$args]
    } else {
        if { [qc::db_connected] || [info commands ns_db] eq "ns_db" } {
            # DB param
            set qry {select param_value from param where param_name=:param_name}
            db_cache_0or1row -ttl $ttl $qry {
                # Not found in DB
            } {
                return $param_value
            } 
            
	    # Check for naviserver params
	    if { [set param_value [ns_config ns/server/[ns_info server] $param_name]] ne "" } {
	        # Naviserver param
	        return $param_value
	    }

        } else {
	    qc::param_datastore_load
	    # Check the datastore
	    if { [info exists ::qc::param::db] && [qc::in [array names ::qc::param::db] $param_name] } {
	        return $::qc::param::db($param_name)
	    }
        }
        # I give up
        error "I don't know how to find param $param_name"
    }
}

proc qc::param_exists { param_name args } {
    #| Check for param existence.
    if { [llength $args] > 0 } {
        if { [qc::param_exists $param_name] } {
            return [dict exists [qc::param_get $param_name] {*}$args]
        } else {
            return false
        }
    } else {

        if { [qc::db_connected] || [info commands ns_config] eq "ns_config" } {
            # Naviserver
            if { [qc::db_connected] || [info commands ns_db] eq "ns_db" } {
                # DB param
                set qry {select param_value from param where param_name=:param_name}
                db_cache_0or1row -ttl 86400 $qry {
                    # Not found in DB
                } {
                    return true
                }
            }

	    # Check for naviserver params
	    if { [ns_config ns/server/[ns_info server] $param_name] ne "" } {
	        # Naviserver param
	        return true
	    }

        } else {

            # Non-naviserver
	    qc::param_datastore_load
	    if { [info exists ::qc::param::db] && [qc::in [array names ::qc::param::db] $param_name] } {
	        return true
	    }
        }
    }
    return false
}

proc qc::param_set { param_name args } {
    #| Attempt to set param_name to param_value
    if { [llength $args] == 0 } {
        error "Usage: qc::param_set name ?key ...? value" 
    } elseif { [llength $args] > 1 } {
        if { [qc::param_exists $param_name] } {
            set param_value [qc::param_get $param_name]
            dict set param_value {*}$args
        } else {
            dict set param_value {*}$args 
        }
        return [qc::param_set $param_name $param_value]
    }
    # 2 args
    set param_value [lindex $args 0]
    # First check this doesn't exist in ns_config in which case it is read-only and we cannot set it here.
    if { [info commands ns_config] eq "ns_config" && [ne [ns_config ns/server/[ns_info server] $param_name] ""] } {
        # We cannot set this param programmatically
        error "Param $param_name is read-only" {} USER
    }

    if { [qc::db_connected] || [info commands ns_db] eq "ns_db" } {
        # Check the DB table
        db_cache_clear [db_qry_parse "select param_value from param where param_name=:param_name"]
        db_0or1row { select param_value as existing_value from param where param_name=:param_name } {
            # new db param
            db_dml {insert into param values(:param_name,:param_value)}
        } {
            # updating an existing db param
            db_dml { update param set param_value=:param_value where param_name=:param_name }
        }
    } else {
	qc::param_datastore_load
	array set ::qc::param::db [list $param_name $param_value]
	qc::param_datastore_save
    }

    return $param_value
}

proc qc::param_datastore_load {} {
    #| Load the datastore if it exists and isn't already in memory and is readable
    if { ![info exists ::env(HOME)] } {
	error "Don't know where your home directory is"
    }
    if { ![info exists ::qc::param::db] && [file exists $::env(HOME)/.muppet/db] } {
        set fh [open $::env(HOME)/.muppet/db r]
        array set ::qc::param::db [read -nonewline $fh]
        close $fh
    }
}

proc qc::param_datastore_save {} {
    #| Write out the params to disk.
    if { ![info exists ::env(HOME)] } {
	error "Don't know where your home directory is"
    }
    set filename "$::env(HOME)/.muppet/db"
    if { ![file exists $filename] } {
	file mkdir [file dirname $filename]
    }
    # Write to datastore
    set fh [open $filename w]

    puts -nonewline $fh [array get ::qc::param::db]
    close $fh
    file attributes $filename -permissions 0600
}


namespace eval qc {
    namespace export sql_set sql_set_varchars_truncate sql_set_with sql_insert sql_insert_with sql_sort sql_select_case_month sql_in sql_array2list sql_list2array sql_where_postcode sql_insert_or_update sql_insert_or_update_with sql_limit
}

proc qc::sql_set {args} {
    foreach name $args {
	lappend set_list "$name=:$name"
    }
    return [join $set_list ,]
}

proc qc::sql_set_varchars_truncate {args} {
    qc::args $args -schema ? -- table args
    if { ! [info exists schema] } {
        lassign [qc::db_qualify_table $table] {*}{
            schema
            table
        }
    }
    foreach name $args {
        set length [db_col_varchar_length $schema $table $name]
        lappend set_list "${name}=:${name}::varchar($length)"
    }
    return [join $set_list ,]
}

proc qc::sql_set_with {args} {
    foreach {name value} $args {
	lappend set_list "\"$name\"=[db_quote $value]"
    }
    return [join $set_list ,]
}

proc qc::sql_insert { args } {
    foreach name $args {
	lappend cols $name
	lappend values ":$name"
    }
    return "( [join $cols ,] ) values ( [join $values ,] )" 
}

proc qc::sql_insert_with { args } {
    #| Construct a SQL INSERT statement using the name value pairs given
    foreach {name value} $args {
	lappend cols "\"$name\""
	lappend values [db_quote $value]
    }
    return "( [join $cols ,] ) values ( [join $values ,] )" 
}

proc qc::sql_sort { args } {
    args $args -paging -limit ? -nulls last -- args
    #| Create the sql for sorting and paging from form_vars
    #| Default sort order can be specified in args

    # Accept args in format col1,col2,col3 DESC,col4 ASC
    # or col1 col2 col3 DESC col4 
    # Returned normal SQL order by clause

    if { [form_var_exists sortCols] } {
        set string [form_var_get sortCols]
    } else {
        set string [join $args " "]
    } 

    # collapse consequetive spaces and commas down to a single space
    set string [regsub -all {[\s,]+} $string " "]

    set list [split $string " "]
    set order_by_list {}
    for {set i 0} {$i<[llength $list]} {incr i} {

	set this_item [lindex $list $i]
        set this_item [string trim $this_item]
        set parts [split $this_item "."]
        switch [llength $parts] {
            1 {
                set this_item [qc::db_quote_identifier $this_item]                
            }
            2 {
                set table [qc::db_quote_identifier [lindex $parts 0]]
                set column [qc::db_quote_identifier [lindex $parts 1]]
                set this_item "${table}.${column}"
            }
            default {
                error "Unable to parse identifier \"$this_item\""
            }
        }

	set next_item [lindex $list [expr {$i+1}]]
        set next_item [string trim $next_item]

	switch -nocase $next_item {
	    ASC {
		if { [string toupper $nulls] eq "FIRST" } {
		    # override default null sorting order
		    lappend order_by_list "$this_item NULLS FIRST"
		} else {
		    # default null sorting order for ASC (NULLS LAST)
		    lappend order_by_list $this_item
		}
		incr i
	    }
	    DESC {
		if { [string toupper $nulls] eq "LAST" } {
		    # override default null sorting order
		    lappend order_by_list "$this_item DESC NULLS LAST"
		} else {
		    # default null sorting order for DESC (NULLS FIRST)
		    lappend order_by_list "$this_item DESC"
		}
		incr i
	    }
	    default {
		if { [string toupper $nulls] eq "FIRST" } {
		    # override default null sorting order
		    lappend order_by_list "$this_item NULLS FIRST"
		} else {
		    # default null sorting order (NULLS LAST)
		    lappend order_by_list $this_item
		}
	    }
	}
    }
    if { [llength $order_by_list]==0 } {
	# postgresql syntax for 1st column
	set sql "1"
    } else {
	set sql [join $order_by_list ,]
    }

    # Paging
    if { [info exists limit] || [info exists paging] } {
        # We are paging
        if { [form_var_exists limit] && [qc::is integer [form_var_get limit]] } {
            #formvar trumps everything
            set limit [form_var_get limit]
        } elseif { [info exists limit] } {
            # -limit option was used and limit is already set
        } elseif { [uplevel 1 {info exists limit}] } {
            #limit is set in callers namespace
            qc::upcopy 1 limit limit
        } else {
            #it's not set anywhere - use a default
            set limit 100
        }
            
        # make sure it's set in caller's namespace
        upset 1 limit $limit

        if { [form_var_exists offset] && [qc::is integer [form_var_get offset]]} {
            set offset [form_var_get offset]
        } else {
            set offset 0
        }
        upset 1 offset $offset

        return "$sql limit [qc::db_quote $limit] offset [qc::db_quote $offset]"
    } else {
	return $sql
    }
}

proc qc::sql_select_case_month { date_col value_col {alt_value 0} {col_names {jan feb mar apr may jun jul aug sep oct nov dec}}} {
    #| SQL case for crosstab style queries
    set alt_value [db_quote $alt_value]
    foreach month {1 2 3 4 5 6 7 8 9 10 11 12} {
	lappend list "CASE WHEN extract(month from $date_col)=$month THEN $value_col ELSE $alt_value END as [lindex $col_names [expr {$month-1}]]"
    }
    return [join $list ,\n]
}

proc qc::sql_in {list {type ""}} {
    #| Return a SQL comma separated list
    set sql {}
    foreach value $list {
	lappend sql [db_quote $value $type]
    }
    if { [llength $sql]==0 } {
	return "(NULL)"
    } else {
	return "([join $sql ,])"
    }
}

proc qc::sql_array2list {array} {
    # Convert Postgresql 1-dimensional Array to a Tcl list
    set list [csv2list [string map [list \{ "" \} "" \\\" \"\"] $array]]
    return [lreplace_values $list NULL ""]
}

proc qc::sql_list2array { args } {
    #| Convert a list into a PostgrSQL array constructor.
    qc::args $args -type "" -- list
    foreach item $list {
	lappend lquoted [db_quote $item $type]
    }
    if { $type ne ""} {
        set sql_cast "::${type}\[\]"
    } else {
        set sql_cast ""
    }
    if {[llength $list]==0} {
	return array\[\]$sql_cast
    } else {
	return array\[[join $lquoted ,]\]$sql_cast
    }
}

proc qc::sql_where_postcode {column postcode} {
    #| Search for rows matching this full or partial UK postcode.
    # Eg. [sql_where_postcode "delivery_postcode" "IV"] matches "IV1 5DZ", "IV10 5DZ" etc.
    #     [sql_where_postcode "delivery_postcode" "I"] matches "I0 5DZ", "I10 5DZ" etc
    set postcode [string toupper $postcode]
    set area_regexp {[A-Z]{1,2}}
    set district_regexp {[0-9][0-9]?[A-Z]?}
    set space_regexp {\s}
    set sector_regexp {[0-9]}
    set unit_regexp {[A-Z]{2}}
    set parse_regexp "
        ^
        ( \$ | ${area_regexp} )
        ( \$ | ${district_regexp} )
        ( \$ | ${space_regexp} )
        ( \$ | ${sector_regexp} )
        ( \$ | ${unit_regexp} )
        $
    "
    
    # Parse postcode extracting area, district, sector and unit.
    if { ! [regexp -expanded $parse_regexp $postcode match area district space sector unit] } {
        error "Unable to parse postcode \"$postcode\""
    }
    
    # Build regexp
    set regexp {^}
    foreach var [list area district space sector unit] {
        if { [set $var] ne "" && $var ne "space" } {
            append regexp [db_escape_regexp [set $var]]
        } else {
            append regexp [set ${var}_regexp]
        }
    }
    append regexp {$}
    
    return [db_qry_parse "$column ~ :regexp"]
}

proc qc::sql_insert_or_update_with {table primary_key_cols dict} {
    #| Run an insert or update query against this table after checking the primary key for existence.
    # Pass column name/value pairs in the dict
    if { [llength $primary_key_cols]==0 } {
        error "You must specify one or more primary key columns"
    }
    set list {}
    foreach col $primary_key_cols {
        lappend list "\"$col\"=[db_quote [dict get $dict $col]]"
    }
    set sql_where [join $list " and "]
    db_1row "
        select count(*) as count
        from \"$table\"
        where $sql_where
    "
    if { $count > 0 } {
        return "
            update \"$table\"
            set [sql_set_with {*}[dict_exclude $dict {*}$primary_key_cols]]
            where $sql_where
        "
    } else {
        return "
            insert into \"$table\"
            [sql_insert_with {*}$dict]
        "
    }
}

proc qc::sql_insert_or_update {table primary_key_cols cols} {
    #| Run an insert or update query against this table after checking the primary key for existence.
    # Pass column name/value pairs in the dict
    set list {}
    foreach col $primary_key_cols {
        lappend list "$col=:$col"
    }
    set sql_where [join $list " and "]
    set qry "
        select count(*) as count
        from \"$table\"
        where $sql_where
    "
    uplevel 1 [list db_1row $qry]
    upvar 1 count count
    if { $count > 0 } {
        return "
            update \"$table\"
            set [sql_set {*}$cols] 
            where $sql_where
        "
    } else {
        return "
            insert into \"$table\"
            [sql_insert {*}$primary_key_cols {*}$cols]
        "
    }
}

proc qc::sql_limit {limit} {
    #| Helper to construct limit clause
    if { $limit ne "" } {
        return "limit [db_quote $limit]"
    } else {
        return ""
    }
}

package require crc32

namespace eval qc {
    namespace export db_*
}

proc qc::db_qry_parse {qry {level 0} } {
    #| parses a SQL query replacing bind variables
    #| (ACS) like :varname with a quoted version
    #| of $varname in the caller level $level's env
    incr level

    # Quoted fields: Escape colons with \0 and []$\\
    # regsub -all won't work because the regexp need to be applied repeatedly to anchor correctly
    set start 0
    while { $start<[set length [string length $qry]] && [regexp -indices -start $start -- {(^|[^'])'(([^']|'')*)'([^']|$)} $qry -> left field . right] } {
	set qry [string replace $qry [lindex $field 0] [lindex $field 1] [string map {: \\u003A [ \\u005B ] \\u005D $ \\u0024 \\ \\u005C} [string range $qry [lindex $field 0] [lindex $field 1]]]]
        # Calculate how many characters we've added to the string
        set offset [expr [string length $qry]-$length]
        # Offset the start of the next iteration by the increase in string length
	set start [expr [lindex $right 0]+$offset]
    }

    # Dollar Quoted fields
    set start 0
    while { $start<[set length [string length $qry]] \
		&& [regexp -indices -start $start -- {(\$[a-zA-Z0-9_]*?\$)(.*?)(\1)} $qry -> left field right] } {
	set qry [string replace $qry [lindex $left 0] [lindex $right 1] [string map {: \\u003A [ \\u005B ] \\u005D $ \\u0024 \\ \\u005C} [string range $qry [lindex $left 0] [lindex $right 1]]]]
        # Calculate how many characters we've added to the string
        set offset [expr [string length $qry]-$length]
        # Offset the start of the next iteration by the increase in string length
	set start [expr {[lindex $right 1]+1+$offset}]
    }
    
    ## SQL Arrays ##
    # May be multi-dimensional, indexed or slices, with numbers, $variables, :variables, or sql_functions()
    # If sliced with the upper bound being a :variable, \: must be used - eg my_array[2\::index] or my_array[:from_index\::to_index]
    # cannot handle nested sql functions, cannot handle more advanced sql expressions as indices
    set start 0
    while { $start<[string length $qry] \
		&& [regexp -indices -nocase -expanded -start $start -- {[a-z_][a-z0-9_]*(
                                                                                         (?:\[
                                                                                          (?: '?[0-9]+'?
                                                                                           | [:$][a-z_][a-z0-9_]*
                                                                                           | [a-z_]+\([^\)]*\)
                                                                                           )
                                                                                          (?:
                                                                                           :'?[0-9]+'?
                                                                                           | :\$[a-z_][a-z0-9_]*
                                                                                           | \\::[a-z_][a-z0-9_]*
                                                                                           | :[a-z_]+\([^\)]*\)
                                                                                           )?
                                                                                          \])+
                                                                                         )} $qry -> match] } {
	set qry [string replace $qry [lindex $match 0] [lindex $match 1] [string map {\[ \\[ \] \\]} [string range $qry [lindex $match 0] [lindex $match 1]]]]
	set start [expr {[lindex $match 1]+1}]
    }

    # ARRAY constructor - eg. array[1, 2, 3]
    regsub -all -nocase {(array)\[([^\]]+)\]} $qry {\1\\[\2\\]} qry

    # array[]
    set re {
	(
	 bigint
	 |int8
	 |bigserial
	 |serial8
	 |double\s+precision
	 |float8
	 |integer
	 |int
	 |int4
	 |numeric(\([^,]+,[^\)]+\))?
	 |decimal(\([^,]+,[^\)]+\))?
	 |real
	 |float4
	 |smallint
	 |int2
	 |serial
	 |serial4
	 |money

	 |bit(\([^\)]+\))?
	 |bit\s+varying(\([^\)]+\))?
	 |varbit
	 |character\s+varying(\([^\)]+\))?
	 |varchar(\([^\)]+\))?
	 |character(\([^\)]+\))?
	 |char(\([^\)]+\))?
	 |text

	 |date
	 |time(\([^\)]+\))?(\s+with(out)?\s+time\s+zone)?
	 |timetz
	 |timestamp(\([^\)]+\))?(\s+with(out)?\s+time\s+zone)?
	 |timestamptz
	 |interval(\s+[a-z0-9]+)*(\([^\)]+\))?

	 |boolean
	 |bool

	 |bytes
	 |bytea	    
	 |inet
	 |cidr
	 |macaddr
	 |point
	 |line
	 |path
	 |lseg
	 |box	 
	 |circle
	 |polygon	    
	 |tsquery
	 |tsvector
	 |txid_snapshot
	 |uuid
	 |xml	  
	 )\[\]
    }
    regsub -all -nocase -expanded $re $qry {\1\\[\\]} qry

    # Escaped \:colon
    set qry [string map {\\: \0} $qry]
    
    # Colon variable substitution
    set type_re {
	(::(
	    bigint
	    |int8
	    |bigserial
	    |serial8
	    |double\s+precision
	    |float8
	    |integer
	    |int
	    |int4
	    |numeric(\([^,]+,[^\)]+\))?
	    |decimal(\([^,]+,[^\)]+\))?
	    |real
	    |float4
	    |smallint
	    |int2
	    |serial
	    |serial4
	    |money

	    |bit(\([^\)]+\))?
	    |bit\s+varying(\([^\)]+\))?
	    |varbit
	    |character\s+varying(\([^\)]+\))?
	    |varchar(\([^\)]+\))?
	    |character(\([^\)]+\))?
	    |char(\([^\)]+\))?
	    |text

	    |date
	    |time(\([^\)]+\))?(\s+with(out)?\s+time\s+zone)?
	    |timetz
	    |timestamp(\([^\)]+\))?(\s+with(out)?\s+time\s+zone)?
	    |timestamptz
	    |interval(\s+[a-z0-9]+)*(\([^\)]+\))?

	    |boolean
	    |bool

	    |bytes
	    |bytea	    
	    |inet
	    |cidr
	    |macaddr
	    |point
	    |line
	    |path
	    |lseg
	    |box	 
	    |circle
	    |polygon	    
	    |tsquery
	    |tsvector
	    |txid_snapshot
	    |uuid
	    |xml	  
	    )(?=[^a-z0-9]|$)
         )?
    }
    set re {
	([^:\\]):
	([a-z_][a-z0-9_]*)  
    }
    append re $type_re
    regsub -all -nocase -expanded $re $qry {\1[::qc::db_quote [set {\2}] {\4}]} qry

    # Eval with uplevel
    set qry [uplevel $level [list subst $qry]]

    # =NULL to IS NULL
    if {[regexp -nocase {^[ \t\r\n]*select[ \t\r\n]} $qry]} {
	# A select query
        set null_re {=NULL}
        append null_re $type_re
	regsub -all -expanded $null_re $qry { IS NULL} qry
    }
    return $qry
}

proc qc::db_quote { value {type ""}} {
    #| quotes SQL values by escaping single quotes with \'
    #| leaves integers and doubles alone
    #| Empty strings are converted to NULL
    if { $type eq ""} {
	set sql_cast ""
    } else {
	set sql_cast "::$type"
    }

    if { [string equal $value ""] } {
	return "NULL${sql_cast}"
    }
    # Reserved keywords
    if { [in [list current_time current_timestamp] $value ] } {
	return "${value}${sql_cast}"
    }
    
    
    set re_numeric_types {
        ^(
          bigint
          |int8
          |bigserial
          |serial8
          |double\s+precision
          |float8
          |integer
          |int
          |int4
          |numeric(\([^,]+,[^\)]+\))?
          |decimal(\([^,]+,[^\)]+\))?
          |real
          |float4
          |smallint
          |int2
          |serial
          |serial4
          |money
          )$
    }

    if { [regexp -nocase -expanded $re_numeric_types $type] } {
        # integer no leading zeros
        # -123456
        if { [regexp {^-?[1-9][0-9]*$} $value] || [string equal $value 0] } {
            return "${value}${sql_cast}"
        }
        # double greater than 1
        if { [regexp {^-?[1-9][0-9]*\.[0-9]+$} $value] } {
            return "${value}${sql_cast}"
        }
        # decimal less than 1
        # in form .23 or 0.23
        if { [regexp {^(-)?0?\.([0-9]+)$} $value -> sign tail] } {
            return "${sign}0.${tail}${sql_cast}"
        }
        # scientific notation        
        if { [regexp {^-?[1-9][0-9]*(\.[0-9]+)?(e|E)(\+|-)?[0-9]{1,2}$} $value] } {
            return "${value}${sql_cast}"
        }
    } 

    # quote everything as a string
    # replace ' with '' and \ with \\ 
    # (tcl also uses slash to escape hence \\ in string map)
    if { [string first \\ $value]==-1 } {
	return "'[string map {' ''} $value]'${sql_cast}"
    } else {
	return "E'[string map {' '' \\ \\\\} $value]'${sql_cast}"
    }
}

proc qc::db_quote_identifier {value} {
    #| Quote a sql identifier (eg. a table or column name)
    return "\"[string map {\" \"\"} $value]\""
}

proc qc::db_escape_regexp { string } {
    # The postgresql parser performs substitution 
    # before passing the regexp to the regexp engine
    # So to find a match for a regex metacharacter
    # we need to escape the preceding slash.
    # For example "foo." becomes "foo\\."
    set list {
	\\ \\\\ 
	^ \\^ 
	. \\. 
	\[ \\\[ 
	\] \\\] 
	\$ \\\$ 
	\( \\\( 
	\) \\\) 
	| \\| 
	* \\* 
	+ \\+ 
	? \\? 
	\{ \\\{
	\} \\\}
    } 

    return [string map $list $string]
}

proc qc::db_get_handle {{poolname DEFAULT}} {
    # Return db handle 
    # Keep one handle per pool for current thread.
    global _db
    if { [info commands ns_db] eq "ns_db" } {
        # Naviserver
        if { $poolname eq "DEFAULT" } {
	    set poolname [ns_config ns/server/[ns_info server]/db defaultpool]  
	} 
        if { ![info exists _db($poolname)] } {
	    set _db($poolname) [ns_db gethandle $poolname]
	}
        return $_db($poolname)
    } else {
        # Should be connected with db_connect
        if { ![info exists _db($poolname)] } {
            error "No database connection"
        }
        return $_db($poolname)
    }
}

proc qc::db_dml { args } {
    args $args -db DEFAULT -- qry
    #| Execute a SQL dml statement
    set db [db_get_handle $db]
    set qry [db_qry_parse $qry 1]
    if { [info commands ns_db] eq "ns_db" } {
        # Naviserver
        ::try {
            ns_db dml $db $qry
        } on error {error_message options} {
            error "Failed to execute dml <code>$qry</code>.<br>[ns_db exception $db]" [dict get $options -errorinfo] [dict get $options -errorcode]
        }
    } else {
        # Connected with db_connect
        ::try {
            pg_execute $db $qry
        } on error {error_message options} {
            error "Failed to execute dml <code>$qry</code>." [dict get $options -errorinfo] [dict get $options -errorcode]
        }
    }
}

proc qc::db_1row { args } {
    # Select 1 row from the database using the qry.
    # Place variables corresponding to column names in the caller's namespace
    # Throw an error if more or less than 1 row is returned.
    args $args -db DEFAULT -- qry
    set table [db_select_table -db $db $qry 1]
    set db_nrows [expr {[llength $table]-1}]
    
    if { $db_nrows!=1 } {
	error "The qry <code>[db_qry_parse $qry 1]</code> returned $db_nrows rows"
    }
    foreach key [lindex $table 0] value [lindex $table 1] { upset 1 $key $value }
    return
}

proc qc::db_0or1row {args} {
    # Select zero or one row from the database using the qry.
    # If zero rows are returned then run no_rows_code else 
    # place variables corresponding to column names in the caller's namespace and execute one_row_body
    args $args -db DEFAULT -- qry {no_rows_code ""} {one_row_code ""}
    set table [db_select_table -db $db $qry 1]
    set db_nrows [expr {[llength $table]-1}]

    if {$db_nrows==0} {
	# no rows
	set return_code [ catch { uplevel 1 $no_rows_code } result options ]
        # Preserve TCL_RETURN
        if { $return_code == 2 && [dict get $options -code] == 0 } {
            dict set options -code return
        } else {
            # Return in parent stack frame instead of here
            dict incr options -level
        }
        return -options $options $result
    } elseif { $db_nrows==1 } { 
	# 1 row
	foreach key [lindex $table 0] value [lindex $table 1] { upset 1 $key $value }
	set return_code [ catch { uplevel 1 $one_row_code } result options ]
        # Preserve TCL_RETURN
        if { $return_code == 2 && [dict get $options -code] == 0 } {
            dict set options -code return
        } else {
            # Return in parent stack frame instead of here
            dict incr options -level
        }
        return -options $options $result
    } else {
	# more than 1 row
	error "The qry <code>[db_qry_parse $qry 1]</code> returned $db_nrows rows"
    }
}

proc qc::db_foreach {args} {
    #| Place variables corresponding to column names in the caller's namespace
    #| for each row returned.
    #| Set special variables db_nrows and db_row_number in caller's namespace to
    #| indicate the number of rows returned and the current row.
    #| Nested foreach statements clean up special variables so they apply to the current scope.
    args $args -db DEFAULT -- qry foreach_code { no_rows_code ""}

    # save special db variables
    qc::upcopy 1 db_nrows      saved_db_nrows
    qc::upcopy 1 db_row_number saved_db_row_number

    set table [db_select_table -db $db $qry 1] 
    set db_nrows [expr {[llength $table]-1}]
    set db_row_number 0

    if { $db_nrows == 0 } {
	upset 1 db_nrows 0
	upset 1 db_row_number 0
	set return_code [ catch { uplevel 1 $no_rows_code } result options ]
        switch $return_code {
            0 {
                # ok
            }
            default {
                # error, return
                
                # Preserve TCL_RETURN
                if { $return_code == 2 && [dict get $options -code] == 0 } {
                    dict set options -code return
                } else {
                    # Return in parent stack frame instead of here
                    dict incr options -level
                }
                return -options $options $result
            }
        }
    } else {
	set masterkey [lindex $table 0]
	foreach list [lrange $table 1 end] {
	    upset 1 db_nrows $db_nrows
	    upset 1 db_row_number [incr db_row_number]
	    foreach key $masterkey value $list {
		upset 1 $key $value
	    }
	    set return_code [ catch { uplevel 1 $foreach_code } result options ]
            switch $return_code {
                0 {
                    # ok
                }
                3 -
                4 {
                    # break, continue
                    return -options $options $result
                }
                default {
                    # error, return

                    # Preserve TCL_RETURN
                    if { $return_code == 2 && [dict get $options -code] == 0 } {
                        dict set options -code return
                    } else {
                        # Return in parent stack frame instead of here
                        dict incr options -level
                    }
                    return -options $options $result
                }
            }

            # Clean up the result variable to prevent Tcl's Copy on Write
            # process from adversely affecting performance
            unset result
        }
    }
    # restore saved variables
    if { [info exists saved_db_nrows] } {
	upset 1 db_nrows      $saved_db_nrows
	upset 1 db_row_number $saved_db_row_number
    }
}

proc qc::db_seq {args} {
    args $args -db DEFAULT -- seq_name
    # Fetch the next value from the sequence named seq_name
    set qry "select nextval(:seq_name) as next_id"
    db_1row -db $db $qry
    return $next_id
}

proc qc::db_select_table {args} {
    # Select results of qry into a table
    # Parse qry at level given
    args $args -db DEFAULT -- qry {level 0}
    incr level
    set qry [db_qry_parse $qry $level]
    set table {}
    set db [db_get_handle $db]
    if { [info commands ns_db] eq "ns_db" } {
        # Naviserver
        ::try {
            set row [ns_db select $db $qry]
            lappend table [ns_set_keys $row]
            while { [ns_db getrow $db $row] } {
                lappend table [ns_set_values $row]
            }
            return $table
        } on error {error_message options} {
            error "Failed to execute qry <code>$qry</code><br>[ns_db exception $db]" [dict get $options -errorinfo] [dict get $options -errorcode]
        }
    } else {
        # Connected with db_connect
        ::try {
            set results [pg_exec $db $qry]
            set error_msg [pg_result $results -error]
            if { $error_msg ne "" } {
                error $error_msg
            }
            lappend table [pg_result $results -attributes]
            set table [concat $table [pg_result $results -llist]]
            pg_result $results -clear
            return $table
        } on error {error_message options} {
	    error "Failed to execute qry <code>$qry</code><br>" [dict get $options -errorinfo] [dict get $options -errorcode]
        }
    }
}

proc qc::db_select_csv { qry {level 0} } {
    #| Select qry into csv report
    #| First row contains column names
    incr level
    set table [db_select_table $qry $level]
    foreach row $table {
        lappend lines [list2csv $row]
    }
    return [join $lines \r\n]
}

proc qc::db_select_ldict { qry } {
    # Select the results of qry into a ldict
    set table [db_select_table $qry 1]
    return [qc::table2ldict $table]
}

proc qc::db_select_dict { qry } {
    # Select 0 or 1 row from the database using the qry.
    # Place result in a dict
    # Throw an error if more than 1 row is returned.

    set table [db_select_table $qry 1]
    set db_nrows [expr {[llength $table]-1}]
    
    if { $db_nrows>1 } {
	error "The qry <code>[db_qry_parse $qry 1]</code> returned $db_nrows rows"
    }
    set dict {}
    foreach key [lindex $table 0] value [lindex $table 1] { lappend dict $key $value }
    return $dict
}

proc qc::db_row_exists {args} {
    #| Check for existance of 1 or more rows in $table
    #| Matching name/value pairs in $args
    qc::args $args -schema "" -- table args
    set dict [qc::args2dict $args]
    if { $schema ne "" } {
        set sql_from [db_quote_identifier $schema].[db_quote_identifier $table]
    } else {
        set sql_from [db_quote_identifier $table]
    }
    db_0or1row {
        select true as exists
        from $sql_from
        where [sql_where {*}$dict]
        limit 1
    } {
        return false
    } {
        return true
    }
}

proc qc::db_connect {args} {
    #| Connect to a postgresql database
    qc::args $args -poolname DEFAULT -- args
    
    package require Pgtcl
    global _db
    if { ![info exists _db($poolname)] } {
        set _db($poolname) [pg_connect -connlist $args]
    }
    return $_db($poolname)
}

proc qc::db_connected {args} {
    #| Is there a database connected?
    qc::args $args -poolname DEFAULT -- args
    
    global _db
    if { [info exists _db($poolname)] } {
        return true
    } else {
        return false
    }
}

proc qc::db_pg_copy_load { args } {
    #| Will load a file of data in pg copy format to the specified database table.
    #| $columns is the list of column names to insert the corresponding column 
    #| into.
    qc::args \
        $args \
        -user "postgres" \
        -password "" \
        -host "localhost" \
        -database "" \
        -db_table "" \
        -columns "" \
        -header 0 \
        -footer 0 \
        -- \
        filename
    
    set qry [db_qry_parse {
        SET client_min_messages TO WARNING;
        
        BEGIN;
        truncate $db_table;
        COPY $db_table ([join $columns ,]) FROM STDIN;
        COMMIT;
    }]
    set ::env(PGPASSWORD) $password 
    set psql [qc::which psql]
    exec cat $filename | $psql -w -U $user -h $host $database -c $qry
}

proc qc::db_disconnect {args} {
    #| Disconect the global handle
    qc::args $args -poolname DEFAULT -- args
    
    global _db
    
    pg_disconnect $_db($poolname)
    unset _db($poolname)
}

proc qc::db_advisory_trans_lock {class object} {
    #| Obtain exclusive transaction level advisory lock on an application-defined resource.
    #  class: identifies the class of resource to be locked
    #  object: identifies the individual resource to be locked

    # convert class and object identifiers to a 32 bit signed integers
    set class_id [expr {[crc::crc32 $class] -  (4294967295 + 1)/2}]
    set object_id [expr {[crc::crc32 $object] -  (4294967295 + 1)/2}]

    # obtain advisory lock
    db_1row {select pg_advisory_xact_lock(:class_id,:object_id)}
}

namespace eval qc {
    namespace export soap_template
}

proc qc::soap_template {xml method {namespace ""} } {
    #| SOAP template
    set soap {<?xml version="1.0"?>
<soap:Envelope
xmlns:soap="http://www.w3.org/2001/12/soap-envelope"
soap:encodingStyle="http://www.w3.org/2001/12/soap-encoding">
<soap:Body>
${request_xml}
</soap:Body>
</soap:Envelope>}
    # Only define namespace prefix if a namespace has been passed

    if { $namespace ne "" } {
        set request_xml [qc::xml "ns:$method" \$xml [dict create xmlns:ns $namespace]]
    } else {
         set request_xml [qc::xml $method \$xml]
    }
    set request_xml [string map [list \$xml $xml] $request_xml]
      
    set soap [string map \
                    [list \${request_xml}  $request_xml] \
                    $soap \
             ]
    return $soap
}

namespace eval qc {
    namespace export upcopy upset
}

proc qc::upcopy { level upname localname } {
    #| Make a local copy of upname
    #| from level $level as $localname
    incr level
    upvar $level $upname value
    upvar 1 $localname localvar
    if { [info exists value] } {
	set localvar $value
    } else {
	if { [info exists localvar] } { unset localvar }
    }
}

proc qc::upset {level upname args} {
    #| Like set in level $level
    if { [regexp {::} $upname] } {
        return -code error "Will not set variable outside specified\
        namespace."
    }
    
    incr level
    upvar $level $upname var
    switch [llength $args] {
        1 {
            return [set var [lindex $args 0]]
        }
        0 {
            if { [info exists var] } {
                return $var
            } else {
                error "can't read \"$upname\" :no such variable" 
            }
        }
        default {
            error "Usage qc::upset level upname ?upvalue?"
        }
    }
}

package require sha1
package require md5
package require base64
package require tdom
package require fileutil
namespace eval qc {
    namespace export s3_xml_select s3_xml_node2dict
}


# Public S3 XML procs
proc qc::s3_xml_select { xmlDoc xpath} {
    #| Returns xml nodes specified by the supplied xpath.
    # any namespace specified in the xmlns attribute is mapped to "ns" for use in the xpath query.
    set doc [dom parse $xmlDoc]
    set root [$doc documentElement]
    if { [$root hasAttribute xmlns] } {
        $doc selectNodesNamespaces "ns [$root getAttribute xmlns]"
    }
    return [$root selectNodes $xpath] 
}

proc qc::s3_xml_node2dict { node } {
    #| Converts an XML tdom node into a dict.
    # Use qc::s3_xml_select to select suitable nodes with non-repeating elements
    set dict ""
    set nodes [$node childNodes]
    foreach node $nodes {
        if { [llength [$node childNodes]] > 1 \
                 || ([llength [$node childNodes]] == 1 \
                         && [ne [[$node firstChild] nodeType] TEXT_NODE] ) } {
            lappend dict [$node nodeName] [qc::s3_xml_node2dict $node]
        }  elseif { [llength [$node childNodes]] == 0 } {
            # empty node
            lappend dict [$node nodeName] {}
        } else {
            lappend dict [$node nodeName] [$node asText]
        }
    }
    return $dict
}

namespace eval qc {
}

proc qc::sass_recompile {src dest} {
    #| Recompiles sass files if they have changed.
    set reload false
    set sass [exec which sass]
    set files [glob -nocomplain $src/*.scss $src/*/*.scss]
    array set md5s {}
    
    if { [info commands nsv_exists] ne "" } {
        # Running in Naviserver. Check if there been any changes before recompiling sass.
        foreach file $files {
            set fh [open $file r]
            set md5 [qc::md5 [read $fh]]
            close $fh
            
            if { !([nsv_exists sass_file_md5 $file] && ($md5 eq [nsv_get sass_file_md5 $file])) } {
                log Notice "Compiling scss file $file"
                set reload true
            }
            # Make a note of md5 for this file
            set md5s($file) $md5
        }
    } else {
        # Not running in Naviserver. Always recompile sass
        set reload true
    }
    
    if { $reload } {
        if { [catch {exec $sass -l --update --style expanded --stop-on-error $src:$dest} errorMessage options] } {
            return -code error -options $options $errorMessage
        } else {
            if { [info commands nsv_exists] ne "" } {
                # Update md5 records
                foreach file $files {
                    nsv_set sass_file_md5 $file $md5s($file)
                }
            }
        }
    }
}
namespace eval qc {
    namespace export socket_*
}

proc qc::socket_open {host port timeout} {
    #| Open a TCP socket with timeout

    set sock [socket -async $host $port]
    # Store socket state in global array
    global sock_state

    # After $timeout seconds flag socket state as timeout.
    set id [after [expr {$timeout*1000}] [list set sock_state($sock,write) "timeout"]]

    # Asynchronous. When $sock becomes writable, set sock_state($sock,write).
    fileevent $sock writable [list set sock_state($sock,write) "writable"]
   
    # Wait for socket state to change
    vwait sock_state($sock,write)

    # Cleanup.
    after cancel $id
    set state $sock_state($sock,write)
    unset sock_state($sock,write)

    # Check if the socket timed-out
    if { $state eq "timeout" } {
        error "Timeout waiting to connect to $host on port $port"
    }

    # Check for socket error
    # fconfigure $sock -error gets and clears any error states so call once only.
    set sock_error [fconfigure $sock -error]
    if { $sock_error ne "" } {
        error "Socket connection error: $sock_error"
    }

    # Config
    fconfigure $sock -blocking 0 -buffering line
    return $sock
}

proc qc::socket_puts {args} {
    #| Write data to a tcp socket with timeout
    args $args -nonewline -- sock string timeout

    # Store socket state in global array
    global sock_state

    # After $timeout seconds flag socket state as timeout.
    set id [after [expr {$timeout*1000}] [list set sock_state($sock,write) "timeout"]]

    # Asynchronous. When $sock becomes writable, set sock_state($sock,write).
    fileevent $sock writable [list set sock_state($sock,write) "writable"]

    # Wait for socket state to change
    vwait sock_state($sock,write)

    # Cleanup. 
    after cancel $id
    set state $sock_state($sock,write)
    unset sock_state($sock,write)

    # Check if the socket timed-out
    if { $state eq "timeout" } {
        error "Timeout waiting to write to socket"
    }
    
    # Check for socket error
    # fconfigure $sock -error gets and clears any error states so call once only.
    set sock_error [fconfigure $sock -error]
    if { $sock_error ne "" } {
        error "Socket error: $sock_error"
    }

    # At this point $sock should be writable - go ahead with puts
    if { [info exists nonewline] } {
	puts -nonewline $sock $string
    } else {
	puts $sock $string
    }
}

proc qc::socket_gets {sock timeout} {
    #| Read data from a tcp socket with timeout

    # Store socket state in global array
    global sock_state sock_data

    # Check Config
    fconfigure $sock -blocking 0 -buffering line

    # After $timeout seconds flag socket state as timeout.
    set id [after [expr {$timeout*1000}] [list set sock_state($sock,read) timeout]]

    # When $sock becomes readable
    fileevent $sock readable [list qc::socket_gets_if_ready $sock]

    # Wait for socket state to change
    vwait sock_state($sock,read)

    # Cleanup.
    after cancel $id
    set state $sock_state($sock,read)
    unset sock_state($sock,read)
    
    # Check if the socket timed-out
    if { $state eq "timeout" } {
	error "Timeout waiting to read from socket"
    }

    # Check for socket error
    # fconfigure $sock -error gets and clears any error states so call once only.
    set sock_error [fconfigure $sock -error]
    if { $sock_error ne "" } {
	error "Socket error: $sock_error"
    }

    # Cleanup
    set data $sock_data($sock)
    unset sock_data($sock)

    return $data
}

proc qc::socket_gets_if_ready { sock } {
    #| Called when a socket is readable
    # Event and data passed back through global variables
    global sock_state sock_data
    gets $sock line
    if { [eof $sock] } {
        set sock_data($sock) ""
        set sock_state($sock,read) "eof"
    } elseif { ![fblocked $sock] } {
        set sock_data($sock) $line
        set sock_state($sock,read) "done"
    } 
}

namespace eval qc {
    namespace export auth auth_check auth_hba auth_hba_check auth_session auth_as_user auth_logout
}

proc qc::auth {} {
    #| Try to authenticate the current user
    #| If successful cache the result in global
    #| On failure throw AUTH error
    global current_user_id session_id
    if { [info exists current_user_id] } {
        return $current_user_id
    }

    # Try session based auth
    if { [info exists ::session_id] || [qc::cookie_exists session_id] } {
        set session_id [qc::session_id]
        if { [qc::session_exists $session_id] } {
            set current_user_id [qc::auth_session $session_id]
            qc::session_update $session_id
            return $current_user_id
        }
    }

    # HBA
    if { [qc::auth_hba_check] } {
        # HBA user without a session, create a session for HBA user.
        set current_user_id [qc::auth_hba]
        set session_id [qc::session_new $current_user_id]
        cookie_set session_id $session_id expires [ns_httptime [clock scan "+365 days"]]
        return $current_user_id
    }

    error "Cannot authenticate you using either session_id or ip address. Please log in." {} AUTH
}

proc qc::auth_as_user {user_id} {
    #| Set the authenticated user
    global current_user_id
    set current_user_id $user_id
}

proc qc::auth_logout {} {
    #| Logout the current user
    global current_user_id
    unset current_user_id
}

proc qc::auth_check {} {
    #| Check if we can authenticate the user
    #| Return true or false
    # session based auth
    global current_user_id session_id
    if { [info exists current_user_id] } {
        return true
    }
    if { [qc::cookie_exists session_id] || [info exists session_id] } {
	set session_id [qc::session_id]
	if { [qc::session_exists $session_id] } {
	    return true
	}
    }
    
    # HBA
    if { [qc::auth_hba_check] } {
	return true
    }
    
    return false
}

proc qc::auth_hba {} {
    #| Try to authenticate who the current user is
    #| based on ip address
    #| On failure throw AUTH error
    set ip [qc::conn_remote_ip]
    set qry "select user_id as user_id from users where ip=:ip"
    db_cache_0or1row $qry { 
	error "Cannot authenticate user on ip $ip" {} AUTH
    } { 
	return $user_id
    } 
}

proc qc::auth_hba_check {} {
    #| Check if the current user can be authenticated
    #| based on ip address
    if { [info command ns_conn] ne "ns_conn" } {
        return false
    }
    set ip [qc::conn_remote_ip]
    set qry "select user_id from users where ip=:ip"
    db_cache_0or1row $qry { 
	return false
    } { 
	return true
    } 
}

proc qc::auth_session { session_id} {
    #| Try to authenticate user based on the session_id given
    #| Return the user_id if successful
    #| On failure throw AUTH error
    if { [session_valid $session_id] } {
        return [session_user_id $session_id]
    } else {
	error "Session authentication failed to identify you." {} AUTH
    }
}


namespace eval qc {
    namespace export bs_form_layout_*
}

proc qc::bs_form_layout_capsule {args} {
    #| Constructs a bootstrap floating label form layout with labels 
    # and form elements
    args $args -- conf
    
    set level 1
    set body [list]
    
    foreach dict $conf {
	array set this $dict	
	default this(type) text

	if { ![info exists this(value)] } {
	    qc::upcopy $level $this(name) value
	    if {[info exists value]} {
		set this(value) $value
	    } else {
		set this(value) ""
	    }
	}
	if {[info procs "::qc::bs_capsule_$this(type)"] ne "::qc::bs_capsule_$this(type)"} {
	    error "No widget proc defined for $this(type)"
	}

	lappend body ["qc::bs_capsule_$this(type)" {*}[array get this]]
	
	unset this
	
    }
    return [join $body "\n"]
}

namespace eval qc {
    namespace export sticky_* sticky2vars
}

proc qc::sticky_save {args} {
    #| Save form_vars for the given sticky_url or referrer
    if { [form_var_exists sticky_url] } {
        set sticky_url [form_var_get sticky_url]
	if { [url_path $sticky_url] ne "" } {
            # extract path from sticky_url eg: "/sales_form.html"
            set url [url_path $sticky_url]
        } else {
            # Use sticky_url as a namespace eg: "sales_form"
            set url $sticky_url
        }
    } else {
	set url [url_path [ns_set iget [ns_conn headers] Referer]]
    }
  
    if { [llength $args] == 0 } {
	# set all vars
	set args [uplevel 1 [lexclude [list info locals] sticky_url]]
    }
    # set vars
    foreach name $args {
	if { [uplevel 1 [list info exists $name]] } {
	    sticky_set -url $url $name [upset 1 $name]
	}
    }
}

proc qc::sticky_get {args} {
    #| Get the saved sticky value for this name
    args $args -url ? -user_id ? name
    default user_id [auth]
    default url [url_path [qc::conn_url]]
    db_1row {select value from sticky where user_id=:user_id and url=:url and name=:name}
    return $value
}

proc qc::sticky_exists {args} {
    #| Test if a sticky value has been saved for this name
    args $args -url ? -user_id ? name
    default user_id [auth]
    default url [url_path [qc::conn_url]]
    db_0or1row {select value from sticky where user_id=:user_id and url=:url and name=:name} {
	return 0
    } {
	return 1
    }
}

proc qc::sticky_set {args} {
    #| Insert or Update the sticky record
    # trim whitespace
    args $args -url ? -user_id ? name value
    default user_id [auth]
    default url [url_path [qc::conn_url]]
    set value [string trim $value]
    db_trans {
        db_1row {select user_id from users where user_id=:user_id for update}
        db_0or1row {select value as old_value from sticky where user_id=:user_id and url=:url and name=:name} {
            db_dml "insert into sticky [sql_insert user_id url name value]"
        } {
            db_dml "update sticky set value=:value where user_id=:user_id and url=:url and name=:name"
        }
    }
    return $value
}

proc qc::sticky2vars { args } {
    #| Set variables corresponding to saved sticky values in the caller's namespace.
    foreach name $args {
	if { [sticky_exists $name] } {
	    upset 1 $name [sticky_get $name]
	} else {
	    if { [uplevel 1 [list info exists $name]] } {
		uplevel 1 [list unset $name]
	    }
	}
    }
}

proc qc::sticky_default {args} {
    #| Set var values in the caller's namespace using sticky values if not passed in form vars
    set url [url_path [ns_set iget [ns_conn headers] Referer]]
    foreach name $args {
	if { ![form_var_exists $name] && [sticky_exists -url $url $name]} {
	    upset 1 $name [sticky_get -url $url $name]
	}
    }
}

namespace eval qc {
    namespace export \
        aws_credentials_set_from_ec2_role \
        aws_credentials_get_from_ec2_role \
        aws_credentials_set \
        aws_region_set
}

proc qc::aws_credentials_set { access_key secret_key {token ""}} {
    #| Set globals containing AWS credentials
    set ::env(AWS_ACCESS_KEY_ID) $access_key
    set ::env(AWS_SECRET_ACCESS_KEY) $secret_key
    if { $token ne "" } {
        set ::env(AWS_SESSION_TOKEN) $token
    }
    return true
}

proc qc::aws_region_set { region } {
    #| Set global containing AWS region
    set ::env(AWS_DEFAULT_REGION) $region
    return true
}

proc qc::aws_credentials_set_from_ec2_role { {region "eu-west-1"} } {
    #| Apply the EC2 role credentials found attached to this EC2 instance

    # Set region
    qc::aws_region_set $region

    # Set role credentials
    lassign [qc::aws_credentials_get_from_ec2_role] access_key secret_access_key token
    qc::aws_credentials_set $access_key $secret_access_key $token
}

proc qc::aws_credentials_get_from_ec2_role { } {
    #| Get role credentials from ec2 role metadata
    set role_name [qc::aws_metadata iam/security-credentials/]
    set role_credentials_json [qc::aws_metadata iam/security-credentials/${role_name}]
    set role_credentials_tson [qc::json2tson $role_credentials_json]
    set access_key [qc::tson_object_get_value $role_credentials_tson AccessKeyId]
    set secret_key [qc::tson_object_get_value $role_credentials_tson SecretAccessKey]
    set token [qc::tson_object_get_value $role_credentials_tson Token]
    return [list \
                $access_key \
                $secret_key \
                $token \
            ]
}

namespace eval qc {
    package require try
    package require tdom
    namespace export html2* html html_tag h h_tag html_escape html_unescape html_hidden html_hidden_set html_list html_a html_a_replace html_id html_menu html_paragraph_layout html_info_tables html_styles2inline html_style2inline html_col_styles_apply2td html_clean strip_html html2tcl tcl_escape tdom_node2tcl html_sanitize element_sanitize attribute_sanitize safe_elements_check safe_attributes_check safe_html_error_report safe_elements_error_report safe_attributes_error_report safe_elements safe_attributes response2html
}

proc qc::html2pdf { args } {
    # | Looks for wkhtmltopdf binary to generate PDF.
    # | Requires wkhtmltopdf 0.12.2.1 (with patched qt)
    # Adds pagesize and margin arguments for label printing.
    # usage html2pdf ?-encoding encoding? ?-timeout timeout(secs)? ?-nomargin boolean? ?-pagesize size? html
    args $args \
        -encoding base64 \
        -pagesize A4 \
        -timeout 20 \
        -nomargin false \
        -orientation portrait \
        -- html
    
    if { ![in {base64 binary} $encoding] } {
        error "HTML2PDF: Unknown encoding $encoding"
    }
    set wkhtmltopdf [qc::which wkhtmltopdf]
    package require fileutil
    set filename [fileutil::tempfile]
    set fh [open $filename r]
    fconfigure $fh -translation binary
    if { $nomargin } {
        set margin_switches [list --margin-bottom 0 --margin-top 0 --margin-left 0 --margin-right 0]
    } else {
        set margin_switches {}
    }
    qc::exec_proxy \
        -timeout [expr {$timeout * 1000}] \
        << $html \
        $wkhtmltopdf \
        {*}$margin_switches \
        --page-size $pagesize \
        --orientation $orientation \
        --encoding UTF-8 \
        --print-media-type \
        -q \
        - $filename
    
    set pdf [read $fh]
    close $fh
    file delete $filename
    if { $encoding eq "base64" } {
        return [::base64::encode $pdf]
    } else {
        return $pdf
    }
}

proc qc::html {tagName nodeValue args} {
    #| Deprecated - use qc::h instead.
    #| Generate an html node
    return "[html_tag $tagName {*}$args]$nodeValue</$tagName>"
}

proc qc::html_tag { tagName args } {
    #| Deprecated - use qc::h_tag instead.
    #| Generate just the opening html tag
    set minimized [list compact checked declare readonly disabled selected defer ismap nohref noshade nowrap multiple noresize]
    set attributes {}
    foreach {name value} [qc::dict_exclude $args {*}$minimized] {
	lappend attributes "$name=\"[string map {< &lt; > &gt; & &amp; \" &#34;} $value]\""
    }
    foreach {name value} [qc::dict_subset $args {*}$minimized] {
	if { [string is true $value] } {
	    lappend attributes "$name=\"$name\""
	}
    }
    if { [llength $attributes]==0 } {
	return "<$tagName>"
    } else {
	return "<$tagName [join $attributes]>"
    }
}

proc qc::h_tag { tag_name args } {
    #| Generate just the opening html tag
    set singleton_tags [list area base br col command embed hr img input link meta param source]
    set minimized [list compact checked declare readonly disabled selected defer ismap nohref noshade nowrap multiple noresize novalidate]
    set attributes {}
   
    foreach {name value} [qc::dict_exclude $args {*}$minimized] {
        if { $name eq "classes" } {
            lappend attributes "class=\"[string map {< &lt; > &gt; & &amp; \" &#34;} [join $value]]\""
        } else {
            lappend attributes "$name=\"[string map {< &lt; > &gt; & &amp; \" &#34;} $value]\""
        }
    }
    foreach {name value} [qc::dict_subset $args {*}$minimized] {
	if { [string is true $value] } {
	    lappend attributes "$name=\"$name\""
	}
    }

    if { $tag_name in $singleton_tags && [llength $attributes]==0 } { 
	return "<$tag_name/>"
    } elseif { $tag_name in $singleton_tags } { 
        return "<$tag_name [join $attributes]/>"
    } elseif { [llength $attributes]==0 } {
	return "<$tag_name>"
    } else {
	return "<$tag_name [join $attributes]>"
    }
}

proc qc::h {tag_name args} {
    #| Generate an html node
    set singleton_tags [list area base br col command embed hr img input link meta param source]
    if { $tag_name in $singleton_tags } {
        return "[qc::h_tag $tag_name {*}$args]"
    } else {
        if { [llength $args]%2 == 0 } {
            return "[qc::h_tag $tag_name {*}$args]</$tag_name>"
        } else {
            return "[qc::h_tag $tag_name {*}[lrange $args 0 end-1]][lindex $args end]</$tag_name>"
        }
    }
}

proc qc::html_escape {html} {
    #| Convert html markup characters to HTML entities
    return [string map [list < "&lt;" > "&gt;" \" "&quot;" ' "&#39;" & "&amp;"] $html]
}

proc qc::html_unescape { text } {
    #| Convert html entities back to text
    return [string map {&lt; < &gt; > &amp; & &\#39; ' &\#34; \" &quot; \"} $text]
}

proc qc::html_hidden { args } {
    #| Create hidden fields from vars
    set html {}
    foreach name $args {
	append html [qc::html_tag input type hidden name $name value [upset 1 $name] id $name]\n
    }
    return $html
}

proc qc::html_hidden_set { args } {
    #| Create hidden fields from list of name value pairs.
    set html {}
    foreach {name value} $args {
	append html [qc::html_tag input type hidden name $name value $value id $name]\n
    }
    return $html
}

proc qc::html_list { list args } {
    #| Convert list into HTML list.
    set html [html_tag ul {*}$args]
    foreach item $list {
	append html [html li $item] \n
    }
    append html "</ul>\n"
    return $html
}

proc qc::html_a { link url args } {
    # Create a HTML hyperlink 
    lappend args href $url
    return [html a $link {*}$args]
}

proc qc::html_a_replace { link url args } {
    # Used to replace browser history state so browser back button will not record these urls.
    lappend args onclick "location.replace(this.href);return false;"
    return [html_a $link $url {*}$args]
}

proc qc::html_id { name {value UNDEF}} {
    #| Wrap value in span tag and give it an ID
    if {[string equal $value UNDEF]} {
	qc::upcopy 1 $name value
    }
    default value ""
    return [html span $value id $name]
}

proc qc::html_menu { lmenu } {
    #| Join items to form a horizontal menu
    return [join $lmenu " &nbsp;<b>|</b>&nbsp; "]
}

proc qc::html_paragraph_layout {args} {
    #| Construct paragraph elements as a bold title with the detail below it by default.
    set html {}
    foreach {label detail} $args {
	append html "<p><b>$label</b><br>$detail</p>"
    }
    return $html
}

proc qc::html2text { html } {
    #| Wrapper for html2text. Input encoding UTF-8, Output encoding UTF-8
    set html [string map [list "&#8209;" -] $html] 
    ::try {
	return [encoding convertfrom utf-8 [exec_proxy [qc::which html2text] -utf8 -nobs << [encoding convertto utf-8 $html]]]
    } on error {} {
	# html2text unable to convert (possibly invalid html).
	# Return text by removing all html tags and any style and script elements.
	if { [regexp -nocase {<body>.*</body>} $html body] } {
	    # try to extract the body element.
	    set html $body
	} 
	return [regsub -all -nocase {<[^>]*>|<script[^>]*>.*?</script>|<style[^>]*>.*?</style>} $html ""]
    }
}

proc qc::html_info_tables {args} {
    # WACKY
    #| Foreach dict in args return a table with 2 columns with name value in each row
    set cols {
	{class bold}
	{}
    }
    set class column
    foreach dict $args {
	set tbody {}
	foreach {name value} $dict {
	    lappend tbody [list $name $value]
	}
	lappend row [qc::html_table cols $cols class $class tbody $tbody]
    }
    return [qc::html_table class "columns-container" tbody [list $row]]
}

proc qc::html_styles2inline {html} {
    #| Applies defined styles in html head as inline styles for relevant elements in body
    set styles [regexp -all -inline {<style[^>]*>[^<]*</style>} $html]
    #regsub -all {<style[^>]*>([^<]*)</style>} $html {} html
    foreach style $styles {
	regexp {<style[^>]*>([^<]*)</style>} $style -> style
	set html [qc::html_style2inline $html $style]
    }
    return $html
}

proc qc::html_style2inline {html style} {
    #| Helper proc for qc::html_styles2inline
    set data [qc::css_parse $style]
    package require tdom
    dom parse -html $html doc
    foreach {selector styles} $data {
        foreach item [split $selector ","] {
            set xpath [qc::css_selector2xpath [string trim $item]]
            set nodes [$doc selectNodes $xpath]

            foreach node $nodes {
                if { [$node hasAttribute style] } {
                    $node setAttribute style [style_set [$node getAttribute style] {*}$styles]
                } else {
                    $node setAttribute style [style_set "" {*}$styles]
                }
            }
        }
    }
    set html [$doc asHTML  -escapeNonASCII -htmlEntities]
    $doc delete
    return $html
}

proc qc::html_col_styles_apply2td {html} {
    #| Applies any relevent col styles as inline styles on td, tr, or tfoot elements
    #| Useful for correct html rendering in email clients
    set data {}
    set styles [regexp -all -inline {<style[^>]*>[^<]*</style>} $html]
    foreach style $styles {
	regexp {<style[^>]*>([^<]*)</style>} $style -> style
	set data [dict merge $data [qc::css_parse $style]]
    }
    # Apply col styles to td elements in all tables
    package require tdom
    dom parse -html $html doc
    foreach table [$doc selectNodes //table] {
	set col_number 1
	foreach col [$table selectNodes colgroup/col] {
	    if { [$col hasAttribute style] } {
		set style [$col getAttribute style]
	    } else {
		set style ""
	    }
	    if { [$col hasAttribute class] } {
		set class [$col getAttribute class] 
		foreach selector [list .$class col.$class] {
		    if { [dict exists $data $selector] } {
			set style [style_set $style {*}[dict get $data $selector]]
		    }
		}
	    }
	    if { [ne $style ""] } {
		foreach td [$table selectNodes "tbody/tr/td\[$col_number\] | tfoot/tr/td\[$col_number\]"] {
		    if { [$td hasAttribute style] } {
			$td setAttribute style [style_set [$td getAttribute style] {*}[qc::css_rule2dict $style]]
		    } else {
			$td setAttribute style [style_set "" {*}[qc::css_rule2dict $style]]
		    }
		}
	    }
	    incr col_number
	}
    }
    set html [$doc asHTML  -escapeNonASCII -htmlEntities]
    $doc delete
    return $html
}

proc qc::html_clean {html} {

    # Get rid of unnecessary tags
    regsub -all -- {<(/?o:|/?st[0-9]|!\[|\?xml)[^>]*>} $html {} html
  
    # remove extra lines
    regsub -all -- {(\n\r){2,}} $html {} html
    regsub -all { +>} $html {>} html
    regsub -all {< +} $html {<} html

    set start 0
    while { $start<[string length $html] && \
		[regexp -indices -start $start -- {<([^ >]+) ([^<]+)>} $html match ielement iattributes] } {
	set element [string range $html [lindex $ielement 0] [lindex $ielement 1]]
	set attributes [string range $html [lindex $iattributes 0] [lindex $iattributes 1]]

	# minimized attributes
	regsub -all {(^| )(checked|selected|disabled)( |$)} $attributes {\1\2="\2"\3} attributes
	
	set count 0
	while { [regsub -all {(^| )([^= ]+) *= *([^\"' ]+)( |$)} $attributes {\1\2="\3"\4} attributes] && $count<10000} {
	    incr count 
	}
	
	set count 0
	while { [regsub -all {(^| )([^= ]+) *= *'([^']+)'( |$)} $attributes {\1\2="\3"\4} attributes] && $count<10000} {
	    incr count
	}
	
	set attributes [join [regexp -all -inline {[^= ]+=\"[^\"]+\"} $attributes]]

	set element [lower $element]
	if { [eq $attributes ""] } {
	    set replace "<$element>"
	} else {
	    set replace "<$element $attributes>"
	}

	set html [string replace $html [lindex $match 0] [lindex $match 1] $replace]

	set start [expr {[lindex $match 0]+[string length $replace]}]
    }

    set start 0
    while { $start<[string length $html] && \
		[regexp -indices -start $start -- {</?[A-Z]+>} $html match] } {
	set replace [lower [string range $html [lindex $match 0] [lindex $match 1]]]
	set html [string replace $html [lindex $match 0] [lindex $match 1] $replace]
	set start [expr {[lindex $match 0]+[string length $replace]}]
    }

    return $html
}

proc qc::html2textlite {html} {
    # Will try to expand this to deal with block and inline elements but initially just preserve whitespace and newlines.
    set html [string map [list <br> \r\n] $html]
    return [qc::strip_html $html]
}

proc qc::strip_html {html} {
    #| Returns a string that is the HTML with all the HTML tags removed
    return [regsub -all -- {<[^>]+>} $html ""]
}

proc qc::tcl_escape {string} {
    #| Escape special tcl characters
    set map  {\" \\\" $ \\$ \\ \\\\ [ \\[ ] \\]}
    return [string map $map $string]
}

proc qc::tdom_node2tcl {node} {
    #| Convert tdom node to tcl 
    set singleton_tags [list area base br col command embed hr img input link meta param source]
    # Node Name
    set command {}
    lappend command h [qc::tcl_escape [$node nodeName]]
    # Node Attributes
    foreach attribute_name [$node attributes] {
        lappend command [qc::tcl_escape $attribute_name] \"[qc::tcl_escape [$node getAttribute $attribute_name]]\"
    }
    # Node Value
    if { [$node nodeName] ni $singleton_tags } {
        set list {}
        foreach child_node [$node childNodes] {
            if { [$child_node nodeName] eq "#text" } {
                lappend list [qc::tcl_escape [$child_node asXML]]
            } else {
                lappend list [qc::tdom_node2tcl $child_node]
            }
        } 
        lappend command \"[join $list ""]\"
    }
    return "\[[join $command]\]"
}

proc qc::html2tcl {html} {
    #| Convert HTML to tcl 
    dom parse -html $html doc    
    if { [llength [$doc childNodes]] > 0 } {
        return [qc::tdom_node2tcl [$doc firstChild]]
    } else {
        return ""
    }         
}

proc qc::html_sanitize {text} {
    #| Sanitizes the given text by removing any html and attributes that are not whitelisted.
    ::try {
        # wrap the text up in <root> to preserve text outwith the html
        set text [h root $text]
        set doc [dom parse -html $text]
        set root [$doc documentElement]
        
        if {$root eq ""} {
            return $text
        } else {
            qc::element_sanitize $root
            qc::attribute_sanitize $root
        }

        set html [$doc asHTML -escapeNonASCII -htmlEntities]
        $doc delete
        # remove the <div> tags we added earlier
        if { [regexp {^<root>(.*)</root>$} $html -> html_fragment]} {
            return $html_fragment
        } else {
            return -code error "The html returned by tDom was not what we expected."
        }
    } on error [list error_message options] {
        return -code error -errorcode HTML_PARSE $error_message
    }
}

proc qc::element_sanitize {node} {
    #| Checks the node and all of it's children for safe html elements removing those that are unsafe.
    if {[$node nodeType] eq "ELEMENT_NODE"} {
        set element [$node nodeName]
        set table_elements [list thead tbody tfoot tr td th]
        if {$element ni [safe_elements]} {
            $node delete
            return
        } elseif {$element eq "li"} {
            # top-level <li> elements are removed because they can break out of containing markup
            set ancestors [$node ancestor all]
            set found false
            foreach ancestor $ancestors {
                if {[$ancestor nodeName] eq "ul" || [$ancestor nodeName] eq "ol"} {
                    set found true
                    break
                }
            }
            if {! $found} {
                $node delete
                return
            }
            
        } elseif {$element in $table_elements} {
            # remove any table elements that are not contained in a table
            set ancestors [$node ancestor all]
            set found false
            foreach ancestor $ancestors {
                if {[$ancestor nodeName] eq "table"} {
                    set found true
                    break
                }
            }
            if {! $found} {
                $node delete
                return
            }
        }
    }
    
    foreach child [$node childNodes] {
        qc::element_sanitize $child
    }
}

proc qc::attribute_sanitize {node} {
    #| Checks the node and all of it's children for safe attributes removing those that are unsafe.
    set safe_attributes [dict get [safe_attributes] all]
    set nodeName [$node nodeName]
    # add a, img, or div attributes to the whitelist if the node is one of them
    if {[dict exists [safe_attributes] $nodeName]} {
        lappend safe_attributes {*}[dict get [safe_attributes] $nodeName]
    }
    # sanitize attributes of this node
    set attributes [$node attributes]
    foreach attribute $attributes {
        if {$attribute ni $safe_attributes} {
            $node removeAttribute $attribute
        } elseif {$attribute eq "href"} {
            set value [$node getAttribute $attribute]
            if {[regexp {^(.+):\/\/} $value] && ! [regexp "^(http|https|mailto):\/\/" $value]} {
                $node removeAttribute $attribute
            }
        } elseif {$attribute eq "src"} {
            set value [$node getAttribute $attribute]
            if {[regexp {^(.+):\/\/} $value] && ! [regexp "^(http|https):\/\/" $value]} {
                $node removeAttribute $attribute
            }
        } elseif {$attribute eq "class"} {
            set value [$node getAttribute $attribute]
            if {! [regexp {^language-} $value]} {
                $node removeAttribute $attribute
            }
        }
    }
    
    foreach child [$node childNodes] {
        qc::attribute_sanitize $child
    }
}

proc qc::safe_elements_check {node} {
    #| Checks the node and all of it's children for unsafe html elements.
    #| Returns true if all elements are safe otherwise false.
    if {[$node nodeType] eq "ELEMENT_NODE"} {
        set element [$node nodeName]
        set table_elements [list thead tbody tfoot tr td th]
        if {$element ni [safe_elements]} {
            return false
        } elseif {$element eq "li"} {
            # top-level <li> elements are considered unsafe because they can break out of containing markup
            set ancestors [$node ancestor all]
            set found false
            foreach ancestor $ancestors {
                if {[$ancestor nodeName] eq "ul" || [$ancestor nodeName] eq "ol"} {
                    set found true
                    break
                }
            }
            if {! $found} {
                return false
            }
            
        } elseif {$element in $table_elements} {
            # check for any table elements that are not contained in a table
            set ancestors [$node ancestor all]
            set found false
            foreach ancestor $ancestors {
                if {[$ancestor nodeName] eq "table"} {
                    set found true
                    break
                }
            }
            if {! $found} {
                return false
            }
        }
    }
    foreach child [$node childNodes] {
        if {! [qc::safe_elements_check $child]} {
            return false
        }
    }
    return true
}

proc qc::safe_attributes_check {node} {
    #| Checks the node and all of it's children for safe attributes.
    #| Returns true if all attributes are safe otherwise false.
    set safe_attributes [dict get [safe_attributes] all]
    set nodeName [$node nodeName]
    # add specific element attributes to the whitelist if the node is one of them
    if {[dict exists [safe_attributes] $nodeName]} {
        lappend safe_attributes {*}[dict get [safe_attributes] $nodeName]
    }
    set attributes [$node attributes]
    foreach attribute $attributes {
        if {$attribute ni $safe_attributes} {
            return false
        } elseif {$attribute eq "href"} {
            set value [$node getAttribute $attribute]
            if {[regexp {^(.+):\/\/} $value] && ! [regexp "^(http|https|mailto):\/\/" $value]} {
                return false
            }
        } elseif {$attribute eq "src"} {
            set value [$node getAttribute $attribute]
            if {[regexp {^(.+):\/\/} $value] && ! [regexp "^(http|https):\/\/" $value]} {
                return false
            }
        } elseif {$attribute eq "class"} {
            set value [$node getAttribute $attribute]
            if {! [regexp {^language-} $value]} {
                return false
            }
        }
    }
    foreach child [$node childNodes] {
        if {! [qc::safe_attributes_check $child]} {
            return false
        }
    }
    return true
}


proc qc::safe_html_error_report {text} {
    #| Reports all occurrences of unsafe html in the given text.
    ::try {
        # wrap the text up in <root> to preserve text outwith the html
        set text [qc::h root $text]
        set doc [dom parse -html $text]
        set root [$doc documentElement]
        if {$root eq ""} {
            $doc delete
            return {}
        } else {
            set unsafe_elements [qc::safe_elements_error_report $root]
            set unsafe_attributes [qc::safe_attributes_error_report $root]
            $doc delete
            return [lappend unsafe_elements {*}$unsafe_attributes]
        }
    } on error [list error_message options] {
        return [list [dict create node_value $text reason "Content could not be parsed. Check for unbalanced tags or quotations."]]
    }
}

proc qc::safe_elements_error_report {node} {
    #| Checks the node and all of it's children for unsafe html elements.
    #| Returns a list of dictionaries that specify unsafe elements.
    set unsafe {}
    if {[$node nodeType] eq "ELEMENT_NODE"} {
        set element [$node nodeName]
        set table_elements [list thead tbody tfoot tr td th]
        if {$element ni [qc::safe_elements]} {
            lappend unsafe [dict create node_value [$node asHTML -escapeNonASCII] element $element reason "Unsafe element: $element"]
        } elseif {$element eq "li"} {
            # top-level <li> elements are considered unsafe because they can break out of containing markup
            set ancestors [$node ancestor all]
            set found false
            foreach ancestor $ancestors {
                if {[$ancestor nodeName] eq "ul" || [$ancestor nodeName] eq "ol"} {
                    set found true
                    break
                }
            }
            if {! $found} {
                lappend unsafe [dict create node_value [$node asHTML -escapeNonASCII] element $element reason "List element without \"<ul>\" or \"<ol>\" ancestor"]
            }
            
        } elseif {$element in $table_elements} {
            # check for any table elements that are not contained in a table
            set ancestors [$node ancestor all]
            set found false
            foreach ancestor $ancestors {
                if {[$ancestor nodeName] eq "table"} {
                    set found true
                    break
                }
            }
            if {! $found} {
                lappend unsafe [dict create node_value [$node asHTML -escapeNonASCII] element $element reason "Table element without \"<table>\" ancestor"]
            }
        }
    }
    foreach child [$node childNodes] {
        foreach item [qc::safe_elements_error_report $child] {
            lappend unsafe $item
        }
    }
    return $unsafe
}

proc qc::safe_attributes_error_report {node} {
    #| Checks the node and all of it's children for unsafe attributes and values.
    #| Returns a list of dictionaries that specify unsafe items.
    set safe_attributes [dict get [qc::safe_attributes] all]
    set element [$node nodeName]
    # add specific element attributes to the whitelist if the node is one of them
    if {[dict exists [qc::safe_attributes] $element]} {
        lappend safe_attributes {*}[dict get [qc::safe_attributes] $element]
    }
    set attributes [$node attributes]
    set unsafe {}
    foreach attribute $attributes {
        if {$attribute ni $safe_attributes} {
            lappend unsafe [dict create node_value [$node asHTML -escapeNonASCII] element $element attribute $attribute reason "Unsafe attribute: $attribute"]
        } elseif {$attribute eq "href"} {
            set value [$node getAttribute $attribute]
            if {[regexp {^(.+):\/\/} $value] && ! [regexp "^(http|https|mailto):\/\/" $value]} {
                lappend unsafe [dict create node_value [$node asHTML -escapeNonASCII] element $element attribute $attribute attribute_value $value reason "Unsafe value \"$value\" for attribute \"$attribute\""]
            }
        } elseif {$attribute eq "src"} {
            set value [$node getAttribute $attribute]
            if {[regexp {^(.+):\/\/} $value] && ! [regexp "^(http|https):\/\/" $value]} {
                lappend unsafe [dict create node_value [$node asHTML -escapeNonASCII] element $element attribute $attribute attribute_value $value reason "Unsafe value \"$value\" for attribute \"$attribute\""]
            }
        } elseif {$attribute eq "class"} {
            set value [$node getAttribute $attribute]
            if {! [regexp {^language-} $value]} {
                lappend unsafe [dict create node_value [$node asHTML -escapeNonASCII] element $element attribute $attribute attribute_value $value reason "Unsafe value \"$value\" for attribute \"$attribute\""]
            }
        }
    }
    foreach child [$node childNodes] {
        foreach item [qc::safe_attributes_error_report $child] {
            lappend unsafe $item
        }
    }
    return $unsafe
}

proc qc::safe_elements {} {
    #| Returns a list of safe html elements.
    set safe_elements {}
    lappend safe_elements a
    lappend safe_elements b
    lappend safe_elements blockquote
    lappend safe_elements br
    lappend safe_elements code
    lappend safe_elements dd
    lappend safe_elements del
    lappend safe_elements div
    lappend safe_elements dl
    lappend safe_elements dt
    lappend safe_elements em
    lappend safe_elements h1
    lappend safe_elements h2
    lappend safe_elements h3
    lappend safe_elements h4
    lappend safe_elements h5
    lappend safe_elements h6
    lappend safe_elements h7
    lappend safe_elements h8
    lappend safe_elements hr
    lappend safe_elements i
    lappend safe_elements img
    lappend safe_elements ins
    lappend safe_elements kbd
    lappend safe_elements li
    lappend safe_elements ol
    lappend safe_elements p
    lappend safe_elements pre
    lappend safe_elements q
    lappend safe_elements ruby
    lappend safe_elements rt
    lappend safe_elements rp
    lappend safe_elements samp
    lappend safe_elements strong
    lappend safe_elements sub
    lappend safe_elements sup
    lappend safe_elements table
    lappend safe_elements tbody
    lappend safe_elements td
    lappend safe_elements tfoot
    lappend safe_elements th
    lappend safe_elements thead
    lappend safe_elements tr
    lappend safe_elements tt
    lappend safe_elements ul
    lappend safe_elements var
    lappend safe_elements root

    return $safe_elements
}

proc qc::safe_attributes {} {
    #| Returns a dict of {element, attributes} where attributes is a list of safe attributes for the given element.
    #| Includes element 'all' that applies to all elements.
    set list {}
    lappend list abbr
    lappend list accept
    lappend list accept-charset
    lappend list accesskey
    lappend list action
    lappend list align
    lappend list alt
    lappend list axis
    lappend list border
    lappend list cellpadding
    lappend list cellspacing
    lappend list char
    lappend list charoff
    lappend list charset
    lappend list checked
    lappend list cite
    lappend list class
    lappend list clear
    lappend list cols
    lappend list colspan
    lappend list color
    lappend list compact
    lappend list coords
    lappend list datetime
    lappend list dir
    lappend list disabled
    lappend list enctype
    lappend list for
    lappend list frame
    lappend list headers
    lappend list height
    lappend list hreflang
    lappend list hspace
    lappend list ismap
    lappend list label
    lappend list lang
    lappend list longdesc
    lappend list maxlength
    lappend list media
    lappend list method
    lappend list multiple
    lappend list name
    lappend list nohref
    lappend list noshade
    lappend list nowrap
    lappend list prompt
    lappend list readonly
    lappend list rel
    lappend list rev
    lappend list rows
    lappend list rowspan
    lappend list rules
    lappend list scope
    lappend list selected
    lappend list shape
    lappend list size
    lappend list span
    lappend list start
    lappend list summary
    lappend list tabindex
    lappend list target
    lappend list title
    lappend list type
    lappend list usemap
    lappend list valign
    lappend list value
    lappend list vspace
    lappend list width
    lappend list itemprop

    set safe_attributes [dict create]
    dict append safe_attributes a [list href]
    dict append safe_attributes img [list src]
    dict append safe_attributes div [list itemscope itemtype]
    dict append safe_attributes code [list class]
    dict append safe_attributes all $list

    return $safe_attributes
}

namespace eval qc {
    namespace export password_hash password_complexity_ok password_complexity_check
}

proc qc::password_hash { password {strength 7} } {
    #| Returns a salted password hash using blowfish with an iteration count of $strength
    # Strength min is 4 and max is 31
    check strength INT
    db_1row {select crypt(:password, gen_salt('bf',:strength)) as password_hash}
    return $password_hash
}

proc qc::password_complexity_ok { args } {
    #| Return true if password meets min, max and minclasses complexity requirements
    #| Otherwise return false
    #| max size is limited to 72
    #| Usage: password_complexity_ok password min 2 max 10 minclasses 4
    qc::args $args password args
    qc::args2vars $args min max minclasses
        
    # Check min password length
    if { [info exists min] && [string length $password] < $min } {
        return false
    }
    # Check max password length
    if { [info exists max] && [string length $password] > $max } {
        return false
    }
    # Check password contains minimum number of character classes: upper, lower, digit, punctuation
    set classes 0
    if { [info exists minclasses] } {
        foreach class [list upper lower digit] {
            if { [regexp "\[\[:${class}:\]\]" $password] } {
                incr classes
            }
        }
        # [[:punct:]] matches !"%&*()_-{}][@:#';?/.,\
        # Using inverse instead to catch £$^+=~><|`¬ in addition
        if {[regexp {[^[:upper:][:lower:][:digit:]]} $password]} {
            incr classes
        }
        if { $classes < $minclasses } {
            return false
        }
    }
    return true
}

proc qc::password_complexity_check { args } {
    #| Checks password against min, max and minclasses complexity requirements
    #| max size is limited to 72
    #| Usage: password_complexity_check password min 2 max 10 minclasses 4
    qc::args $args password args
    qc::args2vars $args min max minclasses

    default min 7
    default max 72
    default minclasses 1 
    
    # Check min password length
    if { ! [qc::password_complexity_ok $password min $min] } {
        error "Your password must be at least $min characters long" {} USER
    }
    # Check max password length
    if { ! [qc::password_complexity_ok $password max $max] } {
        error "Your password must be less than $max characters long" {} USER
    }
    # Check password contains minimum number of character classes: upper, lower, digit, punctuation
    if { ! [qc::password_complexity_ok $password minclasses $minclasses] } {
        error "Your password must contain at least $minclasses of uppercase, lowercase, numeric or punctuation" {} USER
    }
    return true
}

namespace eval qc {
    namespace export sortcols_* sortcols2dict
}

proc qc::sortcols_push { sortCols colname {sortorder ASC} } {
    #| take a list of [colname1 order colname2 order ...]
    #| and reorder putting $colname first with order $sortorder
    set newSortCols [list $colname $sortorder]

    for {set i 0} {$i<[llength $sortCols]} {incr i} {
	set this_item [string trim [lindex $sortCols $i]]
	set next_item [string trim [lindex $sortCols [expr {$i+1}]]]
	if { [ne $this_item $colname]} {
	    lappend newSortCols $this_item
	}
	if { [eq [string toupper $next_item] ASC] } {
	    incr i
	} 
	if { [eq [string toupper $next_item] DESC] } {
	    lappend newSortCols DESC
	    incr i
	} 
    }
    return $newSortCols
}

proc qc::sortcols_toggle { sortCols colname } {
    #| Toggle the sort order on colname
    if { [eq [string toupper [dict get $sortCols $colname]] ASC] } {
	dict set sortCols $colname DESC
    } else {
	dict set sortCols $colname ASC
    }
    return $sortCols
}

proc qc::sortcols_parse { args } {
    #| Accept args in format col1,col2,col3 DESC,col4 ASC
    #| or col1 col2 col3 DESC col4 
    #| Returned list col1 col2 col3 DESC ...
    if { [regexp , $args] } {
	set args [string trim $args]
	set args [split $args " ,"]
    }
    set list {}
    for {set i 0} {$i<[llength $args]} {incr i} {
	set this_item [string trim [lindex $args $i]]
	set next_item [string trim [lindex $args [expr {$i+1}]]]
	if { [eq [string toupper $next_item] ASC] } {
	    lappend list $this_item
	    incr i 
	} elseif {[eq [string toupper $next_item] DESC] } {
	    lappend list $this_item [string toupper $next_item]
	    incr i
	} else {
	    lappend list $this_item
	}
    }
    return $list
}

proc qc::sortcols2dict { args } {
    #| Accept args in format col1,col2,col3 DESC,col4 ASC
    #| or col1 col2 col3 DESC col4 
    #| Returned list col1 ASC col2 ASC col3 DESC ...
    if { [regexp , $args] } {
	set args [string trim $args]
	set args [split $args " ,"]
    }
    set list {}
    for {set i 0} {$i<[llength $args]} {incr i} {
	set this_item [string trim [lindex $args $i]]
	set next_item [string trim [lindex $args [expr {$i+1}]]]
	if { [eq [string toupper $next_item] ASC] } {
	    lappend list $this_item ASC
	    incr i 
	} elseif {[eq [string toupper $next_item] DESC] } {
	    lappend list $this_item DESC
	    incr i
	} else {
	    lappend list $this_item ASC
	}
    }
    return $list
}

proc qc::sortcols_from_cols { cols } {
    #| Extract column names from a cols list-of-lists 
    set sortCols {}
    foreach col $cols {
	if { [dict exists $col name] } {
	    lappend sortCols [dict get $col name]
	}
    }
    return $sortCols
}

proc qc::sortcols_from_qry { qry } {
    #| Look for an order by clause in the qry and return a list version
    if { [regexp -nocase {order by (.+?)(offset|limit|$)} $qry -> orderby ignore] } {
	regsub -all {\n} $qry {} qry
	return [qc::sortcols_parse $orderby]
    } else {
	return ""
    }
}

namespace eval qc {
    namespace export upper lower trim truncate plural singular cmplen levenshtein_distance string_similarity strip_common_leading_whitespace spell_correct spell_index noun_indefinite_determiner
}

proc qc::upper { string } {
    #| Convert string to upper case
    return [string toupper $string]
}

proc qc::lower { string } {
    #| Convert string to lower case
    return [string tolower $string]
}

proc qc::trim { string } {
    #| Removes and leading or trailing white space.
    return [string trim $string]
}

proc qc::truncate {string length} {
    #| Truncate to nearest word boundary to create string of at the most of specified length
    if { [string length $string]<= $length } {
	return $string
    }
    set position [string wordstart $string $length]
    if { $position == 0 } {
	set position $length
    }
    return [string range $string 0 [expr {$position-1}]]
}

# We'll begin with a box, and the plural is boxes;
# but the plural of ox became oxen not oxes.
# One fowl is a goose, but two are called geese,
# yet the plural of moose should never be meese.
# You may find a lone mouse or a nest full of mice;
# yet the plural of house is houses, not hice.
# If the plural of man is always called men,
# why shouldn't the plural of pan be called pen?
# If I spoke of my foot and show you my feet,
# and I give you a boot, would a pair be called beet?
# If one is a tooth and a whole set are teeth,
# why shouldn't the plural of booth be called beeth?
# Then one may be that, and three would be those,
# yet hat in the plural would never be hose,
# and the plural of cat is cats, not cose.
# We speak of a brother and also of brethren,
# but though we say mother, we never say methren.
# Then the masculine pronouns are he, his and him,
# but imagine the feminine, she, shis and shim.

proc qc::plural word {
    #| Attempts to return the plural form of a word.
    #| Assumes the supplied word is not already plural.
    set exceptions {
	man men
	person people
	goose geese
	mouse mice
	nucleus nuclei
	syllabus syllabi
	focus foci
	fungus fungi
	cactus cacti
	phenomenon phenomena
	criterion criteria
	foot feet
        louse lice
	ox oxen
	tooth teeth
	genus genera
        phylum phyla
        radius radii
        cherub cherubim
        mythos mythoi
        formula formulae
	radio radios
	flex flexes
    }
    if { [dict exists $exceptions $word] } {
	return [dict get $exceptions $word]
    }

    if { [in {calf elf half hoof leaf loaf scarf self sheaf thief wolf} $word] } {
	return [string range $word 0 end-1]ves
    }
    if { [in {knife life wife} $word] } {
	return [string range $word 0 end-2]ves
    }
    if { [in {auto kangaroo kilo memo photo piano pimento pro solo soprano studio tattoo video zoo} $word] } {
	return ${word}s
    }
    # unchanged
    if { [in {cod deer fish perch sheep trout species barracks equipment conduit glasses} $word] } {
	return $word
    }

    switch -regexp -- $word {
	{ing$} - {ies$}           {return $word}
	{[ei]x$}                  {return [string range $word 0 end-2]ices}
	{[sc]h$} - {[soxz]$}      {return ${word}es}
	{[bcdfghjklmnprstvwxz]y$} {return [string range $word 0 end-1]ies}
	{child$}                  {return ${word}ren}
	{eau$}                    {return ${word}x}
	{is$}                     {return [string range $word 0 end-2]es}
	{woman$}                  {return [string range $word 0 end-2]en}

    }
    return ${word}s
}

proc qc::singular word {
    # TODO unused
    switch -- $word {
	men   {return man}
	feet  {return foot}
	geese {return goose}
	lice {return louse}
	mice {return mouse}
	oxen    {return ox}
	teeth {return tooth}
	calves - elves - halves - hooves - leaves - loaves - scarves
	- selves - sheaves - thieves - wolves
	{return [string range $word 0 end-3]f}
	knives - lives - wives
	{return [string range $word 0 end-3]fe}
	autos - kangaroos - kilos - memos
	- photos - pianos - pimentos - pros - solos - sopranos - studios
	- tattoos - videos - zoos
	{return [string range $word 0 end-1]}
	cod - deer - fish - offspring - perch - sheep - trout
	- species
	{return $word}
	genera {return genus}
	phyla {return phylum}
	radii {return radius}
	cherubim {return cherub}
	mythoi {return mythos}
	phenomena {return phenomenon}
	formulae {return formula}
	octopodes {return octopus}
	octopi {return octopus}
    }
    switch -regexp -- $word {
	{[ei]ices$}                  {return [string range $word 0 end-4]x}
	{[sc]hes$} - {[soxz]es$}      {return [string range $word 0 end-2]}
	{[bcdfghjklmnprstvwxz]ies$} {return [string range $word 0 end-3]y}
	{children$}                  {return [string range $word 0 end-3]}
	{eaux$}                    {return [string range $word 0 end-1]}
	{ises$}                     {return [string range $word 0 end-4]is}
	{women$}                  {return [string range $word 0 end-2]an}
	{s$}                       {return [string range $word 0 end-1]}
    }
    #Not handled
    error "Don't know singular for \"$word\""
}

proc qc::cmplen {string1 string2} {
    #| Compare length of 2 strings
    if { [string length $string1]<[string length $string2] } {
	return -1 
    } elseif {[string length $string1]==[string length $string2] } {
	return 0
    } else {
	return 1
    }
}

proc qc::levenshtein_distance {s t} {
    # Returns the number of edits required to turn one string into the other.
    # Tcl wiki 
    if {![set n [string length $t]]} {
        return [string length $s]
    } elseif {![set m [string length $s]]} {
        return $n
    }
    for {set i 0} {$i <= $m} {incr i} {
        lappend d 0
        lappend p $i
    }
    for {set j 0} {$j < $n} {} {
        set tj [string index $t $j]
        lset d 0 [incr j]
        for {set i 0} {$i < $m} {} {
            set a [expr {[lindex $d $i]+1}]
            set b [expr {[lindex $p $i]+([string index $s $i] ne $tj)}]
            set c [expr {[lindex $p [incr i]]+1}]
            lset d $i [expr {$a<$b ? $c<$a ? $c : $a : $c<$b ? $c : $b}]
        }
        set nd $p; set p $d; set d $nd
    }
    return [lindex $p end]
}

proc qc::string_similarity {s t} {
    #| Returns a number from 0 to 1 indicating how similar the 2 strings are
    #| using the levenshtein distance.
    # Tcl wiki

    set sl [string length $s]
    set tl [string length $t]

    set ml [max $sl $tl]
    set dn [qc::levenshtein_distance $s $t]

    # -- get match characters number
    set mn [expr $ml - $dn]

    # -- match number != 0? (mn-1)/tl + (1/tl)*(mn/sl)
    return [expr $mn==0?0:($mn-1+double($mn)/$sl)/$tl]
}

proc qc::strip_common_leading_whitespace {text} {
    #| Strip leading whitespace common to all non-zero length lines
    #
    # replace tabs with four spaces
    regsub -all {\t} $text "    " text
    regsub -all {\r\n} $text \n text
    foreach line [split $text \n] {
        if { [regexp {^( *)[^ ]} $line -> whitespace] } {
            set length [string length $whitespace]
            if { ![info exists min] || $length<$min} {
                set min $length
            } 
        }
    }
    if { [info exists min] && $min>0 } {
        regsub -all -line "^ {$min}" $text {} text
    }
    return $text
}

proc qc::spell_index { corpus } {
    #| Takes a string containing a text corpus & creates spell_corpus table
    #| containing each word and its number of entries.
    #| spell_corpus table is:
    #|         create table spell_corpus (
    #|                  word text,
    #|                  entries int
    #|                  );

    # insert into temp table since ts_stat takes a sql query as parameter
    db_dml {drop table if exists temp_corpus}
    db_dml {create table temp_corpus(words text)}

    foreach line [split $corpus \n] {
        db_dml { insert into temp_corpus (words) values(:line) }
    }

    # Use ts_stat to extract lexemes and statistical data from the tsvectorised
    # text corpus
    db_dml {truncate spell_corpus}
    db_dml {
        insert
        into
        spell_corpus
        select
        word,
        nentry
        from
        ts_stat('select to_tsvector(''simple'',words) from temp_corpus')
    }
    db_dml {drop table temp_corpus}
}

proc qc::spell_correct { word } {
    #| Returns the most frequently occuring valid word from spell_corpus which
    #| is within a levenshtein distance of 2 of the supplied word.
    #| Returns an empty string if no alternative suggestion is found.
    #| The spell_corpus table is populated by qc::spell_index

    set qry {
        select 
        word as suggestion
        from
        spell_corpus 
        where 
        levenshtein(lower(word),lower(:word))<=2
        order by levenshtein(lower(word),lower(:word)) asc,entries desc 
        limit 1
    }
    db_0or1row $qry {
        set suggestion ""
    } 

    if { $suggestion ne $word } {
        return $suggestion
    } else {
        return "" 
    }
}

proc qc::noun_indefinite_determiner {noun} {
    #| Returns the indefinite determiner for a noun.
    if { [string tolower [string index $noun 0]] in [list a e i o u] } {
        return "an"
    } else {
        return "a"
    }
}

namespace eval qc {
    namespace export date_* dates years iso_year_iso_weeks year_months year_quarters time_hour
}

proc qc::date_month_start {date} {
    #| Return the date on the 1st day of the month for date given.
    #| The date can be any valid date expression.
    return [clock format [cast epoch $date] -format "%Y-%m-01"]
}

proc qc::date_month_end {date} {
    #| Return the date on the last day of the month for the date given
    return [cast date "[qc::date_month_start $date] +1 month -1 day"]
}

proc qc::date_year_start {date} {
    #| Return the date on the first day of the year
    return [clock format [cast epoch $date] -format "%Y-01-01"]
}

proc qc::date_year_end {date} {
    #| Return the date on the last day of the year
    return [clock format [cast epoch $date] -format "%Y-12-31"]
}

proc qc::date_year_iso_start {date} {
    #| Return the date on the 1st day of the ISO year for the date given.
    #| ISO Week numbers start on Monday.
    #| The first week of the year includes the first Thursday.
    #| The first week also includes the 4th Jan
    set year [qc::date_year $date]
    return [cast date [clock scan "3 days ago" -base [clock scan "Thursday" -base [clock scan "$year-01-01"] ]]]
}

proc qc::date_year_iso_end {date} {
    #| Return the date on the last day of the ISO year for the date given.
    #| ISO Week numbers start on Monday
    #| The first week of the year includes the first Thursday
    set next_year [date_year "$date + 1 year"]
    return [cast date [clock scan "4 days ago" -base [clock scan "Thursday" -base [clock scan "$next_year-01-01"]]]]
}

proc qc::date_iso_year { date } {
    #| Return the ISO year.
    return [qc::cast integer [clock format [cast epoch $date] -format "%G"]]
}

proc qc::date_iso_week_start {date} {
    #| Returns the date of the start of the week in which $date falls
    if { [eq [date_day_name $date] Monday] } {
	return [cast date $date]
    } else {
	return [cast date [clock scan "last monday" -base [cast epoch $date]]]
    }
}

proc qc::date_iso_week_end {date} {   
    #| Returns the date of the end of the week in which $date falls
    return [cast date [clock scan "sunday" -base [cast epoch $date]]]
}

proc qc::date_iso_week { date } {
    #| Return the ISO week number.
    return [qc::cast integer [clock format [cast epoch $date] -format "%V"]]
}

proc qc::date_yesterday { date } {
    #| Return yesterday's date
    return [cast date "$date -1day"]
}

proc qc::date_tomorrow { date } {
    #| Return tomorrow's date
    return [cast date "$date +1day"]
}

proc qc::date_month { date } {
    #| Return the month number.
    if { [regexp {^(\d{4})-(\d{2})-(\d{2})$} $date -> year month day] } {
        return [qc::cast integer $month]
    }
    return [qc::cast integer [clock format [cast epoch $date] -format "%m"]]
}

proc qc::date_doy { date } {
    #| Return the day of year number.
    return [qc::cast integer [clock format [cast epoch $date] -format "%j"]]
}

proc qc::date_year { date } {
    #| Return the year.
    if { [regexp {^(\d{4})-(\d{2})-(\d{2})$} $date -> year month day] } {
        return $year
    }
    return [clock format [cast epoch $date] -format "%Y"]
}

proc qc::date_dom { date } {
    #| Return the day of the month.
    if { [regexp {^(\d{4})-(\d{2})-(\d{2})$} $date -> year month day] } {
        return [qc::cast integer $day]
    }
    return [qc::cast integer [clock format [cast epoch $date] -format "%d"]]
}

proc qc::date_dow { date } {
    #| Return the day of the week.
    return [qc::cast integer [clock format [cast epoch $date] -format "%u"]]
}

proc qc::date_day_name { date } {
    #| Return the full day of the week as Monday,Tuesday etc
    return [clock format [cast epoch $date] -format "%A"]
}

proc qc::date_day_shortname { date } {
    #| Return the day of the week as Mon,Tue,Wed,Thu,Fri,Sat,Sun
    return [clock format [cast epoch $date] -format "%a"]
}

proc qc::date_month_shortname { date } {
    #| Return the short month name Jan,Feb,Mar etc.
    return [clock format [cast epoch $date] -format "%b"]
}

proc qc::date_month_name { date } {
     # Return the full month name January,February,March etc.
    return [clock format [cast epoch $date] -format "%B"]
}

proc qc::date_compare { date1 date2 } {  
    #| Compare 2 date expressions and return 1,0,-1 if date1 is greater,equal or less than date2
    set epoch1 [cast epoch $date1]
    set epoch2 [cast epoch $date2]
    if { $epoch1 > $epoch2 } {
	return 1
    } elseif { $epoch1 == $epoch2 } {
	return 0
    } else {
	return -1
    }
}

proc qc::date_quarter {date} {
    #| Return the quarter the date is in
    return [qc::cast integer [expr {ceil(double([date_month $date])/3)}]]
}

proc qc::date_quarter_start {date} {
    #| Return the date at the beginning of the quarter
    set year  [date_year $date]
    set month [qc::cast integer [expr {ceil(double([date_month $date])/3)*3-2}]]
    return [cast date "$year-$month-01"]
}

proc qc::date_quarter_end {date} {
    #| Return the date at the end of the quarter
    set year  [clock format [cast epoch $date] -format %Y]
    set month [qc::cast integer [expr {ceil(double([date_month $date])/3)*3}]]
    return [cast date [clock scan "+1 month -1 day" -base [clock scan "$year-$month-01"]]]
}

proc qc::dates {from_date to_date} {
    #| Return a list of dates from $from_date to $to_date inclusive.
    set dates {}
    set date [cast date $from_date]
    while {[date_compare $date $to_date]<1} {
	lappend dates $date
	set date [cast date "$date +1 day"]
    }
    return $dates
}

proc qc::date_days {from_date to_date} {
    #| Return the number of days between from_date and to_date
    # Rounding to cope with daylight saving time, leap-seconds, etc.
    return [qc::round [expr {([cast epoch $to_date]-[cast epoch $from_date])/(60.0*60*24)}] 0]
}

proc qc::years {from_date to_date} {
    #| Returns list of years between from_date & to_date
    set years {}
    for {set year [date_year $from_date]} {$year <= [date_year $to_date]} {incr year} {
	lappend years $year
    }
    return $years
}

proc qc::iso_year_iso_weeks {from_date to_date} {
    # Return list of iso years/iso week pairs between from_date & to_date
    set year_weeks {}
    set week [date_iso_week $from_date]
    set year [date_iso_year $from_date]
    set to_week [date_iso_week $to_date]
    set to_year [date_iso_year $to_date]

    while {$year<$to_year || ($year==$to_year && $week<=$to_week) } {
	lappend year_weeks $year $week
	set last_week [date_iso_week [date_year_iso_end ${year}-01-01]]
	if { $week==$last_week } { set week 1; incr year } else { incr week }
    }
    return $year_weeks
}

proc qc::year_months {from_date to_date} {
    #| Returns list of iso year/iso month pairs between from_date & to_date
    set year_months {}
    set month [date_month $from_date]
    set year [date_year $from_date]
    set to_month [date_month $to_date]
    set to_year [date_year $to_date]

    while {$year<$to_year || ($year==$to_year && $month<=$to_month) } {
	lappend year_months $year $month
	if { $month==12 } { set month 1; incr year } else { incr month }
    }
    return $year_months
}

proc qc::year_quarters {from_date to_date} {
    #| Returns list of iso year/quarter pairs between from_date & to_date
    set year_quarters {}
    set quarter [date_quarter $from_date]
    set year [date_year $from_date]
    set to_quarter [date_quarter $to_date]
    set to_year [date_year $to_date]

    while {$year<$to_year || ($year==$to_year && $quarter<=$to_quarter) } {
	lappend year_quarters $year $quarter
	if { $quarter==4 } { set quarter 1; incr year } else { incr quarter }
    }
    return $year_quarters
}

proc qc::time_hour { datetime } {
    #| Return the hour
    return [qc::cast integer [clock format [cast epoch $datetime] -format "%H"]]
}

proc qc::date_year_is_leap {year} {
    #| Check if a year is a leap year
    if { $year % 4 == 0
         && ( $year % 100 > 0
              ||
              $year % 400 == 0 ) } {
        return true
    }
    return false
}

namespace eval qc {
    namespace export \
        aws_metadata \
}

proc qc::aws_metadata { category } {
    #| Simple wrapper which queries AWS instance metadata for the requested category
    # eg. a request for 
    # http://169.254.169.254/latest/meta-data/placement/availability-zone
    # would become
    # qc::aws_metadata placement/availability-zone
    # See http://docs.aws.amazon.com/AWSEC2/latest/UserGuide/AESDG-chapter-instancedata.html
    # for a full list of cetegories supported.

    # Token init
    if { ![info exists ::env(AWS_METADATA_TOKEN)] } {
        qc::aws_metadata_token_refresh ::env(AWS_METADATA_TOKEN)
    }

    # Get cached token
    set token $::env(AWS_METADATA_TOKEN)

    ::try {
        return [qc::aws_metadata_get $token $category]
    } trap {IMDSV2_TOKEN_EXPIRED} {} {
        set token [qc::aws_metadata_token_refresh ::env(AWS_METADATA_TOKEN)]
        # Retry
        return [qc::aws_metadata_get $token $category]
    }
}

proc qc::aws_metadata_get { token category } {
    #| Get IMDS request
    set result [qc::http_get \
            -headers [list "X-aws-ec2-metadata-token" $token] \
            -noproxy \
            -response_code true \
            -valid_response_codes [list 200 401] \
            http://169.254.169.254/latest/meta-data/$category \
        ]
    set http_response_code [dict get $result code]
    if { $http_response_code == 401 } {
        # expired token
        error "IMDSv2 token expired." {} {IMDSV2_TOKEN_EXPIRED}
    }
    return [dict get $result body]
}

proc qc::aws_metadata_token_refresh { token_cache } {
    #| Refresh and cache metadata token
    set token [qc::http_put \
                -data "" \
                -headers [list X-aws-ec2-metadata-token-ttl-seconds 21600] \
                http://169.254.169.254/latest/api/token \
              ]
    set $token_cache $token
    return $token
}

namespace eval qc {
    namespace export html_table_db html_table_db_flexgrid
}

proc qc::html_table_db {args} {
    # cols tbody tfoot class height id initialFocus enabled addURL updateURL deleteURL dataURL
    set varNames [qc::args2vars $args]
 
    if { [info exists data] && ![info exists tbody] } {
	set tbody [qc::html_table_tbody_from_ldict $data $cols]
	ldelete varNames [lsearch $varNames data]
	lappend varNames tbody
    }

    if { [info exists table] && ![info exists tbody] } {
	set tbody [lrange $table 1 end]
	ldelete varNames [lsearch $varNames data]
	lappend varNames tbody
    }
    
    # col types
    # encode all text columns by escaping html
    set colIndex 0
    foreach col $cols {
	if { ![dict exists $col type] || [eq [dict get $col type] text] } {
	    dict set col type text
	    for {set rowIndex 0} {$rowIndex < [llength $tbody]} {incr rowIndex} {
		# Try to test if the cell exists but lindex only returns empty string
		if { [ne [set cell [lindex $tbody $rowIndex $colIndex]] ""] } {
		    lset tbody $rowIndex $colIndex [html_escape $cell]
		}
	    }
	}
	incr colIndex
    }
    if { ![info exists class] } {
	set class [list "db-grid" "fixed" "grid" "status-framed"]
    } else {
	if { "db-grid" ni $class } {
	    lappend class "db-grid"
	}
	if { "fixed" ni $class && "flex" ni $class } {
	    lappend class "fixed"
	}
	if { "grid" ni $class } {
	    lappend class "grid"
	}
        if { "status-framed" ni $class } {
            lappend class "status-framed"
        }
    }
    # rowHeight
    if { [info exists rowHeight] } {
	append html "<style type=\"text/css\">table.$class tr { height:${rowHeight}px;vertical-align:top}</style>"
    }
    default height 500
    default id oDbGrid
    default enabled true
    #
    if { [string is true $enabled] } {
	set headers [ns_conn outputheaders]
	ns_set update $headers Expires "Fri, 01 Jan 1990 00:00:00 GMT"
	ns_set update $headers Pragma no-cache
	ns_set update $headers Cache-Control "no-cache, no-store, max-age=0, must-revalidate"
    }

    set scrollHeight $height
    lappend varNames scrollHeight

    lappend varNames class id
    append html [qc::html_table [dict_from {*}[lsort -unique $varNames]]]
    return $html
}

proc qc::html_table_db_flexgrid {args} {
    # cols tbody tfoot class height id initialFocus enabled addURL updateURL deleteURL dataURL sortable
    set varNames [qc::args2vars $args]
    if { [info exists rowHeight] } {
        append html "<style type=\"text/css\">table.db-flex-grid tr { height:${rowHeight}px;vertical-align:top}</style>"
    }
    default id oDbFlexGrid
    default enabled true
    #
    if { [string is true $enabled] } {
        set headers [ns_conn outputheaders]
        ns_set update $headers Pragma no-cache
        ns_set update $headers Cache-Control no-cache
    }
    append html  "<div class=\"db-flex-grid-wrapper\">\n"

    if { ![info exists class] } {
	set class [list "db-grid" flex grid]
    } else {
        if { "db-grid" ni $class } {
            lappend class "db-grid"
        }
        if { "flex" ni $class } {
            lappend class flex
        }
	if { "grid" ni $class } {
	    lappend class grid
	}
    }

    lappend varNames class id
    set dict [dict_from {*}[lsort -unique $varNames]]

    append html [qc::html_table $dict]
    append html "</div>\n"
    return $html
}

namespace eval qc {
    namespace export session_* anonymous_session_id anonymous_user_id
}

proc qc::session_new { user_id } {
    #| Create a new session
    # Grab some random entropy
    set file [open /dev/urandom]
    fconfigure $file -translation binary
    set entropy1 [read $file 50]
    set entropy2 [read $file 50]
    close $file
    set uuid [qc::uuid]
    set session_id [qc::sha1 "$uuid $entropy1"]
    set authenticity_token [qc::sha1 $entropy2]
    if { [qc::conn_open]} {
        set ip [qc::conn_remote_ip]
    } else {
        set ip ""
    }
    db_dml "insert into session [sql_insert session_id ip user_id authenticity_token]"    
    return $session_id
}

proc qc::session_authenticity_token {session_id} {
    #| Return the authenticity token
    db_1row {select authenticity_token from session where session_id=:session_id}
    return $authenticity_token
}

proc qc::session_update { session_id } {
    #| Update a session
    db_0or1row {select hit_count from session where session_id=:session_id} {
	return -code error -errorcode USER "Invalid Session $session_id"
    } {
	incr hit_count
	set ip [qc::conn_remote_ip]
	set time_modified [cast timestamp now]
	db_dml "update session set [sql_set hit_count ip time_modified] where session_id=:session_id"
	return $session_id
    }
}

proc qc::session_sudo_logout {session_id} {
    #| Go back to using own id
    db_dml {update session set effective_user_id=NULL where session_id=:session_id}
    global current_user_id
    if { [info exists current_user_id] } {
        unset current_user_id
    }
}

proc qc::session_kill {session_id} {
    #| Kill a session
    db_dml "update session set deleted=true where session_id=:session_id"
}

proc qc::session_exists {session_id} {
    #| Test if a session exists
    db_1row {select count(*) as count from session where session_id=:session_id and not deleted}
    if { $count==1 } {
	return true
    } else {
	return false
    }
}

proc qc::session_valid {args} {
    #| Check session exists and has not expired.
    # Anonomous user sessions never expire
    qc::args $args -age_limit ? -idle_timeout ? -- session_id
    
    # Check config parameters for session age limit and idle timeout.
    if { [qc::param_exists session_age_limit] } {
        default age_limit [qc::param_get session_age_limit]
    } else {
        default age_limit "12 hours"
    }
    if { [qc::param_exists session_idle_timeout] } {
        default idle_timeout [qc::param_get session_idle_timeout]
    } else {
        default idle_timeout "1 hour"
    }
    
    set qry {
        select
        user_id,
        CASE 
        WHEN (current_timestamp-time_modified) >:idle_timeout::interval
        THEN true ELSE false
        END as idle_timeout_exceeded,
        CASE
        WHEN (current_timestamp-time_created) >:age_limit::interval
        THEN true ELSE false
        END as age_limit_exceeded
        from session
        where
        session_id=:session_id
        and not deleted
    }
    db_0or1row $qry {
	return false
    } {
        if { $user_id == [qc::anonymous_user_id] } {
            # Anonomous sessions never expire
            return true
        } elseif { $age_limit_exceeded || $idle_timeout_exceeded } {
            return false
        } else {
            return true
        }
    }
}

proc qc::session_user_id {session_id} {
    #| Return the user_id owner of this session
    db_1row {select coalesce(effective_user_id,user_id) as user_id from session where session_id=:session_id}
    return $user_id
}

proc qc::session_sudo {session_id effective_user_id} {
    #| Sudo enables a user to use the priviliges of another user.
    db_dml {update session set effective_user_id=:effective_user_id where session_id=:session_id}
    global current_user_id
    set current_user_id $effective_user_id
}

proc qc::session_purge { {timeout_secs 0 } } {
    #| Purge all sessions older than time_out_secs
    log Notice "session purge older than $timeout_secs secs"
    db_dml {
        delete from session
        where (extract(epoch from current_timestamp)-extract(epoch from time_modified)) > :timeout_secs
    }
}

proc qc::session_id {} {
    #| Return the current user's session_id
    global session_id
    if { [info exists session_id] } {
        return $session_id
    }
    if { [cookie_exists session_id] } {
        return [set session_id [cookie_get session_id]]
    }
    return -code error "No known session_id"
}

proc qc::anonymous_session_id {} {
    #| Return the anonymous session ID.
    set user_id [qc::anonymous_user_id]
    qc::db_0or1row {
        SELECT session_id,(current_timestamp-time_created)>'1 hour'::interval as old 
        FROM session
        WHERE user_id=:user_id
        and not deleted
        order by time_created DESC LIMIT 1
    } {
        # No session found
        return [qc::session_new $user_id] 
    } {
        # Session found
        if { $old } {
            return [qc::session_new $user_id] 
        } else {
            return $session_id
        }
    }
}

proc qc::anonymous_user_id {} {
    #| Return the anonymous user ID.
    return -1
}

namespace eval qc {
    namespace export register validate
}

proc qc::register {args} {
    #| Register a URL handler.
    #| Switch for -no-auth has been deprecated and replaced by -no-authenticate 
    if { [llength $args] < 2 || [llength $args] > 6 } {
        return -code error "Usage: qc::register ?-no-authenticate? ?-no-authorise? method path ?args? ?body?"
    }
    qc::args $args -no-auth -no-authenticate -no-authorise -- args
    if { [info exists no-auth] } {
        # Continue to support deprecated switch
        set no-authenticate true
    }
    
    set method [string toupper [lindex $args 0]]
    set path [lindex $args 1]
    # Add to registered nsv dict if not already registered
    if { ! [qc::registered $method $path] } {
        qc::nsv_dict lappend registered $method $path
    }

    # Check authentication exemption
    if { ! [info exists no-authenticate] && ! [qc::registered authenticate $method $path]} {
        qc::nsv_dict lappend authenticate $method $path
    } elseif { [info exists no-authenticate] && [qc::registered authenticate $method $path] } {
        # Previously registered but now exempt from authentication - remove from nsv dict
        set paths [qc::nsv_dict get authenticate $method]
        set paths [qc::lexclude $paths $path]
        qc::nsv_dict set authenticate $method $paths
    }

    # Check authorisation exemption
    if { ! [info exists no-authorise] && ! [qc::registered authorise $method $path]} {
        qc::nsv_dict lappend authorise $method $path
    } elseif { [info exists no-authorise] && [qc::registered authorise $method $path] } {
        # Previously registered but now exempt from authorisation - remove from nsv dict
        set paths [qc::nsv_dict get authorise $method]
        set paths [qc::lexclude $paths $path]
        qc::nsv_dict set authorise $method $paths
    }
    
    if { [llength $args] >= 3 } {
        set proc_args [lindex $args 2]
        # Separate arg names and default values
        set arg_names {}
        set defaults {}
        foreach arg $proc_args {
            if {[llength $arg] == 2} {
                lappend arg_names [lindex $arg 0]
                dict set defaults [lindex $arg 0] [lindex $arg 1]
            } else {
                lappend arg_names $arg
            }
        }

        # Check that colon variables in the path appear in the args
        set path_parts [split $path /]
        foreach path_part $path_parts {
            if { [string first : $path_part] == 0 && [lsearch -exact $arg_names [string range $path_part 1 end]] == -1} {
                return -code error "Colon variable \"$path_part\" missing from args."
            }
        }
    }

    if { [llength $args] == 4 } {
        # Prepend call to arg simplifier to the procedure body
        set proc_body "qc::args_qualified2unqualified; [lindex $args 3]"
        # Create the proc
        namespace eval ::${method} {}
        set proc_name "::${method}::$path"
        {*}[list proc $proc_name $proc_args $proc_body]

        # Update the handlers nsv dict
        if { ![nsv_exists "handlers.patterns" $method]
             || $path ni [nsv_get "handlers.patterns" $method] } {
            nsv_lappend "handlers.patterns" $method $path
        }
        nsv_set "handlers.${method}.proc_names" $path $proc_name
        nsv_set "handlers.${method}.args" $path $arg_names
        nsv_set "handlers.${method}.defaults" $path $defaults
    }
}

proc qc::validate {method path proc_args proc_body} {
    #| Register a URL handler for custom validation.
    set method [string toupper $method]
    set proc_name "::${method}::VALIDATE::$path"
    namespace eval ::${method}::VALIDATE {}
    {*}[list proc $proc_name $proc_args "qc::args_qualified2unqualified; $proc_body"]

    # Separate arg names and default values
    set arg_names {}
    set defaults {}
    foreach arg $proc_args {
        if {[llength $arg] == 2} {
            lappend arg_names [lindex $arg 0]
            dict set defaults [lindex $arg 0] [lindex $arg 1]
        } else {
            lappend arg_names $arg
        }
    }
    
    # Update the handlers nsv array.
    if { ![nsv_exists "handlers.VALIDATE.patterns" $method]
         || $path ni [nsv_get "handlers.VALIDATE.patterns" $method] } {
        nsv_lappend "handlers.VALIDATE.patterns" $method $path
    }
    nsv_set "handlers.VALIDATE.${method}.proc_names" $path $proc_name
    nsv_set "handlers.VALIDATE.${method}.args" $path $arg_names
    nsv_set "handlers.VALIDATE.${method}.defaults" $path $defaults
}

proc qc::registered {args} {
    #| Checks if the given method url_path is registered.
    if { [llength $args] < 2 || [llength $args] > 3 } {
        return -code error "Usage: qc::registered ?service? method path"
    }
    
    if { [llength $args] == 3 } {
        set service [lindex $args 0]
        set method [string toupper [lindex $args 1]]
        set path [lindex $args 2]
    } else {
        set service registered
        set method [string toupper [lindex $args 0]]
        set path [lindex $args 1]
    }
    
    if { [qc::nsv_dict exists $service $method] } {
        return [qc::path_matches $path [qc::nsv_dict get $service $method]]
    } else {
        return false
    }
}

proc qc::path_matches {path patterns} {
    #| Checks if the given path matches any of the given patterns.
    foreach pattern $patterns {
        set path_parts [split $path /]
        set pattern_parts [split $pattern /]
        # check number of parts in each path are equal
        if { [llength $path_parts] == [llength $pattern_parts] } {
            # check that each part matches
            set parts_equal true
            foreach path_part $path_parts pattern_part $pattern_parts {
                if {[string index $pattern_part 0] eq ":"} {
                    # if the item part is a colon variable
                    continue
                } elseif {$path_part ne $pattern_part} {
                    set parts_equal false
                    break
                }
            }
            if {$parts_equal} {
                return true
            }
        }
    }
    return false
}

proc qc::path_best_match {path patterns} {
    #| Finds the best match to path from the given patterns.
    set path_parts [split $path /]
    set matches {}
    foreach pattern $patterns {
        set pattern_parts [split $pattern /]
        # Check if path and pattern have the same number of parts
        if { [llength $path_parts] != [llength $pattern_parts] } {
            continue
        }
        set parts_equal true
        foreach path_part $path_parts pattern_part $pattern_parts {
            if { $path_part ne $pattern_part && [string first : $pattern_part] != 0 } {
                # Parts are not the same and the pattern part is not a colon variable
                set parts_equal false
            }
        }

        if { ! $parts_equal } {
            continue
        }

        # count the number of colon variables in the pattern
        set colon_variable_count [regexp -all {[^:]:[^:]+(?=/|$)} $pattern]
        # Add it to the matches dict
        dict lappend matches $colon_variable_count $pattern
    }
    
    # Get the lowest count of colon variables in a matched pattern.
    foreach count [dict keys $matches] {
        if { ! [info exists best_match_count] || $count < $best_match_count} {
            set best_match_count $count
        }
    }
    
    # If there were matches found then return one with the least amount of colon variables.
    if { [dict size $matches] > 0 } {
        return [lindex [dict get $matches $best_match_count] 0]
    } else {
        return -code error "No pattern matched path: \"$path\""
    }
}

proc qc::path_variables {path pattern} {
    #| Gets variables from the path that corresponds to colon variables in the pattern.
    set path_parts [split $path /]
    set pattern_parts [split $pattern /]
    set variables {}
    foreach path_part $path_parts pattern_part $pattern_parts {
        if { [string first : $pattern_part ] == 0 } {
            # The pattern part is a colon variable - store the path_part
            dict set variables [string range $pattern_part 1 end] $path_part
        }
    }
    return $variables
}

namespace eval qc {
    namespace export format_date format_date_*
}

proc qc::format_date { date } {
    #| Format a date for the user
    #| Will be customizable in future but at present chooses the ISO format.
    return [string map [list - "&#8209;"] [clock format [cast epoch $date] -format "%Y-%m-%d"]]
}

proc qc::format_date_iso { date } {
    #| Format a date as an ISO 8601 string like 2006-04-28
    return [clock format [cast epoch $date] -format "%Y-%m-%d"]
}

proc qc::format_date_uk { date } {
    #| Format a date in UK format e.g. 27/03/07
    return [clock format [cast epoch $date] -format "%d/%m/%y"]
}

proc qc::format_date_uk_long { date } {
    #| Format a date in UK format with a 4 digit year e.g. 27/03/2007
    return [clock format [cast epoch $date] -format "%d/%m/%Y"]
}

proc qc::format_date_rel { date } {
    #| Format the date relatively depending on age
    #| dates this month -> Wed 3rd
    #| dates this year -> JUN 3rd
    set epoch [cast epoch $date]
    set epoch_now [clock seconds]
    # Today 
    if { [string equal [clock format $epoch_now -format "%Y-%m-%d"] [clock format $epoch -format "%Y-%m-%d"]] } {
        return "Today"
    }
    if { [string equal [clock format $epoch_now -format "%Y"] [clock format $epoch -format "%Y"]] } {
	# this year
	if { [string equal [clock format $epoch_now -format "%m"] [clock format $epoch -format "%m"]] } {
	    # same month
	    set dom [clock format $epoch -format "%e"]
	    set dow [clock format $epoch -format "%a"]
	    return "$dow [format_ordinal $dom]"; # Wed 3rd
	} else {
	    set dom [clock format $epoch -format "%e"] 
	    set mon [string toupper [clock format $epoch -format "%b"]]
	    return "$mon [format_ordinal $dom]"; # JUN 3rd
	}
    } else {
	return [clock format $epoch -format "%Y-%m-%d"]; # 2007-05-06
    }
}

proc qc::format_date_letter { date } {
    #| Format a date as would be used on a letter
    # 2007-04-12 -> 12th April 2007
    set epoch [cast epoch $date]
    set dom [clock format $epoch -format "%e"] 
    return "[format_ordinal $dom] [clock format $epoch -format "%B %Y"]"
}


namespace eval qc::aws {
    namespace export s3
    namespace ensemble create
}
namespace eval qc::aws::s3 {
    namespace export ls
    namespace ensemble create

    proc ls { args } {
        #| Lists the contents of a bucket optionally filtered by objects 
        #| starting with object_key_prefix
        #| Usage: qc::aws s3 ls mybucket ?prefix?
        qc::args $args -max_keys 1000 -timeout 60 -- bucket {object_key_prefix ""}

        set query_params [dict create \
                            max-keys $max_keys \
                            prefix $object_key_prefix \
                         ]
        set s3_uri [qc::cast s3_uri $bucket]
        set response [qc::aws s3 rest_api http_get -timeout $timeout $s3_uri $query_params]
        return [qc::aws s3 xml2ldict $response {/ns:ListBucketResult/ns:Contents}]
    }
}

namespace eval qc {
    namespace export db_init
}

proc qc::db_init {} {
    #| Initialises the database objects required by qcode-tcl

    # param table
    db_dml {
	CREATE TABLE IF NOT EXISTS param (
					  param_name varchar(50) PRIMARY KEY,
					  param_value text
					  );
    }

    # plain_text
    if { ![qc::db_domain_exists -no-cache plain_text]} {
	db_dml {
	    create domain plain_text as text
	    check(
		  value !~ '<([^>]+)>'
		  );
	}
    }

    # plain_string
    if {![qc::db_domain_exists -no-cache plain_string]} {
	db_dml {
	    create domain plain_string as varchar(255)
	    check(
		  value !~ '<([^>]+)>'
		  );
	}
    }

    # url
    if {![qc::db_domain_exists -no-cache url]} {
	db_dml {
	    create or replace function is_url(text)
	    returns boolean as \$\$
	    select \$1 ~ '^https?://[a-z0-9\-\.]+(:[0-9]+)?([a-zA-Z0-9_\-\.~+/%]+)?(\?[a-zA-Z0-9_\-\.~+/%=&]+)?(\#[a-zA-Z0-9_\-\.~+/%]+)?$';
	    \$\$ language sql immutable strict;

	    create domain url as text
	    check(
		  is_url(value)
		  );
	}
    }

    # url_path
    if {![qc::db_domain_exists -no-cache url_path]} {
	db_dml {
	    create or replace function is_url_path(text)
	    returns boolean as \$\$
	    select \$1 ~ '/([a-zA-Z0-9\-._~]|%[0-9a-fA-F]{2}|[!$&''()*+,;=:@]|/)*$';
	    \$\$ language sql immutable strict;

	    create domain url_path as text
	    check(
		  is_url_path(value)
		  );
	}
    }   
    
    # user_state
    if {![qc::db_enum_exists -no-cache user_state]} {
	db_dml {
	    CREATE TYPE user_state AS ENUM ('ACTIVE', 'DISABLED');
	}    
    }

    # perm_method
    if {![qc::db_enum_exists -no-cache perm_method]} {
	db_dml {
	    CREATE TYPE perm_method AS ENUM ('READ', 'WRITE', 'APPEND');
	}
    }

    
    # sequence
    foreach sequence {
	user_id_seq 
	file_id_seq 
	perm_category_id_seq 
	perm_class_id_seq 
	perm_id_seq
    } {
	if {![qc::db_sequence_exists -no-cache $sequence]} {
	    db_dml {
		CREATE SEQUENCE [db_quote_identifier $sequence];
	    }	
	}
    }

    # Make sure pgcrypto extension exists
    if {![qc::db_extension_exists pgcrypto]} {
        # We need a superuser, or the nearest RDS equivalent
        if { [qc::db_user_is_superuser [qc::db_user]]
             || [qc::db_user_is_member [qc::db_user] "rds_superuser"]
         } {
            db_dml {
	        CREATE EXTENSION pgcrypto;
            }
        } else {
            error "Database user \"[qc::db_user]\" is not superuser so pgcrypto extension cannot be created"
        }
    }
    
    # pgcryto
    if {[qc::db_extension_exists pgcrypto]} {
	db_dml {
	    CREATE OR REPLACE FUNCTION sha1(bytea) returns text AS $$
	    SELECT encode(digest($1, 'sha1'), 'hex')
	    $$ LANGUAGE SQL STRICT IMMUTABLE;
	}    
    } else {
	error "Extension \"pgcrypto\" is not installed" 
    }


    # qcode-tcl required tables
    db_dml {
	CREATE TABLE IF NOT EXISTS users (
					  user_id int PRIMARY KEY,
					  firstname varchar(255) NOT NULL,
					  surname varchar(255) NOT NULL,
					  email varchar(255) NOT NULL,
					  password_hash varchar(60) NOT NULL,
					  user_state user_state NOT NULL DEFAULT 'ACTIVE',
					  ip varchar(15)
					  );

	CREATE TABLE IF NOT EXISTS file (
					 file_id integer PRIMARY KEY DEFAULT nextval('file_id_seq'::regclass),
					 user_id integer NOT NULL REFERENCES users,
					 filename text NOT NULL,
					 data bytea,
                                         s3_location text,
					 upload_date timestamp without time zone DEFAULT now(),
					 mime_type text NOT NULL
					 );

        CREATE TABLE IF NOT EXISTS image (
					  file_id int primary key references file on delete cascade,
					  width int,
					  height int
					  );

	CREATE TABLE IF NOT EXISTS validation_messages (
							table_name text NOT NULL,
							column_name text NOT NULL,
							message varchar(255) NOT NULL
							);

	

	


	CREATE TABLE IF NOT EXISTS session (
					    session_id text PRIMARY KEY,
					    time_created timestamp(0) WITHOUT time zone DEFAULT now() NOT NULL,
					    time_modified timestamp(0) WITHOUT time zone DEFAULT now() NOT NULL,
					    hit_count integer DEFAULT 0 NOT NULL,
					    ip varchar(15),
					    user_id integer NOT NULL REFERENCES users,
					    effective_user_id integer,
					    authenticity_token varchar(100)
					    );

	CREATE TABLE IF NOT EXISTS schema (
					   version int NOT NULL
					   );
    }

    # Initialise schema version.
    db_0or1row {select version from schema} {
	db_dml "INSERT INTO schema VALUES(1);"
    }
    
    db_dml {
	CREATE TABLE IF NOT EXISTS perm_category (
						  perm_category_id int PRIMARY KEY,
						  description text
						  );


	CREATE TABLE IF NOT EXISTS perm_class (
					       perm_class_id int PRIMARY KEY,
					       perm_name varchar(50) NOT NULL UNIQUE,
					       description text,
					       perm_category_id int REFERENCES perm_category
					       );


	CREATE TABLE IF NOT EXISTS perm (
					 perm_id int PRIMARY KEY,
					 perm_class_id int NOT NULL REFERENCES perm_class ON DELETE CASCADE,
					 method perm_method NOT NULL,
					 unique(perm_class_id, method)
					 );

	CREATE TABLE IF NOT EXISTS user_perm (
					      user_id int NOT NULL REFERENCES users ON DELETE CASCADE,
					      perm_id int NOT NULL REFERENCES perm ON DELETE CASCADE,
					      unique(user_id, perm_id)
					      );


	CREATE TABLE IF NOT EXISTS sticky (
					   user_id integer NOT NULL REFERENCES users(user_id) ON DELETE CASCADE,
					   name varchar(255),
					   value varchar(255) ,
					   url varchar(255)
					   );

	CREATE TABLE IF NOT EXISTS file_alias_path (
						    url_path text PRIMARY KEY,
						    file_id int REFERENCES file(file_id) NOT NULL
						    );
    }
    
    # Utility tables for validation
    db_dml {
	CREATE TABLE IF NOT EXISTS form ();
	CREATE TABLE IF NOT EXISTS required ();
	CREATE TABLE IF NOT EXISTS optional ();
    }

    # Create anonymous user if none exists
    qc::db_0or1row {SELECT user_id FROM users WHERE user_id='-1'} {
	set password_hash [qc::password_hash TheAnonymousUserCreepsSilentlyIntoTheDatabase]
	qc::db_dml {
	    INSERT INTO
	    users (user_id, firstname, surname, email, password_hash)
	    VALUES
	    (-1, 'anonymous', 'anonymous', 'anonymous@nowhere.co.uk', :password_hash);
	}
    }
}

namespace eval qc {
    namespace export json2tson
}

# Taken from tclib code by andreas_kupries

proc qc::json2tson {json} {
    return [qc::json::json2tson $json]
}

namespace eval qc::json {
    # Regular expression for tokenizing a JSON text (cf. http://json.org/)

    # tokens consisting of a single character
    variable singleCharTokens { "{" "}" ":" "\\[" "\\]" "," }
    variable singleCharTokenRE "\[[join $singleCharTokens {}]\]"

    # quoted string tokens
    variable escapableREs { "[\\\"\\\\/bfnrt]" "u[[:xdigit:]]{4}" }
    variable escapedCharRE "\\\\(?:[join $escapableREs |])"
    variable unescapedCharRE {[^\\\"]}
    variable stringRE "\"(?:$escapedCharRE|$unescapedCharRE)*\""

    # (unquoted) words
    variable wordTokens { "true" "false" "null" }
    variable wordTokenRE [join $wordTokens "|"]

    # number tokens
    # negative lookahead (?!0)[[:digit:]]+ might be more elegant, but
    # would slow down tokenizing by a factor of up to 3!
    variable positiveRE {[1-9][[:digit:]]*}
    variable cardinalRE "-?(?:$positiveRE|0)"
    variable fractionRE {[.][[:digit:]]+}
    variable exponentialRE {[eE][+-]?[[:digit:]]+}
    variable numberRE "${cardinalRE}(?:$fractionRE)?(?:$exponentialRE)?"

    # JSON token
    variable tokenRE "$singleCharTokenRE|$stringRE|$wordTokenRE|$numberRE"

    # 0..n white space characters
    set whiteSpaceRE {[[:space:]]*}

    # Regular expression for validating a JSON text
    variable validJsonRE "^(?:${whiteSpaceRE}(?:$tokenRE))*${whiteSpaceRE}$"
}

# Validate JSON text
# @param jsonText JSON text
# @return 1 iff $jsonText conforms to the JSON grammar
#           (@see http://json.org/)
proc qc::json::validate {jsonText} {
    variable validJsonRE

    return [regexp -- $validJsonRE $jsonText]
}

proc qc::json::json2tson {jsonText} {
    #| Convert json to tson
    variable tokenRE

    set tokens [regexp -all -inline -- $tokenRE $jsonText]
    set nrTokens [llength $tokens]
    set tokenCursor 0
    return [parseValue $tokens $nrTokens tokenCursor]
}

# Throw an exception signaling an unexpected token
proc qc::json::unexpected {tokenCursor token expected} {
    return -code error -errorcode [list JSON UNEXPECTED $tokenCursor $expected] \
	"unexpected token \"$token\" at position $tokenCursor; expecting $expected"
}

# Get rid of the quotes surrounding a string token and substitute the
# real characters for escape sequences within it
# @param token
# @return unquoted unescaped value of the string contained in $token
proc qc::json::unquoteUnescapeString {token} {
    set unquoted [string range $token 1 end-1]
    return [subst -nocommands -novariables $unquoted]
}

# Parse an object member
# @param tokens list of tokens
# @param nrTokens length of $tokens
# @param tokenCursorName name (in caller's context) of variable
#                        holding current position in $tokens
# @param objectDictName name (in caller's context) of dict
#                       representing the JSON object of which to
#                       parse the next member
proc qc::json::parseObjectMember {tokens nrTokens tokenCursorName objectDictName} {
    upvar $tokenCursorName tokenCursor
    upvar $objectDictName objectDict

    set token [lindex $tokens $tokenCursor]
    incr tokenCursor

    set leadingChar [string index $token 0]
    if {$leadingChar eq "\""} {
        set memberName [unquoteUnescapeString $token]

        if {$tokenCursor == $nrTokens} {
            unexpected $tokenCursor "END" "\":\""
        } else {
            set token [lindex $tokens $tokenCursor]
            incr tokenCursor

            if {$token eq ":"} {
                set memberValue [parseValue $tokens $nrTokens tokenCursor]
                dict set objectDict $memberName $memberValue
            } else {
                unexpected $tokenCursor $token "\":\""
            }
        }
    } else {
        unexpected $tokenCursor $token "STRING"
    }
}

# Parse the members of an object
# @param tokens list of tokens
# @param nrTokens length of $tokens
# @param tokenCursorName name (in caller's context) of variable
#                        holding current position in $tokens
# @param objectDictName name (in caller's context) of dict
#                       representing the JSON object of which to
#                       parse the next member
proc qc::json::parseObjectMembers {tokens nrTokens tokenCursorName objectDictName} {
    upvar $tokenCursorName tokenCursor
    upvar $objectDictName objectDict

    while true {
        parseObjectMember $tokens $nrTokens tokenCursor objectDict

        set token [lindex $tokens $tokenCursor]
        incr tokenCursor

        switch -exact $token {
            "," {
                # continue
            }
            "\}" {
                break
            }
            default {
                unexpected $tokenCursor $token "\",\"|\"\}\""
            }
        }
    }
}

# Parse an object
# @param tokens list of tokens
# @param nrTokens length of $tokens
# @param tokenCursorName name (in caller's context) of variable
#                        holding current position in $tokens
# @return parsed object (Tcl dict)
proc qc::json::parseObject {tokens nrTokens tokenCursorName} {
    upvar $tokenCursorName tokenCursor

    if {$tokenCursor == $nrTokens} {
        unexpected $tokenCursor "END" "OBJECT"
    } else {
        set result [dict create]

        set token [lindex $tokens $tokenCursor]

        if {$token eq "\}"} {
            # empty object
            incr tokenCursor
        } else {
            parseObjectMembers $tokens $nrTokens tokenCursor result
        }

        return $result
    }
}

# Parse the elements of an array
# @param tokens list of tokens
# @param nrTokens length of $tokens
# @param tokenCursorName name (in caller's context) of variable
#                        holding current position in $tokens
# @param resultName name (in caller's context) of the list
#                   representing the JSON array
proc qc::json::parseArrayElements {tokens nrTokens tokenCursorName resultName} {
    upvar $tokenCursorName tokenCursor
    upvar $resultName result

    while true {
        lappend result [parseValue $tokens $nrTokens tokenCursor]

        if {$tokenCursor == $nrTokens} {
            unexpected $tokenCursor "END" "\",\"|\"\]\""
        } else {
            set token [lindex $tokens $tokenCursor]
            incr tokenCursor

            switch -exact $token {
                "," {
                    # continue
                }
                "\]" {
                    break
                }
                default {
                    unexpected $tokenCursor $token "\",\"|\"\]\""
                }
            }
        }
    }
}

# Parse an array
# @param tokens list of tokens
# @param nrTokens length of $tokens
# @param tokenCursorName name (in caller's context) of variable
#                        holding current position in $tokens
# @return parsed array (Tcl list)
proc qc::json::parseArray {tokens nrTokens tokenCursorName} {
    upvar $tokenCursorName tokenCursor

    if {$tokenCursor == $nrTokens} {
        unexpected $tokenCursor "END" "ARRAY"
    } else {
        set result {}

        set token [lindex $tokens $tokenCursor]

        set leadingChar [string index $token 0]
        if {$leadingChar eq "\]"} {
            # empty array
            incr tokenCursor
        } else {
            parseArrayElements $tokens $nrTokens tokenCursor result
        }

        return $result
    }
}

# Parse a value
# @param tokens list of tokens
# @param nrTokens length of $tokens
# @param tokenCursorName name (in caller's context) of variable
#                        holding current position in $tokens
# @return parsed value (dict, list, string, number)
proc qc::json::parseValue {tokens nrTokens tokenCursorName} {
    #| Return tson object
    upvar $tokenCursorName tokenCursor

    if {$tokenCursor == $nrTokens} {
        unexpected $tokenCursor "END" "VALUE"
    } else {
        set token [lindex $tokens $tokenCursor]
        incr tokenCursor

        set leadingChar [string index $token 0]
        switch -exact -- $leadingChar {
            "\{" {
                return [linsert [parseObject $tokens $nrTokens tokenCursor] 0 object]
            }
            "\[" {
                return [linsert [parseArray $tokens $nrTokens tokenCursor] 0 array]
            }
            "\"" {
                # quoted string
                return [list string [unquoteUnescapeString $token]]
            }
            "t" -
            "f" -
            "n" {
                # bare word: true, false or null
                return $token
            }
            default {
                # number?
                if {[string is double -strict $token]} {
                    return $token
                } else {
                    unexpected $tokenCursor $token "VALUE"
                }
            }
        }
    }
}

proc qc::data2json {} {
    #| Deprecated
    return [qc::response2json]
}


namespace eval qc {
    namespace export dict_exists dict_subset dict_exclude dict_sort dict2xml dict_from dict2vars dict_default dicts_equal dict_intersect dict_is_subset
}

proc qc::dict_exists { args } {
    #| Return true if the given key (or path to key) exists. 
    #| Otherwise return false.

    #| Unlike dict exists command, do not fail if path to key does not exist.
    #| Eg: dict_exists [dict create a 1 b 2 c 3] a a1
    #
    # This is fixed apparently in Tcl8.5.12
    args $args dict args

    set key [lindex $args 0]
    set index [lsearch -exact $dict $key]
    while { $index%2!=0 && $index!=-1 && $index<[llength $dict]} {
	incr index
	set index [lsearch -exact -start $index $dict $key]
    }
    if { $index%2 == 0 && [llength $args] > 1 } {
        # key has a value but there are more keys, so recurse with value and rest of keys
        return [dict_exists [lindex $dict [expr {$index+1}]] {*}[lrange $args 1 end]]
    } elseif { $index%2 == 0 && [llength $dict]%2 == 0} {
        # key has value, this is a valid dict and there are no more keys. the end.
        return 1
    } else {
        # key doesn't have value
        return 0
    }
}

proc qc::dict_subset {dict args} {
    #| Return a dict made up of the keys given.
    set result {}
    foreach key $args {
	if { [dict exists $dict $key] } {
	    lappend result $key [dict get $dict $key]
	}
    }
    return $result
}

proc qc::dict_exclude {dict args} {
    #| Return an dict excluding the keys given.
    set temp {}
    foreach {key value} $dict {
	if { ![in $args $key] } {
	    lappend temp $key $value
	}
    }
    return $temp
}

proc qc::dict_sort {dictVariable} {
    #| Sort the top level dict contained in dictVariable by ascending key values.
    #| Write the resulting dict back to dictVariable and return the sorted dict.
    upvar 1 $dictVariable dict
    set llist {}
    foreach {name value} $dict {
	lappend llist [list $name $value]
    }
    set llist [lsort -index 1 $llist]
    set dict {}
    foreach item $llist {
	lappend dict [lindex $item 0] [lindex $item 1] 
    }
    return $dict
}
    

proc qc::dict2xml { dict } {
    #| Convert top level {key value} pairs in dict value to xml elements.
    #| Return xml.
    set list {}
    foreach {name value} $dict {
        lappend list [qc::xml $name $value]
    }
    return [join $list \n]
}

proc qc::dict_from { args } {
    #| Take a list of var names and return a dict.
    set dict {}
    foreach name $args {
	upvar 1 $name value
	if { [info exists value] } {
	    lappend dict $name $value
	} else {
	    error "Can't create dict with $name: No such variable"
	}
    }
    return $dict
}

proc qc::dict2vars { dict args } {
    #| Set all or a subset of the {key value} pairs in dict as variables in the caller.
    #|
    #| If a list of keys is provided only set corresponding variables.
    #| If any of the keys do not exist in the dict unset the variable in the caller if it exists.
    if { [llength $args]==0 } {
	# set all variables
	foreach {name value} $dict {
            upset 1 $name $value
        }
    } else {
	# only set named variables
	foreach name $args {
            if { [dict exists $dict $name] } {
		upset 1 $name [dict get $dict $name]
	    } else {
		if { [uplevel 1 [list info exists $name]] } {
		    uplevel 1 [list unset $name]
		}
	    }
	}
    }
}

proc qc::dict_default {dictVar args} {
    #| Set default values in the dict if they do not exist
    foreach {name value} $args {
        upvar 1 $dictVar dict
        if { ![dict exists $dict $name] } {
            dict set dict $name $value
        }
    }
    return $dict
}

proc qc::dict_zipper {keys values} {
    #| Creates a dictionary by pairing up each key and value from the provided lists.
    set zipped ""
    foreach key $keys value $values {
        dict set zipped $key $value
    }
    return $zipped
}

proc qc::dicts_equal {dict1 dict2} {
    #| Compare 2 dicts for equivalence
    if { [dict size $dict1] != [dict size $dict2] } {
        return false
    }
    foreach {key1 value1} $dict1 {key2 value2} $dict2 {
        if { $key1 ne $key2
             ||
             $value1 ne $value2 } {
            return false
        }
    }
    return true
}

proc qc::dict_intersect {dict1 dict2} {
    #| Returns a dict with key value pairs that
    # are present in both dict1 and dict2
    set result [dict create]

    foreach {key value} $dict1 {
        if { [dict exists $dict2 $key] && [dict get $dict2 $key] eq $value } {
            dict set result $key $value
        }
    }

    return $result    
}

proc qc::dict_is_subset {dict1 dict2} {
    #| Check if dict1 is a subset of dict2
    return [dicts_equal [dict_intersect $dict1 $dict2] $dict1]
}

namespace eval qc {
    namespace export csv2list csv2ll
}

proc qc::csv2list {csv} {
    return [lindex [qc::csv2ll $csv] 0]
}

proc qc::csv2ll {csv } {
    #| Convert csv data to a list of lists
    #| Accepts csv quoted fields separated by commas with records terminated by newlines.
    #| Commas and newlines may appear inside a quoted field.
    # Windows -> UNIX
    regsub -all {\r\n} $csv \n csv
    
    # Optimized for csv with no quotes
    if {[string first \" $csv]==-1} {
	# trim whitespace around fields
	regsub -all {(^|\n|,) +} $csv {\1} csv
	regsub -all { +($|\n|,)} $csv {\1} csv
	set result {}
	foreach line [split $csv \n] {
	    lappend result [split $line ,]
	}
	return $result
    }
    # Escape commas embedded in quoted fields with \0
    # Escape newlines embedded in quoted fields with \1
    # Escape double quotes embedded in quoted fields with \2
    # regsub -all won't work because the regexp need to be applied repeatedly to anchor correctly
    set start 0
    while { $start<[string length $csv] && \
		[regexp -expanded -indices -start $start -- {
		    (?:^|\n|,)\s*
		    \"
		    (
		     ([^\"]|\"\")*
		     (,|\n|"")
		     ([^\"]|\"\")*
		     )
		    \"
		    \s*(?:,|\n|$)
		} $csv -> field] } {
	set field_value [string range $csv [lindex $field 0] [lindex $field 1]]
	set escaped_value [string map [list , \0 \n \1 \"\" \2] $field_value]
	set csv [string replace $csv [lindex $field 0] [lindex $field 1] $escaped_value]
	set start [expr {[lindex $field 0]+[string length $escaped_value]+1}]
    }

    set csv [string map {, \0 \0 ,} $csv] 
    # trim whitespace
    regsub -all {(^|\n|\0) +} $csv {\1} csv
    regsub -all { +($|\n|\0)} $csv {\1} csv
    # remove quotes - no longer needed
    set csv [string map {\" {}} $csv] 

    set result {}
    foreach line [split $csv \n] {
	lappend result [split [string map {\1 \n \2 \"} $line] \0]
    }
    return $result
}

proc qc::csv_file_foreach { filename code } {
    #| Loop through a CSV file setting local variables with values in the row.
    #| Var names are based on the CSV header line.
    set in [open $filename r]
    fconfigure $in -buffering line

    # Load header
    gets $in line
    set keys [lindex [qc::csv2ll $line] 0]

    # Read file line by line
    while { [gets $in line] >= 0 } {
	foreach key $keys value [lindex [qc::csv2ll $line] 0] {
	    upset 1 $key $value
	}
	uplevel 1 $code
    }
    close $in
}

namespace eval qc::handlers {

    namespace export call exists validate2model validation
    namespace ensemble create

    proc call {method path} {
        #| Call the registered handler that matches the given path and method.
        set method [string toupper $method]
        set pattern [qc::path_best_match $path [get $method]]
        # Gather a list of form variables we might be interested in and add in
        # path variables to the form data.
        set args [args $method $pattern]
        foreach arg $args {
            set shortname [qc::arg_shortname $arg]
            if { $shortname ne $arg } {
                lappend args $shortname
            }
        }
        set form [dict merge [qc::form2dict {*}$args] [qc::path_variables $path $pattern]]
        # Cast to model
        set dict [qc::cast_values2model {*}[data $form $method $pattern]]
        # Call handler
        return [qc::call_with [proc_name $method $pattern] {*}$dict]
    }

    proc exists {method path} {
        #| Check if a handler exists for the given path and method.
        if { $method eq "" } {
            return false
        } else {
            return [qc::path_matches $path [get [string toupper $method]]]
        }
    }

    proc validate2model {method path} {
        #| Validates the args of the handler registered for the given path and method.
        set method [string toupper $method]
        set pattern [qc::path_best_match $path [get $method]]
        # Gather a list of form variables we might be interested in and add in
        # path variables to the form data.
        set args [args $method $pattern]
        foreach arg $args {
            set shortname [qc::arg_shortname $arg]
            if { $shortname ne $arg } {
                lappend args $shortname
            }
        }
        set form [dict merge [qc::form2dict {*}$args] [qc::path_variables $path $pattern]]
        # Validate the data.
        return [qc::validate2model [data $form $method $pattern]]
    }

    ##################################################
    #
    # Private procs
    #
    ##################################################

    proc get {{method ""}} {
        #| Get all patterns.
        #| If method has been given then return only patterns that match the given method.
        if { ! [nsv_array exists "handlers.patterns"] } {
            return [list]
        } elseif { $method ne "" } {
            if { [nsv_exists "handlers.patterns" $method] } {
                return [nsv_get "handlers.patterns" $method]
            } else {
                return [list]
            }
        } else {
            return [concat {*}[dict values [nsv_array get "handlers.patterns"]]]
        }
    }
    
    proc proc_name {method pattern} {
        #| Get the proc name for the handler identified by $method $pattern.
        return [nsv_get "handlers.${method}.proc_names" $pattern]
    }

    proc args {method pattern} {
        #| Get all arguments for the given handler identified by $method $pattern.
        return [nsv_get "handlers.${method}.args" $pattern]
    }

    proc default {method pattern arg} {
        #| Get the default value of the given arg for the handler identified by $method $pattern.
        return [qc::nsv_dict get "handlers.${method}.defaults" $pattern $arg]
    }

    proc default_exists {method pattern arg} {
        #| Check if a default argument exists for the given argument for handler identified by $method $pattern.
        return [qc::nsv_dict exists "handlers.${method}.defaults" $pattern $arg]
    }

    proc data {form method pattern} {
        #| Returns a dictionary of data for the handler identified by $method $pattern.
        set method [string toupper $method]
        set args [args $method $pattern]
        set unambiguous [qc::args_unambiguous {*}$args]
        set result {}
        foreach arg $args {
            # Check if a form variable exists for $arg
            if { [dict exists $form $arg] } {
                lappend result $arg [dict get $form $arg]
                continue
            }

            # Check if shortname of $arg is unambiguous and exists as a form variable
            set shortname [qc::arg_shortname $arg]
            if { [llength [lsearch -all $unambiguous $arg]] && [dict exists $form $shortname] } {
                lappend result $arg [dict get $form $shortname]
                continue
            }

            # Check if default value exists for $arg
            if { [default_exists $method $pattern $arg] } {
                lappend result $arg [default $method $pattern $arg]
                continue
            }
            
            # $arg wasn't optional and didn't appear in form
            return \
                -code error \
                -errorcode INVALID_REQUEST \
                "No matching arg value for \"$arg\" in form."
        }
        return $result
    }


    ##################################################
    #
    # qc::handlers validation
    #
    ##################################################
    
    namespace eval validation {

        namespace export call exists
        namespace ensemble create

        proc call {method path} {
            #| Calls the registered handler that matches the given method and path.
            set method [string toupper $method]
            set pattern [qc::path_best_match $path [get $method]]
            # Gather a list of form variables we might be interested in and add in
            # path variables to the form data.
            set args [args $method $pattern]
            foreach arg $args {
                set shortname [qc::arg_shortname $arg]
                if { $shortname ne $arg } {
                    lappend args $shortname
                }
            }
            set form [dict merge [qc::form2dict {*}$args] [qc::path_variables $path $pattern]]
            # Grab relevant arg data from form.
            set dict [qc::cast_values2model {*}[data $form $method $pattern]]
            # Call the validation handler.
            return [qc::call_with [proc_name $method $pattern] {*}$dict]
        }

        proc exists {method path} {
            #| Checks if a validation handler exists for the given path and method.
            return [qc::path_matches $path [get [string toupper $method]]]
        }

        ##################################################
        #
        # qc::handlers validation
        # Private procs
        #
        ##################################################

        proc get {{method ""}} {
            #| Get all the validation handler paths.
            #| If method has been given then return only handlers that match method.
            if { ! [nsv_array exists "handlers.VALIDATE.patterns"] } {
                return [list]
            } elseif { $method ne "" } {
                if { [nsv_exists "handlers.VALIDATE.patterns" $method] } {
                    return [nsv_get "handlers.VALIDATE.patterns" $method]
                } else {
                    return [list]
                }
            } else {
                return [concat {*}[dict values [nsv_array get "handlers.VALIDATE.patterns"]]]
            }
        }
        
        proc proc_name {method pattern} {
            #| Get the validation proc_name for the handler identified by $method $pattern.
            return [nsv_get "handlers.VALIDATE.${method}.proc_names" $pattern]
        }

        proc args {method pattern} {
            #| Get all arguments for the handler identified by $method $pattern.
            return [nsv_get "handlers.VALIDATE.${method}.args" $pattern]
        }

        proc default {method pattern arg} {
            #| Get the default value of the given arg for the handler identified by $method $pattern.
            return [qc::nsv_dict get "handlers.VALIDATE.${method}.defaults" $pattern $arg]
        }

        proc default_exists {method pattern arg} {
            #| Check if a default argument exists for the given argument for handler identified by $method $pattern.
            return [qc::nsv_dict exists "handlers.VALIDATE.${method}.defaults" $pattern $arg]
        }

        proc data {form method pattern} {
            #| Returns a dictionary of args for the handler identified by $method $pattern.
            set method [string toupper $method]
            set args [args $method $pattern]
            set unambiguous [qc::args_unambiguous {*}$args]
            set result {}
            foreach arg $args {
                # Check if a form variable exists for $arg
                if { [dict exists $form $arg] } {
                    lappend result $arg [dict get $form $arg]
                    continue
                }

                # Check if shortname of $arg is unambiguous and exists as a form variable
                set shortname [qc::arg_shortname $arg]
                if { [llength [lsearch -all $unambiguous $arg]] && [dict exists $form $shortname] } {
                    lappend result $arg [dict get $form $shortname]
                    continue
                }

                # Check if default value exists for $arg
                if { [default_exists $method $pattern $arg] } {
                    lappend result $arg [default $method $pattern $arg]
                    continue
                }
                
                # arg wasn't optional and didn't appear in form
                return \
                    -code error \
                    -errorcode INVALID_REQUEST \
                    "No matching arg value for \"$arg\" in form."
            }
            return $result
        }
    }
}

namespace eval qc {
}

proc qc::my {args} {
    #| Multifunction proc to return information about the local OS.
    #| Written to be debian specific but may work on some other Linux distributions.
    switch [lindex $args 0] {
	hostname {
	    return [qc::exec_proxy hostname -s]
	}
	fqdn {
	    return [qc::exec_proxy hostname -f]
	}
	domain {
	    return [qc::exec_proxy hostname -d]
	}
        ip {
	    regexp {inet ([^/]*)/} [qc::exec_proxy ip addr show scope global | fgrep inet] -> ip
	    return $ip
        }
        arch {
            switch [qc::exec_proxy getconf LONG_BIT] {
                64 { return "amd64" }
                32 { return "i386" }
            }
        }
        total_memory {
            #| Returned in kB
            set fp [open /proc/meminfo]
            set meminfo [read $fp]
            close $fp
            
            regexp -all -linestop -lineanchor -- {^\s*MemTotal:\s+([^[:space:]]+) ([^[:space:]]+)$} $meminfo -> total_mem units
            return $total_mem
        }
        instance_id {
            # AWS only
            return [qc::aws_metadata instance-id]
        }
        availability_zone {
            # AWS only
            return [qc::aws_metadata placement/availability-zone]
        }
    }
}


namespace eval qc {
    namespace export reload
}
package require fileutil

proc qc::reload {args} {
    #| Source files that have changed in this project
    if {[llength $args]==0} {
        set args [nsv_array names tcl_libs]
    }
    set reloaded false
    foreach dir $args {
        nsv_set tcl_libs $dir 1
        set files [fileutil::findByPattern $dir "*.tcl"]
        foreach file $files {
            set fh [open $file r]
            set md5 [qc::md5 [read $fh]]
            close $fh
            
            if { ![nsv_exists tcl_lib_md5 $file] } {
                log Notice "Loading $file"
                namespace eval :: [list ns_eval -sync source $file]
                nsv_set tcl_lib_md5 $file $md5
                set reloaded true

            } elseif { $md5 ne [nsv_get tcl_lib_md5 $file] } {
                namespace eval :: [list ns_eval -sync source $file]
                nsv_set tcl_lib_md5 $file $md5
                log Notice "Reloading $file"
                set reloaded true
            } 
        }
    }
    if { $reloaded } {
        ns_memoize_flush
    }
}

namespace eval qc {
    namespace export sql_where sql_where_like sql_where_cols_start sql_where_col_starts sql_where_combo sql_where_compare sql_where_compare_set sql_where_or sql_where_word_in
}

proc qc::sql_where { args } {
    qc::args $args -nocase -type "" -- args
    #| Construct part of SQL WHERE clause using varNames
    #| in a pass-by-name list or a dict.
    #| Any empty values or non-existent variables are ignored
    set dict [qc::args2dict $args]
    set list {}
    foreach {name value} $dict {
        if { [ne $value ""] } {
         if { [string equal $value NULL] } {
                lappend list "$name IS NULL"
         } elseif { [string equal $value "NOT NULL"] } {
                lappend list "$name IS NOT NULL"
         } else {
                if { [info exists nocase] } {
                 lappend list "lower($name)=lower([db_quote $value $type])"
                } else {
                 lappend list "$name=[db_quote $value $type]"
                }
         }
        }
    }
    if { [llength $list]==0 } {
        return true
    } else {
        return [join $list " and "]
    }
}

proc qc::sql_where_like { args } {
    #| Construct part of SQL WHERE clause using varNames
    #| in a pass-by-name list or a dict.
    #| Any empty values or non-existent variables are ignored
    set dict [args2dict $args]
    set list {}
    foreach {name value} $dict {
	if { [info exists value] && [ne $value ""] } {
	    foreach word [split $value] {
		lappend list "$name ~~* [db_quote "%$word%"]"
	    }
	}
    }
    if { [llength $list]==0 } {
	return true
    } else {
	return [join $list " and "]
    }
}

proc qc::sql_where_cols_start { args } {
    #| Construct part of SQL WHERE clause using varNames
    #| corresponding to column names with a regexp matching the
    #| start of the col value.
    #| Any empty values or non-existent variables are ignored
    qc::args $args -nocase -- args
    set dict [args2dict $args]
    if { [info exists nocase] } {
	set operator ~*
    } else {
	set operator ~
    }
    set list {}
    foreach {name value} $dict {
	if { [info exists value] && [ne $value ""] } {
	    lappend list "$name $operator [db_quote "^[db_escape_regexp $value]"]"
	}
    }
    if { [llength $list]==0 } {
	return true
    } else {
	return [join $list " and "]
    }
}

proc qc::sql_where_col_starts { args } {
    #| Construct part of SQL WHERE clause with a regexp matching 
    #| any given values to the start of the col value.
    #| Any empty values or non-existent variables are ignored
    qc::args $args -nocase -not -- col_name args
    if { [info exists nocase] } {
	set operator ~*
    } else {
	set operator ~
    }
    if { [info exists not] } {
	set operator "!$operator"
	set logic and
    } else {
	set logic or
    }
    set list {}
    foreach value $args {
	if { [ne $value ""] } {
	    lappend list "$col_name $operator [db_quote "^[db_escape_regexp $value]"]"
	}
    }
    if { [llength $list]==0 } {
	return true
    } else {
	return ([join $list " $logic "])
    }
}

proc qc::sql_where_combo { col_name value } {
    if { [ne $value ""] } {
	return "$col_name ~* [db_quote "^[db_escape_regexp $value]"]"
    } else {
	return true
    }
}

proc qc::sql_where_compare { args } {
    # WACKY
    # list of varNames
    # assume operator is stored in names variable named varName_op
    set list {}
    foreach name $args {
	if { ![regexp {^([^.]+)\.([^.]+)$} $name sql_name relation var_name] } {
	    set sql_name $name
	    set var_name $name
	}
	qc::upcopy 1 $var_name value
	qc::upcopy 1 "${var_name}_op" operator
	if { [info exists value] && [info exists operator] && [ne $value ""] } {
	    # check operator
	    if { ![in [list < = > <> <= >=] $operator] } { error "Unknown operator $operator" }
	    if { [eq $value NULL] && [eq $operator =] } {
		lappend list "$sql_name IS NULL"
	    } else {
		lappend list "${sql_name}${operator}[db_quote $value]"
	    }
	}
    }
    if { [llength $list]==0 } {
	return true
    } else {
	return [join $list " and "]
    }
}

proc qc::sql_where_compare_set { args } {
    qc::args $args -type "" --  name operator value
    if { [ne $value ""] } {
	# check operator
	if { ![in [list < = > <> <= >=] $operator] } { error "Unknown operator $operator" }
	if { [eq $value NULL] && [eq $operator =] } {
	    return "$name IS NULL"
	} else {
	    return "${name}${operator}[db_quote $value $type]"
	}
    } else {
	return true
    }
}

proc qc::sql_where_or { args } {
    #| Return a SQL where clause. Any empty values or non-existent variables are ignored
    set dict [args2dict $args]
    set list {}
    foreach {name value} $dict {
	if { [ne $value ""] } {
	    if { [string equal $value NULL] } {
		lappend list "$name IS NULL"
	    } elseif { [string equal $value "NOT NULL"] } {
		lappend list "$name IS NOT NULL"
	    } else {
		lappend list "$name=[db_quote $value]"
	    }
	}
    }
    if { [llength $list]==0 } {
	return true
    } else {
	return ([join $list " or "])
    }
}

proc qc::sql_where_word_in { name value } {
    #| Deprecated proc name - use qc::sql_where_words_in 
    return [sql_where_words_in $name $value]
}

proc qc::sql_where_phrase_words_in { args } {
    #| Where clause to match the words in each phrase
    qc::args $args -all -- name args
    set list {}
    foreach phrase $args {
	lappend list "([qc::sql_where_words_in $name $phrase])"
    }
    if { [llength $list] > 0 } {
        if { [info exists all] } {
            return "[join $list " and "]"
        } else {
            return "([join $list " or "])"
        }
    } else {
        return true
    }
}

proc qc::sql_where_words_in { name phrase } {
    #| Where clause to match all words in the phrase in any order.
    set list {}
    if { [ne $phrase ""] } {
        foreach word [split $phrase] {
            lappend list "$name ~ [db_quote "( |^)[db_escape_regexp $word]( |$)"]"
	}
        return [join $list " and "]
    } else {
	return true
    } 
}

proc qc::sql_where_phrases_in { args } {
    #| Where clause to evaluate to true if $phrase occurrs in sql expression $name
    qc::args $args -all -- name args
    set list {}
    foreach phrase $args {
        if { [ne $phrase ""] } {
            lappend list "$name ~ [db_quote "( |^)[db_escape_regexp $phrase]( |$)"]"
        }
    }
    if { [llength $list] > 0 } {
        if { [info exists all] } {
            return [join $list " and "]
        } elseif { [llength $list] > 1 } {
            return ([join $list " or "])
        } else {
            return [lindex $list 0]
        }
    } else {
        return true
    }
}


namespace eval qc {
    namespace export form_db form_db_plain
}

proc qc::form_db { content args } {
    #| Produce HTML to be used with JavaScript behavior dbForm
    array set this $args
    set this(class) "db-form"
    default this(id) oDbForm
    default this(method) POST
    if { ![info exists this(formType)] } {
	if {[info exists this(submitURL)]} {
	    set this(formType) submit
	} elseif {[info exists this(updateURL)]} {
	    set this(formType) update
	}
    }
    append html [html_tag form {*}[array get this]] \n
    append html "<div class=\"db-form-wrapper\">\n"
    append html $content \n
    append html "</div>\n"
    append html "<div class=\"db-form-status\">\n"
    append html "</div>\n"
    append html [qc::form_authenticity_token]
    append html "</form>"

    return $html
}

proc qc::form_db_plain { content args } {
    #| Produce HTML to be used with JavaScript behavior dbForm
    array set this $args
    set this(class) "db-form"
    default this(id) oDbForm
    return [html form $content {*}[array get this]]
}

namespace eval qc {
    namespace export url url_*
}

proc qc::url { url args } {
    #| Builds a URL from a given base and name value pairs.
    #| Substitutes and encodes any colon variables from the name value pairs into the path and fragment
    #| with any remaining name value pairs treated as parameters for the query string.
    #| NOTE: Only supports root-relative URLs.
    # TODO: Support escape sequences for fragments or paths beginning with colons.
    set dict [qc::args2dict $args]

    qc::dict2vars [qc::url_parts $url] params hash protocol domain port segments

    # decode params, segments, and hash
    set params [qc::lapply url_decode $params]
    set segments [qc::lapply url_decode $segments]
    set hash [url_decode $hash]

    # look for colon vars in the URL path segments and substitute the matching value given in the args
    set substituted_segments [list]
    foreach segment $segments {
        if { [string index $segment 0] eq ":" } {
            # remove the colon
            set segment [string range $segment 1 end]
            # check if caller has provided a substitution for the segment
            if { [dict exists $dict $segment] } {
                set value [dict get $dict $segment]
                if { [string index $value 0] eq ":" } {
                    return -code error "The substitute value \"$value\" for colon variable \"$segment\" begins with a colon."
                }
                lappend substituted_segments $value
                # remove the dict entry so that it isn't reused
                set dict [dict remove $dict $segment]
            } else {
                return -code error "Key \"$segment\" not found in args."
            }
        } else {
            lappend substituted_segments $segment
        }
    }

    # check if the fragment identifier requires substitution
    if { [string index $hash 0] eq ":" } {
        # remove the colon
        set temp [string range $hash 1 end]
        if { [dict exists $dict $temp] } {
            set value [dict get $dict $temp]
            if { [string index $value 0] eq ":" } {
                return -code error "The substitute value \"$value\" for colon variable \"$temp\" begins with a colon."
            }
            set hash $value
            # remove the dict entry so that it isn't reused
            set dict [dict remove $dict $temp]
        } else {
            return -code error "Key \"$temp\" not found in args."
        }
    }

    # rest of args are form vars
    dict for {key value} $dict {
        # overwrite any existing form vars
        dict set params $key $value
    }
    
    return [qc::url_make [dict create protocol $protocol domain $domain port $port segments $substituted_segments params $params hash $hash]]
}

proc qc::url_unset { url var_name } {
    #| Unset a url encoded variable in url
    if { ![qc::is uri $url] } {
        error "\"$url\" is not a valid URI."
    }  
    if { [regexp {([^\?\#]+)(?:\?([^\#]*))?(\#.*)?} $url -> path query_string fragment] } {
	foreach {name value} [split $query_string &=] {
	    set this([qc::url_decode $name]) [qc::url_decode $value]
	}
    } else {
       	array set this {}
    }
    # Unset required value
    if { [info exists this($var_name)] } {
	unset this($var_name)
    }
    # Recontruct the query string
    set pairs {}
    foreach name [array names this] {
	lappend pairs "[url_encode $name]=[url_encode $this($name)]"
    }
    if { [llength $pairs] != 0 } {
	return "$path?[join $pairs &]$fragment"
    } else {
	return ${path}${fragment}
    }
}

proc qc::url_to_html_hidden { url } {
    #| Convert a url with form vars into html hidden input tags
    if { ![qc::is uri $url] } {
        error "\"$url\" is not a valid URI."
    }  
    set html ""
    if { [regexp {([^\?\#]+)(?:\?([^\#]*))?} $url -> path query_string] } {
	foreach {name value} [split $query_string &=] {
	    set this([qc::url_decode $name]) [qc::url_decode $value]
	}
    } else {
	array set this {}
    }
    append html [html_hidden_set {*}[array get this]]
    return $html
}

proc qc::url_back { url args } {
    #| Creates a link to url with a formvar next_url which links back to the current page.
    #| Preserve vars passed in via GET or POST
    # TODO Aolserver only
    if { ![qc::is uri $url] } {
        error "\"$url\" is not a valid URI."
    }  
    foreach name $args {
	set value($name) [upset 1 $name]
    }
    set value(next_url) [qc::url_here]
    return [url $url {*}[array get value]]
}

proc qc::url_here {} {
    #| Encode form_vars from GET and POST in this url.
    # TODO Aolserver only
    return [qc::form2url [qc::conn_url]]
}

proc qc::url_encoding_init {} {
    #| Initialise url encode/decode maps in the qc namespace
    variable url_encode_map {}
    variable url_decode_map {}
    for {set i 0} {$i < 256} {incr i} {
        set char [format %c $i]
        set hex %[format %02x $i]
        # RFC 3986 - Encode all characters except ( a-z A-Z 0-9 . - ~ _ )
        if { ! [string match {[-a-zA-Z0-9.~_]} $char] } {
            lappend url_encode_map $char $hex
        }
        # Decode any percent hex encoded characters
        lappend url_decode_map $hex $char
    }
    # Decode + char as a space
    lappend url_decode_map + " " 
}

proc qc::url_encode {string {charset utf-8}} { 
    #| Return url-encoded string with option to specify charset    
    # Conforms to RFC 3986
    variable url_encode_map
    if { ! [info exists url_encode_map] } { 
        url_encoding_init 
    }
    return [string map $url_encode_map [encoding convertto $charset $string]]
}

proc qc::url_decode {string {charset utf-8}} { 
    #| Return url-decoded string with option to specify charset
    # Conforms to RFC 3986
    variable url_decode_map
    if { ! [info exists url_decode_map] } { 
        url_encoding_init 
    }
    return [encoding convertfrom $charset [string map -nocase $url_decode_map $string]]
}

proc qc::url_path {url} {
    # Return just the url path
    if { ![qc::is uri $url] } {
        error "\"$url\" is not a valid URI."
    }  
    if { [regexp {^https?://[a-z0-9_][a-z0-9_\-]*(?:\.[a-z0-9_\-]+)+(?::[0-9]+)?(/[^\?]*)} $url -> path] } {
	return $path
    } elseif { [regexp {^(/?[^\?]*)} $url -> path] } {
	return $path 
    } else {
	return ""
    }
}

proc qc::url_root {url} {
    # Return the root of an url without GET string or anchor
    if { ![qc::is uri $url] } {
        error "\"$url\" is not a valid URI."
    }  
    if { [regexp {^https?://[a-z0-9_][a-z0-9_\-]*(?:\.[a-z0-9_\-]+)+(?::[0-9]+)?(/[^\?\#]*)?} $url root] } {
	return $root
    } else {
        error "Url \"$url\" is not a valid URL"
    }
}

proc qc::url_match {canonical_url test_url} {
    #| Test if a test url matches a "canonical" url.
    # To match, they must have the same base, if the canonical url has a hash then
    # the test url must have the same hash, and every name/value pair in the
    # canonical url's query must appear at least the same number of times in the
    # test url's query.

    set c_parts [url_parts $canonical_url]
    set t_parts [url_parts $test_url]
    if { [dict get $c_parts base] ne [dict get $t_parts base] } {
        return false
    }
    if { [dict get $c_parts hash] ni [list "" [dict get $t_parts hash]] } {
        return false
    }

    set c_params [dict get $c_parts params]
    set t_params [dict get $t_parts params]
    foreach {c_name c_value} $c_params {
        set matched false
        foreach {t_name t_value} $t_params {
            if { $c_name eq $t_name
                 && $c_value eq $t_value } {
                multimap_unset_first t_params $c_name $t_value
                set matched true
                break
            }
        }
        if { ! $matched } {
            return false
        }
    }

    return true
}

proc qc::url_parts {url} {
    #| Return a dict containing the base, params (as a multimap), hash, protocol, domain, port,
    #| and path of url.
    #| qc::url_parts is not responsible for decoding any url part.
    if { ![qc::is uri $url] } {
        error "\"$url\" is not a valid URI."
    }  
    set pchar {[a-zA-Z0-9\-._~]|%[0-9a-fA-F]{2}|[!$&'()*+,;=:@]}
    set query_char "(?:${pchar}|/|\\?)"
    set hash_char "(?:${pchar}|/|\\?)"
    set path_char "(?:${pchar}|/)"

    set pattern [subst -nocommands -nobackslashes {^
        # base with protocol, domain, port (optional), and abs_path (optional)
        (
         (https?)://
         ([a-z0-9\-\.]+)
         (?::([0-9]+))?
         (/${path_char}*)?
         )
        
        # query (optional)
        (?:\?(${query_char}*))?
        
        # hash (optional)
        (?:\#(${hash_char}*))?
        $}]
    if { [regexp -expanded $pattern $url -> base protocol domain port path query hash] } {
        set pairs [split $query &]
        set params [list]
        foreach pair $pairs {
            set list [split $pair =]
            lappend params [lindex $list 0] [lindex $list 1]
        }
        set segments [split [string trimleft $path "/"] "/"]
        return [qc::dict_from base params hash protocol domain port path segments]
    }

    set pattern [subst -nocommands -nobackslashes {^
        # base with path (abs or rel) only
        (${path_char}*)
        
        # query (optional)
        (?:\?(${query_char}*))?
        
        # hash (optional)
        (?:\#(${hash_char}*))?
        $}]
    if { [regexp -expanded $pattern $url -> path query hash] } {
        lassign [list "" "" ""] protocol domain port
        set base $path
        set pairs [split $query &]
        set params [list]
        foreach pair $pairs {
            set list [split $pair =]
            lappend params [lindex $list 0] [lindex $list 1]
        }
        set segments [split [string trimleft $path "/"] "/"]
        return [qc::dict_from base params hash protocol domain port path segments]
    }

    error "Unable to parse url $url"
}

proc qc::url_request_path {request} {
    #| Return the path of an http request line
    # (eg. takes "GET /homepage?foo=bar HTTP/1.1", returns "/homepage")
    set request_regexp {^
        ([A-Z]+)
        \s
        ([^ ]+)
        \s
        HTTP/([0-9]+\.[0-9]+)
        $}
    if { ! [regexp -expanded $request_regexp $request \
                -> request_method request_uri http_version] } {
        error "bad request line"
    }
    set parts [qc::url_parts $request_uri]
    return [dict get $parts path]
}

proc qc::url_make {dict} {
    #| Construct an url from a dict of parts
    # eg.
    # url_make {
    #     protocol https
    #     domain qcode.co.uk
    #     port 80
    #     segments {posts 123 "hello world"}
    #     path /posts/123/hello+world
    #     params {tags tcl tags psql author peter}
    #     hash comments
    # }
    # => https://qcode.co.uk:80/posts/123/hello+world?tags=tcl&tags=psql&author=peter#comments
    # all keys are optional, but if protocol, domain, and/or port are specified,
    #   protocol and domain must be specified)
    # "segments" is a list
    # "params" is a multimap
    # If a non empty path is specified then segments will be ignored
    # If segments is specified (even as an empty list), the url path will be absolute
    dict2vars $dict protocol domain port path segments params hash

    # Construct url root (eg. http://qcode.co.uk:80).
    if {
        ([info exists protocol] && $protocol ne "")
        || ([info exists domain] && $domain ne "")
        || ([info exists port] && $port ne "")
    } {
        if { ! ([info exists protocol] && $protocol ne "") } {
            error "Invalid url dict - domain or port without protocol"
        }
        if { ! ([info exists domain] && $domain ne "") } {
            error "Invalid url dict - protocol or port without domain"
        }
        if { ! [regexp {https?} $protocol] } {
            error "Invalid url protocol $protocol"
        }
        if { ! [regexp {[a-z0-9\-\.]+} $domain] } {
            error "Invalid url domain $domain"
        }
        if { ([info exists port] && $port ne "") } {
            if { ! [regexp {[0-9]+} $port] } {
                error "Invalid url port $port"
            }
            set root "${protocol}://${domain}:${port}"
        } else {
            set root "${protocol}://${domain}"
        }
    }

    # Construct url path (eg. /posts/123/hello-world).
    if { (![info exists path] || $path eq "") && [info exists segments] } {
        set segments_escaped [list]
        foreach segment $segments {
            lappend segments_escaped [url_encode $segment]
        }
        set path "/[join $segments_escaped "/"]"
    }

    # Construct url query (eg. tags=tcl&tags=psql&author=peter).
    if { [info exists params] && [llength $params] > 0 } {
        if { [llength $params] % 2 != 0 } {
            error "Invalid params multimap $params"
        }
        set pairs [list]
        foreach {name value} $params {
            lappend pairs "[url_encode $name]=[url_encode $value]"
        }
        set query "[join $pairs "&"]"
    }

    # Encode hash into url "fragment"
    if { [info exists hash] && $hash ne "" } {
        set fragment [url_encode $hash]
    }

    # Bring the parts together into an url
    set url ""
    if { [info exists root] } {
        append url $root
    }
    if { [info exists path] } {
        append url $path
    }
    if { [info exists query] } {
        append url "?$query"
    }
    if { [info exists fragment] } {
        append url "#$fragment"
    }

    return $url
}

proc qc::url_maker {args} {
    #| Deprecated - see qc::url
    # Usage url_maker [$base] [: $port] [/ [segment ...]] [? [param_name param_value ...]] [# [hash]]
    qc::args $args -multimap_form_vars -- args
    default multimap_form_vars false
    set usage_error {Usage: url_maker [$base [: $port]] [/ [segment ...]] [? [param_name param_value ...]] [# [hash]] }
    set url_dict [dict create]
    set section "base"
    set sections [list "base" "port" "segments" "params" "hash"]
    set delimiters {
        ":" "port"
        "/" "segments"
        "?" "params"
        "#" "hash"
    }
    set end_of_section false
    set arg_is_param_value false
    set current_param_name ""
    foreach arg $args {
        if { [dict exists $delimiters $arg] } {
            if { $arg_is_param_value } {
                error $usage_error
            }
            set section [dict get $delimiters $arg]
            switch $section {
                "port" {
                    dict set url_dict port ""
                    set delimiters {
                        "/" "segments"
                        "?" "params"
                        "#" "hash"
                    }
                }
                "segments" {
                    dict_default url_dict segments [list]
                    set delimiters {
                        "?" "params"
                        "#" "hash"
                    }
                }
                "params" {
                    if { $multimap_form_vars } {
                        dict_default url_dict params [list]
                    } else {
                        dict_default url_dict params [dict create]
                    }
                    set delimiters {
                        "#" "hash"
                    }
                }
                "hash" {
                    dict set url_dict hash ""
                    set delimiters {}
                }
            }
            set end_of_section false
        } else {
            if { $end_of_section } {
                error $usage_error
            }
            if { [string index $arg 0] eq ":" } {
                set value [uplevel set [string range $arg 1 end]]
            } else {
                set value $arg
            }
            switch $section {
                "base" {
                    if { $value ne "" } {
                        set url_dict [qc::url_parts $value]
                    }
                    set end_of_section true
                }
                "port" {
                    dict set url_dict port $value
                    set end_of_section true
                }
                "segments" {
                    dict lappend url_dict segments $value
                }
                "params" {
                    if { $multimap_form_vars } {
                        dict lappend url_dict params $value
                    } else {
                        if { $arg_is_param_value } {
                            dict set url_dict params $current_param_name $value
                            set arg_is_param_value false
                        } else {
                            set current_param_name $value
                            set arg_is_param_value true
                        }
                    }
                }
                "hash" {
                    dict set url_dict hash $value
                    set end_of_section true
                }
            }
        }
    }
    if { $arg_is_param_value } {
        error $usage_error
    }
    return [qc::url_make $url_dict]
}

namespace eval qc {
    namespace export html_table_sort html_table_sort_header
}

proc qc::html_table_sort {args} {
    qc::args2vars $args
    default id tScroll
    default height 600

    # QRY
    if { [info exists qry] && ![info exists tbody] } {
	if { ![info exists sortCols] } {
	    if { [qc::form_var_exists sortCols] } {
		set sortCols [qc::form_var_get sortCols]
	    } else {
		set sortCols [qc::sortcols_from_qry $qry]
	    }
	}
	# 
	regsub -nocase {order by (.+?)(offset|limit|$)} $qry "order by [sql_sort {*}$sortCols] \\2" qry
	set table [qc::db_select_table [qc::db_qry_parse $qry 1]]	    
    }
    # table 
    if { [info exists table] && ![info exists tbody] } {
	default cols {}
	set cols [qc::html_table_cols_from_table table $cols]
	set tbody [lrange $table 1 end]
    }
    # sortCols from cols
    if { ![info exists sortCols] } {
	if { [form_var_exists sortCols] } {
	    set sortCols [form_var_get sortCols]
	} else {
	    set sortCols [qc::sortcols_from_cols $cols]
	}
    }
    # Highlight th sorted
    if { [info exists cols] && [info exists sortCols] } {
	set index [qc::ldict_search cols name [lindex $sortCols 0]]
	if { $index !=-1 } {
	    qc::ldict_set cols $index thClass sorted
	}
    }
    set thead [qc::html_table_sort_header $cols $sortCols]

    set div_style ""
    set div_class [list scroll]
    if { [lower $height] eq "max" } {
	lappend div_class "dynamic-resize"
    } else {
	set div_style [style_set $div_style height ${height}px]
    }

    append html [html_tag div class $div_class style $div_style]
    append html [qc::html_table ~ class cols thead tbody tfoot data rowClasses id]    
    append html "</div>\n"

    return $html
}

proc qc::html_table_sort_header { cols sortCols } {
    set sortCols [qc::sortcols2dict {*}$sortCols]
    set row {}
    foreach col $cols {
	if { [dict exists $col label] } {
	    set label [dict get $col label]
	} else {
	    set label ""
	}
	if { [dict exists $col name] } {
	    if { [eq [dict get $col name] [lindex $sortCols 0]] } {
		# Primary sort col
		if { [eq [upper [dict get $sortCols [dict get $col name]]] DESC] } {
		    set sort_order DESC
		} else {
		    set sort_order ASC
		}
		if { [dict exists $col class] } {
		    set class [dict get $col class]
		} else {
		    set class ""
		}
		if { [eq $sort_order ASC] } {
		    if { "rank" in $class } {
			set indicator "Sorted Top to Bottom"
		    } elseif { "number" in $class || "money" in $class || "perct" in $class } {
			set indicator "Sorted Low to High"
		    } elseif { "date" in $class } {
			set indicator "Sorted Old to New"
		    } else {
			set indicator "Sorted A-Z"
		    }
		    lappend row "[html span $label class sort][html div $indicator class asc]"
		} else {
		    if { "rank" in $class } {
			set indicator "Sorted Bottom to Top"
		    } elseif { "number" in $class || "money" in $class || "perct" in $class } {
			set indicator "Sorted High to Low"
		    } elseif { "date" in $class } {
			set indicator "Sorted New to Old"
		    } else {
			set indicator "Sorted Z-A"
		    }
		    lappend row "[html span $label class sort][html div $indicator class desc]"
		}
	    } else {
		# Sortable col
		lappend row [html span $label class sort]
	    }
	} else {
	    # Not sortable col
	    lappend row $label
	}
    }
    lappend thead $row
    return $thead
}

namespace eval qc {
    namespace export form_layout_*
}

proc qc::form_layout_table { args } {
    #| Construct an html table with 2 columns for labels and form elements
    args $args -sticky -class "form-layout-table" -- conf
    set cols {
	{class label}
	{}
    }
    if { [info exists sticky] } {
	set tbody [uplevel 1 [list qc::form_layout_tbody -sticky -- $conf]]
    } else {
	set tbody [uplevel 1 [list qc::form_layout_tbody $conf]]
    }
    return [qc::html_table cols $cols class $class tbody $tbody]
}

proc qc::form_layout_tables { args } {
    #| Construct multi-column layout with a form table in each column
    args $args -sticky -- args
    set cols {
	{class label}
	{}
    }
    set class "form-layout-table"
    foreach conf $args {
	if { [info exists sticky] } {
	    set tbody [uplevel 1 [list qc::form_layout_tbody -sticky -- $conf]]
	} else {
	    set tbody [uplevel 1 [list qc::form_layout_tbody $conf]]
	}
	lappend row [qc::html_table cols $cols class $class tbody $tbody]
    }
    set class "columns-container"
    set tbody [list $row]
    return [qc::html_table class $class tbody $tbody]
}
    
proc qc::form_layout_tbody { args } {
    #| Construct a 2-column tbody for form elements.
    #| For most input elements the label is shown on the left and input element on the right.
    #| Exceptions for checkboxes and radiogroup.
    args $args -sticky -- conf
    set level 1
    set tbody {}
    foreach dict $conf {
	array set this $dict
	default this(label) Unknown
	default this(type) text
	# sticky used if set for the widget or globally
	# can be turned off for the widget by setting "sticky no"
	if { [info exists sticky] } {
	    default this(sticky) yes
	} else {
	    default this(sticky) no
	}
	if {[true $this(sticky)] && [sticky_exists $this(name)]} {
	    # sticky values
	    set this(value) [sticky_get $this(name)]
	}
	if { ![info exists this(value)]} {
	    # check caller variable for value
	    qc::upcopy $level $this(name) value
	    if { [info exists value] } {
		set this(value) $value
	    } else {
		set this(value) ""
	    }
	}
    
	
	switch $this(type) {
	    checkbox -
	    bool {
		lappend tbody [list "" "[qc::widget {*}[array get this]] [qc::widget_label {*}[array get this]]"]
	    }
	    radiogroup {
		lappend tbody [list [qc::widget_label {*}[array get this]] [qc::widget_radiogroup {*}[array get this]]]
	    }
	    default {
		set cell [qc::widget {*}[array get this]]
		if { [info exists this(units)] } {
		    append cell [html span " $this(units)"]
		} 
		lappend tbody [list [qc::widget_label {*}[array get this]] $cell]
	    }
	}

	unset this
    }
    return $tbody
}

proc qc::form_layout_list {conf} {
    #| Layout the form elements in the conf with input elements below labels.
    set level 1
    set html {}
    foreach dict $conf {
	array set this $dict
	default this(label) Unknown
	default this(type) text
	if { ![info exists this(value)]} {
	    qc::upcopy $level $this(name) value
	    if { [info exists value] } {
		set this(value) $value
	    } else {
		set this(value) ""
	    }
	}

	set label [qc::widget_label {*}[array get this]]

	set widget [qc::widget {*}[array get this]]

	if { [in {checkbox bool} $this(type)] } {
	    append html "<div style=\"padding-bottom:1em;\">$widget $label</div>"
	} else {
	    set lines {}
	    if { [ne $this(label) ""] } { lappend lines $label }
	    if { [info exists this(note)] } { lappend lines $this(note) }
	    lappend lines $widget
	    append html "<div style=\"padding-bottom:1em;\">[join $lines "<br>"]</div>"
	}
	unset this
    }
    return $html
}


namespace eval qc {
    namespace export ldict_* ldict2tbody
}

proc qc::ldict_set {ldictVar index key value} {
    #| Takes a list of dicts and sets the value for $key in the dict at $index
    upvar 1 $ldictVar ldict
    set dict [lindex $ldict $index]
    dict set dict $key $value
    lset ldict $index $dict
}

proc qc::ldict_sum { ldictVar key } {
    #| Traverse a dict list and sum all dict values for given key.
    # TODO: could be de-aolserverised, and perhaps should error if we try to add
    # non-decimal strings rather than return 0.
    set sum 0
    upvar 1 $ldictVar ldict
    foreach dict $ldict {
	set value [dict get $dict $key]
	set value [qc::strip_html $value]
	regsub -all {[, ]} $value {} value
	if { [string is double -strict $value] } {
	    set sum [expr {$sum + $value}]
	}
    }
    return $sum
}

proc qc::ldict_max { ldictVar key } {
    set max {}
    upvar 1 $ldictVar ldict
    foreach dict $ldict {
	set value [dict get $dict $key]
	if { [string equal $max ""] } {
	    set max $value
	} else {
	    if { $value > $max } {
		set max $value
	    }
	}
    }
    return $max
}

proc qc::ldict_values { ldictVar key } {
    #| Return a list of the values of this key
    #| in each dict in the ldict
    upvar 1 $ldictVar ldict
    set list {}
    foreach dict $ldict {
	lappend list [dict get $dict $key]
    }
    return $list
}

proc qc::ldict_exists {ldict key} {
    #| Return the first index of the dict that contains the the key $key
    set index 0
    foreach dict $ldict {
	if { [dict exists $dict $key]} {
	    return $index
	}
	incr index
    }
    return -1
}

proc qc::ldict_search {ldictVar key value} {
    #| Return the first index of the dict that contains the value $value for the key $key
    upvar 1 $ldictVar ldict
    set index 0
    foreach dict $ldict {
	if { [dict exists $dict $key] && [dict get $dict $key]=="$value" } {
	    return $index
	}
	incr index
    }
    return -1
}

proc qc::ldict_exclude { ldict key } {
    #| Remove all occurances of $key from the dicts in $ldict
    set newldict {}
    foreach dict $ldict {
	lappend newldict [dict_exclude $dict $key]
    }
    return $newldict
}

proc qc::ldict2tbody {ldict colnames} {
    #| Take a ldict and a list of col names to convert into tbody
    set tbody {}
    foreach dict $ldict {
	set row {}
	foreach colname $colnames {
	    if { [dict exists $dict $colname] } {
		lappend row [dict get $dict $colname]
	    } else {
		lappend row ""
	    }
	}
	lappend tbody $row
    }
    return $tbody
}

proc qc::ldict_mappings_equal {ldict1 ldict2} {
    #| Compare 2 ldicts, for dicts with same mappings
    # (independant of key ordering, but dependant on ldict ordering)
    
    if { ! [string is list $ldict1]
         || ! [string is list $ldict2]
     } {
        return false
    }

    if { [llength $ldict1] != [llength $ldict2] } {
        return false
    }

    # compare each element
    foreach dict1 $ldict1 dict2 $ldict2 {
        if { ! [qc::dicts_equal [dict_sort dict1] [dict_sort dict2]] } {
            return false
        }
    }
    return true
}

proc qc::ldict_subsets {ldict args} {
    #| Return subsets of the given dicts
    set result {}
    foreach dict $ldict {
        lappend result [qc::dict_subset $dict {*}$args]
    }
    return $result
}

package require sha1
package require md5
package require base64
package require tdom
package require fileutil
namespace eval qc {
    namespace export s3
}

# Deprecated.
# See qc::aws s3
# Public S3 procs
proc qc::s3 { args } {
    #| Access Amazon S3 buckets via REST API
    # Usage: s3 subcommand {args}

    if { (![info exists ::env(AWS_ACCESS_KEY_ID)] || ![info exists ::env(AWS_SECRET_ACCESS_KEY)]) &&
         [lindex $args 0] ni [list "md5" "uri" "uri_bucket_object_key"] } {
        error "No AWS credentials set."
    }

    switch [lindex $args 0] {
        md5 {
            #| Just print the base64 md5 of a local file for reference
            # usage: s3 md5 filename
            lassign $args -> filename 
            return [qc::_s3_base64_md5 -file $filename]
        }
        ls {
            # usage: s3 ls 
            set nodes [qc::s3_xml_select [qc::_s3_get "" ""] {/ns:ListAllMyBucketsResult/ns:Buckets/ns:Bucket}]
            return [qc::lapply qc::s3_xml_node2dict $nodes]
        }
        lsbucket {
            # usage: s3 lsbucket bucket {prefix}
            # s3 lsbucket myBucket Photos/
            if { [llength $args] == 1 || [llength $args] > 3 } {
                error "Missing argument. Usage: qc::s3 lsbucket mybucket {prefix}"
            } elseif { [llength $args] == 3 }  {
                # prefix is specified
                set xmlDoc [qc::_s3_get [lindex $args 1] "?prefix=[lindex $args 2]"]
            } else {
                set xmlDoc [qc::_s3_get [lindex $args 1] ""]
            }
	    return [qc::lapply qc::s3_xml_node2dict [qc::s3_xml_select $xmlDoc {/ns:ListBucketResult/ns:Contents}]]
        }
        get {
            # usage:
            # qc::s3 get bucket remote_filename {local_filename}
            # qc::s3 get s3_uri local_filename
            if { [llength $args] < 3 || [llength $args] > 4 } {
                error "Wrong number of arguments. Usage: \"qc::s3 get mybucket remote_filename ?local_filename?\" or \"qc::s3 get s3_uri local_filename\"."
            } elseif { [llength $args] == 3 } {
                # Test if args are $bucket and $remote_filename or $s3_uri and $local_filename
                lassign $args -> arg0 arg1
                if { [qc::is s3_uri $arg0] } {
                    # qc::s3 get s3_uri local_filename
                    set s3_uri $arg0
                    lassign [qc::s3 uri_bucket_object_key $s3_uri] bucket object_key
                    set local_filename $arg1
                } else {
                    # qc::s3 get bucket remote_filename
                    set bucket $arg0
                    # Strip the starting "/" from the remote_filename
                    set object_key [string range $arg1 1 end]
                    set s3_uri [qc::s3 uri $bucket $object_key]
                    # No local filename, assume same as remote_filename
                    set local_filename "./[file tail $remote_filename]"
                }
            } else {
                # qc::s3 get bucket remote_filename {local_filename}
                lassign $args -> bucket remote_filename local_filename
                set object_key [string range $remote_filename 1 end]
                set s3_uri [qc::s3 uri $bucket $object_key]
            }
            if { [file exists $local_filename] } {
                error "File $local_filename already exists."
            }
            set head_dict [qc::s3 head $s3_uri]
            set file_size [dict get $head_dict Content-Length]
            # set timeout - allow 1Mb/s
            set timeout_secs [expr {max( (${file_size}*8)/1000000 , 60)} ]
            log Debug "Timeout set at $timeout_secs seconds"
            qc::_s3_save -timeout $timeout_secs $bucket $object_key $local_filename
        }
        exists {
            # usage:
            # qc::s3 exists bucket remote_path
            # qc::s3 exists s3_uri
            if { [llength $args] == 3 } {
                # qc::s3 exists bucket remote_path
                lassign $args -> bucket remote_path
                set object_key [string range $remote_path 1 end]
            } elseif { [llength $args] == 2 } {
                # qc::s3 exists s3_uri
                set s3_uri [qc::cast s3_uri [lindex $args 1]]
                lassign [qc::s3 uri_bucket_object_key $s3_uri] bucket object_key
            } else {
                error "qc::s3 exists: Wrong number of args. Usage \"qc::s3 exists bucket remote_path\" or \"qc::s3 head s3_uri\"."
            }
            qc::_s3_exists $bucket $object_key

        }
        head {
            # usage:
            # qc::s3 head bucket remote_path
            # qc::s3 head s3_uri

            if { [llength $args] == 3 } {
                # qc::s3 head bucket remote_path
                lassign $args -> bucket remote_path
                set object_key [string range $remote_path 1 end]
            } elseif { [llength $args] == 2 } {
                # qc::s3 head s3_uri
                set s3_uri [qc::cast s3_uri [lindex $args 1]]
                lassign [qc::s3 uri_bucket_object_key $s3_uri] bucket object_key
            } else {
                error "qc::s3 head: Wrong number of args. Usage \"qc::s3 head bucket remote_path\" or \"qc::s3 head s3_uri\"."
            }

            qc::_s3_head $bucket $object_key
        }       
        copy {
            # usage:
            # qc::s3 copy bucket remote_filename_to_copy remote_filename_copy
            # E.G. qc::s3 copy "my-bucket" "file_to_copy" "file_copy"
            # qc::s3 copy s3_uri_to_copy s3_uri_copy

            if {[llength $args] == 4} {
                # qc::s3 copy bucket remote_filename_to_copy remote_filename_copy
                lassign $args -> bucket remote_filename remote_filename_copy
                set file_to_copy "${bucket}${remote_filename}"
                set object_key_copy [string range $remote_filename_copy 1 end]
            } elseif {[llength $args] == 3} {
                # qc::s3 copy s3_uri_to_copy s3_uri_copy
                lassign $args -> s3_uri_to_copy s3_uri_copy
                lassign [qc::s3 uri_bucket_object_key $s3_uri_to_copy] bucket object_key
                set file_to_copy "${bucket}/${object_key}"
                lassign [qc::s3 uri_bucket_object_key $s3_uri_copy] bucket_to object_key_copy
                if { $bucket ne $bucket_to } {
                    error "qc::s3 copy: The s3_uri to copy to must be in the same bucket as the s3_uri to copy from."
                }
            } else {
                error "qc::s3 copy: Wrong number of args. Usage \"qc::s3 copy bucket remote_filename_to_copy remote_filename_copy\" or \"qc::s3 copy s3_uri_to_copy s3_uri_copy\"."
            }

            qc::_s3_put -s3_copy $file_to_copy $bucket $object_key_copy
        }
        put {
            # usage:
            # qc::s3 put bucket local_path ?remote_filename?
            # qc::s3 put s3_uri local_filename
            # 5MB limit
            if { [llength $args] < 3 || [llength $args] > 4 } {
                error "Wrong number of arguments. Usage: \"qc::s3 put mybucket local_filename ?remote_filename?\" or \"qc::s3 put s3_uri local_filename\"."
            } elseif { [llength $args] == 3 } {
                # Test if args are $bucket and $remote_filename or $s3_uri and $local_filename
                lassign $args -> arg0 arg1
                if { [qc::is s3_uri $arg0] } {
                    # qc::s3 put s3_uri local_filename
                    set s3_uri $arg0
                    lassign [qc::s3 uri_bucket_object_key $s3_uri] bucket object_key
                    set local_filename $arg1
                } else {
                    # qc::s3 put bucket local_path
                    set bucket $arg0
                    set local_filename $arg1
                    # No remote filename, assume same as local_filename
                    set object_key [file tail $local_filename]
                    set s3_uri [qc::s3 uri $bucket $object_key]
                }
            } else {
                # qc::s3 put bucket local_path remote_filename
                lassign $args -> bucket local_filename remote_filename
                set object_key [string range $remote_filename 1 end]
                set s3_uri [qc::s3 uri $bucket $object_key]
            }

            if { [file size $local_filename] > [expr {1024*1024*5}]} { 
                # Use multipart upload
                qc::s3 upload $s3_uri $local_filename
            } else {
                qc::_s3_put -infile $local_filename $bucket $object_key
            }
        }
        restore {
            # usage:
            # qc::s3 restore bucket remote_path days
            # qc::s3 restore s3_uri Days
            # Requests restore of object from Glacier storage to S3 storage for $days days
            lassign $args -> bucket remote_path Days
            if { [llength $args] == 4  } {
                 # qc::s3 restore bucket remote_path days
                lassign $args -> bucket remote_path Days
                set object_key [string range $remote_path 1 end]
            } elseif { [llength $args] == 3 } {
                # qc::s3 restore s3_uri Days
                lassign $args -> s3_uri Days
                lassign [qc::s3 uri_bucket_object_key $s3_uri] bucket object_key
            } else {
                error "Invalid number of arguments. Usage: \"qc::s3 restore bucket remote_path days\" or \"qc::s3 restore s3_uri days\"."
            }
            set data "<RestoreRequest>[qc::xml_from Days]</RestoreRequest>"
            qc::_s3_post $bucket "${object_key}?restore" $data
        }
        upload {
            switch [lindex $args 1] {
                init {
                    # Usage:
                    # qc::s3 upload init bucket local_file remote_file {content_type}
                    # qc::s3 upload init s3_uri local_file {content_type}

                    set content_type ""
                    if { [llength $args] == 6 } {
                        # qc::s3 upload init bucket local_file remote_file content_type
                        lassign $args -> -> bucket local_file remote_file content_type
                        set object_key [string range $remote_file 1 end]
                    } elseif { [llength $args] == 5 } {
                        lassign $args -> -> arg0 local_file arg1
                        if { [qc::is s3_uri $arg0] } {
                            # qc::s3 upload init s3_uri local_file content_type
                            set s3_uri $arg0
                            lassign [qc::s3 uri_bucket_object_key $s3_uri] bucket object_key
                            set content_type $arg1
                        } else {
                            # qc::s3 upload init bucket local_file remote_file
                            set bucket $arg0
                            set remote_file $arg1
                            set object_key [string range $remote_file 1 end]
                        }
                    } elseif { [llength $args] == 4 } {
                        # qc::s3 upload init s3_uri local_file
                        lassign $args -> -> s3_uri local_file
                        lassign [qc::s3 uri_bucket_object_key $s3_uri] bucket object_key
                    } else {
                        error "Invalid number of arguments. Usage: \"qc::s3 upload init bucket local_file remote_file {content_type}\" or \"qc::s3 upload init s3_uri local_file {content_type}\"."
                    }
                    
                    set content_md5 [qc::_s3_base64_md5 -file $local_file]
                    if {$content_type eq ""} {
                        set content_type [qc::mime_type_guess $local_file]
                    }
                    set upload_dict [qc::s3_xml_node2dict [qc::s3_xml_select [qc::_s3_post -content_type $content_type -amz_headers [list x-amz-meta-content-md5 $content_md5] $bucket "${object_key}?uploads"] {/ns:InitiateMultipartUploadResult}]]
                    set upload_id [dict get $upload_dict UploadId]
                    log Debug "Upload init for $object_key to $bucket."
                    log Debug "Upload_id: $upload_id"
                    return $upload_id
                }
                abort {
                    # Usage:
                    # qc::s3 upload abort bucket remote_path upload_id
                    # qc::s3 upload abort s3_uri upload_id
                    if {[llength $args] == 5 } {
                        # qc::s3 upload abort bucket remote_path upload_id
                        lassign $args -> -> bucket remote_path upload_id
                        set object_key [string range $remote_path 1 end]
                    } elseif {[llength $args] == 4 } {
                        # qc::s3 upload abort s3_uri upload_id
                        lassign $args -> -> s3_uri upload_id
                        lassign [qc::s3 uri_bucket_object_key $s3_uri] bucket object_key
                    } else {
                        error "Invalid number of arguments. Usage: \"qc::s3 upload abort bucket remote_path upload_id\" or \"qc::s3 upload abort s3_uri upload_id\"."
                    }
                    return [_s3_delete $bucket "${object_key}?uploadId=$upload_id"]
                }
                ls {
                    # usage: s3 upload ls bucket 
                    lassign $args -> -> bucket
                    return [qc::lapply qc::s3_xml_node2dict [qc::s3_xml_select [qc::_s3_get $bucket "?uploads"] {/ns:ListMultipartUploadsResult/ns:Upload}]]
                }
                lsparts {
                    # usage:
                    # qc::s3 upload lsparts bucket remote_path upload_id
                    # qc::s3 upload lsparts s3_uri upload_id
                    if {[llength $args] == 5 } {
                        # qc::s3 upload lsparts bucket remote_path upload_id
                        lassign $args -> -> bucket remote_path upload_id
                        set object_key [string range $remote_path 1 end]
                    } elseif {[llength $args] == 4 } {
                        # qc::s3 upload lsparts s3_uri upload_id
                        lassign $args -> -> s3_uri upload_id
                        lassign [qc::s3 uri_bucket_object_key $s3_uri] bucket object_key
                    } else {
                        error "Invalid number of arguments. Usage: \"qc::s3 upload lsparts bucket remote_path upload_id\" or \"qc::s3 upload lsparts s3_uri upload_id\"."
                    }
                    
                    return [qc::lapply qc::s3_xml_node2dict [qc::s3_xml_select [qc::_s3_get $bucket "${object_key}?uploadId=$upload_id"] {/ns:ListPartsResult/ns:Part}]]
                }
                cleanup {
                    # usage: s3 upload cleanup bucket 
                    # aborts any unfinished uploads for bucket
                    lassign $args -> -> bucket 
                    foreach dict [qc::s3 upload ls $bucket] {
                        qc::s3 upload abort $bucket "/[dict get $dict Key]" [dict get $dict UploadId]
                    }
                }
                complete {
                    # usage:
                    # qc::s3 upload complete bucket remote_path upload_id etag_dict
                    # qc::s3 upload complete s3_uri upload_id etag_dict
                    if {[llength $args] == 6} {
                        # qc::s3 upload complete bucket remote_path upload_id etag_dict
                        lassign $args -> -> bucket remote_path upload_id etag_dict
                        set object_key [string range $remote_path 1 end]
                    } elseif {[llength $args] == 5} {
                        # qc::s3 upload complete s3_uri upload_id etag_dict
                        lassign $args -> -> s3_uri upload_id etag_dict
                        lassign [qc::s3 uri_bucket_object_key $s3_uri] bucket object_key
                    } else {
                        error "Invalid number of arguments. Usage: \"qc::s3 upload complete bucket remote_path upload_id etag_dict\" or \"qc::s3 upload complete s3_uri upload_id etag_dict\"."
                    }
                    set xml {<CompleteMultipartUpload>}
                    foreach PartNumber [dict keys $etag_dict] {
                        set ETag [dict get $etag_dict $PartNumber]
                        lappend xml "<Part>[qc::xml_from PartNumber ETag]</Part>"
                    }
                    lappend xml {</CompleteMultipartUpload>}
                    log Debug "Completing Upload to $object_key in $bucket."
                    return [qc::_s3_post $bucket "${object_key}?uploadId=$upload_id" [join $xml \n]]
                }
                send {
                    # Perform upload
                    # usage:
                    # qc::s3 upload send bucket local_path remote_path upload_id
                    # qc::s3 upload send s3_uri local_path upload_id
                    if {[llength $args] == 6} {
                        # qc::s3 upload send bucket local_path remote_path upload_id
                        lassign $args -> -> bucket local_path remote_path upload_id
                        set object_key [string range $remote_path 1 end]
                    } elseif {[llength $args] == 5} {
                        # qc::s3 upload send s3_uri local_path upload_id
                        lassign $args -> -> s3_uri local_path upload_id
                        lassign [qc::s3 uri_bucket_object_key $s3_uri] bucket object_key
                    } else {
                        error "Invalid number of arguments. Usage: \"qc::s3 upload send bucket local_path remote_path upload_id\" or \"qc::s3 upload send s3_uri local_path upload_id\"."
                    }
                    
                    # bytes
                    set part_size [expr {1024*1024*5}]
                    set part_index 1
                    set etag_dict [dict create]
                    set file_size [file size $local_path]
                    # Timeout - allow 10 KB/s
                    global s3_timeout
                    set s3_timeout($upload_id) false
                    # Set timeout based on file size, but don't make it less than 10,000ms
                    set timeout_ms [expr { max(round((double($file_size)/10240)*1000),10000) }]
                    set max_attempt 10
                    log Debug "Timeout set as $timeout_ms ms"
                    set id [after $timeout_ms [list set s3_timeout($upload_id) true]]
                    set num_parts [expr {round(ceil($file_size/double($part_size)))}]
                    set fh [open $local_path r]
                    fconfigure $fh -translation binary
                    while { !$s3_timeout($upload_id) &&  $part_index <= $num_parts } {

                        # Use temp file to upload part from - inefficient, but posting binary data directly from http_put not yet working.
                        set tempfile [::fileutil::tempfile]
                        set tempfh [open $tempfile w]
                        fconfigure $tempfh -translation binary
                        log Debug "Uploading ${local_path}: Sending part $part_index of $num_parts"
                        puts -nonewline $tempfh [read $fh $part_size]
                        close $tempfh

                        set success false 
                        set attempt 1
                        while { !$s3_timeout($upload_id) && $attempt<=$max_attempt && !$success } {
                            try {
                                set response [qc::_s3_put -header 1 -nochecksum -infile $tempfile $bucket "${object_key}?partNumber=${part_index}&uploadId=$upload_id"]
                                set success true
                            } {
                                log Debug "Failed - retrying part $part_index of ${num_parts}... "
                                incr attempt
                            }
                        }

                        if { $s3_timeout($upload_id) || $attempt>$max_attempt } {
                            #TODO should we abort or leave for potential recovery later?
                            ::try {
                                qc::s3 upload abort $bucket $object_key $upload_id
                            } on error [list error_message options] {
                                # error when attempting to abort upload; do nothing
                            }

                            set message "Failed to upload file to bucket $bucket."

                            if { $s3_timeout($upload_id) } {
                                append message " Upload timed out."
                            }

                            if { $attempt > $max_attempt } {
                                append message " Number of attempts exceeded max attempts (${max_attempt})."
                            }

                            error $message
                        }
                        regexp -line -- {^ETag: "(\S+)"\s*$} $response match etag
                        
                        dict set etag_dict $part_index $etag
                        file delete $tempfile
                        incr part_index
                    }
                    close $fh
                    after cancel $id
                    unset s3_timeout($upload_id)
                    return $etag_dict

                }
                default {
                    # Top level multipart upload
                    # usage:
                    # qc::s3 upload bucket local_file remote_file {content_type}
                    # qc::s3 upload s3_uri local_file {content_type}
                    # TODO could be extended to retry upload part failures
                    set content_type ""
                    if {[llength $args] == 5} {
                        # qc::s3 upload bucket local_file remote_file content_type
                        lassign $args -> bucket local_file remote_file content_type
                        set object_key [string range $remote_file 1 end]
                    } elseif {[llength $args] == 4} {
                        lassign $args -> arg0 local_file arg1
                        if { [qc::is s3_uri $arg0] } {
                            # qc::s3 upload s3_uri local_file content_type
                            set s3_uri $arg0
                            lassign [qc::s3 uri_bucket_object_key $s3_uri] bucket object_key
                            set content_type $arg1
                        } else {
                            # qc::s3 upload bucket local_file remote_file
                            set bucket $arg0
                            set remote_file $arg1
                            set object_key [string range $remote_file 1 end]
                            set content_type ""
                        }
                    } elseif {[llength $args] == 3} {
                        # qc::s3 upload s3_uri local_file
                        lassign $args -> s3_uri local_file
                        lassign [qc::s3 uri_bucket_object_key $s3_uri] bucket object_key
                        set content_type ""
                    } else {
                        error "Invalid number of arguments. Usage: \"qc::s3 upload bucket local_file remote_file {content_type}\" or \"qc::s3 upload s3_uri local_file {content_type}\"."
                    }
                    
                    set upload_id [qc::s3 upload init [qc::s3 uri $bucket $object_key] $local_file $content_type]
                    set etag_dict [qc::s3 upload send [qc::s3 uri $bucket $object_key] $local_file $upload_id]
                    qc::s3 upload complete [qc::s3 uri $bucket $object_key] $upload_id $etag_dict
                }
            }
        }
        delete {
            # usage:
            # qc::s3 delete bucket remote_file
            # qc::s3 delete s3_uri
            if {[llength $args] == 3} {
                # qc::s3 delete bucket remote_file
                lassign $args -> bucket remote_file
                set object_key [string range $remote_file 1 end]
            } elseif {[llength $args] == 2} {
                # qc::s3 delete s3_uri
                lassign $args -> s3_uri
                lassign [qc::s3 uri_bucket_object_key $s3_uri] bucket object_key
            } else {
                error "Invalid number of arguments. Usage: \"qc::s3 delete bucket remote_filename\" or \"qc::s3 delete s3_uri\"."
            }
            
            qc::_s3_delete $bucket $object_key
        }
        uri {
            #| Return an s3_uri in the format s3://bucket/object_key
            # Usage:
            # qc::s3 uri bucket ?object_key?
            if {[llength $args] < 2 || [llength $args] > 3} {
                error "Invalid number of arguments. Usage: \"qc::s3 uri bucket ?object_key?\"."
            }
            lassign $args -> bucket object_key
            return [qc::cast s3_uri "${bucket}/${object_key}"]
        }
        uri_bucket_object_key {
            #| Return a list of the bucket and object key for the given s3_uri
            #| Usage:
            #| qc::s3 uri_bucket_object_key s3_uri
            if {[llength $args] != 2} {
                error "Invalid number of arguments. Usage: \"qc::s3 uri_bucket_object_key s3_uri\"."
            }
            regexp {^(?:[sS]3://|/)?([^/]+)/?(.*)} [lindex $args 1] -> bucket object_key
            return [list $bucket $object_key]
        }
        default {
            error "Unknown s3 command \"$args\""
        }
    }
}

namespace eval qc {
}

proc qc::validate2model {dict} {
    #| Validates dictionary against the data model and sets up the record in the global data structure.
    #| First checks all types. If all are valid, proceeds to check constraints.
    set all_valid true
    set cast_values [dict create]

    dict for {name value} $dict {
        # Resolve name to column, table, and schema
        lassign [qc::memoize qc::db_resolve_field_name $name] {*}{
            schema
            table
            column
        }
        set message [qc::memoize qc::db_validation_message $table $column]
        set data_type [qc::memoize qc::db_column_type \
                           -qualified -- $schema $table $column]
        set nullable [qc::memoize qc::db_column_nullable $schema $table $column]

        # Check if this is a sensitive data type - values should not be echo'd back to the client
        if { $data_type in [list "password" "card_number"] } {
            # Mark field as being sensitive in the global data structure
            qc::response record sensitive $column
        }
        # Initialise valid flag
        qc::response record valid $column $value
        # Check if nullable
        if {! $nullable && $value eq ""} {
            qc::response record invalid $column $value $message
            set all_valid false
            continue
        } elseif {$nullable && $value eq ""} {
            dict set cast_values $schema $table $column ""
            continue
        }
        # Check value against data type
        if { ![qc::castable $data_type $value] } {
            qc::response record invalid $column $value $message
            set all_valid false
            continue
        } 

        dict set cast_values $schema $table $column [qc::cast $data_type $value]
    }

    if { $all_valid } {
        dict for {schema tables} $cast_values {
            dict for {table columns} $tables {
                set results [qc::db_table_check_constraints_eval $schema $table $columns]
                dict for {constraint columns} [dict get $results failed] {
                    dict for {column value} $columns {
                        set message [qc::memoize qc::db_validation_message $table $column]
                        qc::response record invalid $column $value $message
                        set all_valid false
                    }
                }
            }
        }
    }

    if { ! $all_valid } {
        qc::response status invalid
    } else {
        qc::response status valid
    }

    return $all_valid
}

namespace eval qc {}

proc qc::binary_convert_unit_prefix2mult {prefix} {
    #| Return multiplier for a binary unit prefix.
    switch $prefix {
        "" {
            set power 0
        }
        K -
        Ki -
        kilo -
        Kilo -
        kibi -
        Kibi {
            set power 1
        }
        M -
        Mi -
        mega -
        Mega -
        mebi -
        Mebi {
            set power 2
        }
        G -
        Gi -
        giga -
        Giga -
        gibi -
        Gibi {
            set power 3
        }
        T -
        Ti -
        tera -
        Tera -
        tebi -
        Tebi {
            set power 4
        }
        P -
        Pi -
        peta -
        Peta -
        pebi -
        Pebi {
            set power 5
        }
        E -
        Ei -
        exa -
        Exa -
        exbi -
        Exbi {
            set power 6
        }
        Z -
        Zi -
        zetta -
        Zetta -
        zebi -
        Zebi {
            set power 7
        }
        Y -
        Yi -
        yotta -
        Yotta -
        yobi -
        Yobi {
            set power 8
        }        
        default {
            error "Unknown binary unit prefix \"$prefix\""
        }
    }
    
    return [expr {pow(1024,$power)}]
}

proc qc::binary_convert {args} {
    #| Convert binary file size units.
    #| Usage: qc::binary_convert size from_unit to_unit
    #|        qc::binary_convert size to_unit
    
    # Parse arguments
    if { [llength $args] == 3 } {
        lassign $args size from_units to_units
    } elseif { [llength $args] == 2 } {
        if { ! [regexp {^([0-9]+) *([^ 0-9]+)$} [lindex $args 0] -> size from_units] } {
            error "Invalid argument \"[lindex $args 0]\". Size must specify units eg. \"1000B\"."
        }
        set to_units [lindex $args 1]        
    } else {
        error "Invalid number of arguments.\nUsage:\nqc::binary_convert size from_unit to_unit\nqc::binary_convert size to_unit"
    }

    # Convert to bits
    if { [regexp {^(.*)(b|(?:bit|Bit)s?)$} $from_units -> unit_prefix] } {
        set unit_prefix_value [qc::binary_convert_unit_prefix2mult $unit_prefix]
        set size [expr {$size * $unit_prefix_value}]
    } elseif { [regexp {^(.*)(B|(byte|Byte)s?)$} $from_units -> unit_prefix] } {
        set unit_prefix_value [qc::binary_convert_unit_prefix2mult $unit_prefix]
        set size [expr {$size * $unit_prefix_value * 8}]
    } else {
        error "Unable to convert from unit \"$from_units\". Units must be in bits or bytes."
    } 
    
    # Convert to to_units
    if { [regexp {^(.*)(b|(?:bit|Bit)s?)$} $to_units -> unit_prefix] } {
        set unit_prefix_value [qc::binary_convert_unit_prefix2mult $unit_prefix]
        set size [expr {double($size) / $unit_prefix_value}]
    } elseif { [regexp {^(.*)(B|(byte|Byte)s?)$} $to_units -> unit_prefix] } {
        set unit_prefix_value [qc::binary_convert_unit_prefix2mult $unit_prefix]
        set size [expr {double($size) / $unit_prefix_value / 8}]
    } else {
        error "Unable to convert to unit \"$to_units\". Units must be in bits or bytes."
    } 

    return $size
}

proc qc::binary_format {args} {
    #| Return formatted string to display a binary file size in the most appropriate units.
    #| Usage: qc::binary_format ?-sigfigs sigfigs? size
    #|        qc::binary_format ?-sigfigs sigfigs? size units

    # Parse arguments
    args $args -sigfigs 3 args
    if { [llength $args] == 2 } {
        lassign $args size units
    } elseif { [llength $args] == 1 } {
        if { ! [regexp {^([0-9]+) *([^ 0-9]+)$} [lindex $args 0] -> size units] } {
            error "Invalid argument \"[lindex $args 0]\". Size must specify units eg. \"1000B\"."
        }
    } else {
        error "Invalid number of arguments.\nUsage:\nqc::binary_format ?-sigfigs sigfigs? size\nqc::binary_format ?-sigfigs sigfigs? size units"
    }

    # Convert to prefixed bits or bytes to unit bits or bytes
    if { [regexp {^(.*)(b|(?:bit|Bit)s?)$} $units -> unit_prefix] } {
        set size [qc::binary_convert $size $units b]
        set units b
    } elseif { [regexp {^(.*)(B|(byte|Byte)s?)$} $units -> unit_prefix] } {
        set size [qc::binary_convert $size $units B]
        set units B
    } else {
        error "Unable to parse to unit \"$units\". Units must be in bits or bytes."
    } 

    # Determine the most appropriate unit prefix
    switch [expr {int(log($size)/log(1024))}] {
        0 {
            set unit_prefix ""
        }
        1 {
            set unit_prefix K
        }
        2 {
            set unit_prefix M
        }
        3 {
            set unit_prefix G
        }
        4 {
            set unit_prefix T
        }
        5 {
            set unit_prefix P
        }
        6 {
            set unit_prefix E
        }
        7 {
            set unit_prefix Z
        }
        8 {
            set unit_prefix Y
        }
        default {
            # Make Y the default prefix 
            set unit_prefix Y
        }
    }

    # Convert to desired unit prefix, round to sigfigs and append units to the string
    return "[qc::sigfigs [qc::binary_convert $size $units ${unit_prefix}${units}] $sigfigs] ${unit_prefix}${units}"
}


namespace eval qc {
    namespace export multimap_*
}

proc qc::multimap_get_first {args} {
    #| Return the value for the first matching key
    args $args -nocase -glob -regexp -- multimap key
    set switches {}
    foreach switch {nocase glob regexp} {
        if { [info exists $switch] } {
            lappend switches -${switch}
        }
    }
    set index [lsearch {*}$switches [multimap_keys $multimap] $key]

    if { $index > -1 } {
	return [lindex $multimap [expr {($index*2)+1}]]
    } else {
	error "multimap does not contain the key:$key it contains \"$multimap\""
    }
}

proc qc::multimap_set_first {args} {
    #| Set the value of the first matching key
    args $args -nocase -glob -regexp -- multimapVariable key value
    upvar 1 $multimapVariable multimap
    set switches {}
    foreach switch {nocase glob regexp} {
        if { [info exists $switch] } {
            lappend switches -${switch}
        }
    }
    set index [lsearch {*}$switches [multimap_keys $multimap] $key]

    if { $index > -1 } {
	lset multimap [expr {($index*2)+1}] $value
    } else {
	lappend multimap $key $value
    }
}

proc qc::multimap_unset_first { multimapVariable key args } {
    #| Delete the first matching key/value pair from the multimap
    if { [llength $args] > 1 } {
        error "Usage: qc::multimap_unset_first multimapVariable key ?value?"
    }
    upvar 1 $multimapVariable multimap
    if { [llength $args] == 1 } {
        set search_value [lindex $args 0]
        set index 0
        foreach {name value} $multimap {
            if { $name eq $key && $value eq $search_value } {
                set multimap [lreplace $multimap $index [expr {$index+1}]]
                return $multimap
            }
            incr index 2
        }
    } else {
        set search_value [lindex $args 0]
        set index 0
        foreach {name value} $multimap {
            if { $name eq $key } {
                set multimap [lreplace $multimap $index [expr {$index+1}]]
                return $multimap
            }
            incr index 2
        }
    }
}

proc qc::multimap_exists { args } {
    #| Check if a value exists for this key
    args $args \
        -nocase \
        -glob \
        -regexp \
        -- \
        multimap \
        key

    set switches [list]

    foreach switch {nocase glob regexp} {
        if { [info exists $switch] } {
            lappend switches -${switch}
        }
    }

    if { [lsearch {*}$switches [multimap_keys $multimap] $key] > -1 } {
	return 1
    } else {
	return 0
    }
}

proc qc::multimap_keys {multimap} {
    #| Return a list of keys in the multimap
    set keys {}
    foreach {key value} $multimap {
	lappend keys $key
    }
    return $keys
}

proc qc::multimap_get_all { multimap key } {
    #| Return all value for this key
    set list {}
    foreach index [lsearch -all $multimap $key] {
	if { $index%2 == 0 } {
	    lappend list [lindex $multimap [expr {$index+1}]]
	}
    }
    return $list
}


namespace eval qc::response {

    namespace export status record message action extend data
    namespace ensemble create -unknown {
        response_subcommand_map
    }

    proc data {tson} {
        #| Sets the JSON response object data property using tson
        global data
        dict set data data $tson
    }

    proc extend {name args} {
        #| Extends the JSON response with an object named $name with properties defined in $args.
        #| Deprecated in favour of qc::response data
        if { [llength $args] % 2 != 0 } {
            return -code error "Usage: qc::response extend name key value ?key value ...?"
        }
        global data
        dict for {key value} $args {
            dict set data $name $key $value
        }
    }

    ##################################################
    #
    # Response Status
    #
    ##################################################
    namespace eval status {

        namespace export invalid valid get
        namespace ensemble create
        
        proc invalid {} {
            #| Set the status as "invalid" for the JSON Response.
            global data
            dict set data status "invalid"
        }

        proc valid {} {
            #| Set the status as "valid" for the JSON Response.
            global data
            dict set data status "valid"
        }

        proc get {} {
            #| Gets the current status for the JSON Response.
            global data
            return [dict get $data status]
        }
    }
    
    ##################################################
    #
    # Response Record
    #
    ##################################################
    namespace eval record {

        namespace export invalid valid remove all_valid sensitive
        namespace ensemble create

        proc valid {name value {message ""}} {
            #| Adds the given field to the record as valid. If the field already exists then updates it.
            global data
            dict set data record $name valid true
            dict set data record $name value $value
            dict set data record $name message $message
        }

        proc invalid {name value message} {
            #| Adds the given field to the record as invalid. If the field already exists then updates it.
            #| Also sets the status of the response to invalid.
            global data
            dict set data record $name valid false
            dict set data record $name value $value
            dict set data record $name message $message
            qc::response status invalid
        }

        proc remove {name} {
            #| Removes the field with the given name from the record.
            global data
            dict unset data record $name
        }
        
        proc all_valid {} {
            #| Return whether the record is all valid.
            global data
            set values {}
            if {[info exists data] && [dict exists $data record]} {
                dict for {field dict} [dict get $data record] {
                    lappend values [dict get $dict valid]
                }
                return [expr [join $values " && "]]
            } else {
                return true
            }
        }

        proc sensitive {name} {
            #| Marks the given field as being sensitive.
            # Values of sensitive fields will not be included in the response sent back to the client.
            global data
            dict set data record $name sensitive true
        }
    }

    ##################################################
    #
    # Response Message
    #
    ##################################################
    namespace eval message {

        namespace export notify alert error
        namespace ensemble create

        proc notify {message} {
            #| Sets the notify property with the given message.
            global data
            dict set data message notify value $message
        }

        proc alert {message} {
            #| Sets the alert property with the given message.
            global data
            dict set data message alert value $message
        }

        proc error {message} {
            #| Sets the error property with the given message.
            global data
            dict set data message error value $message
        }
    }

    ##################################################
    #
    # Response Action
    #
    ##################################################
    namespace eval action {
        
        namespace export redirect external_redirect resubmit login
        namespace ensemble create

        proc redirect { url } {
            #| Sets the redirect property with the given internal URL.
            global data
            reset
            
            dict set data action redirect value [qc::cast next_url $url]
        }

        proc external_redirect {url} {
            #| Sets the redirect property with the given external URL.
            global data
            reset
                        
            # check for malicious mal-formed url
            if { ![qc::is url $url] } {
                error "\"[html_escape $url]\" is not a valid url."
            }

            dict set data action redirect value $url
        }

        proc resubmit {} {
            #| Sets the resubmit action - used to resubmit the form after updating client's session/authenticity token 
            global data
            reset
            dict set data action resubmit value true
        }

        proc login { url } {
            #| Sets the login property with the given internal URL.
            global data
            reset

            dict set data action login value [qc::cast next_url $url]
        }

        proc reset {} {
            #| Resets the action property of the data structure.
            global data
            if {[info exists data] && [dict exists $data action]} {
                dict unset data action
            }
        }
    }
}

proc qc::response2tson {} {
    #| Convert the global data structure into TSON.
    global data
    set record_objects [list object]
    set action_objects [list object]
    set message_objects [list object]
    set data_tson [list]
    set extensions [list]
    # Default status is "valid"
    set status "valid"
    ::try {
        dict for {key value} $data {
            switch $key {
                record {
                    foreach {name values} $value {
                        set object [list object]
                        lappend object valid [dict get $values valid]
                        lappend object message [list string [dict get $values message]]
                        if { ! [dict exists $values sensitive] || [dict get $values sensitive] == false } {
                            # Not a sensitive field so we can safely echo back the value to the client.
                            lappend object value [list string [dict get $values value]]
                        }
                                                                   
                        lappend record_objects $name $object
                    }
                }
                message {
                    foreach {type val} $value {
                        lappend message_objects $type [list object value [list string [dict get $val value]]]
                    }
                }
                action {
                    foreach {type val} $value {
                        lappend action_objects $type [list object value [list string [dict get $val value]]]
                    }
                }
                status {
                    set status $value
                }
                data {
                    set data_tson $value
                }
                default {
                    set object [list object]
                    foreach {name val} $value {
                        lappend object $name [list string $val]
                    }
                    lappend extensions $key $object
                }
            }
        }
        set result [list object status $status record $record_objects message $message_objects action $action_objects data $data_tson {*}$extensions]
        return $result
    } on error [list error_message options] {
        return -code error "Malformed data: $error_message"
    }
}

proc qc::response2json {} {
    #| Convert the global data structure into JSON.
    return [qc::tson2json [qc::response2tson]]
}

proc qc::response2xml {} {
    #| Convert the global data structure into XML.
    return [qc::tson2xml [qc::response2tson]]
}

proc qc::response2html_snippet {} {
    #| Convert the global data structure into HTML snippet.
    global data
    set record_elements {}
    set action_elements {}
    set message_elements {}
    set extended_elements {}
    set data_html ""
    # Default status is "valid"
    set status "valid"
    ::try {
        dict for {key value} $data {
            switch $key {
                record {                   
                    foreach {name values} $value {
                        set field_status "valid"
                        if { ! [dict get $values valid] } {
                            set field_status "invalid"
                        }
                        
                        set temp {}
                        # value
                        if { ! [dict exists $values sensitive] || [dict get $values sensitive] == false } {
                            # Not a sensitive field so we can safely echo back the value to the client.
                            lappend temp [h div \
                                              class "value" \
                                              [html_escape [dict get $values value]] \
                                             ]
                        }
                        # message
                        lappend temp [h div \
                                          class "message" \
                                          [dict get $values message] \
                                         ]
                        
                        lappend record_elements [h div \
                                                     id $name \
                                                     class "field $field_status" \
                                                     [join $temp \n] \
                                                    ]                                                 
                    }
                }
                message {
                    foreach {type val} $value {
                        lappend message_elements [h div \
                                                      class "$type" \
                                                      [dict get $val value] \
                                                     ]
                    }
                }
                action {
                    foreach {type val} $value {
                        lappend action_elements [h div \
                                                     class "$type" \
                                                     [html_escape [dict get $val value]] \
                                                    ]
                    }
                }
                status {
                    set status $value
                }
                data {
                    set data_html [qc::response_tson2html_snippet $value]
                }
                default {
                    foreach {name val} $value {
                        lappend extended_elements [h div \
                                                       id "$key" \
                                                       [html_escape $val] \
                                                      ]
                    }
                }
            }
        }
    } on error [list error_message options] {
        return -code error "Malformed data: $error_message"
    }
    
    return [h div \
                id "validation_response" \
                class "validation-response" \
                [join [list \
                           [h div class status $status] \
                           [h div class message [join $message_elements \n]] \
                           [h div class record [join $record_elements \n]] \
                           [h div class action [join $action_elements \n]] \
                           [h div class data $data_html] \
                           [h div class extended [join $extended_elements \n]]
                          ] \n] \
               ]
}

proc qc::response_tson2html_snippet {tson} {
    #| Converts tson into a HTML snippet recursively
    switch -- [lindex $tson 0] {
        object {
            set elements [list]
            foreach {name value} [lrange $tson 1 end] {
                lappend elements \
                    [h div \
                         class object-property \
                         data-name $name \
                         [qc::response_tson2html_snippet $value]]
            }
            return [join $elements \n]                        
        }
        array {
            set elements [list]
	    foreach value [lrange $tson 1 end] {
                lappend elements \
                    [h div \
                         class array-element \
                         [qc::response_tson2html_snippet $value]]
            }
            return [join $elements \n]
        }
        string -
        number -
        boolean {
            return \
                [h span \
                     class [lindex $tson 0] \
                     [qc::html_escape [lindex $tson 1]]]
        }
        default {
            return [qc::html_escape $tson]
        }
    }
}

proc qc::response2html {} {
    #| Convert the global data structure into HTML.
    
    if {  [qc::response status get] eq "valid" } {
        # data validation passed
        set title "Data Validation Passed"

        # hide all response elements
        set css {
            .validation-response * {
                display: none;
            }
        }
    } else {
        # data validation failed
        set title "Missing or Invalid Data"
        
        # hide response elements that do not relate to invalid user input
        set css {
            .validation-response > .status,
            .validation-response > .record .field.valid,
            .validation-response > .record .field.invalid .value,
            .validation-response > .action,
            .validation-response > .data,
            .validation-response > .extended {
                display: none;
            }
            
            .validation-response > .record .field.invalid {
                display: list-item;
                list-style-position: inside;
                margin-left: 0em;        
            }
            
            .validation-response > .message,
            .validation-response > .record,
            .validation-response-advise {
                margin-bottom: 10px;
            }
        }
    }
    
    set content ""
    append content [h h1 class "validation-response-page-title" $title]
    append content [qc::response2html_snippet]
    append content [h div class "validation-response-advice" "Please go back and try again."]
    
    set html [h_tag "!doctype html"]
    append html [h html \
                     lang "en" \
                     [join [list \
                                [h head \
                                     [join [list \
                                                [h title $title] \
                                                [h style $css] \
                                               ] \n] \
                                     ] \
                                [h body \
                                     $content \
                                    ] \
                               ] \n] \
                     ]                      

    return $html
}

namespace eval qc {
    namespace export file_temp file_write file_upload
}

proc qc::file_temp {args} {
    #| Write the text $text out into a temporary file
    #| and return the name of the file.
    qc::args $args -binary -- contents {mode 0600}
    default binary false
    package require fileutil
    set filename [fileutil::tempfile]
    set switches {}
    if { $binary } {
        lappend switches -binary
    }
    qc::file_write {*}$switches -- $filename $contents $mode
    return $filename
}

proc qc::file_write {args} {
    # Return true if file has changed by writing to it.
    qc::args $args -binary -- filename contents {perms ""}
    default binary false
    if { $perms ne "" } {
	set perms [qc::format_right0 $perms 5]
    }
    if { ![file exists $filename]
         || [qc::cat $filename] ne $contents
         || [file attributes $filename -permissions]!=$perms
     } { 
        log Debug "writing ${filename} ..."
        set handle [open $filename w+ 00600]
        if { $binary } {
            fconfigure $handle -translation "binary"
        }
        puts -nonewline $handle $contents
        close $handle
        if { $perms ne "" } {
            # set file permissions
            file attributes $filename -permissions $perms
        }
        log Debug "written"
        return true
    } else {
        return false
    }
}

proc qc::file_upload {name chunk chunks file} {
    #| Keeps uploaded file parts sent by chunked upload and concat them once all parts have been sent.
    # returns filename of concat file when upload is complete ("" otherwise)
    set user_id [qc::auth]
    set id $name
    set tmp_file /tmp/$id.$chunk

    # Move the AOLserver generated tmp file to one we will keep
    file rename [ns_getformfile file] $tmp_file

    if { ![nsv_exists locks $user_id] } {
        nsv_set locks $user_id [ns_mutex create]
    }

    ns_mutex eval [nsv_get locks $user_id] {
        if { [nsv_exists uploads $user_id] } {
            set dict [nsv_get uploads $user_id]
        } else {
            set dict {}
        }
        dict set dict $id $chunk $tmp_file
        dict set dict $id chunks $chunks
        dict set dict $id filename $file

        set complete true
        foreach chunk [.. 0 $chunks-1] {
            if { ![dict exists $dict $id $chunk] } {
                set complete false
                break
            } else {
                lappend files /tmp/$id.$chunk
            }
        }

        if { $complete } {
            # Join parts together
            ::try {
                exec_proxy cat {*}$files > /tmp/$id
            } finally {
                # Clean up
                foreach file $files {
                    file delete $file
                }
            }
            dict unset dict $id
            nsv_set uploads $user_id $dict
            set filename "/tmp/$id"
        } else {
            nsv_set uploads $user_id $dict
            set filename ""
        }
    }

    return $filename
}

proc qc::cat {filename} {
    set handle [open $filename r]
    set contents [read $handle]
    close $handle
    return $contents
}

proc qc::file2url {file} {
    # Takes a file and returns url path relative to www root.
    set root [ns_pagepath]
    if { [string equal -length [string length $root] $root $file]
         && [string length $file] > [string length $root]
     } {
        return [string range $file [string length $root] end]
    } else {
        error "$file is outside page root $root"
    }
}

proc qc::file_cache_create {file_id file_path} {
    #| Create a file in the disk cache
    set file [qc::db_file_export $file_id]

    # Place files in cache
    if { ! [file exists [file dirname $file_path]] } {
        file mkdir [file dirname $file_path]
    }

    file rename -force $file $file_path

    return $file_path
}

proc qc::file_cache_data {file_id file_path} {
    #| Return dict containing canonical url of cached file.
    #| Return an empty dict if cached file does not exist.

    if { [nsv_exists file_cache_data $file_id] } {
        # File has been recorded in NSV file_cache_data.

        return [nsv_get file_cache_data $file_id]
    } elseif { [file exists $file_path]
               && [file isfile $file_path] } {
        # File exists on disk but hasn't been recorded in NSV file_cache_data.

        set url [qc::file2url $file_path]
        set data [dict_from url]
        nsv_set file_cache_data $file_id $data

        return $data
    } else {
        # File doesn't exist and hasn't been recorded in NSV file_cache_data.

        return [dict create]
    }
}

proc qc::file_data {file_id file_path} {
    #| Return dict containing url of file.
    #| Generates file cache if it doesn't already exist.
    if { ! [qc::file_cache_exists $file_id $file_path] } {
        qc::file_cache_create $file_id $file_path
    }

    return [qc::file_cache_data $file_id $file_path]
}

proc qc::file_cache_exists {file_id file_path} {
    #| Return true if this file has been cached on disk.
    if { [llength [qc::file_cache_data $file_id $file_path]] > 0 } {
        return true
    } else {
        return false
    }
}

namespace eval qc {
    namespace export table_foreach table2ldict table2array table_sum
}

proc qc::table_foreach { table code } {
    # Loop through a table setting local variables with values in the row
    # named using the first row.
    # Then execute code given for every row.
    set keys [lindex $table 0]
    foreach list [lrange $table 1 end] {
	foreach key $keys value $list {
	    upset 1 $key $value
	}
	uplevel 1 $code
    }
}

proc qc::table2ldict { table } {
    #| Convert a table to an ldict (list of dicts)
    set ldict {}
    set keys [lindex $table 0]
    foreach list [lrange $table 1 end] {
	set dict {}
	foreach key $keys value $list {
	    lappend dict $key $value
	}
	lappend ldict $dict
    }
    return $ldict
}

proc qc::table2array { table args } {
    # Attempt at cross tab stuff 
    array set data {}
    set keys [lindex $table 0]
    table_foreach $table {
	foreach varname [lexclude $keys {*}$args] {
	    set pkeys {}
	    foreach key $args {
		lappend pkeys [set $key]
	    }
	    lappend pkeys $varname
	    set pk [join $pkeys ,]
	    set data($pk) [set $varname]
	}
    }
    return [array get data]
}

proc qc::table_sum { table col_name } {
    #| Return the sum of values in a column
    set sum 0
    set col_index [lsearch [lindex $table 0] $col_name]
    foreach row_index [.. 1 [expr {[llength $table]-1}]] {
	set value [lindex $table $row_index $col_index]
	if { [qc::is decimal $value] } {
	    incr0 sum $value
	}
    }
    return $sum
}

namespace eval qc {
    namespace export db_file_* file_upload plupload.html
}

proc qc::db_file_insert {args} {
    #| Insert a file into the file db table
    args $args -user_id ? -filename ? -mime_type ? -- file_path

    if { ![info exists user_id] } {
        set user_id [auth]
    }
    default filename [file tail $file_path]

    if { ! [info exists mime_type] } {
        set mime_type [qc::mime_type_guess $filename]
    }
   
    set file_id [db_seq file_id_seq]
    qc::aws_credentials_set_from_ec2_role
    set s3_location [qc::s3 uri [qc::param_get s3_file_bucket] $file_id]
    # upload file to amazon s3
    qc::s3 put $s3_location $file_path
    
    set qry {
	insert into file 
	(file_id,user_id,filename,mime_type,s3_location)
	values 
	(:file_id,:user_id,:filename,:mime_type,:s3_location)
    }
    db_dml $qry
    return $file_id

}

proc qc::db_file_copy {file_id} {
    #| Make a copy of this file
    set new_file_id [db_seq file_id_seq]
    db_dml {
        insert into file 
        (file_id,user_id,filename,data,mime_type,s3_location)
	select
        :new_file_id,
        user_id,
        filename,
        data,
        mime_type,
        s3_location
        from file
        where file_id=:file_id
    }
    return $new_file_id
}

proc qc::db_file_export {args} {
    # Export a file in the db file table to a tmp local file
    args $args -tmp_file ? -- file_id

    default tmp_file /tmp/[qc::uuid]
    db_1row {
        select
        filename,
        encode(data,'base64') as base64,
        s3_location
        from file
        where file_id=:file_id
    }

    if { $s3_location eq "" } {
        # file does not exists on amazon s3
        if { $base64 eq "" } {
            error "Missing data or s3_location for file $file_id"
        }
        set id [open $tmp_file w]
        fconfigure $id -translation binary
        puts -nonewline $id [base64::decode $base64]
        close $id
    } else {
        # file exists on amazon s3
        qc::aws_credentials_set_from_ec2_role
        qc::s3 get $s3_location $tmp_file
    }
    return $tmp_file
}

proc qc::plupload.html {name chunk chunks file} {
    # Deprecated
    return [db_file_upload $name $chunk $chunks $file]
}

proc qc::db_file_upload {name chunk chunks file {filename ""} {mime_type ""}} {
    #| Upload a file chunk, insert the completed file into the db when complete.
    if { $filename eq "" } {
        set filename $file
    }
    set flags [dict create -filename $filename]
    dict set flags -user_id [qc::auth]
    if { $mime_type ne "" } {
        dict set flags -mime_type $mime_type
    }
    set tmp_file [qc::file_upload $name $chunk $chunks $file]
    if { $tmp_file ne "" } {
        ::try {
            set file_id [qc::db_file_insert {*}$flags $tmp_file]
            if { [qc::file_is_valid_image $tmp_file] } {
                dict2vars [qc::image_file_info $tmp_file] width height
                db_dml {
                    insert into image
                    (file_id, width, height)
                    values
                    (:file_id, :width, :height)
                }
            }
        } finally {
            file delete $tmp_file
        }
        return $file_id
    } else {
        return ""
    }
}

proc qc::db_file_migrate_to_s3 {file_id} {
    set tmp_file [db_file_export $file_id]
    set s3_location [qc::s3 uri [qc::param_get s3_file_bucket] $file_id]
    qc::s3 put $s3_location $tmp_file
    file delete $tmp_file
    
    set qry {
	update file
        set s3_location=:s3_location, data=NULL
        where file_id=:file_id
    }
    db_dml $qry
    
    return $s3_location
}

proc qc::db_file_delete {file_id} {
    #| Delete a file
    db_1row {
        select
        s3_location
        from file
        where
        file_id=:file_id
    }

    if { $s3_location ne "" } {
        qc::aws_credentials_set_from_ec2_role
        qc::s3 delete $s3_location
    }
    db_dml {delete from file where file_id=:file_id}
}

namespace eval qc {
    # Tcl 8.5 only
    namespace import ::tcl::mathop::eq
    namespace import ::tcl::mathop::ne

    namespace export {*}{
        K
        default
        setif
        sset
        sappend
        coalesce
        incr0
        call
        margin
        breakpoint
        trunc
        iif
        ?
        true
        false
        escapeHTML
        unescapeHTML
        xsplit
        mcsplit
        perct
        subsets
        permutations
        split_pair
        min_nz
        max_nz
        key_gen
        ..
        debug
        log
        exec_proxy
        info_proc
        which
        string2hex
        not_null
        eq
        ne
        commonmark2html
        regexp_escape
        split2
        split_pair2
        string_is_escaped
        control_port_source
        ./
    }
}

proc qc::K {a b} {set a}

proc qc::try { try_code { catch_code ""} } {
    #| Try to execute the code try_code and catch any error. 
    #| If an error occurs then run catch_code.
    set return_code [ catch { uplevel 1 $try_code } result options]
    switch $return_code {
        1 {
            # Error
            set return_code [ catch { uplevel 1 $catch_code } catch_result options]
            # Preserve TCL_RETURN
            if { $return_code == 2 && [dict get $options -code] == 0 } {
                dict set options -code return
            } else {
                # Return in parent stack frame instead of here
                dict incr options -level
            }
            return -options $options $catch_result
	}
        default {
            # ok, return, break, continue

            # Preserve TCL_RETURN
            if { $return_code == 2 && [dict get $options -code] == 0 } {
                dict set options -code return
            } else {
                # Return in parent stack frame instead of here
                dict incr options -level
            }
            return -options $options $result
        }
    }
}

proc qc::default { args } {
    #| If a variable does not exists then set its value to defaultValue
    foreach {varName defaultValue} $args {
	upvar 1 $varName value
	if { ![info exists value] || [string equal $value UNDEF] } {
	    set value $defaultValue
	} 
    }
}

proc qc::setif { varName ifValue defaultValue } {
    #| Set varName to be defaultValue if varName is set to ifValue or does not exist
    upvar 1 $varName value
    if { ![info exists value] || [string equal $value $ifValue] } {
	set value $defaultValue
    } 
}

proc qc::sset { varName value } {
    #| Set varName to value after having performed a subst.
    # strip leading newline
    regsub -all {^\n +} $value {} value
    # strip trailing newline
    regsub -all {\n$} $value {} value
    # strip leading whitespace
    regsub -all {\n[ \t]+} $value {\n} value
    # subst in uplevel namespace
    uplevel "set $varName \[[list subst $value]\]"
}

proc qc::sappend { varName value } {
    #| Append value to the contents of varName having first performed a subst

    # strip leading newline
    regsub -all {^\n +} $value {} value
    # strip trailing newline
    regsub -all {\n$} $value {} value
    # strip leading whitespace
    regsub -all {\n[ \t]+} $value {\n} value
    # subst in uplevel namespace
    uplevel "append $varName \[[list subst $value]\]"
}

proc qc::coalesce { varName altValue } {
    #| If varName exists then return its value
    #| else return the altvalue
    upvar 1 $varName value
    if { [ info exists value] } {
	return $value
    } else {
	return $altValue
    }
}


proc qc::incr0 { varName amount } {
    #| Increment the value of varName by amount
    upvar 1 $varName var
    if { [info exists var] } {
	set var [add $var $amount]
    } else {
	set var $amount
    }
    return $var
}

proc qc::call { proc_name args } {
    #| Calls a procedure using local variables as arguments.
    #| Useful when a large number of arguments are required.
    if { [info args $proc_name] eq "args" } {
	# Call the proc using a dict of corresponding local vars
	if { $args ne "" } {
	    return [uplevel 1 "$proc_name {*}\[dict_from {*}$args\]"]
	} else {
	    return [uplevel 1 $proc_name]
	} 
    } else {
	# Call a proc using args of matching names to local variables.
        set largs {}
	foreach arg [info args $proc_name] {
	    upvar 1 $arg temp
	    if { [info default $proc_name $arg default_value] && ![info exists temp]} {
		# No corresponding variable exists 
		lappend largs $default_value
	    } elseif { [info exists temp] } {
		lappend largs $temp
	    } else {
		error "Cannot use variable \"$arg\" to call proc qc::\"$proc_name\":no such variable \"$arg\""
	    }
	}
	return [uplevel 1 $proc_name $largs]
    }
}

proc qc::call_with {proc_name args} {
    #| Calls proc_name by mapping name value pairs to named args.
    if { [llength $args]%2 != 0 } {
        return -code error "usage qc::call_with proc_name ?name value?"
    }
    
    set proc_args [info args $proc_name]
    set largs {}
    foreach arg $proc_args {
	if { [dict exists $args $arg] } {
            set value [dict get $args $arg]
            if {$arg eq "args"} {
                lappend largs {*}$value
            } else {
                lappend largs $value
            }
	} else {
            if { [info default $proc_name $arg default_value] } {
                lappend largs $default_value
            } else {
                return -code error "Missing argument: \"$arg\"."
            }
        }
    }
    return [uplevel 0 $proc_name $largs]    
}

proc qc::margin { cost price {dec_places 1} } {
    #| Calculates the gross margin on supplied cost and revenue
    if { $price==0 } {
	return ""
    } else {
	return [qc::round [expr {double($price-$cost)/$price*100}] $dec_places]
    }
}

proc qc::breakpoint {{s {}}} {
    # From tcl cookbook
    if {![info exists ::bp_skip]} {
	set ::bp_skip [list]
    } elseif {[lsearch -exact $::bp_skip $s]>=0} {
	return
    }
    set who [info level -1]
    while 1 {
	# Display prompt and read command.
	puts -nonewline "$who/$s> "; flush stdout
	gets stdin line

	# Handle shorthands
	if {$line=="c"} {puts "continuing.."; break}
	if {$line=="i"} {set line "info locals"}

	# Handle everything else.
	catch {uplevel 1 $line} res
	puts $res
    }
}

proc qc::trunc {string length} {
    #| Truncates string to specified length
    return [string range $string 0 [expr {$length-1}]]
}

proc qc::iif { expr true false } {
    #| Inline if statement
    if { [uplevel 1 eval expr $expr] } {
	return $true
    } else {
	return $false
    }
}

proc qc::? { expr true false } {
    #| Shorthand version of qc::iif
    if { [uplevel 1 eval expr $expr] } {
	return $true
    } else {
	return $false
    }
}

proc qc::true { string {true true} {false false} } {
    #| Test if string is true. Recognised forms are "yes/no" "true/false" or 1/0.
    #| Optionally set the values to return for each case.
    if { [string is true -strict $string] } {
	return $true
    } else {
	return $false
    }
}

proc qc::false { string {true true} {false false} } {
    #| Test if string is false. Recognised forms are "yes/no" "true/false" or 1/0.
    #| Optionally set the values to return for each case.
    if { [string is false -strict $string] } {
	return $true
    } else {
	return $false
    }
}

proc qc::escapeHTML { html } {
    #| TODO Deprecate for html_escape: Convert reserved HTML characters in a string into entities
    return [string map {< &lt; > &gt; & &amp; \" &quot; ' &\#39;} $html]
}

proc qc::unescapeHTML { text } {
    #| Convert HTML entities back to their ascii characters
    return [string map {&lt; < &gt; > &amp; & &\#39; ' &\#34; \" &quot; \"} $text]
}

proc qc::xsplit [list str [list regexp "\[\t \r\n\]+"]] {
    # TODO unused
    set list  {}
    while {[regexp -indices -- $regexp $str match submatch]} {
	lappend list [string range $str 0 [expr [lindex $match 0] -1]]
	if {[lindex $submatch 0]>=0} {
	    lappend list [string range $str [lindex $submatch 0]\
                              [lindex $submatch 1]] 
	}	
	set str [string range $str [expr [lindex $match 1]+1] end] 
    }
    lappend list $str
    return $list
}

proc qc::mcsplit {string splitString} {
    #| Split the string on the supplied string which can be of arbitrary length (unlike split).
    set mc \x00
    return [split [string map [list $splitString $mc] $string] $mc]
}

proc qc::perct {x n {p 1}} {
    # TODO unused
    return [qc::round [expr {double($x)/$n*100}] $p]
}

proc qc::subsets {l n} {
    #| Returns all possible subsets of length n from list l
    set subsets [list [list]]
    set result [list]
    foreach e $l {
	foreach subset $subsets {
	    lappend subset $e
	    if {[llength $subset] == $n} {
		lappend result $subset
	    } else {
		lappend subsets $subset
	    }
	}
    }
    return $result
}

proc qc::permutations {list} {
    #| Returns all permuations of the supplied list
    set res [list [lrange $list 0 0]]
    set posL {0 1}
    foreach item [lreplace $list 0 0] {
	set nres {}
	foreach pos $posL {
	    foreach perm $res {
		lappend nres [linsert $perm $pos $item]
	    }
	}
	set res $nres
	lappend posL [llength $posL]
    }
    return $res
}

proc qc::split_pair {string delimiter} {
    #| split a string into 2 parts at the first occurence of the delimiter
    set list {}
    if {[set index [string first $delimiter $string]]!=-1} {
	lappend list [string trim [string range $string 0 [expr {$index-1}]]] 
	lappend list [string trim [string range $string [expr {$index+[string length $delimiter]}] end]]
    } else {
	error "Delimiter \"$delimiter\" was not found in the string \"$string\""
    }
    return $list
}

proc qc::min_nz {args} {
    # TODO Unused
    #| Return the minimum supplied value which is non zero 
    set list {}
    foreach value $args {
	if { [string is double -strict $value] && $value>0 } {
	    lappend list $value
	}
    }
    if { [llength $list] > 0 } {
        return [min {*}$list]
    } else {
        return ""
    }
}

proc qc::max_nz {args} {
    #| Return the maximum supplied value which is non zero 
    set list {}
    foreach value $args {
	if { [string is double -strict $value] && $value>0 } {
	    lappend list $value
	}
    }
    if { [llength $list] > 0 } {
        return [max {*}$list]
    } else {
        return ""
    }
}

package require md5
proc qc::md5 {string} {
    #| Returns the md5 hash of supplied string.
    return [string tolower [::md5::md5 -hex $string]]
}

package require sha1
proc qc::sha1 {string} {
    # Return the sha1 hash
    return [sha1::sha1 $string]
}

proc qc::key_gen { args } {
    args $args -lower -upper -int -- length

    set alphabet_lower [list a b c d e f g h i j k l m n o p q r s t u v w x y z]
    set alphabet_upper [list A B C D E F G H I J K L M N O P Q R S T U V W X Y Z]
    set alphabet_int [list 0 1 2 3 4 5 6 7 8 9]

    if { [info exists lower] } {
	lappend alphabet {*}$alphabet_lower
    }
    if { [info exists upper] } {
	lappend alphabet {*}$alphabet_upper
    }
    if { [info exists int] } {
	lappend alphabet {*}$alphabet_int
    }
    default alphabet [concat $alphabet_lower $alphabet_upper $alphabet_int]

    set key ""
    while { [string length $key] < $length } {
	append key [lindex $alphabet [expr int(rand()*[llength $alphabet]) ]]
    } 

    return $key
}

proc qc::.. {from to {step 1} {limit ""}} {
    #| List all values from $from to $to. Will attempt to guess the input type.
    set result {}
    # Check month lists
    set lists {}
    lappend lists {January February March April May June July August September October November December}
    lappend lists {Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec}
    lappend lists {Mon Tue Wed Thu Fri Sat Sun}
    lappend lists {Monday Tuesday Wednesday Thursday Friday Saturday Sunday}
    foreach list $lists {
	lappend lists [upper $list] [lower $list]
    }
    foreach list $lists {
	if { [in $list $from] && [in $list $to] } {
	    set index [lsearch $list $from]
	    while { ($limit eq "" && [lindex $list $index] ne $to && [llength $result]<[llength $list]) \
			|| ([llength $result]<$limit)} {
		lappend result [lindex $list $index]
		set index [expr {($index+$step)%[llength $list]}]
	    }
	    if { $limit eq "" && [lindex $list $index] eq $to} {
		lappend result $to
	    }
	    return $result
	}
    }
    # Dates
    if { [qc::is date $from] && [qc::is date $to] } {
	if {![regexp {(-)?([0-9]+) (day|month|year)s?} $step -> sign scaler unit] } {
	    # default step 1 day
	    set sign +;set scaler 1; set unit day
	}
	for {set i $from} {([ne $sign -] && [date_compare $i $to]<=0) || ([eq $sign -] && [date_compare $i $to]>=0)} {set i [cast date "$i $sign $scaler $unit"]} {
	    lappend result $i
	}
	return $result
    }
    # Expression
    set from [eval expr $from]
    set to [eval expr $to]
    if { [qc::is decimal $from] && [qc::is decimal $to] && [qc::is decimal $step]} {
	for {set i $from} {($step>0 && $i<=$to) || ($step<0 && $i>=$to)} {set i [expr {$i+$step}]} {
	    lappend result $i
	}
	return $result
    }
}

proc qc::debug {message} {
    #| If running in naviserver and debugging is switched on then write message to nsd log.
    #| Otherwise write message to stdout.
    #| Filter message by masking anything that looks like a card number.
    log Debug [qc::format_cc_masked_string $message]
}

proc qc::log {args} {
    #| If running in naviserver then write message to nsd log using App: prefix. 
    #| Otherwise write message to stout or stderr.
    #| Default severity argument to "Notice". 
    #| Filter message by masking anything that looks like a card number.
    #| Usage: qc::log ?Severity? message
    
    # Parse args
    set list [list Debug Notice Error]
    if { [llength $args]==1 } {
	set severity Notice
	set message [lindex $args 0]
    } elseif { [llength $args]==2 } {
	set severity [lindex $args 0]
        if { $severity ni $list } {
            error "Severity must be one of \"$list\""
        }
	set message [lindex $args 1]
    } else {
	error "Invalid args: usage log ?severity? message"
    }    
    
    # Mask anything in message that looks like a card number 
    set message [qc::format_cc_masked_string $message]

    # Output message
    if { [info commands ns_log] eq "ns_log" } {
        # Write to naviserver's nsd log
        if { "App:$severity" ni [ns_logctl severities] } {
            # turn on this log level.
            ns_logctl severity App:$severity on
        }
        ns_log "App:$severity" $message
    } elseif { $severity eq "Error" } {
        # Write to stderr
        puts stderr $message
    } else {
        # Write to stdout
        puts stdout  $message
    }
}

proc qc::exec_proxy {args} {
    #| Execute the given command.
    #| If running on aolserver will use ns_proxy, otherwise the command is executed directly.
    qc::args $args -timeout 1000 -pool exec -- args
    if { [info commands ns_proxy] eq "ns_proxy" } {
	set handle [ns_proxy get $pool]
	::try {
	    set result [ns_proxy eval $handle [list exec {*}$args] $timeout]
	    ns_proxy release $handle
	    return $result
	} on error {error_message options} {
	    ns_proxy release $handle
	    error $error_message [dict get $options -errorinfo] [dict get $options -errorcode]
	}
    } else {
	# No ns_proxy
	exec {*}$args
    }
}

proc qc::info_proc { proc_name } {
    #| Return the Tcl source code definition of a Tcl proc.
    if { [eq [info procs $proc_name] ""] && [eq [info procs ::$proc_name] ""] } {
	error "The proc $proc_name does not exist"
    }
    set proc_name [namespace which $proc_name]
    set largs {}
    foreach arg [info args $proc_name] {
	if { [info default $proc_name $arg value] } {
	    lappend largs [list $arg $value]
	} else {
	    lappend largs $arg
	}
    }
    set body [info body $proc_name]
    
    return "proc [string trimleft $proc_name :] \{$largs\} \{$body\}"
}



proc qc::which {command} {
    #| Return path of unix command - cache result in nsv on AOLserver if present
    if { [info commands nsv_exists] eq "nsv_exists" } {
	if { ![nsv_exists which $command] } {
	    nsv_set which $command [exec_proxy which $command]
	}
	set which [nsv_get which $command]
    } else {
	set which [exec which $command]
    }
    return $which
}

proc qc::string2hex {string} {
    #| Convert string to hex
    binary scan [encoding convertto utf-8 $string] H* hex
    return [regsub -all (..) $hex {\\x\1}]    
}

proc qc::not_null {var} {
    #| Test if this variable exists and is not the empty string
    upvar $var value
    return [expr {[info exists value] && $value ne "" }]
}

proc qc::commonmark2html {args} {
    #| Converts Commonmark Markdown (http://commonmark.org) to  HTML.
    qc::args $args -unsafe -- markdown
    set markdown [string map {\r ""} $markdown] ;#removes ^M (alternative carriage return) characters
    set html [exec cmark << $markdown]
    if {[info exists unsafe]} {
        return $html
    } elseif { [qc::is safe_html $html] } {
        return $html
    } else {
        return -code error -errorcode CAST "Markdown contains unsafe HTML."
    }
}

proc qc::glob_recursive {args} {
    #| Return names of files that match patterns, including files in sub-directories
    # Usage: same as glob (see http://www.tcl.tk/man/tcl8.5/TclCmd/glob.htm),
    # with additional max_depth switch
    qc::args $args -join -nocomplain -tails -path ? -directory ? -types ? -max_depth 10 -- args

    # Use a single glob call, but extend each pattern passed in with */ prefixes
    # (eg. *.html becomes *.html */*.html */*/*.html etc.)
    if { [info exists join] } {
        set base_patterns [list [join $args /]]
    } else {
        set base_patterns $args
    }
    set patterns $base_patterns
    foreach depth [.. 1 $max_depth] {
        foreach pattern $base_patterns {
            lappend patterns [join [list {*}[lrepeat $depth *] $pattern] /]
        }        
    }

    # Switches and options to be passed on to glob
    # switches
    set switches [list]
    foreach var {nocomplain tails} {
        if { [info exists $var] } {
            lappend switches -${var}
        }
    }
    # options
    set options [list]
    foreach var {path directory types} {
        if { [info exists $var] } {
            lappend options -${var} [set $var]
        }
    }

    # Call glob and return the results
    return [glob {*}$switches {*}$options -- {*}$patterns]
}

proc qc::args_qualified2unqualified {} {
    #| Creates an unqualified variable in the caller's stack frame for each qualified argument
    #| present in the caller but only if the resulting variable would be unambiguous.
    #| E.G. Caller's args: { table.foo }
    #|      -> Variable named "foo" is set in caller.
    #|      Caller's args: { table.foo other.foo }
    #|      -> No new variables set in caller because "foo" would be ambiguous.
    set caller_args [info args [dict get [info frame -1] proc]]
    foreach item [qc::args_unambiguous {*}$caller_args] {
        upvar 1 $item value
        qc::upset 1 [qc::arg_shortname $item] $value
    }
}

proc qc::arg_shortname {name} {
    #| Returns the shortname of a name if it is qualified otherwise returns the name.
    if { [regexp {^([^\.]+)\.([^\.]+)$} $name -> table column] } {
        return $column
    } else {
        return $name
    }
}

proc qc::args_unambiguous {args} {
    #| Returns items from the given list that would be unambiguous if shortnened by unqualifying them.
    #| E.G. unamibguous table.foo table.bar table.baz other.baz
    #|      -> table.foo table.bar
    set unambiguous {}
    set shortnames [map {x {qc::arg_shortname $x}} $args]
    foreach shortname $shortnames arg $args { 
        if { [llength [lsearch -all $shortnames $shortname]]==1 } {
            lappend unambiguous $arg
        }
    }
    return $unambiguous
}

proc qc::memoize {args} {
    #| Wrapper that caches the result of script evaluation to improve performance for future evaluations.
    #| Usage: `qc::memoize -timeout ? -expires ? script args`
    #| Example: `qc::memoize -expires 2520 -timeout 1 deep_tought "meaning of life"`
    # The script result remains valid until the supplied expire time passes, or forever if not specified. 
    # The value for -expires can be expressed either as an absolute time (large values intrepreted as seconds since epoch) 
    # or as an seconds offset from the current time.
    # If two threads execute the same script and args, one will wait for the other to compute the result and store it in the cache. 
    # The -timeout option specifies how long to wait in seconds.
    # nb. currently requires ns_memoize to perform memoization.
  
    if { [info commands ns_memoize] eq "ns_memoize" } {
        # ns_memoize available
        return [eval [list ns_memoize {*}$args]]
    } else {
        # ns_memoize is unavailable - evaluate script in the caller without memoization
        qc::args $args -expires ? -timeout ? -- args

        return [eval {*}$args]
    }
}

proc qc::postcode_parse { postcode } {
    #| Parse postcode into dict of constituent parts

    set postcode [string toupper $postcode]
    set area_regexp {[A-Z]{1,2}}
    set district_regexp {[0-9][0-9]?[A-Z]?}
    set space_regexp {\s}
    set sector_regexp {[0-9]}
    set unit_regexp {[A-Z]{2}}
    set parse_regexp "
        ^
        ( \$ | ${area_regexp} )
        ( \$ | ${district_regexp} )
        ( \$ | ${space_regexp} )
        ( \$ | ${sector_regexp} )
        ( \$ | ${unit_regexp} )
        $
    "
    # Parse postcode extracting area, district, sector and unit.
    if { ! [regexp -expanded $parse_regexp $postcode match area district space sector unit] } {
        error "Unable to parse postcode \"$postcode\""
    }
    return [qc::dict_from area district space sector unit]
}

proc qc::regexp_escape {string} {
    #| Escape string for use inside TCL regular expression
    set list {
	\\ \\\\ 
	^ \\^ 
	. \\. 
	\[ \\\[ 
	\] \\\] 
	\$ \\\$ 
	\( \\\( 
	\) \\\) 
	| \\| 
	* \\* 
	+ \\+ 
	? \\? 
	\{ \\\{
	\} \\\}
    } 

    return [string map $list $string]
}

proc qc::string_is_escaped {string index escape} {
    #| Check if char at index of string is preceded by an odd number of escapes
    set preceding_escapes 0
    incr index -1
    foreach index [.. $index 0 -1] {
        if { [string index $string $index] eq $escape } {
            incr preceding_escapes
        } else {
            break
        }
    }
    if { $preceding_escapes % 2 == 1 } {
        return true
    } else {
        return false
    }
}

proc qc::uuid {} {
    #| Generate type 4 UUID
    #| Tcllib's uuid liable to collisions if used under NaviServer so
    #| use a native implementation if ns_crypto is available.
    if { [info commands "::ns_crypto::randombytes"] eq "::ns_crypto::randombytes" } {
        # Use Naviserver implementation
        set b [ns_crypto::randombytes 16]
        set time_hi_and_version [string replace [string range $b 12 15] 0 0 4]
        set clk_seq_hi_res      [string range $b 16 17]
        set clk_seq_hi_res2     [format %2x [expr {("0x$clk_seq_hi_res" & 0x3f) | 0x80}]]
        return [format %s-%s-%s-%s%s-%s \
                    [string range $b 0 7] \
                    [string range $b 8 11] \
                    $time_hi_and_version \
                    $clk_seq_hi_res2 \
                    [string range $b 18 19] \
                    [string range $b 20 31] \
                ]
    } else {
        # Use tcllib implementation
        package require uuid
        return [uuid::uuid generate]
    }
}

proc qc::control_port_source {filename} {
    #| Capture and return stdout while sourcing filename
    global stdout
    set stdout ""
    rename ::puts ::_puts
    proc puts { args } {
        global stdout
        if { [llength $args ] == 1 || ([llength $args]==2 && [lindex $args 0] eq "stdout") } {           
            append stdout [lindex $args end]\n
        } elseif {
                  ([llength $args] == 2 && [lindex $args 0] eq "-nonewline")
                  || ([llength $args ]== 3 && [lindex $args 0] eq "-nonewline" && [lindex $args 1] eq "stdout")
              } {
            append stdout [lindex $args end]
        } else {
            ::_puts {*}$args
        }
    }      
    source $filename
    rename ::puts {}
    rename ::_puts ::puts
    return $stdout
}

proc qc::time_start {tag} {
    global timing
    set timing($tag) [::tcl::clock::microseconds]
}

proc qc::time_end {tag} {
    global timing
    qc::log "Time for $tag : [expr {double([expr {[::tcl::clock::microseconds] - $timing($tag)}])/1000000}] sec"
}

proc qc::./ {path} {
    #| Returns the path relative to this file.
    return [file join [file dirname [info script]] $path]
}

namespace eval qc {
    namespace export ssh scp ssh_call_proc
}

proc qc::ssh {args} {
    #| Execute an ssh command with some default options 
    set args [linsert $args 0 -o UserKnownHostsFile=/dev/null -o StrictHostKeyChecking=no -q]
    return [exec -ignorestderr ssh {*}$args]
}

proc qc::scp {args} {
    #| Execute an scp command with some default options 
    set args [linsert $args 0 -o UserKnownHostsFile=/dev/null -o StrictHostKeyChecking=no -q]
    return [exec -ignorestderr scp {*}$args]
}

proc qc::ssh_call_proc {args} {
    #| Take a proc definition and args and write a Tcl script that can be run on the remote host
    # eg. ssh_call_proc user@remote link_checker /var/www/www.example.com
 
    if {![regexp {^([^@]+)@([^@]+)$} [lindex $args 0] -> username host] } {
	error "Usage ssh_call_proc user@host proc_name args"
    }
    qc::lshift args
    set proc_name [qc::lshift args]
    set script [qc::info_proc $proc_name]
    append script \n "puts \[$proc_name $args\]"
    set filename_local  [qc::file_temp $script]
    set filename_remote "/tmp/tmp-[qc::uuid]"
    ::try {
	scp $filename_local ${username}@${host}:$filename_remote
	set out [ssh ${username}@${host} /usr/bin/tclsh $filename_remote]
	ssh ${username}@${host} rm $filename_remote
    } on error {} {
	ssh ${username}@${host} rm -f $filename_remote
    }
    file delete $filename_local
    return $out
}

namespace eval qc {
    namespace export style_set css_parse css_rule2dict
}

proc qc::style_set {rule args} {
    # example 
    # > qc::style_set "color:pink;font-weight:bold;" color green
    # color:green;font-weight:bold
    set rule [string trim $rule ";"]
    set dict [split $rule ";:"]
    foreach {property value} $args {
	if {[in {width height top left right bottom} $property] && [qc::is integer $value]} {
	    # add units px for bare integers
	    append value px
	}
	dict set dict $property $value
    }
    set list {}
    foreach {property value} $dict {
	lappend list "$property:$value"
    }
    return [join $list ";"]
}
    
proc qc::css_parse {css} {
    # remove comments
    regsub -all {/\*((\*[^/])|[^*])*\*?\*/} $css {} css
    # Remove @ media sections 
    # may handle them at a later date 
    regsub -all {@media [^\{]+\{([^\{]+\{[^\}]+\}[ \t\r\n]*)+\}} $css {} css
    set data {}
    foreach rule [regexp -all -inline {[^\{]+\{[^\}]+\}} $css] {
	regexp {([^\{]+)\{([^\}]+)\}} $rule -> selector declaration
	regsub -all {[\t\r\n]} $declaration {} declaration
	regsub -all { *: *} $declaration {:} declaration
	regsub -all { *; *} $declaration {;} declaration
	set declaration [string trim $declaration "; "]
	set selector [string trim $selector]
	set dict [split $declaration ";:"]
	lappend data $selector $dict
    }
    return $data
}
    
proc qc::css_rule2dict {rule} {
    set rule [string trim $rule ";"]
    set dict [split $rule ";:"]
}

namespace eval qc::aws {
    namespace export s3
    namespace ensemble create
}

namespace eval qc::aws::s3 {
    namespace export xml2ldict
    namespace ensemble create

    proc _xml_select { xmlDoc xpath} {
        #| Returns xml nodes specified by the supplied xpath.
        set doc [dom parse $xmlDoc]
        set root [$doc documentElement]
        if { [$root hasAttribute xmlns] } {
            # Any namespace specified in the xmlns attribute is mapped 
            # to "ns" for use in the xpath query.
            $doc selectNodesNamespaces "ns [$root getAttribute xmlns]"
        }
        return [$root selectNodes $xpath] 
    }

    proc _xml_node2dict { node } {
        #| Converts an XML tdom node into a dict.
        set dict [list]
        set nodes [$node childNodes]
        foreach node $nodes {
            if { [llength [$node childNodes]] > 1 \
                 || ([llength [$node childNodes]] == 1 \
                     && [ne [[$node firstChild] nodeType] TEXT_NODE] ) } {
                lappend dict [$node nodeName] [_xml_node2dict $node]
            }  elseif { [llength [$node childNodes]] == 0 } {
                # empty node
                lappend dict [$node nodeName] {}
            } else {
                lappend dict [$node nodeName] [$node asText]
            }
        }
        return $dict
    }
        
    proc xml2ldict { xmlDoc xpath } {
        #| Returns ldict for xml nodes selected by xpath.
        set result [list]
        foreach node [_xml_select $xmlDoc $xpath] {
            lappend result [_xml_node2dict $node]
        }
       return $result
    }
}


namespace eval qc {
    namespace export html_table html_table_row html_table_row_head html_table_colgroup html_table_thead_from_cols html_table_tfoot_sums html_table_wants_* html_table_cols_from_table html_table_tbody_* html_table_format html_table_format_* html_tbody_row page_html_table columns_show_hide_toolbar
}

proc qc::html_table { args } {
    set argnames [qc::args2vars $args]
    # Special args are:- cols thead tbody tfoot table qry rowClasses scrollHeight sortable data tfoot_data
    # Some col keys have special meaning :- label class format thClass tfoot sum

    if { [info exists sortCols] } {
	default sortable yes
    } else {
	default sortable no
    }
    if { [true $sortable] && [form_var_exists sortCols] } {
	set sortCols [form_var_get sortCols]
    }

    if { [info exists classes] } {
        set class [join $classes " "]
        lappend argnames class
    }

    if { [info exists class] && "db-grid" in $class } {
        set headers [ns_conn outputheaders]
	ns_set update $headers Expires "Fri, 01 Jan 1990 00:00:00 GMT"
        ns_set update $headers Pragma no-cache
        ns_set update $headers Cache-Control "no-cache, no-store, max-age=0, must-revalidate"
    }
   
    
    # QRY
    if { [info exists qry] && ![info exists table] && ![info exists tbody] } {
	set table [qc::db_select_table [qc::db_qry_parse $qry 1]]	    
    }
    # cols from table
    if { [info exists table] } {
	default cols {}
	set cols [qc::html_table_cols_from_table table $cols]
    }
    # sortable Header
    if { [info exists cols] && [true $sortable] && ![info exists thead] } {
	default sortCols ""
	set thead [qc::html_table_sort_header $cols $sortCols]
    }
    # Highlight th sorted
    if { [info exists cols] && [info exists sortCols] } {
	set index [qc::ldict_search cols name [lindex $sortCols 0]]
	if { $index !=-1 } {ldict_set cols $index thClass sorted}
    }
    # thead from labels
    if { [info exists cols] && [qc::html_table_wants_col_labels $cols] && ![info exists thead] } {
	set thead [qc::html_table_thead_from_cols $cols]
    }

    # tbody from table
    if { [info exists table] && ![info exists tbody] } {
	set tbody [qc::html_table_tbody_from_table $table $cols]
    }
    # tbody from data
    if { [info exists data] && ![info exists tbody] } {
	set tbody [qc::html_table_tbody_from_ldict $data $cols]
    }
    # format tbody
    if { [info exists tbody] && [info exists cols] && [qc::html_table_wants_format $cols] } { 
	set tbody [qc::html_table_format $tbody $cols]
    }
    # tfoot sum
    if { [info exists cols] && [qc::html_table_wants_sum $cols] } {
	if { [info exists tbody] } {
	    set tfoot [qc::html_table_tfoot_sums $cols tbody]
	}
    }
    # tfoot data
    if { [info exists tfoot_data] && [info exists cols] } {
        set tfoot [list]
        foreach dict $tfoot_data {
            set row [list]
            foreach col $cols {
                set name [dict get $col name]
                if { [dict exists $dict $name] } {
                    lappend row [dict get $dict $name]
                } else {
                    lappend row ""
                }
            }
            lappend tfoot $row
        }
    }
    # format tfoot
    if { [info exists tfoot] && [info exists cols] && [qc::html_table_wants_format $cols] } { 
	set tfoot [qc::html_table_format $tfoot $cols]
    }

    # Scrollable - vertical
    if { [info exists scrollHeight] } {
	if { [info exists class] } {
	    lappend class scrollable
	} else {
	    set class scrollable
	    lappend argnames class
	}
	if { [lower $scrollHeight] eq "max" } {
	    lappend class "maximize-height"
	    unset scrollHeight
	    set argnames [lexclude $argnames scrollHeight]
	}
    }
    # Write table tag
    set html [qc::html_tag table {*}[dict_from {*}[lexclude $argnames {*}{
        table thead tbody tfoot
        cols data tfoot_data qry
        classes sortable rowClasses
        height
    }]]]

    append html \n
    # Create colgroup and col children
    if { [info exists cols]} {
        append html [qc::html_table_colgroup $cols]
    }
    # Create thead
    if { [info exists thead] } {
        append html "<thead>\n"
	set rowNumber 0
        foreach row $thead {
	    if { [info exists headerRowClasses] } {
		set rowClass [lindex $headerRowClasses [expr {$rowNumber%[llength $headerRowClasses]}]]
	    } else {
		set rowClass {}
	    }
	    if { [info exists cols]} {
		append html [qc::html_table_row_head $row $rowClass $cols]
	    } else {
		append html [qc::html_table_row_head $row $rowClass]
	    }
	    incr rowNumber
        }
        append html "</thead>\n"
    }
    # Create tbody
    if { [info exists tbody] } {
        append html "<tbody>\n"
	set rowNumber 0
        foreach row $tbody {
	    if { [info exists rowClasses] } {
		set rowClass [lindex $rowClasses [expr {$rowNumber%[llength $rowClasses]}]]
	    } else {
		set rowClass {}
	    }
	    append html [qc::html_table_row $row $rowClass]
	    incr rowNumber
        }
        append html "</tbody>\n"
    }
    # Create tfoot
    if { [info exists tfoot] } {
        append html "<tfoot>\n"
	foreach row $tfoot {
	    append html [qc::html_table_row $row]
        }
        append html "</tfoot>\n"
    }
    append html "</table>\n"
    return $html
}

proc qc::html_table_row { row {rowClass ""} } {
    #| Return HTML for a table row
    if { $rowClass == "" } {
	set html "<tr>\n"
    } else {
	set html "<tr class=\"$rowClass\">\n"
    }
    foreach cell $row {
	append html "<td>$cell</td>\n"
    }
    append html "</tr>\n"
    return $html 
}

proc qc::html_table_row_head { row rowClass {cols ""} } {
    #| Return HTML for th row
    if { $rowClass == "" } {
	set html "<tr>\n"
    } else {
	set html "<tr class=\"$rowClass\">\n"
    }
    # look for thClass in col config
    for {set i 0} {$i<[llength $row]} {incr i} {
	set cell [lindex $row $i]
	set col [lindex $cols $i]
	if {[dict exists $col thClass]} {
	    append html "<th class=\"[dict get $col thClass]\">$cell</th>\n"
	} else {
	    append html "<th>$cell</th>\n"
	}
    }
    append html "</tr>\n"
    return $html 
}

proc qc::html_table_colgroup { cols } {
    #| Create the html for the colgroup
    #| col elements wrapped in a colgroup
    set html "<colgroup>\n"
    foreach col $cols {
	if { [dict exists $col width] } {
	    set width [dict get $col width]
	    if [qc::is integer $width] {
		append width "px"
	    }
	    if { [dict exists $col style] } {
		dict set col style [qc::style_set [dict get $col style] width $width]
	    } else {
		dict set col style "width:${width};"
	    }
	}
	append html [html_tag col {*}[dict_exclude $col width label sum format tfoot thClass]]\n
    }
    append html "</colgroup>\n"
    return $html
}

proc qc::html_table_thead_from_cols {cols} {
    #| Helper. Write thead using cols object.
    set thead {}
    set row {}
    foreach col $cols {
        if { [dict exists $col label] } {
            lappend row [dict get $col label]
        } else {
            lappend row ""
        }
    }
    lappend thead $row
    return $thead
}

proc qc::html_table_tfoot_sums {cols tbodyVar} {
    #| Check if "sum yes" or "avg yes" is specified in the cols object and
    #| add column sum or average.
    # If "tfoot value" is given in col definition then use the value in the tfoot.
    upvar 1 $tbodyVar tbody
    set tfoot {}
    set row {}
    for {set i 0} {$i < [llength $cols]} { incr i } {
	set col [lindex $cols $i]
        if { [dict exists $col sum] && [string is true [dict get $col sum]] } {
	    #lappend row [qc::ll_sum ll $i]
	    lappend row [qc::html_table_tbody_sum tbody $i]
	} elseif { [dict exists $col avg] && [string is true [dict get $col avg]] } {
	    # Average of numeric values
	    lappend row [qc::html_table_tbody_avg tbody $i]
	} elseif {[dict exists $col tfoot]} {
	    lappend row [dict get $col tfoot]
        } else {
            lappend row ""
        }
    }
    lappend tfoot $row
    return $tfoot
}

proc qc::html_table_wants_sum {cols} {
    # look for "sum yes" or "tfoot value" in any col
    foreach col $cols {
	if { [dict exists $col sum] && [string is true [dict get $col sum]] } {
	    return 1
	}
        if { [dict exists $col tfoot] } {
            return 1
        }
    }
    return 0
}

proc qc::html_table_wants_col_labels {cols} {
    #| Look through the list of cols to see if any of them
    #| have used a "label text" pair to indicate a column heading
    foreach col $cols {
	if { [dict exists $col label] } {
	    return 1
	}
    }
    return 0
}

proc qc::html_table_wants_format {cols} {
    #| Look through the list of cols to see if any of them
    #| have format or class set
    foreach col $cols {
	if { [dict exists $col format] || [dict exists $col class] } {
	    return 1
	}
    }
    return 0
}

proc qc::html_table_cols_from_table { tableVar cols } {
    #| Create or update cols from table data structure
    #| use table to set label and name if not already set in col
    upvar 1 $tableVar table
    set index 0
    foreach colname [lindex $table 0] col $cols {
	if { ![dict exists $col label] } {
	    dict set col label $colname
	}
	if { ![dict exists $col name] } {
	    dict set col name $colname
	}
	lappend newcols $col
    }
    # use first row of data for class
    set index 0
    foreach value [lindex $table 1] col $newcols {
	if { ![dict exists $col class] } {
	    if { [qc::is integer $value] } {
		dict set col class number
	    } elseif { [qc::is decimal $value] } {
		dict set col class money
	    }
	}
	lset newcols $index $col
	incr index
    }
    return $newcols
}

proc qc::html_table_tbody_from_ldict {ldict cols} {
    #| Convert ldict into a tbody
    set tbody {}
    foreach dict $ldict {
	set row {}
	foreach col $cols {
	    set name [dict get $col name]
	    if { [dict exists $dict $name] } {
		lappend row [dict get $dict $name]
	    } else {
		lappend row {}
	    }
	}
	lappend tbody $row
    }
    return $tbody
}

proc qc::html_table_tbody_from_table {table cols} {
    #| Convert table data structure into tbody
    set tbody {}
    set indices {}
    set keys [lindex $table 0]
    set col_names [ldict_values cols name]
    if { [eq $keys $col_names] } {
	# Special case 
	return [lrange $table 1 end]
    }
    foreach col  $col_names { 
	lappend indices [lsearch $keys $col]
    }
    foreach list [lrange $table 1 end] {
	set row {}
	foreach index $indices {
	    lappend row [lindex $list $index]
	}
	lappend tbody $row
    }
    return $tbody
}

proc qc::html_table_tbody_sum { tbodyVar index } {
    #| Calculate the sum of cells in the column given by index
    set sum 0
    upvar 1 $tbodyVar tbody
    foreach list $tbody {
	set value [lindex $list $index]
	set value [qc::strip_html $value]
	regsub -all {[, ]} $value {} value
	if { [string is double -strict $value] } {
	    set sum [expr {$sum + $value}]
	}
    }
    return $sum
}

proc qc::html_table_tbody_avg { tbodyVar index } {
    #| Calculate the average value of cells in the column given by index
    set sum 0
    set count 0
    upvar 1 $tbodyVar tbody
    foreach list $tbody {
	set value [lindex $list $index]
	set value [qc::strip_html $value]
	regsub -all {[, ]} $value {} value
	if { [string is double -strict $value] } {
	    set sum [expr {$sum + $value}]
	    incr count
	}
    }
    if { $count>0 } {
	return [expr {double($sum)/$count}]
    } else {
	return ""
    }
}

proc qc::html_table_format {table cols} {
    #| Format columns based on the column class or column format
    foreach col $cols colIndex [.. 0 [llength $cols]] {
	dict2vars $col class format dp zeros sigfigs commify percentage
	if { [info exists class] } {
	    switch -glob -- $class {
		money* {default dp 2;default commify yes}
		integer* {default dp 0;default commify yes}
		number* {default commify yes}
		bool* {default format "qc::format_bool"}
		perct {default percentage yes;default commify yes}
	    }
	}

	if { [info exists dp] || [info exists sigfigs] || [info exists zeros] || [info exists commify] || [info exists percentage]} {
	    default dp ""
	    default sigfigs ""
	    default commify no
	    default zeros yes
	    default percentage no
	    foreach rowIndex [.. 0 [llength $table]] {
		set cell [lindex $table $rowIndex $colIndex]
		# Can't find a way to test if the cell exists to prevent an error when trying to lset
		# Best try is to skip empty strings
		if { [eq $cell ""] } {continue}
 		lset table $rowIndex $colIndex [qc::html_table_format_cell_if_number $cell $dp $sigfigs $zeros $commify $percentage]
	    }
	}
	if { [info exists format] } {
	    foreach rowIndex [.. 0 [llength $table]] {
		set cell [lindex $table $rowIndex $colIndex]
		if { [eq $cell ""] } {continue}
		lset table $rowIndex $colIndex [$format $cell]
	    }
	}
    }
    return $table
}

proc qc::html_table_format_cell_if_number {html dp sigfigs zeros commify percentage} {
    if { [regexp {(<[^>]+>)([^<]+)(<[^>]+>)} $html] } {
	regsub -all {[][$\\]} $html {\\&} html
	regsub -all {(<[^>]+>)([^<]+)(<[^>]+>)} $html "\\1\[qc::html_table_format_if_number {\\2} {$dp} {$sigfigs} {$zeros} {$commify} {$percentage}]\\3" html
	return [subst $html]
    } elseif { [qc::is decimal $html] } {
	return [qc::html_table_format_if_number $html $dp $sigfigs $zeros $commify $percentage]
    } else {
	return $html
    }
}

proc qc::html_table_format_if_number {value dp sigfigs zeros commify percentage} {
    #| If value is a number then commify
    if { [qc::is decimal $value] } {
	if { [true $percentage] } {
	    set value [expr {$value*100}]
	}
	if { [info exists sigfigs] && [qc::is integer $sigfigs]} {
	    set value [qc::sigfigs $value $sigfigs]
	}
	if { [info exists dp] && [qc::is integer $dp] } {
	    set value [qc::round $value $dp]
	}
	if { !$zeros && $value==0 } {
	    set value ""
	}
	if { [true $commify] } {
	    set value [format_commify $value]
	} 
	if { [true $percentage] && $value ne "" } {
	    set value $value%
	}
    }
    return $value
}

proc qc::html_tbody_row {cols} {
    #| Return a list to append to a tbody using variables corresponding to col names in the correct sequence.
    set list {}
    foreach col $cols {
	if { [dict exists $col name] } {
	    set name [dict get $col name]
	    lappend list [upset 1 $name]
	} else {
	    lappend list ""
	}
    }
    return $list
}

proc qc::page_html_table { args } {
    #| Return the HTML for a Heading and Menu of a paged report.
    #sql_sort -paging will have set the correct vars in caller's namespace
    args $args -limit ? -offset ? -count ? -- heading menu table
    default limit  [upset 1 limit]
    default offset [upset 1 offset]
    if { [uplevel 1 {info exists db_nrows}] } {
	default count [upset 1 db_nrows]
    }

    append heading " [expr {$offset +1}]...[expr {$offset+$count}]"

    if { $offset - $limit >=0 } {
	if { $menu ne "" } {
	    append menu " &nbsp;|&nbsp; "
	}
	append menu [qc::html_a_replace "Previous Page [expr {$offset-$limit+1}]...[expr {$offset}]" [url [qc::url_here] offset [expr {$offset-$limit}]]]
    }
    if { $count == $limit } {
	if { $menu ne "" } {
	    append menu " &nbsp;|&nbsp; "
	}
	append menu [qc::html_a_replace "Next Page [expr {$offset+$limit+1}]...[expr {$offset+2*$limit}]" [url [qc::url_here] offset [expr {$offset+$limit}]]]
    }

    set html "
<h3>$heading</h3>
<div>$menu</div>
$table
"
    return $html
}

proc qc::columns_show_hide_toolbar {args} {
    #| Construct a toolbar of column show/hide controls.
    #
    # Arguments: -title (optional) and conf (ldict describing show/hide controls).
    # show/hide control mandatory dict keys: label, name, col_selector
    # show/hide control optional dict keys: value, width, sticky, sticky_url, table_selector (reduce the scope to certain tables)
    # dividers: empty dicts create dividers that can be used to spearate groups of show/hide controls.
    #
    # Usage: set conf {}
    #        foreach year [list 2012 2013 2014] {
    #            lappend conf [list label $year name $year_control col_selector ".$year" value true width 120 table "#sales_by_year" sticky true sticky_url sales_by_year_sticky]
    #        }
    #        append html [columns_show_hide_toolbar -title "Show/Hide Years: " $conf]

    args $args -title "Show/Hide Columns: " -table_selector "table:not(:has(.columns-show-hide-toolbar),.columns-show-hide-toolbar)" -- conf

    # Column show/hide controls
    set show_hide_controls {}
    foreach dict $conf {
        if { [llength $dict] == 0 } {
            # Divider
            lappend show_hide_controls [html div "" class "columns-show-hide-control-divider"]
        } else {
            # Show/Hide control
            dict_default dict table_selector $table_selector width "auto" sticky true value true

            if { ! ([dict exists $dict label]
                    && [dict exists $dict name]
                    && [dict exists $dict col_selector])
             } {
                error "Missing config values in columns_show_hide_toolbar: [lexclude {label name col_selector} {*}[dict keys $dict]]"
            }

            # Label
            set show_hide_control [widget_label {*}[dict_subset $dict label name]]
            
            # Checkbox
            append show_hide_control [widget_bool {*}[dict_subset $dict name col_selector value table_selector sticky sticky_url]]

            lappend show_hide_controls [html div $show_hide_control class "columns-show-hide-control" style [qc::style_set "" width [dict get $dict width]]]
        }        
    }
    
    # Contruct Column Show/Hide toolbar
    set row {}
    lappend row [html div $title class "title"]
    lappend row [html div [join $show_hide_controls " "]]
    return [html_table tbody [list $row] class "columns-show-hide-toolbar"]
}

namespace eval qc {
    namespace export form_var_* form2* form_proc form form_authenticity_token
}

proc qc::form_var_names {} {
    #| Return a list of form variable names
    return [ns_set_keys [ns_getform]]
}

proc qc::form2vars {args}  {
    #| Create variables in the caller's namespace corresponding to the form data.
    if { [llength $args] == 0 } {
	# set all vars
	set args [ns_set_keys [ns_getform]]
    }
    # set vars
    foreach name $args {
	if { [form_var_exists $name] } {
	    upset 1 $name [form_var_get $name]
	}
    }
}

proc qc::form_var_get { var_name } {
    #| If the form variable exists return its value otherwise throw an error.
    #| A repeated form variable will return a list of corresponding values.
    #| PHP style repeated form variables foo[]=1 foo[]=2 treated as a list.
    set set_id [ns_getform]
    if { [string equal $set_id ""] } {
	error "No such form variable \"$var_name\""
    }
    if { [ns_set find $set_id $var_name] != -1 } {
	if { [ns_set unique $set_id $var_name] } {
	    return [string trim [ns_set get $set_id $var_name]]
	} else {
            set values [list]
	    foreach raw_value [qc::ns_set_getall $set_id $var_name] {
                lappend values [string trim $raw_value]
            }
            return $values
	}	
    }
    # Look for PHP style repeated form variables
    set array_name "${var_name}\[\]";
    if { [ns_set find $set_id $array_name] != -1 } {
        set values [list]
        foreach raw_value [qc::ns_set_getall $set_id $array_name] {
            lappend values [string trim $raw_value]
        }
        return $values
    }
    error "No such form variable \"$var_name\""
}

proc qc::form_var_exists { var_name } {
    #| Test whether a form variable exists or not.
    # Also check for PHP style repeated variables foo[]=1 foo[]=2 using name foo
    if { [info commands ns_conn] eq "ns_conn"
	 && [ns_conn isconnected]
	 && [ne [set set_id [ns_getform]] ""]
	 && ( [ns_set find $set_id $var_name] != -1 || [ns_set find $set_id "${var_name}\[\]"] != -1 )
     } {
	return 1
    } else {
	return 0
    }
}

proc qc::form2dict {args}  {
    #| Create dict corresponding to the form data.
    set dict {}
    if { [llength $args] == 0 } {
	# set all vars
	set args [ns_set_keys [ns_getform]]
    }
    # set vars
    foreach name $args {
        if { [form_var_exists $name] } {
	    dict set dict $name [form_var_get $name]
	}
    }
    return $dict
}

proc qc::form2url { url } {
    #| Encode the names and values of a form in an url

    # Currently qc::args2dict will throw an error where duplicate args are passed
    # so we protect against that with the "dict create"
    return [qc::url $url {*}[dict create {*}[qc::ns_set_to_multimap [ns_getform]]]]
}

proc qc::form_proc { proc_name } {
    #| Call proc_name using corresponding form variables
    set largs {}
    if { [eq [lindex [info args $proc_name] end] args] } {
	set args [lrange [info args $proc_name] 0 end-1]
    } else {
	set args [info args $proc_name]
    }
    foreach arg $args {
	if { [qc::form_var_exists $arg] } {
	    lappend largs [qc::form_var_get $arg]
	} else {
	    if { [info default $proc_name $arg default_value] } {
		lappend largs $default_value
	    } else {
		error "The form variable \"$arg\" was missing from the request" {} USER
	    }
	}
    }
    if { [eq [lindex [info args $proc_name] end] args] } {
	lappend largs [qc::form2dict {*}[lexclude [ns_set_keys [ns_getform]] {*}[info args $proc_name]]]
    }
    return [uplevel 0 $proc_name $largs]
}

proc qc::form {args} {
    #| Helper proc to create html form with authenticity tokens and method overload.
    set html [lindex $args end]
    set args [lrange $args 0 end-1]
    
    if { [llength $args]%2!=0 } {
        error "Uneven number of name/value pairs in \"$args\""
    }
    if { [dict exists $args method] } {
        # Make the method is uppercase
        dict set args method [string toupper [dict get $args method]]
    } else {
        # Default method
        dict set args method POST
    }
    set method [dict get $args method]
    if { $method ne "GET"} {
        # Add an authenticity token
        append html [qc::form_authenticity_token]
    }
    if { $method ni [list "POST" "GET"] } {
        # Overload the method.
        append html [h input type hidden value $method name _method]
        dict set args method POST
    }
    return [h form {*}$args $html]
}

proc qc::form_authenticity_token {} {
    #| Generate the hidden input for authenticity token
    set authenticity_token [qc::session_authenticity_token [qc::session_id]]
    return [h input type hidden name _authenticity_token value $authenticity_token]
}


package require sha1
package require md5
package require base64
package require tdom
package require fileutil

namespace eval qc {}

# Private S3 procs
proc qc::_s3_endpoint { args } {
    #| Return an s3 endpoint
    #| Usage:
    #| qc::_s3_endpoint bucket object_key
    #| qc::_s3_endpoint s3_uri
    #| qc::_s3_endpoint bucket

    if { [llength $args] == 1 } {
        if { [qc::is s3_uri [lindex $args 0]] } {
            # qc::_s3_endpoint s3_uri
            lassign [qc::s3 uri_bucket_object_key [lindex $args 0]] bucket object_key
            set object_key_exists true
        } else {
            # qc::_s3_endpoint bucket
            set bucket [lindex $args 0]
            set object_key_exists false
        }
    } elseif { [llength $args] == 2 } {
        # qc::_s3_endpoint bucket object_key
        lassign $args bucket object_key
        set object_key_exists true
    } else {
        error "Invalid number of arguments:\
               Usage: \"qc::_s3_endpoint bucket object_key\"\
               or \"qc::_s3_endpoint s3_uri\"."
    }

    # Use regional endpoint if default region is set - otherwise default to apex domain.
    if { [info exists ::env(AWS_DEFAULT_REGION)] } {
        set endpoint "s3.${::env(AWS_DEFAULT_REGION)}.amazonaws.com"
    } else {
        set endpoint "s3.amazonaws.com"
    }

    if { $bucket ne "" } {
        set endpoint [join [list $bucket $endpoint] .]
    }
    if { $object_key_exists } {
        append endpoint / $object_key
    }
    return $endpoint
}

proc qc::_s3_auth_headers { args } {
    #| Constructs the required s3 authentication header for the request type in question.
    #| See: http://docs.aws.amazon.com/AmazonS3/latest/dev/RESTAuthentication.html
    qc::args $args \
        -amz_headers "" \
        -content_type "" \
        -content_md5 "" \
        -- \
        verb \
        object_key \
        bucket

    # check for security token
    if { [info exists ::env(AWS_SESSION_TOKEN)] && $::env(AWS_SESSION_TOKEN) ne "" } {
        # We're using temporary security credentials - add token header
        lappend amz_headers x-amz-security-token $::env(AWS_SESSION_TOKEN)
    }

    set date [qc::format_timestamp_http now]

    if { $bucket ne "" } {
        # Is there a subresource specified?
        set subresources [list "acl" "lifecycle" "location" "logging" "notification" \
                              "partNumber" "policy" "requestPayment" "torrent" \
                              "uploadId" "uploads" "versionId" "versioning" "versions" \
                              "website" "restore"]
        if { [regexp {^[^\?]*\?([A-Za-z]+).*$} $object_key -> resource]
             && [qc::in $subresources $resource] } {
            set canonicalized_resource "/${bucket}/${object_key}"
        } else {
            # otherwise, drop the query part
            set canonicalized_resource "/${bucket}/[qc::url_path $object_key]"
        }
    } else {
        set canonicalized_resource "/"
    }

    # amz_headers format {header value header value ...}
    if { $amz_headers eq "" } {
        set canonicalized_amz_headers  ""
    } else {
        foreach {header value} $amz_headers {
            if { [info exists amz_header_array([qc::lower $header])] } {
                lappend amz_header_array([qc::lower $header]) $value
            } else {
                set amz_header_array([qc::lower $header]) $value
            }
        }
        set canonicalized_amz_headers  ""
        foreach key [lsort [array names amz_header_array]] {
            lappend canonicalized_amz_headers \
                "${key}:[join $amz_header_array($key) ,]\u000A"
        }
        set canonicalized_amz_headers [join $canonicalized_amz_headers ""]
    }

    # Contruct string for hmac signing
    set string_to_sign "$verb"
    lappend string_to_sign "$content_md5"
    lappend string_to_sign "$content_type"
    lappend string_to_sign "$date"
    lappend string_to_sign "${canonicalized_amz_headers}${canonicalized_resource}"
    set string_to_sign [join $string_to_sign \n]
    set signature [::base64::encode [::sha1::hmac \
                                         -bin ${::env(AWS_SECRET_ACCESS_KEY)} \
                                         $string_to_sign]]
    set authorization "AWS ${::env(AWS_ACCESS_KEY_ID)}:$signature"

    set return_headers [list \
                            Host [qc::_s3_endpoint $bucket] \
                            Date $date \
                            Authorization $authorization]

    if { [dict exists $amz_headers "x-amz-security-token"] } {
        lappend return_headers \
            "x-amz-security-token" [dict get $amz_headers "x-amz-security-token"]
    }

    return $return_headers
}

proc qc::_s3_get { bucket object_key } {
    #| Construct the http GET request to S3 including auth headers
    set headers [_s3_auth_headers GET $object_key $bucket]
    set result [qc::http_get \
                    -headers $headers \
                    "https://[qc::_s3_endpoint $bucket $object_key]"]
    return $result
}

proc qc::_s3_exists { bucket object_key } {
    #| Returns boolean true/false for 200/404 responses, anything else
    #| returns an error.
    set timeout 60
    set url "https://[qc::_s3_endpoint $bucket $object_key]"

    set httpheaders [list]
    foreach {name value} [qc::_s3_auth_headers HEAD $object_key $bucket] {
	lappend httpheaders [qc::http_header $name $value]
    }

    dict2vars [qc::http_curl \
                    -nobody 1 \
                    -httpheader $httpheaders \
                    -url $url \
                    -sslverifypeer 0 \
                    -sslverifyhost 0 \
                    -timeout $timeout \
                    -followlocation 1 \
               ] responsecode curlErrorNumber

    switch $curlErrorNumber {
	0 {
	    switch $responsecode {
		200 {
		    # OK
		    return true
		}
		404 {
                    return false
                }
		500 {return -code error -errorcode CURL "SERVER ERROR $url"}
		default {
                    return \
                        -code error \
                        -errorcode CURL \
                        "RESPONSE $responsecode while contacting $url"
                }
	    }
	}
	28 {
	    return \
                -code error \
                -errorcode TIMEOUT \
                "Timeout after $timeout seconds trying to contact $url"
	}
	default {
	    return -code error -errorcode CURL [curl::easystrerror $curlErrorNumber]
	}
    }
}

proc qc::_s3_head { bucket object_key } {
    #| Construct the http HEAD request to S3 including auth headers
    set headers [_s3_auth_headers HEAD $object_key $bucket]
    set result [qc::http_head \
                    -headers $headers \
                    "https://[qc::_s3_endpoint $bucket $object_key]"]
    return $result
}

proc qc::_s3_post { args } {
    #| Construct the http POST request to S3 including auth headers
    qc::args $args \
        -amz_headers "" \
        -content_type {application/xml} \
        -- \
        bucket \
        object_key \
        {data ""}

    if { $data ne "" } {
        # Used for posting XML
        set content_md5 [qc::_s3_base64_md5 -data $data]
        set headers [_s3_auth_headers \
                         -content_type $content_type \
                         -content_md5 $content_md5 \
                         POST $object_key $bucket]
        lappend headers Content-MD5 $content_md5
        lappend headers Content-Type $content_type
        set result [qc::http_post \
                        -valid_response_codes {100 200 202} \
                        -headers $headers \
                        -data $data \
                        "https://[_s3_endpoint $bucket $object_key]"]
    } else {
        if { $amz_headers ne "" } {
            set headers [_s3_auth_headers \
                             -amz_headers $amz_headers \
                             -content_type $content_type \
                             POST $object_key $bucket]
            lappend headers {*}$amz_headers
            lappend headers Content-Type $content_type
            set result [qc::http_post \
                            -headers $headers \
                            "https://[_s3_endpoint $bucket $object_key]"]
        } else {
            set headers [_s3_auth_headers \
                             -content_type $content_type \
                             POST $object_key $bucket]
            lappend headers Content-Type $content_type
            set result [qc::http_post \
                            -headers $headers \
                            "https://[_s3_endpoint $bucket $object_key]"]
        }
    }
    return $result
}

proc qc::_s3_delete { bucket object_key } {
    #| Construct the http DELETE request to S3 including auth headers
    set headers [_s3_auth_headers DELETE $object_key $bucket]
    set result [qc::http_delete \
                    -headers $headers \
                    "https://[_s3_endpoint $bucket $object_key]"]
    return $result
}

proc qc::_s3_save { args } {
    #| Construct the http SAVE request to S3 including auth headers
    qc::args $args -timeout 60 -- bucket object_key filename
    set tmp_file "/tmp/s3-[qc::uuid]"
    set headers [_s3_auth_headers GET $object_key $bucket]
    qc::http_save \
        -timeout $timeout \
        -headers $headers \
        -return_headers_var return_headers \
        "https://[_s3_endpoint $bucket $object_key]" \
        $tmp_file
    if { [dict exists $return_headers x-amz-meta-content-md5] } {
        set base64_md5 [dict get $return_headers x-amz-meta-content-md5]
        if { [qc::_s3_base64_md5 -file $tmp_file] ne $base64_md5 } {
            file delete -force $tmp_file
            error "qc::_s3_save: md5 of downloaded file does not match\
                   x-amz-meta-content-md5 ($base64_md5)."
        }
    } else {
        log Notice "qc::_s3_save: unable to verify downloaded file md5: $filename"
    }
    file copy $tmp_file $filename
    file delete -force $tmp_file
}

proc qc::_s3_put { args } {
    #| Construct the http PUT request to S3 including auth headers
    # _s3_put ?-header 0 ?-infile ? ?-s3_copy ?bucket object_key
    qc::args $args -nochecksum -header 0 -s3_copy ? -infile ? bucket object_key
    if { [info exists infile]} {
        set content_type [qc::mime_type_guess $infile]
        set content_md5 [qc::_s3_base64_md5 -file $infile]
        set data_size [file size $infile]
        if { [info exists nochecksum] } {
            # Dont send metadata for upload parts
            set headers [_s3_auth_headers \
                             -content_type $content_type \
                             -content_md5 $content_md5 \
                             PUT $object_key $bucket]
        } else {
            # content_md5 header allows AWS to return an error if the file received
            # has a different md5
            # Authentication value needs to use content_* values for hmac signing
            set headers [_s3_auth_headers \
                             -amz_headers [list "x-amz-meta-content-md5" $content_md5] \
                             -content_type $content_type \
                             -content_md5 $content_md5 \
                             PUT $object_key $bucket]
            lappend headers x-amz-meta-content-md5 $content_md5
        }
        lappend headers Content-Length $data_size
        lappend headers Content-MD5 $content_md5
        lappend headers Content-Type $content_type
        # Stop tclcurl from stending Transfer-Encoding header
        lappend headers Transfer-Encoding {}
        lappend headers Expect {}
        # Have timeout values roughly in proportion to the filesize
        # In this case allowing 10 KB/s
        set timeout [expr {$data_size/10240}]
        return [qc::http_put \
                    -header $header \
                    -headers $headers \
                    -timeout $timeout \
                    -infile $infile \
                    "https://[_s3_endpoint $bucket $object_key]"]
    } elseif { [info exists s3_copy] } {
        # s3_copy must be in the format "bucket/object_key"
        # we're copying a S3 file - skip the data processing and send the PUT
        # request with x-amz-copy-source header
        set headers [_s3_auth_headers \
                         -content_type {} \
                         -amz_headers [list "x-amz-copy-source" $s3_copy] \
                         PUT $object_key $bucket]
        lappend headers x-amz-copy-source $s3_copy
        lappend headers Content-Type {}
        return [qc::http_put \
                    -header $header \
                    -headers $headers \
                    -data {} \
                    "https://[_s3_endpoint $bucket $object_key]"]
    } else {
        error "qc::_s3_put: 1 of -infile or -s3_copy must be specified"
    }

}

proc qc::_s3_base64_md5 { args } {
    qc::args $args -file ? -data ? --
    #| Returns the base64 encoded binary md5 digest of a file or data
    if { [info exists file] && [info exists data] } {
        error "qc::_s3_base64_md5: specify only 1 of -file or -data"
    }
    if { [info exists data] } {
        # Just use ::md5 since we don't process chunks of data large enough to
        # cause problems
        return [::base64::encode [::md5::md5 $data]]
    } elseif {[info exists file]} {
        # Will not use ::md5 if Trf isn't installed due to incorrect results &
        # long runtimes for large files
        if { [qc::in [package names] "Trf"] } {
            return [::base64::encode [::md5::md5 -file $file]]
        } else {
            set openssl [exec which openssl]
            return [exec $openssl dgst -md5 -binary $file | $openssl enc -base64]
        }
    } else {
        error "qc::_s3_base64_md5: 1 of -file or -data must be specified"
    }
}

namespace eval qc {
    namespace export conn_*
}

proc qc::conn_remote_ip {} {
    #| Return the remote IP address of the current connection
    return [ns_conn peeraddr]
}

proc qc::conn_url {args} {
    #| Try to construct the full url of this request.
    # Return the encoded version of the path by default unless the -decoded flag is present
    args $args -decoded -- args
    if { [info exist decoded] } {
        return [qc::conn_location][qc::conn_path -decoded]
    } else {
        return [qc::conn_location][qc::conn_path]
    }
}

proc qc::conn_path {args} {
    #| Return the path of the current connection
    #| Return the encoded version of the path by default unless the -decoded flag is present
    # Note: using "ns_conn url" instead of "ns_conn request" as the latter is not updated for "ns_internalredirect"
    args $args -decoded -- args
    if { [info exist decoded] } {
        return [ns_conn url]
    } else {
        # re-ns_urlencode path since Naviserver decodes it by default
        set temp {}
        foreach url_part [split [ns_conn url] /] {
            lappend temp [ns_urlencode -part path $url_part]
        }
        return [join $temp /]
    }
}

proc qc::conn_location { args } {
    #| Try to construct the location
    qc::args $args -conn_protocol ? -conn_host ? -conn_port ? --

    if { ![info exists conn_protocol] } {
        set conn_protocol [qc::conn_protocol]
    }
    if { ![info exists conn_host]} {
        set conn_host [qc::conn_host]
    }
    if { ![info exists conn_port] } {
        set conn_port [qc::conn_port]
    }

    set url "${conn_protocol}://${conn_host}"
    # Legacy support where port doesn't appear in host header
    if { $conn_port ne "" && $conn_port != 80 && $conn_port != 443 } {
        # if Host header doesn't contain port, append it.
        if { [string first ":" $conn_host] == -1 } {
            set url "${conn_protocol}://${conn_host}:${conn_port}"
        }
    }

    if { [qc::is url $url] } {
        return $url
    } else {
        error "conn_location: cannot construct location string"
    }
}

proc qc::conn_protocol { } {
    #| Return protocol used by the client
    set conn_headers [ns_conn headers]
    
    # The connection protocol here may be different to the one initiated
    # by the client. Check X-Forward-Proto for initiating protocol.
    set client_protocol [ns_set iget $conn_headers "x-forwarded-proto"]
    if { $client_protocol in [list "https" "http"] } {
        return $client_protocol
    } else {
        return [ns_conn protocol]
    }
}

proc qc::conn_port { } {
    #| Try to detect connection port
    set conn_headers [ns_conn headers]
   
    # The connection port here may be different to the one initiated
    # by the client. Check X-Forward-Port for initiating port.
    set client_port [ns_set iget $conn_headers "x-forwarded-port"]
    if { $client_port ne "" && [qc::is int $client_port] } {
        return $client_port
    } elseif { [set port_header [ns_set iget $conn_headers "port"]] ne "" } {
        # Legacy support where Port is never included in host header and has its
        # own header value.
        return $port_header
    } else {
        # Otherwise take it from the Host header
        if { [regexp {[a-z0-9_][a-z0-9_\-]*(?:\.[a-z0-9_\-]+)+(?::([0-9]+))} [qc::conn_host] -> port] } {
            return $port
        } else {
            return [expr {[qc::conn_protocol] eq "http" ? "80" : "443"}]
        }
    }
}

proc qc::conn_host { } {
    #| Return the host indicated in the HTTP/1.1 headers
    set conn_headers [ns_conn headers]

    set host [ns_set iget $conn_headers Host]
    if { $host eq "" } {
        # Fallback to NaviServer supplied value which could fallback to driver default
        # TODO Preserves legacy
        if { [regexp {^https?://([a-z0-9_][a-z0-9_\-]*(?:\.[a-z0-9_\-]+)+(?::[0-9]+)?)} [ns_conn location] -> host] } {
            return $host
        } else {
            error "conn_host: Cannot determine host"
        }
    } else {
        return $host
    }
}

proc qc::conn_ie {} {
    if { ![ns_conn isconnected] } {
	return false
    }
    set header_set [ns_conn headers]
    set ui_string [ns_set iget $header_set "User-Agent"]
    if { [regexp {[\s;]MSIE\s([1-9][0-9]*)\.[0-9]+[bB]?;} $ui_string] } {
        # MSIE token only specified for IE < 11
    	return true
    } elseif { [regexp {Trident/[1-9][0-9]*\.[0-9]+} $ui_string] } {
        # Trident layout engine for IE >= 9
        return true
    } else {
        return false
    }
}

proc qc::conn_request_is_valid {request} {
    #| Test if the given request string is valid
    set pchar {
        [a-zA-Z0-9\-._~]|%[0-9a-fA-F]{2}|[!$&'()*+,;=:@]
    }
    set query_char [subst -nobackslashes {
        (${pchar}|/|\?)
    }]
    set path_char [subst {
        (${pchar}|/)
    }]
    set abs_uri [subst -nocommands -nobackslashes {
        https?://([a-z0-9\-\.]+)(:[0-9]+)?
        (/${path_char}*)?
        (\?${query_char}+)?
    }]
    set abs_path [subst -nobackslashes {
        /${path_char}*
        (\?${query_char}+)?
    }]
    set method {
        ([A-Z]+)
    }
    set http_version {
        HTTP/([0-9]+\.[0-9]+)
    }
    set re [subst -nocommands -nobackslashes {
        ^
        $method
        \s
        (\*|${abs_uri}|${abs_path})
        \s
        $http_version
        $
    }]
    return [regexp -expanded $re $request]
}

proc qc::conn_if_modified_since {} {
    #| Return the value of the current If-Modified-Since header, if one exists, or "" otherwise.
    # discard Netscape-style additional params
    set headers [ns_conn headers]
    if { [ns_set ifind $headers If-Modified-Since]!=-1 } {
        return [lindex [split [ns_set iget [ns_conn headers] If-Modified-Since] ";"] 0]
    } else {
        return ""
    }
}

proc qc::conn_open {} {
    #| Check if the client connection is open
    set NS_CONN_CLOSED 0x1
    return [expr {
                  [ns_conn isconnected] 
                  && ($NS_CONN_CLOSED & [ns_conn flags]) == 0 
              }]
}

proc qc::conn_response_headers_sent {} {
    #| Inspect ns_conn flags to determine whether response headers have been sent
    set NS_CONN_SENT_HEADERS 0x10
    return [expr {($NS_CONN_SENT_HEADERS & [ns_conn flags]) != 0}]
}

proc qc::conn_served {} {
    #| Return true if a response has been served to the client (connection closed or response headers returned)
    #| Otherwise return false
    return [expr {![qc::conn_open] || [qc::conn_response_headers_sent]}]
    
}

proc qc::conn_method {} {
    #| Return the method of the current connection.
    if {[qc::form_var_exists _method]} {
        set method [qc::form_var_get _method]
    } else {
        set method [ns_conn method]
    }
    return $method
}

namespace eval qc {
    namespace export chart_*
}

proc qc::chart_sales {x_labels values tips key_text} {
    set y_max [sigfigs_ceil [max {*}$values] 2]
    set x_labels [chart_list2csv $x_labels]
    set values [chart_list2csv $values]
    set tips [chart_list2csv $tips]
    
    set html {
	<script type="text/javascript" src="OpenFlashChart/swfobject.js"></script>
	<div id="my_chart" style="width:100%;margin-right:20px;margin-top:10px;margin-bottom:10px;"></div>
 	
	<script type="text/javascript">
	var so = new SWFObject("OpenFlashChart/open-flash-chart.swf", "ofc", "100%", "250", "9", "#FFFFFF");
	so.addVariable("variables","true");
	//so.addVariable("title","Sales,{font-size: 22px;}");
	so.addVariable("bg_colour","#FFFFFF");
	
	so.addVariable("x_labels","$x_labels");
	so.addVariable("x_label_style","11,#474438,0,3"); // font-size,color,orientation (0,1,2), x_label_step
	so.addVariable("x_axis_steps","1"); // draw x grid line every step
	so.addVariable("x_axis_colour","#837d69");
	so.addVariable("x_grid_colour","#c1bfb5");
	
	so.addVariable("y_label_style","11,#474438"); // y_labels font-size,color
	so.addVariable("y_ticks","5,10,4");// length of small tick,length of big tick,number of ticks
	so.addVariable("y_axis_colour","#837d69");
	so.addVariable("y_grid_colour","#c1bfb5");
	so.addVariable("y_min","0");
	so.addVariable("y_max","$y_max");
	so.addVariable("values","$values");
	
	//so.addVariable("area_hollow","4,4,15,#0066FF,Sales,12"); // line_width, dot_size, alpha_shaded, color, key_text,key_font_size
	so.addVariable("line_dot","4,#0066FF,$key_text,14,5"); // line_width,color,key_text,key_text_fontsize,dot_size
	
	so.addVariable("tool_tip","#tip#");
	so.addVariable("tool_tips_set","$tips");
	
	so.addParam("allowScriptAccess", "always" );
	so.write("my_chart");
	</script>
    }
    return [subst -nocommands -nobackslashes $html]
}

proc qc::chart_list2csv {list} {
    set csv {}
    foreach item $list {
	lappend csv [url_encode [regsub -all {,} $item "#comma#"]]
    }
    return [join $csv ,]
}

namespace eval qc {
    namespace export barcode128 barcode128partB barcode128partC barcode_charcode barcode_charcode_html
}

proc qc::barcode128 {string} {
    #| Make a code128 barcode
    # encode 4 or more consecutive numbers in code C
    # data is char value char value ..... pairs
    set data {}
    while { [string length $string] } {
	if {[regexp {^(([^0-9]+[0-9]{0,3}[^0-9]+)+)(.*)$} $string -> partB . string]} {
	    # e.g A123A
	    if { [llength $data] } {
		# already started so change-to-B char needed
		lappend data [barcode_charcode_html 200] 100
	    } else {
		# Start char for 128 B has value 104 
		lappend data [barcode_charcode_html 204] 104
	    }
	    set data [concat $data [qc::barcode128partB $partB]]
	} elseif {[regexp {^(([0-9]{0,3}[^0-9]+)+)(.*)$} $string -> partB . string]} {
	    # e.g. 123A
	    if { [llength $data] } {
		# already started so change-to-B char needed
		lappend data [barcode_charcode_html 200] 100
	    } else {
		# Start char for 128 B has value 104 
		lappend data [barcode_charcode_html 204] 104
	    }
	    set data [concat $data [barcode128partB $partB]]
	} elseif { [regexp {^([0-9]+)(.*)$} $string -> partC string] } {
	    if { [llength $data] } {
		# already started so change-to-C char needed
		lappend data [barcode_charcode_html 199] 99
	    } else {
		# Start char for 128 C has value 105
		lappend data [barcode_charcode_html 205] 105
	    }
	    set data [concat $data [barcode128partC $partC]]
	} else {
	    error "No match"
	}
    }
    # checksum
    set i 0
    set sum [lindex $data 1]
    set barcode ""
    foreach {char value} $data {
	set sum [expr {$sum + $i*$value}]
	incr i
	append barcode $char
    }
    set check_value [expr {$sum%103}]
    append barcode [barcode_charcode_html $check_value]
    # End char is unicode 206
    append barcode [barcode_charcode_html 206]
    return $barcode
}

proc qc::barcode128partB { string } {
    # Encode ASCII as 128B
    set lstring [split $string ""]
    set data {}
    foreach char $lstring {
	# map ASCII to barcode 128
	binary scan $char c charcode
	set value [expr {$charcode-32}]
	lappend data [barcode_charcode_html $value] $value
    }
    return $data
}

proc qc::barcode128partC {string} {
    #| Encode as 128C
    # If the number does not have an even number of digits
    # then prepend a zero
    if { [string length $string]%2 != 0 } {
	set string "0$string"
    }   
    foreach {a b} [split $string ""] {
	set value "$a$b"
	# Convert eg 08 to 8
	scan $value %d value
	lappend data [barcode_charcode_html $value] $value
    }
    return $data
}

proc qc::barcode_charcode { value } {
    #| Return the corresponding charcode. This is specific to the font set used.
    if { $value == 0 } {
	# space
	return 194
    }
    if { $value > 0 && $value < 95 } {
	return [expr {$value+32}]
    }
    if { $value > 94 && $value < 106} {
	return [expr {$value+100}]
    }
}

proc qc::barcode_charcode_html { value } {
    #| Translate to HTML entities 
    if { $value == 0 } {
	# space
	return "&#194;"
    }
    if { $value > 0 && $value < 95 } {
	#return [format %c [expr {$value+32}]]
	return "&#[expr {$value+32}];"
    }
    if { $value > 94 && $value < 106} {
	return "&#[expr {$value+100}];"
    }
}

namespace eval qc {
    namespace export authorise_token_create authorise_token
}

proc qc::authorise_token_create {args} {
    #| Create an authorisation token for this url
    # TODO change referrer to source
    qc::args2vars $args expires user_id referrer target 
    default expires "24 hours"
    default referrer [qc::conn_url]
    if { ! [info exist user_id] } {
        set user_id [auth]
    }
    if { ![regexp {^https?://} $referrer] } {
        set referrer [url_root "[qc::conn_location]/[string trimleft $referrer /]"]
    }
    if { ![regexp {^https?://} $target] } {
        set target [url_root "[qc::conn_location]/[string trimleft $target /]"]
    }
    set expiration_epoch [clock scan $expires]
    db_1row {
        select 
        sha1(concat(authorisation_key,:target,:referrer,:expiration_epoch)) as hash
        from users
        where user_id=:user_id
    }
    return "$user_id $expiration_epoch $hash"    
}

proc qc::authorise_token {} {
    #| Check the authorisation token of the current request
    #| and return the user_id if authorised
    if { ![form_var_exists authorisation_token] } {
        error "Authorisation token missing" {} AUTHORISATION
    }
    set token [form_var_get authorisation_token]

    lassign $token user_id expiration_epoch hash
    check user_id INT
    check expiration_epoch INT

    if { $expiration_epoch < [clock seconds] } {
        error "Authorisation failed. Token has expired." {} AUTHORISATION
    }

    # referrer
    set header_set_id [ns_conn headers]
    if { [ns_set ifind $header_set_id Referer] == -1 } {
        set referrer ""
    } else {
        set referrer [url_root [ns_set iget [ns_conn headers] Referer]]
    }

    set target [qc::conn_url]
    
    set qry {
        select 
        sha1(concat(authorisation_key,:target,:referrer,:expiration_epoch)) as check_hash
        from users
        where user_id=:user_id
    }
    db_0or1row $qry {
        error "Authorisation failed. No such user_id \"$user_id\"" {} AUTHORISATION
    } {
        if { $hash eq $check_hash } {
            return $user_id
        } else {
            error "Authorisation failed. Invalid hash value." {} AUTHORISATION
        }
    }
}

namespace eval qc {
    namespace export bs_capsule_*
}

proc qc::bs_capsule_text {args} {
    #| Returns bootstrap floating label form group for text input
    args_check_required $args name value label
    array set this $args
    set this(type) text
    default this(id) $this(name)

    set group [list]
    lappend group [h label for $this(id) $this(label)]
    lappend group [h input type $this(type) class "form-control" placeholder \
		       $this(label) value $this(value) name $this(name) \
		       id $this(id)]
    if {[info exist this(unit)]} {
	lappend group [h span \
			   class "help-block help-block-inline help-block-background" \
			   $this(unit)]
    }
    return [h div class "form-group" [join $group \n]]

}

proc qc::bs_capsule_password {args} {
    #| Returns bootstrap floating label form group for password input
    args_check_required $args name value label
    array set this $args
    set this(type) password
    default this(id) $this(name)

    set group [list]
    lappend group [h label for $this(id) $this(label)]
    lappend group [h input type $this(type) class "form-control" \
		       placeholder $this(label) value $this(value) \
		       name $this(name) id $this(id)]
    return [h div class "form-group" [join $group \n]]

}


proc qc::bs_capsule_select {args} {
    #| Returns bootstrap floating label form group for dropdown
    args_check_required $args name value options label
    array set this $args
    default this(id) $this(name)
    default this(null_option) no

    set group [list]
    lappend group [h label for $this(id) $this(label)]
    set dropdown [qc::widget_select name $this(name) id $this(id) \
		      value $this(value) options $this(options) \
		      class "form-control" null_option $this(null_option)]
    lappend group [h div class "select-wrapper" $dropdown]

    return [h div class "form-group form-group-select" [join $group \n]]
}

proc qc::bs_capsule_textarea {args} {
    #| Returns bootstrap floating label form group for textarea
    args_check_required $args name value label
    array set this $args
    default this(id) $this(name)
    default this(rows) 25

    set group [list]
    lappend group [h label for $this(id) $this(label)]
    lappend group [h textarea class "form-control" rows $this(rows) \
		       placeholder $this(label)  name $this(name) id $this(id) \
		       $this(value)]
    return [h div class "form-group form-group-textarea" [join $group \n]]
}

proc qc::bs_capsule_button {args} {
    #| Returns bootstrap flavoured button 
    array set this $args
    default this(type) submit
    default this(label) Submit
    
    return [h button type $this(type) class "btn btn-default" $this(label)]    
}

proc qc::bs_capsule_markdown {args} {
    #| Returns bootstrap floating label form group for markdown textarea
    args_check_required $args name value label
    array set this $args
    default this(id) $this(name)

    set group [list]
    lappend group [h label for $this(id) $this(label)]
    lappend group [h textarea class "form-control" placeholder $this(label) \
		       name $this(name) id $this(id) $this(value)]

    return [h div class "form-group form-group-markdown" [join $group \n]]
}

proc qc::bs_capsule_checkbox {args} {
    #| Returns bootstrap flavoured checkbox
    args_check_required $args name value label
    array set this $args
    default this(id) $this(name)
    set this(type) checkbox
    set this(checked) [true $this(value)]

    set checkbox [h input {*}[qc::dict_exclude [array get this] label]]
    return [h div class checkbox [h label ${checkbox}$this(label)]]
}
namespace eval qc {
    namespace export error_handler error_report error_report_*
}

proc qc::error_handler {{error_message "NULL"} args} {
    #| Return custom error responses depending value of error_code.
    # Pass in error message, info and code if available
    # Otherwise will take a copy of the global error* variables for backward compatiblity
    if { [llength $args] == 2 } {
        set error_info [lindex $args 0]
        set error_code [lindex $args 1]
    } elseif { [llength $args] == 1 } {
        set error_info [dict get [lindex $args 0] -errorinfo]
        set error_code [dict get [lindex $args 0] -errorcode]
    } else {
        return -code error "Usage: qc::error_handler ?error_message? ?(error_info error_code | options)?"
    }
    
    default error_info "NULL" error_code "NULL"
    foreach {local_var global_var} [list error_message errorMessage error_info errorInfo error_code errorCode ] {
        if { [set $local_var] eq "NULL" } {
            global $global_var
            set $local_var [set $global_var]
        }
    }

    set mime_types [list "text/html" "application/json" "application/xml" "text/xml"]
    set mime_type [qc::http_content_negotiate $mime_types]
    set media_type [lindex [split $mime_type "/"] 1]
    if { $media_type eq "" } {
        # Couldn't negotiate an acceptable response type.
        return [return2client code 406 text "Couldn't respond with an acceptable content type. Available content types: [join $mime_types ", "]"]
    }
    
    switch -glob -- $error_code {
	USER* {
	    switch $media_type {
                xml {
                    set body [qc::xml error $error_message]
                }
                json {
                    qc::response status invalid
                    qc::response message error $error_message
                    set body [response2json]
                }
                * -
                html {
                    set media_type html
                    set title "Missing or Invalid Data"
                    set head [h head [h title $title]]
                    set contents [h h2 $title]
                    append contents [h hr]
                    append contents $error_message
                    append contents [h p "Please back up and try again."]
                    append contents [h hr]
                    set body [h html "${head}[h body $contents]"]
                }
            }
            return [return2client code 200 $media_type $body filter_cc yes]
        }
        PERM* {
            switch $media_type {
                xml {
                    set body [qc::xml error "Not Authorised: $error_message"]
                }
                json {
                    qc::response status invalid
                    qc::response message error "Not Authorised: $error_message"
                    set body [response2json]
                }
                * -
                html {
                    set media_type html
                    set body [h p "Not Authorised: $error_message"]
                }
            }
            return [return2client code 401 $media_type $body]
        }
        AUTH* {
            switch $media_type {
                xml {
                    set body [qc::xml error "Authentication Failed: $error_message"]
                }
                json {
                    qc::response status invalid
                    qc::response message error "Authentication Failed: $error_message"
                    set body [response2json]
                }
                * -
                html {
                    set media_type html
                    set body [h p "Authentication Failed: $error_message"]
                }
            }
            return [return2client code 401 $media_type $body]
        }
        NOT_FOUND* {
            switch $media_type {
                xml {
                    set body [qc::xml error "Not Found: $error_message"]
                }
                json {
                    qc::response status invalid
                    qc::response message error "Not Found: $error_message"
                    set body [response2json]
                }
                * -
                html {
                    set media_type html
                    set body [h p "Not Found: $error_message"]
                }
            }
            return [return2client code 404 $media_type $body]
        }
        BAD_REQUEST* {
            switch $media_type {
                xml {
                    set body [qc::xml error "Bad Request: $error_message"]
                }
                json {
                    qc::response status invalid
                    qc::response message error "Bad Request: $error_message"
                    set body [response2json]
                }
                * -
                html {
                    set media_type html
                    set body [h p "Bad Request: $error_message"]
                }
            }
            return [return2client code 400 $media_type $body]
        }
	default {
	    log Error $error_info
            if {  [info exists ::env(ENVIRONMENT)] && $::env(ENVIRONMENT) ne "LIVE" } {
                switch $media_type {
                    xml {
                        set body [qc::xml error "Software Bug - [string range $error_message 0 75]"]
                    }
                    json {
                        qc::response status invalid
                        qc::response message error "Software Bug - [string range $error_message 0 75]"
                        set body [response2json]
                    }
                    * -
                    html {
                        set media_type html
                        set body [qc::error_report $error_message $error_info $error_code]
                    }
                }
            } else {
	        # LIVE
                switch $media_type {
                    xml {
                        set body [qc::xml error "Internal Server Error. An email report has been sent to our engineers"]
                    }
                    json {
                        qc::response status invalid
                        qc::response message error "Internal Server Error. An email report has been sent to our engineers"
                        set body [response2json]
                    }
                    * -
                    html {
                        set media_type html
                        set body [h h2 "Internal Server Error"]
                        append body [h p "An email report has been sent to our engineers."]
                    }
                }
            }

            return2client code 500 $media_type $body filter_cc yes
            
	    if { [qc::param_exists email_support] } {
		set subject "[string toupper [ns_info server]] Bug - [string range $error_message 0 75]"
		qc::email_support subject $subject html [qc::error_report $error_message $error_info $error_code] 
	    }
	}
    }
}

proc qc::error_report {{error_message "NULL"} {error_info "NULL"} {error_code "NULL"}} {
    #| Return html error report. If there was a http connection when error occurred report any 
    #| relevant information about http request.
    # Pass in error message, info and code if available
    # Otherwise will take a copy of the global error* variables for backward compatiblity
    foreach {local_var global_var} [list error_message errorMessage error_info errorInfo error_code errorCode ] {
        if { [set $local_var] eq "NULL" } {
            global $global_var
            set $local_var [set $global_var]
        }
    }
    if { [ns_conn isconnected] } {
	sset html {
	    <html>
	    <h2>Software Bug</h2>
	    An error has occurred while processing your request.
	    <p>
	    <b>hostname:</b>[ns_info hostname]<br>
	    <b>url:</b>[html_escape [qc::conn_path]]<br>
	    <b>request:</b>[html_escape [ns_conn request]]<br>
	    <b>remoteip:</b>[qc::conn_remote_ip]<br>
	    <b>time:</b>[qc::format_timestamp now]<br>
	    <b>errorMessage:</b> [html_escape $error_message]<br>
	    <b>errorInfo:</b> <pre>[html_escape $error_info]</pre><br>
	    <b>errorCode:</b> $error_code
	    <p>
	    <h3>Form Variables:</h3>
	    [qc::error_report_form_vars]
	    <h3>Cookies</h3>
	    [qc::error_report_cookies]
            <h3>HTTP Headers</h3>
            [qc::error_report_headers]
	    </html>
	}
    } else {
	sset html {
	    <html>
	    <h2>Software Bug</h2>
	    <p>
	    <b>hostname:</b>[ns_info hostname]<br>
	    <b>time:</b>[qc::format_timestamp now]<br>
	    <b>errorMessage:</b> [html_escape $error_message] <br>
	    <b>errorInfo:</b> <pre>[html_escape $error_info]</pre><br>
	    <b>errorCode:</b> $error_code
	    <p>
	    </html>
	}
    }
    return $html
}

proc qc::error_report_no_conn { message info code } {
    #| Return html error report, used when there was no http connection when error occurred.
    
    set html {
        <html>
        <h2>Software Bug</h2>
        <p>
        <b>hostname:</b>[ns_info hostname]<br>
        <b>time:</b>[qc::format_timestamp now]<br>
        <b>errorMessage:</b> $message <br>
        <b>errorInfo:</b> <pre>[html_escape $info]</pre><br>
        <b>errorCode:</b> $code
        <p>
        </html>
    }
    return [subst $html]
}

proc qc::error_report_form_vars {} {
    #| Return preformated html indicating values of all form variables when error occurred.
  
    set set_id [ns_getform]
    if { [string equal $set_id ""] } {
	set size 0
    } else {
	set size [ns_set size $set_id]
    } 
    set report {}
    set i 0
    while {$i<$size} {
	set name [ns_set key $set_id $i]
	# mask anything that looks like a card number.
	set value [ns_set value $set_id $i]
	append report "<b>$name</b>\n"
	# Truncate value if too long
	if { [string bytelength $value] > 1024 } {
	    append report "<pre>[string range $value 0 1023]....</pre>"
	} else {
	    append report <pre>$value</pre>
	}
	append report \n
        incr i
    }
    return $report
}

proc qc::error_report_locals {} {
    #| Return preformated html indicating values of all local variables when error occurred.
   
    set report {}
    foreach name [uplevel 1 {info locals}] {
	# mask anything that looks like a card number.
	set value [upset 1 $name]
	append report "<b>$name</b>\n"
	# Truncate value if too long
	if { [string bytelength $value] > 1024 } {
	    append report "<pre>[string range $value 0 1023]....</pre>"
	} else {
	    append report <pre>$value</pre>
	}
	append report \n
    }
    return $report
}

proc qc::error_report_cookies {} {
    set headers [ns_conn headers]
    set cookies [ns_set iget $headers Cookie]
    set report {}
    foreach pair [split $cookies ;] {
	lassign [split $pair =] name value
	set name [qc::url_decode $name]
	set value [string trimright $value "; "]
	set value [qc::url_decode $value]
	append report "<b>$name</b> $value <br>"
    }
    return $report
}

proc qc::error_report_headers {} {
    set headers [ns_conn headers]
    set report {}
    foreach {name value} [ns_set array $headers] {
	append report "[h b $name]: [h span $value]<br>"
    }
    return $report
}

namespace eval qc {
    namespace export handler_restful handler_files handler_db_files
}

proc qc::handler_restful {} {
    #| Handler for registered restful URL handlers.
    set url_path [qc::conn_path]
    set method [qc::conn_method]
    
    if {[qc::handlers exists $method $url_path]} {
        set result [qc::handlers call $method $url_path]
        # Check if conn is still waiting to be served
        if { ![qc::conn_served] } {
            # If a non-GET method then return the global data structure otherwise return the result as text/html.
            if {$method ne "GET"} {
                return [qc::return_response]
            } else {
                return [ns_return 200 text/html $result]
            }
        }
    }
}

proc qc::handler_files {} {
    #| Handler for file requests that haven't been resolved by fastpath.
    set url_path [qc::conn_path]
    if { [file exists [ns_pagepath]$url_path] } {
        # static file
        ns_register_fastpath GET $url_path
        ns_register_fastpath HEAD $url_path
        ns_returnfile 200 [ns_guesstype [ns_pagepath]$url_path] [ns_pagepath]$url_path
    }
}

proc qc::handler_db_files {} {
    #| Handle requests for files stored in the database that have yet to be cached on disk
    set url_path [qc::conn_path]
    if { [regexp {^/image/} $url_path] } {
        # images
        qc::image_handler [ns_pagepath]/image
    } elseif { [regexp {^/file/} $url_path] } {
        qc::file_handler [ns_pagepath]/file
    }
}
namespace eval qc {
    namespace export tson* json_quote
}


proc qc::tson_string {value} {
    #| Returns a TSON string.
    return [list string $value]
}

proc qc::tson_number {value} {
    #| Returns a TSON number.
    if { [string tolower $value] in [list "" "null"] } {
        return "null"
    }
    return [list number $value]
}

proc qc::tson_boolean {value} {
    #| Returns a TSON boolean.
    if { [string tolower $value] in [list "" "null"] } {
        return "null"
    }
    return [list boolean [qc::cast boolean $value true false]]
}

proc qc::tson_array {args} {
    #| Return a tson array from list of values.
    if { [llength $args] == 0 } {
        return "null"
    }
    
    set tson [list array]
    
    foreach value $args {
        if { [string is list $value]
             && [lindex $value 0] in [list object array string number boolean]
         } {
            lappend tson $value
        } elseif { ([qc::is decimal $value] && [qc::upper $value] ni [list NAN INF]) } {
            lappend tson [list number $value]
        } elseif { $value in [list true false] } {
            lappend tson [list boolean $value]
        } elseif { $value eq "null" } {
            lappend tson $value
        } else {
            lappend tson [list string $value]
        }
    }

    return $tson
}

proc qc::tson_object { args } {
    #| Return a tson object from list of name value pairs in args
    # Cannot be used to create nested objects
    # EXAMPLE:  
    # % tson_object firstname "Daniel" surname "Clark" age 23
    # object firstname {string Daniel} surname {string Clark} age {number 23}
    if { [llength $args] == 0 } {
        return "null"
    }
    
    set tson [list object]
    
    foreach {name value} $args {
        if { [string is list $value]
             && [lindex $value 0] in [list object array string number boolean] } {
            lappend tson $name $value
        } else {
            if { ([qc::is decimal $value] && [qc::upper $value] ni [list NAN INF]) } {
                lappend tson $name [list number $value]
            } elseif { $value in [list true false] } {
                lappend tson $name [list boolean $value]
            } elseif { $value eq "null" } {
                lappend tson $name $value
            } else {
                lappend tson $name [list string $value]
            }
        }
    }

    return $tson
}

proc qc::json_quote {value} {
    #| Return a json string literal with appropriate escapes
    return "\"[string map {\" \\\" \\ \\\\ \n \\n \r \\r \f \\f \b \\b \t \\t} $value]\""
}

proc qc::tson2json { tson } {
    #| Convert tson to json
    switch -- [lindex $tson 0] {
	object {
	    set list {}
	    foreach {name value} [lrange $tson 1 end] {
		lappend list "[json_quote $name]: [tson2json $value]"
	    }
	    return "\{\n[join $list ",\n"]\n\}"
	}
	array {
	    set list {}
	    foreach value [lrange $tson 1 end] {
		lappend list [tson2json $value]
	    }
	    return "\[[join $list ,]\]"
	}
	string {
	    return [json_quote [lindex $tson 1]]
	}
        number {
            return [lindex $tson 1]
        }
        boolean {
            return [lindex $tson 1]
        }
	default {
	    if { ([string is double -strict $tson] && [qc::upper $tson] ni [list NAN INF]) || [in {true false null} $tson]} {
		return $tson
	    } else {
		return [json_quote $tson]
	    }
	}
    }
}

proc qc::tson_object_from { args } {
    #| Take a list of var names and return a tson object
    set dict {}
    foreach name $args {
	upvar 1 $name value
	if { [info exists value] } {
	    lappend dict $name $value
	} else {
	    error "Can't create dict with $name: No such variable"
	}
    }
    return [tson_object {*}$dict]
}

proc qc::tson2xml { tson } {
    # not dealt with attributes 
    # prefered technique
    # http://www.ibm.com/developerworks/xml/library/x-xml2jsonphp/
    
    switch -- [lindex $tson 0] {
	object {
	    set list {}
	    foreach {name value} [lrange $tson 1 end] {
		lappend list <$name>[qc::tson2xml $value]</$name>
	    }
	    return [join $list "\n"]
	}
	array {
	    set list {}
	    foreach value [lrange $tson 1 end] {
		lappend list <item>[qc::tson2xml $value]</item>
	    }
	    return [join $list ""]
	}
	string {
	    return [qc::xml_escape [lindex $tson 1]]
	}
        number {
            return [lindex $tson 1]
        }
        boolean {
            return [lindex $tson 1]
        }
	default {
	    if { ([string is double -strict $tson] && [qc::upper $tson] ni [list NAN INF]) || [in {true false null} $tson]} {
		return $tson
	    } else {
		return [qc::xml_escape $tson]
	    }
	}
    }
}

proc qc::tson_get {tson args} {
    #| Returns the value at the specified path in the TSON.
    #| Requires PostgreSQL 9.3 or later.

    # Construct a PostgreSQL array literal from the path.
    set path [qc::sql_list2array -type text $args]
    # Convert TSON to JSON.
    set json [qc::tson2json $tson]
    # Use PostgreSQL JSON operators to get the value at the path.
    qc::db_cache_1row -ttl 86400 {
        select :json::json#>>$path as value
    }

    return $value
}

proc qc::tson_exists {tson args} {
    #| Determines if a value exists at the specified path in the TSON.
    #| Requires PostgreSQL 9.3 or later.
    if { [llength $args] == 0 } {
        # No path specified.
        error "Usage: qc::tson_exists tson key ?key ...?"
    }

    # Check if the outermost key in the path exists.
    set json [qc::tson2json $tson]
    set key [lindex $args 0]
    
    switch [lindex $tson 0] {
        "object" {
            qc::db_cache_1row -ttl 86400 {
                select :key in (select json_object_keys(:json::json)) as key_exists
            }
        }
        "array" {
            qc::db_cache_1row -ttl 86400 {
                select json_array_length(:json::json) as array_length
            }
            
            if { $array_length == 0 || $key > $array_length - 1 } {
                # Array has no elements or key is out of bounds.
                return f
            }

            set key_exists t
        }
        default {
            error "TSON must be an object or an array."
        }
    }
    
    if { !$key_exists || [llength $args] == 1 } {
        # Key doesn't exist or no more keys in the path to check.
        return $key_exists
    }

    return [qc::tson_exists \
                [qc::json2tson [qc::tson_get $tson $key]] \
                {*}[lrange $args 1 end]]
}

proc qc::tson_type {tson args} {
    #| Returns the type of the value at the specified path in the TSON.
    #| Requires PostgreSQL 9.3 or later.
    set value_tson $tson
    
    if { [llength $args] > 0 } {
        # A path was specified so get the value at the path.

        if { ![qc::tson_exists $tson {*}$args] } {
            # Path doesn't existin the TSON.
            error "Path \"$args\" not found in TSON."
        }
        
        # Construct a PostgreSQL array literal from the path.
        set path [qc::sql_list2array -type text $args]
        
        # Convert TSON to JSON.
        set json [qc::tson2json $tson]
        
        # Use PostgreSQL JSON operators to get the value at the path.
        qc::db_cache_1row -ttl 86400 {
            select :json::json#>$path as value
        }
        
        # Convert the JSON to TSON.
        set value_tson [qc::json2tson $value]
    }

    if { $value_tson in [list "true" "false"] } {
        return "boolean"
    } elseif { [qc::is decimal $value_tson] } {
        return "number"
    } elseif { $value_tson eq "null" } {
        return "null"
    } elseif { [lindex $value_tson 0] in [list "object" "array" "string"] } {
        return [lindex $value_tson 0]
    } else {
        error "Invalid TSON."
    }
}

proc qc::tson_array_foreach {varName tson code} {
    #| Iterator for a tson array.
    foreach element [lrange $tson 1 end] {
        upset 1 $varName $element
        uplevel 1 $code
    }
}

proc qc::tson_object_get {tson args} {
    #| Get named property of a TSON object node,
    #| optionally nested
    set value [dict get [lrange $tson 1 end] [lindex $args 0]]
    if { [llength $args] > 1 } {
        return [tson_object_get \
                    $value \
                    {*}[lrange $args 1 end]]
    }
    return $value
}

proc qc::tson_object_exists {tson args} {
    #| Check if named property of a TSON object node exists
    if { ! [dict exists [lrange $tson 1 end] [lindex $args 0]] } {
        return false
    }
    if { [llength $args] > 1 } {
        return [tson_object_exists \
                    [tson_object_get $tson [lindex $args 0]] \
                    {*}[lrange $args 1 end]]
    }
    return true
}

proc qc::tson_value {tson} {
    #| Get tson value
    if { [llength $tson] != 2 } {
        return $tson
    }
    switch [lindex $tson 0] {
        "string" -
        "number" -
        "boolean" {
            return [lindex $tson 1]
        }
        default {
            return $tson
        }
    }
}

proc qc::tson_object_get_value {tson args} {
    #| Get named property of a TSON object node as a flat value
    return [tson_value [tson_object_get $tson {*}$args]]
}

namespace eval qc {
    namespace export barcode_ean13 barcode_ean13_check_digit
}

proc qc::barcode_ean13 {number_to_encode} {
    #| *Untested* Create an EAN barcode
    if { $number_to_encode == "" || $number_to_encode == 0  } {
	return 1;
    }
    set extra_addon ""
    set encoded_text_addon ""
    set encoded_text ""

    if {[string length $number_to_encode] < 12 || [string length $number_to_encode] == 16 } {
	set number_to_encode "0000000000000"
    }
    
    if {[string length $number_to_encode] == 12 || [string length $number_to_encode] == 13 } {
	set number_to_encode [string range $number_to_encode 0 11]
    }

    if {[string length $number_to_encode] == 14 } {
	set extra_addon [string range $number_to_encode 12 13]
	set number_to_encode [string range $number_to_encode 0 11]
    }
    if {[string length $number_to_encode] == 15 } {
	set extra_addon [string range $number_to_encode 13 14]
	set number_to_encode [string range $number_to_encode 0 11]
    }
    if {[string length $number_to_encode] == 17 } {
	set extra_addon [string range $number_to_encode 12 16]
	set number_to_encode [string range $number_to_encode 0 11] 
    }
    if {[string length $number_to_encode] >= 18 } {
	set extra_addon [string range $number_to_encode 13 17]
	set number_to_encode [string range $number_to_encode 0 11] 
    }
    
    # calculate the check digit
    set check_digit [barcode_ean13_check_digit $number_to_encode]
    
    # calculate leading digit;  
    # convert first char(leading_digit) to ASCII
    scan [string index $number_to_encode 0]  %c leading_digit 
    set leading_digit [expr $leading_digit - 48]
    
    # set the encoding type
    if {$leading_digit == 0} {
	set encoding AAAAAACCCCCC
    } elseif {$leading_digit == 1} {
	set encoding AABABBCCCCCC
    } elseif {$leading_digit == 2} {
	set encoding AABBABCCCCCC
    } elseif {$leading_digit == 3} {
	set encoding AABBBACCCCCC
    } elseif {$leading_digit == 4} {
	set encoding ABAABBCCCCCC
    } elseif {$leading_digit == 5} {
	set encoding ABBAABCCCCCC
    } elseif {$leading_digit == 6} {
	set encoding ABBBAACCCCCC
    } elseif {$leading_digit == 7} {
	set encoding ABABABCCCCCC
    } elseif {$leading_digit == 8} {
	set encoding ABABBACCCCCC
    } elseif {$leading_digit == 9} {
	set encoding AAAAAACCCCCC
    }
    
    # add the check digit to the end of the barcode & remove the leading digit
    set actual_data_to_encode "[string range $number_to_encode 1 11]$check_digit"
    # now we have the total number including the check digit
    set i 1
    foreach digit [split $actual_data_to_encode ""] {
	set current_char [scan $digit %c]
	set current_encoding [string index $encoding [expr $i-1]]
        		
	if {[string match $current_encoding A]} {
	    set current_char [format %c $current_char]
	} elseif {[string match $current_encoding B]} {
	    set current_char [format %c [expr $current_char + 17]]
	} else {
	    # In case of type C
	     set current_char [format %c [expr $current_char + 27]]
	}
	set encoded_text "$encoded_text$current_char"

	if {$i == 1} {
	    # first character
	    set temp $leading_digit
	    if {$leading_digit > 4} {
		set temp_ascii "[expr [scan $temp %c] + 64]"
		set encoded_text "[format %c $temp_ascii]\($current_char"
	    } else {
		set temp_ascii "[expr [scan $temp %c] + 37]"
		set encoded_text "[format %c $temp_ascii]\($current_char"
	    }
	} elseif {$i == 6} {
	    # six character
	    set encoded_text "$encoded_text*"
	} elseif {$i == 12} {
	    # last character
	    set encoded_text "$encoded_text\("
	}
	incr i;
    }

     # process 2 digit extra addon if it exits;
    if {[string length $extra_addon] == 2} {
	set i 0;
	# get encoding type for extra add on
	while {$i <= 99} {
	    if {$extra_addon == $i} {
		set encoding AA
	    }
	    if {$extra_addon == [expr $i +1]} {
		set encoding AB
	    }
	    if {$extra_addon == [expr $i + 2]} {
		set encoding BA
	    }
	    if {$extra_addon == [expr $i + 3]} {
		set encoding BB
	    }
	    set i [expr $i + 4]
	}
	set i 1
	# loop runs upto 2(length of extra addon)
	while {$i <= 2} {
       	    if {$i == 1} {
		set encoded_text_addon " [format %c 43]"
	    }
	    set current_char [expr [scan [string index $extra_addon [expr $i -1] ] %c] - 48]
	    set current_encoding [string index $encoding [expr $i - 1]]
	    
	    if {[string match $current_encoding A]} {
		if {$current_char == 0} {
		    set encoded_text_addon "$encoded_text_addon[format %c 34]"
		} elseif {$current_char == 1} {
		    set encoded_text_addon "$encoded_text_addon[format %c 35]"
		} elseif {$current_char == 2} {
		    set encoded_text_addon "$encoded_text_addon[format %c 36]"
		} elseif {$current_char == 3} {
		    set encoded_text_addon "$encoded_text_addon[format %c 37]"
		} elseif {$current_char == 4} {
		    set encoded_text_addon "$encoded_text_addon[format %c 38]"
		} elseif {$current_char == 5} {
		    set encoded_text_addon "$encoded_text_addon[format %c 44]"
		} elseif {$current_char == 6} {
		    set encoded_text_addon "$encoded_text_addon[format %c 46]"
		} elseif {$current_char == 7} {
		    set encoded_text_addon "$encoded_text_addon[format %c 47]"
		} elseif {$current_char == 8} {
		    set encoded_text_addon "$encoded_text_addon[format %c 58]"
		} elseif {$current_char == 9} {
		    set encoded_text_addon "$encoded_text_addon[format %c 59]"
		}
	    }
	    if {[string match $current_encoding B]} {
		if {$current_char == 0} {
		    set encoded_text_addon "$encoded_text_addon[format %c 122]"
		} elseif {$current_char == 1} {
		    set encoded_text_addon "$encoded_text_addon[format %c 61]"
		} elseif {$current_char == 2} {
		    set encoded_text_addon "$encoded_text_addon[format %c 63]"
		} elseif {$current_char == 3} {
		    set encoded_text_addon "$encoded_text_addon[format %c 64]"
		} elseif {$current_char == 4} {
		    set encoded_text_addon "$encoded_text_addon[format %c 91]"
		} elseif {$current_char == 5} {
		    set encoded_text_addon "$encoded_text_addon[format %c 92]"
		} elseif {$current_char == 6} {
		    set encoded_text_addon "$encoded_text_addon[format %c 93]"
		} elseif {$current_char == 7} {
		    set encoded_text_addon "$encoded_text_addon[format %c 95]"
		} elseif {$current_char == 8} {
		    set encoded_text_addon "$encoded_text_addon[format %c 123]"
		} elseif {$current_char == 9} {
		    set encoded_text_addon "$encoded_text_addon[format %c 125]"
		}
	    }
	    # add in the space & add-on guard pattern
	    if {$i == 1} {
		set encoded_text_addon "$encoded_text_addon[format %c 33]"
	    }
	    incr i;
	}
    }
    

    # process 5 digit extra addon if it exits;
    if {[string length $extra_addon] == 5} {
	# calculate check digit for extra addon 
	set factor 3
	set weighted_total 0
	set i 5
	while {$i >= 1} {
	    set current_char [expr [scan [string index $extra_addon [expr $i -1] ] %c] - 48] 
	    if {$factor == 3} {
		set weighted_total [expr $weighted_total + [expr $current_char * 3]]
	    } elseif {$factor == 1} {
		set weighted_total [expr $weighted_total + [expr $current_char * 9]]
	    }
	    set factor [expr 4 -$factor]
	    set i [expr $i -1]
	}
	set check_digit [expr [scan [string index $weighted_total [expr [string length $weighted_total]-1]]  %c] - 48 ]
	        
	# set the encoding type for extra addon
	if {$check_digit == 0} {
	    set encoding BBAAA
	} elseif {$check_digit == 1} {
	    set encoding BABAA
	} elseif {$check_digit == 2} {
	    set encoding BAABA
	} elseif {$check_digit == 3} {
	    set encoding BAAAB
	} elseif {$check_digit == 4} {
	    set encoding ABBAA
	} elseif {$check_digit == 5} {
	    set encoding AABBA
	} elseif {$check_digit == 6} {
	    set encoding AAABB
	} elseif {$check_digit == 7} {
	    set encoding ABABA
	} elseif {$check_digit == 8} {
	    set encoding ABAAB
	} elseif {$check_digit == 9} {
	    set encoding AABAB
	}
	set i 1;
	# loop runs upto 5(extra add on length)
	while {$i <= 5} {
	    set current_char [expr [scan [string index $extra_addon [expr $i -1] ] %c] - 48]
	    set current_encoding [string index $encoding [expr $i - 1]]
	    if {$i == 1} {
		set encoded_text_addon "[format %c 32][format %c 43]"
	    }
	    
	    if {[string match $current_encoding A]} {
		if {$current_char == 0} {
		    set encoded_text_addon "$encoded_text_addon[format %c 34]"
		} elseif {$current_char == 1} {
		    set encoded_text_addon "$encoded_text_addon[format %c 35]"
		} elseif {$current_char == 2} {
		    set encoded_text_addon "$encoded_text_addon[format %c 36]"
		} elseif {$current_char == 3} {
		    set encoded_text_addon "$encoded_text_addon[format %c 37]"
		} elseif {$current_char == 4} {
		    set encoded_text_addon "$encoded_text_addon[format %c 38]"
		} elseif {$current_char == 5} {
		    set encoded_text_addon "$encoded_text_addon[format %c 44]"
		} elseif {$current_char == 6} {
		    set encoded_text_addon "$encoded_text_addon[format %c 46]"
		} elseif {$current_char == 7} {
		    set encoded_text_addon "$encoded_text_addon[format %c 47]"
		} elseif {$current_char == 8} {
		    set encoded_text_addon "$encoded_text_addon[format %c 58]"
		} elseif {$current_char == 9} {
		    set encoded_text_addon "$encoded_text_addon[format %c 59]"
		}
	    }
	    if {[string match $current_encoding B]} {
		if {$current_char == 0} {
		    set encoded_text_addon "$encoded_text_addon[format %c 122]"
		} elseif {$current_char == 1} {
		    set encoded_text_addon "$encoded_text_addon[format %c 61]"
		} elseif {$current_char == 2} {
		    set encoded_text_addon "$encoded_text_addon[format %c 63]"
		} elseif {$current_char == 3} {
		    set encoded_text_addon "$encoded_text_addon[format %c 64]"
		} elseif {$current_char == 4} {
		    set encoded_text_addon "$encoded_text_addon[format %c 91]"
		} elseif {$current_char == 5} {
		    set encoded_text_addon "$encoded_text_addon[format %c 92]"
		} elseif {$current_char == 6} {
		    set encoded_text_addon "$encoded_text_addon[format %c 93]"
		} elseif {$current_char == 7} {
		    set encoded_text_addon "$encoded_text_addon[format %c 95]"
		} elseif {$current_char == 8} {
		    set encoded_text_addon "$encoded_text_addon[format %c 123]"
		} elseif {$current_char == 9} {
		    set encoded_text_addon "$encoded_text_addon[format %c 125]"
		}
	    }
	     # add in the space & add-on guard pattern
	    if {$i < 5} {
		set encoded_text_addon "$encoded_text_addon[format %c 33]"
	    }
	    incr i
	}
    }
    return "$encoded_text$encoded_text_addon"
  
    
}

proc qc::barcode_ean13_check_digit { number } {
    # helper proc
     set odd 1
     set sum 0
     foreach digit [split $number ""] {
         set odd [expr {!$odd}];
         #puts "$sum += ($odd*2+1)*$digit :: [expr {($odd*2+1)*$digit}]"
         incr sum [expr {($odd*2+1)*$digit}]
     }
     set check [expr {$sum % 10}]
     if { $check > 0 } {
         return [expr {10 - $check}]
     }
  return $check
}

namespace eval qc {
    namespace export mime_type_guess
}

set qc::mime_mappings {
    ".adp" "text/html"
    ".ai" "application/postscript"
    ".aif" "audio/aiff"
    ".aifc" "audio/aiff"
    ".aiff" "audio/aiff"
    ".ani" "application/x-navi-animation"
    ".art" "image/x-art"
    ".asc" "text/plain"
    ".au" "audio/basic"
    ".avi" "video/x-msvideo"
    ".bcpio" "application/x-bcpio"
    ".bin" "application/octet-stream"
    ".bmp" "image/bmp"
    ".cdf" "application/x-netcdf"
    ".cgm" "image/cgm"
    ".class" "application/octet-stream"
    ".cpio" "application/x-cpio"
    ".cpt" "application/mac-compactpro"
    ".css" "text/css"
    ".csv" "application/csv"
    ".dci" "text/html"
    ".dcr" "application/x-director"
    ".der" "application/x-x509-ca-cert"
    ".dir" "application/x-director"
    ".dll" "application/octet-stream"
    ".dms" "application/octet-stream"
    ".doc" "application/msword"
    ".docx" "application/vnd.openxmlformats-officedocument.wordprocessingml.document"
    ".dp" "application/commonground"
    ".dvi" "applications/x-dvi"
    ".dwf" "drawing/dwf"
    ".dwg" "image/vnd.dwg"
    ".dxf" "image/vnd.dxf"
    ".dxr" "application/x-director"
    ".elm" "text/plain"
    ".eml" "message/rfc822"
    ".etx" "text/x-setext"
    ".exe" "application/octet-stream"
    ".ez" "application/andrew-inset"
    ".fm" "application/vnd.framemaker"
    ".gbt" "text/plain"
    ".gif" "image/gif"
    ".gtar" "application/x-gtar"
    ".gz" "application/x-gzip"
    ".hdf" "application/x-hdf"
    ".htm" "text/html"
    ".html" "text/html"
    ".hpgl" "application/vnd.hp-hpgl"
    ".hqx" "application/mac-binhex40"
    ".ice" "x-conference/x-cooltalk"
    ".ief" "image/ief"
    ".ies" "application/octet-stream"
    ".igs" "image/iges"
    ".iges" "image/iges"
    ".jpg" "image/jpeg"
    ".jpeg" "image/jpeg"
    ".jfif" "image/jpeg"
    ".jpe" "image/jpeg"
    ".js" "application/x-javascript"
    ".kar" "audio/midi"
    ".latex" "application/x-latex"
    ".ldt" "application/octet-stream"
    ".lha" "application/octet-stream"
    ".ls" "application/x-javascript"
    ".lxc" "application/vnd.ms-excel"
    ".lzh" "application/octet-stream"
    ".man" "application/x-troff-man"
    ".map" "application/x-navimap"
    ".me" "application/x-troff-me"
    ".mesh" "model/mesh"
    ".mid" "audio/x-midi"
    ".midi" "audio/x-midi"
    ".mif" "application/vnd.mif"
    ".mocha" "application/x-javascript"
    ".mov" "video/quicktime"
    ".movie" "video/x-sgi-movie"
    ".mp2" "audio/mpeg"
    ".mp3" "audio/mpeg"
    ".mpe" "video/mpeg"
    ".mpeg" "video/mpeg"
    ".mpg" "video/mpeg"
    ".mpga" "audio/mpeg"
    ".ms" "application/x-troff-ms"
    ".msh" "model/mesh"
    ".nc" "application/x-netcdf"
    ".nvd" "application/x-navidoc"
    ".nvm" "application/x-navimap"
    ".oda" "application/oda"
    ".pbm" "image/x-portable-bitmap"
    ".pcl" "application/vnd.hp-pcl"
    ".pclx" "application/vnd.hp-pclx"
    ".pdb" "chemical/x-pdb"
    ".pdf" "application/pdf"
    ".pgm" "image/x-portable-graymap"
    ".pgn" "application/x-chess-pgn"
    ".pic" "image/pict"
    ".pict" "image/pict"
    ".pnm" "image/x-portable-anymap"
    ".png" "image/png"
    ".pot" "application/vnd.ms-powerpoint"
    ".ppm" "image/x-portable-pixmap"
    ".pps" "application/vnd.ms-powerpoint"
    ".ppt" "application/vnd.ms-powerpoint"
    ".ps" "application/postscript"
    ".qt" "video/quicktime"
    ".ra" "audio/x-realaudio"
    ".ram" "audio/x-pn-realaudio"
    ".rar" "application/x-rar-compressed"
    ".ras" "image/x-cmu-raster"
    ".rgb" "image/x-rgb"
    ".rm" "audio/x-pn-realaudio"
    ".roff" "application/x-troff"
    ".rpm" "audio/x-pn-realaudio-plugin"
    ".rtf" "application/rtf"
    ".rtx" "text/richtext"
    ".sda" "application/vnd.stardivision.draw"
    ".sdc" "application/vnd.stardivision.calc"
    ".sdd" "application/vnd.stardivision.impress"
    ".sdp" "application/vnd.stardivision.impress"
    ".sdw" "application/vnd.stardivision.writer"
    ".sgl" "application/vnd.stardivision.writer-global"
    ".sgm" "text/sgm"
    ".sgml" "text/sgml"
    ".sh" "application/x-sh"
    ".shar" "application/x-shar"
    ".sht" "text/html"
    ".shtml" "text/html"
    ".silo" "model/mesh"
    ".sit" "application/x-stuffit"
    ".skd" "application/vnd.stardivision.math"
    ".skm" "application/vnd.stardivision.math"
    ".skp" "application/vnd.stardivision.math"
    ".skt" "application/vnd.stardivision.math"
    ".smf" "application/vnd.stardivision.math"
    ".smi" "application/smil"
    ".smil" "application/smil"
    ".snd" "audio/basic"
    ".spl" "application/x-futuresplash"
    ".sql" "application/x-sql"
    ".src" "application/x-wais-source"
    ".stc" "application/vnd.sun.xml.calc.template"
    ".std" "application/vnd.sun.xml.draw.template"
    ".sti" "application/vnd.sun.xml.impress.template"
    ".stl" "application/x-navistyle"
    ".stw" "application/vnd.sun.xml.writer.template"
    ".svg" "image/svg+xml"
    ".swf" "application/x-shockwave-flash"
    ".sxc" "application/vnd.sun.xml.calc"
    ".sxd" "application/vnd.sun.xml.draw"
    ".sxg" "application/vnd.sun.xml.writer.global"
    ".sxl" "application/vnd.sun.xml.impress"
    ".sxm" "application/vnd.sun.xml.math"
    ".sxw" "application/vnd.sun.xml.writer"
    ".t" "application/x-troff"
    ".tar" "application/x-tar"
    ".tcl" "x-tcl"
    ".tex" "application/x-tex"
    ".texi" "application/x-texinfo"
    ".texinfo" "application/x-texinfo"
    ".text" "text/plain"
    ".tgz" "application/x-gtar"
    ".tif" "image/tiff"
    ".tiff" "image/tiff"
    ".tr" "application/x-troff"
    ".tsv" "text/tab-separated-values"
    ".txt" "text/plain"
    ".ustar" "application/x-ustar"
    ".vcd" "application/x-cdlink"
    ".vor" "application/vnd.stardivision.writer"
    ".vrml" "model/vrml"
    ".wav" "audio/x-wav"
    ".wbmp" "image/vnd.wap.wbmp"
    ".webp" "image/webp"
    ".wkb" "application/vnd.ms-excel"
    ".wks" "application/vnd.ms-excel"
    ".wml" "text/vnd.wap.wml"
    ".wmlc" "application/vnd.wap.wmlc"
    ".wmls" "text/vnd.wap.wmlscript"
    ".wmlsc" "application/vnd.wap.wmlscript"
    ".wrl" "model/vrml"
    ".xbm" "image/x-xbitmap"
    ".xls" "application/vnd.ms-excel"
    ".xlsx" "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"
    ".xlw" "application/vnd.ms-excel"
    ".xpm" "image/x-xpixmap"
    ".xht" "application/xhtml+xml"
    ".xhtml" "application/xhtml+xml"
    ".xml" "text/xml"
    ".xsl" "text/xml"
    ".xyz" "chemical/x-pdb"
    ".xwd" "image/x-xwindowdump"
    ".z" "application/x-compress"
    ".zip" "application/zip"
}

proc qc::mime_type_guess { filename } {
    #| Lookup a mimetype based on a file extension. Case insensitive.
    # Based on ns_guesstype.
    # Defaults to "*/*".    
    set default_type "*/*"
    set ext [file extension [string tolower $filename]]
    if { [dict exists $qc::mime_mappings $ext] } {
        return [dict get $qc::mime_mappings $ext]
    }
    return $default_type
}

proc qc::mime_has_unique_file_extension { mime_type } {
    #| Check if mime_type maps to a unique file extension
    set mime_types [dict values $qc::mime_mappings]
    set first_index [lsearch -exact $mime_types $mime_type]
    if { $first_index == -1 } {
        # No mapping for mime_type to file extension
        return false
    }
    set second_index [lsearch \
                          -exact \
                          -start ${first_index}+1 \
                          $mime_types \
                          $mime_type]
    if { $second_index == -1 } {
        # Multiple (non-unique) mappings for mime_type to
        # file extension
        return true
    }
    return false
}

proc qc::mime_file_extension { mime_type } {
    #| Look up a file extension based on a mime_type
    return [dict get [lreverse $qc::mime_mappings] $mime_type]
}

namespace eval qc {
    namespace export excel_*
}

proc qc::excel_file_create {args} {
    #| Creates an xls or xlsx file using the information provided
    # type: "xls" or "xlsx", defaults to "xlsx"
    # data: a list of lists containing a grid of cell values
    # formats: a dict of class definitions
    # class definitions: a dict of name-value pairs
    # column_meta: nested dict, eg. {1 {class "foo" width 20} 5 {width 50}}
    # row_meta: nested dict, eg. {0 {class "bar" height 20} 3 {class "foo"}}
    # cell_meta: nested dict, eg. {
    #   {5 2} {class "baz" type "string|number|formula|url"}
    #   {1 1} {class "bar"}
    # }
    # images: ldict, eg. {
    #   {filename /tmp/foo.png row 1 col 5}
    #   {filename /tmp/bar.png row 2 col 3 x 5 y 10 x_scale 1.2 y_scale 1.3}
    # }
    qc::args2vars $args {*}{
        data
        formats
        column_meta
        row_meta
        cell_meta
        timeout
        type
        images
    }
    default {*}{
        data {}
        formats {}
        column_meta {}
        row_meta {}
        cell_meta {}
        timeout 1000
        type "xlsx"
        images {}
    }
    switch $type {
        "xls" {
            set writer "Spreadsheet::WriteExcel"
        }
        "xlsx" {
            set writer "Excel::Writer::XLSX"
        }
        default {
            error "Unsupported excel type"
        }
    }

    set filename [qc::file_temp ""]
    ########################################
    # Template
    ########################################
    set template {#!/usr/bin/perl -w
        use strict;
        use <writer>;

        ####################
        # Data
        ####################

        # Insert filename here
        my $workbook = <writer>->new('<filename>');
        my $worksheet = $workbook->add_worksheet();

        # Cell data, eg.
        # my $data = [
        #    ["name", "DOB", "Telephone"],
        #    ["Bob", "12/12/12", "1234 5678"],
        #    ["Jane", "01/01/01", "0123 9876"]
        # ];
        my $data = <cell_data>;

        # Workbook formats, eg.
        # my $formats = {
        #    highlight {
        #        bold => 1,
        #        color => 'green'
        #    },
        #    error => {
        #        color => 'red'
        #    }
        # };
        my $formats = <formats>;

        # Column meta-data, eg.
        # my $column_meta = {
        #    0 => {
        #        class => "red",
        #        width => 30
        #    }
        # };
        my $column_meta = <column_meta>;

        # Row meta-data, eg.
        # my %row_meta = {
        #    0 => {
        #        class => "bold",
        #        height => 25
        #    }
        # };
        my $row_meta = <row_meta>;

        # Cell meta-data, eg.
        # my %cell_meta = {
        #    2 => {
        #        1 => {
        #            class => "red",
        #            type => "string"
        #        },
        #        2 => {
        #            class => "bold"
        #        }
        #    }
        # };
        my $cell_meta = <cell_meta>;

        # Images, eg.
        # my $images = [
        #   {
        #     filename => /tmp/foo.png,
        #     row => 1,
        #     col => 5,
        #   },
        #   {
        #     filename => /tmp/bar.png,
        #     row => 2,
        #     col => 3,
        #     x => 5,
        #     y => 10,
        #     x_scale => 1.2,
        #     y_scale => 1.3
        #   }
        # ];
        my $images = <images>;

        ####################
        # Logic
        ####################
        # Formats
        my $format_objects = {};
        foreach my $class ( keys %{$formats} ) {
            $format_objects->{$class} = $workbook->add_format(%{$formats->{$class}});
        }

        # Column meta
        foreach my $i ( keys %{$column_meta} ) {
            my $format;
            if ( exists $column_meta->{$i}{"class"} ) {
                $format = $format_objects->{$column_meta->{$i}{"class"}};
            }
            my $width = $column_meta->{$i}{"width"};
            $worksheet->set_column($i, $i, $width, $format);
        }

        # Row meta
        foreach my $i ( keys %{$row_meta} ) {
            my $format;
            if ( exists $row_meta->{$i}{"class"} ) {
                $format = $format_objects->{$row_meta->{$i}{"class"}};
            }
            my $height = $row_meta->{$i}{"height"};
            $worksheet->set_row($i, $height, $format);
        }

        # Cells
        $worksheet->keep_leading_zeros();
        for my $i ( 0 .. $#{$data} ) {
            for my $j ( 0 .. $#{ $data->[$i] } ) {
                my $format;
                my $type;
                if ( exists $cell_meta->{$i}{$j}{"class"} ) {
                    $format = $format_objects->{$cell_meta->{$i}{$j}{"class"}};
                }
                if ( exists $column_meta->{$j}{"type"} ) {
                    $type = $column_meta->{$j}{"type"};
                }
                if ( exists $row_meta->{$i}{"type"} ) {
                    $type = $row_meta->{$i}{"type"};
                }
                if ( exists $cell_meta->{$i}{$j}{"type"} ) {
                    $type = $cell_meta->{$i}{$j}{"type"};
                }

                if ( ! defined $type ) {
                    # Guess type based on value
                    $worksheet->write($i, $j, $data->[$i][$j], $format);

                } elsif ( $type eq "string" ) {
                    $worksheet->write_string($i, $j, $data->[$i][$j], $format);

                } elsif ( $type eq "number" ) {
                    $worksheet->write_number($i, $j, $data->[$i][$j], $format);

                } elsif ( $type eq "formula" ) {
                    $worksheet->write_formula($i, $j, $data->[$i][$j], $format);

                } elsif ( $type eq "url" ) {
                    my $url = $cell_meta->{$i}{$j}{"url"};
                    $worksheet->write_url($i, $j, $url, $data->[$i][$j], $format);
                }
            }
        }

        # Images
        for my $i ( 0 .. $#{$images} ) {
            if ( not ( exists $images->[$i]{"x"} && exists $images->[$i]{"y"} ) ) {
                $worksheet->insert_image(
                                         $images->[$i]{"row"},
                                         $images->[$i]{"col"},
                                         $images->[$i]{"filename"});
            } elsif ( not ( exists $images->[$i]{"x_scale"}
                            && exists $images->[$i]{"y_scale"} )
                      ) {
                $worksheet->insert_image(
                                         $images->[$i]{"row"},
                                         $images->[$i]{"col"},
                                         $images->[$i]{"filename"},
                                         $images->[$i]{"x"},
                                         $images->[$i]{"y"});                
            } else {
                $worksheet->insert_image(
                                         $images->[$i]{"row"},
                                         $images->[$i]{"col"},
                                         $images->[$i]{"filename"},
                                         $images->[$i]{"x"},
                                         $images->[$i]{"y"},
                                         $images->[$i]{"x_scale"},
                                         $images->[$i]{"y_scale"});
            }
        }
    }
    # End of template

    ########################################
    # Perl Data
    ########################################
    # Cell data - A list of list becomes an array of arrays
    set cell_data [llist2perl_aarray $data]
    
    # Formats - a dict of dicts becomes a hash of format objects added to the current workbook
    set formats [qc::excel_formats $formats]

    # Colum meta-data - a dict of dicts becomes a hash of hashes
    set column_meta [qc::ddict2perl_hhash $column_meta]

    # Row meta-data - a dict of dicts becomes a hash of hashes
    set row_meta [qc::ddict2perl_hhash $row_meta]

    # Cell meta-data - a dict of dicts with lists for keys becomes a hash of hashes of hashes
    set cell_meta [qc::cell_meta2perl $cell_meta]

    # Images - An ldict becomes an array of hashes
    set images [qc::ldict2perl_ahash $images]

    ########################################
    # Variable substitution
    ########################################
    set token_vars {
        writer
        filename
        cell_data
        formats
        column_meta
        row_meta
        cell_meta
        images
    }
    set map {}
    foreach token $token_vars {
        lappend map <$token> [set $token]
    }
    set script [string map $map $template]

    ########################################
    # Script execution
    ########################################
    set script_filename [qc::file_temp $script]
    ::try {
        log Debug $script_filename
        exec_proxy -timeout $timeout perl $script_filename
        file delete $script_filename
    } on error {error_message options} {
        file delete $script_filename
        error $error_message [dict get $options -errorinfo] [dict get $options -errorcode]
    }
    return $filename
}

proc qc::excel_formats {class_defs} {
    #| Converts a nested dict of class definitions to a perl hash of excel formats
    set format_dict {}
    dict for {class format} $class_defs {
        dict set format_dict $class [qc::excel_format $format]
    }
    return [ddict2perl_hhash $format_dict]
}

proc qc::excel_format {format} {
    #| Prepare a format defn for the Perl Lib
    set format_list {}
    dict for {attribute value} $format {
        switch $attribute {
            "font-family" {
                dict set attribute_dict font $value
            }
            "font-size" {
                dict set attribute_dict size $value
            }
            "font-weight" {
                if {$value eq "bold"} {
                    dict set attribute_dict bold 1
                } else {
                    error "$css_attribute $css_value is unsupported"
                }
            }
            "font-style" {
                if {$value eq "italic"} {
                    dict set attribute_dict italic 1
                } else {
                    error "$css_attribute $css_value is unsupported"
                }
            }
            "text-decoration" {
                switch $value {
                    "underline" {
                        dict set attribute_dict underline 1
                    }
                    "line-through" {
                        dict set attribute_dict font_strikeout 1
                    }
                    default {
                        error "$css_attribute $css_value is unsupported"
                    }
                }
            }
            "text-align" {
                dict set attribute_dict align $value
            }
            "vertical-align" {
                if {$value eq "middle"} {
                    dict set attribute_dict valign vcenter
                } else {
                    dict set attribute_dict valign $value
                }
            }
            "background" -
            "background-color" {
                dict set attribute_dict bg_color $value
            }
            "border" -
            "border-top" -
            "border-right" -
            "border-bottom" -
            "border-left" {
                set side [string range $attribute 7 end]
                if {$side eq ""} {
                    set side "border"
                }
                set values [split $value " "]
                set color [lindex $values 2]
                switch [lindex $values 1] {
                    "solid" {
                        switch [lindex $values 0] {
                            "1px" {
                                set index 1
                            }
                            "2px" {
                                set index 2
                            }
                            "3px" {
                                set index 5
                            }
                            default {
                                error "$css_attribute $css_value is unsupported"
                            }
                        }
                    }
                    "dashed" {
                        switch [lindex $values 0] {
                            "1px" {
                                set index 3
                            }
                            "2px" {
                                set index 8
                            }
                            default {
                                error "$css_attribute $css_value is unsupported"
                            }
                        }
                    }
                    "dotted" {
                        switch [lindex $values 0] {
                            "1px" {
                                set index 7
                            }
                            default {
                                error "$css_attribute $css_value is unsupported"
                            }
                        }
                    }
                    "double" {
                        switch [lindex $values 0] {
                            "1px" {
                                set index 6
                            }
                            default {
                                error "$css_attribute $css_value is unsupported"
                            }
                        }
                    }
                    default {
                        error "$css_attribute $css_value is unsupported"
                    }
                }
                dict set attribute_dict $side $index
                dict set attribute_dict ${side}_color $color
            }
            default {
                dict set attribute_dict $attribute $value
            }
        }
    }
    return $attribute_dict
}

proc qc::cell_meta2perl {cell_meta} {
    #| Converts excel cell meta data from a nested dict to a perl hash of hashes
    set cell_meta_nested {}
    dict for {indices cell} $cell_meta {
        set row [lindex $indices 0]
        set column [lindex $indices 1]
        dict set cell_meta_nested $row $column $cell
    }
    set row_hash_list {}
    dict for {row row_data} $cell_meta_nested {
        lappend row_hash_list "$row => [ddict2perl_hhash $row_data]"
    }
    return \{[join $row_hash_list ", "]\}
}

namespace eval qc {
    namespace export schedule schedule_id schedule_exists schedule_running schedule_stop schedule_start
}

proc qc::schedule {args} {
    #| Schedule proc for execution unless already scheduled. Start schedule if it is not already running.
    # Usage:
    # schedule -thread "50 seconds" a_proc_name ?arg? ?arg?
    # schedule -thread "5 minutes" a_proc_name ?arg? ?arg?
    # schedule "1 hour" a_proc_name ?arg? ?arg?
    # schedule "10:15" a_proc_name arg arg
    # schedule "Monday 10:15" a_proc_name arg arg
    args $args -thread -- schedule proc_name args
    
    set switches {}
    if { [info exists thread] } {
	lappend switches -thread
    }

    if { ![qc::schedule_exists $proc_name] } {
	switch -regexp -nocase -matchvar match -- $schedule {
	    {^([0-9]+ +(seconds?|minutes?|hours?)($| +))+} {
		set interval 0
		foreach {value unit} $schedule {
		    if { [regexp -nocase {seconds?} $unit] } {
			set interval [expr {$interval + $value}]
		    } elseif { [regexp -nocase {minutes?} $unit] } {
			set interval [expr {$interval + ($value * 60)}]
		    } elseif { [regexp -nocase {hours?} $unit] } {
			set interval [expr {$interval + ($value * 60 * 60)}]
		    }		    
		}
		ns_schedule_proc {*}$switches $interval $proc_name {*}$args
	    }
	    {^([0-9]{1,2}):([0-9]{2})$} {
		set hour [qc::cast integer [lindex $match 1]]
		set minute [qc::cast integer [lindex $match 2]]
		ns_schedule_daily {*}$switches $hour $minute $proc_name {*}$args
	    }
	    {^(Monday|Tuesday|Wednesday|Thursday|Friday|Saturday|Sunday) +([0-9]{1,2}):([0-9]{2})$} {
		set day [lindex $match 1]
		set dow_map {Sunday 0 Monday 1 Tuesday 2 Wednesday 3 Thursday 4 Friday 5 Saturday 6} 
		set dow [string map -nocase $dow_map $day]
		set hour [qc::cast integer [lindex $match 2]]
		set minute [qc::cast integer [lindex $match 3]]
		ns_schedule_weekly {*}$switches $dow $hour $minute $proc_name {*}$args
	    }
	    default {
		error "Invalid syntax for schedule variable, \"$schedule\""
	    }
	}
	log Notice "Scheduled $proc_name" 
    }
    if { ![schedule_running $proc_name] } {
	schedule_start $proc_name
    }
}

proc qc::schedule_id {proc_name} {
    #| Return id for schedule proc_name.    
    foreach schedule [ns_info scheduled] {
	lassign $schedule id . . . . . . . arg
	if { [regexp "^${proc_name}( .*)?" $arg] } {
	    return $id
	}
    }
    error "Could not determine id for scheduled proc \"$proc_name\""
}

proc qc::schedule_exists {proc_name} {
    #| Return true if proc has been scheduled.
    foreach schedule [ns_info scheduled] {
	lassign $schedule . . . . . . . . arg
	if { [regexp "^${proc_name}( .*)?" $arg] } {
	    return true
	}
    }
    return false
}

proc qc::schedule_running {proc_name} {
    #| Return true if schdeule is running.
    foreach schedule [ns_info scheduled] {
	lassign $schedule . flag . . . . . . arg
	if { [regexp "^${proc_name}( .*)?" $arg] } {
	    # Check paused flag has not been set.
	    if { ! ($flag & 16) } {
		return true
	    } 
	    break
	}
    }
    return false
}

proc qc::schedule_stop {proc_name} {
    #| Stop schedule 
    if { [schedule_running $proc_name] } {	
	ns_pause [schedule_id $proc_name]
	log Notice "Schedule $proc_name has been stopped" 
    }
}

proc qc::schedule_start {proc_name} {
    #| Start schedule 
    if { ! [schedule_running $proc_name] } {	
	ns_resume [schedule_id $proc_name]
	log Notice "Schedule $proc_name has been started" 
    }
}

namespace eval qc {
    namespace export user_agent_parse
}

proc qc::user_agent_parse { user_agent } {
    #| Parse a user agent to a dict of device, os, and browser
    # Example:
    #     % set user_agent {Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_2) AppleWebKit/601.3.9 (KHTML, like Gecko) Version/9.0.2 Safari/601.3.9}
    #     % return [user_agent_parse $user_agent]
    #     device PC os MacOS browser Safari
    
    set user_agent [lower $user_agent]

    # Device
    set device "Unknown"    
    # PC
    if { [regexp {\y(windows nt|macintosh|linux|cros)\y} $user_agent] } {
        set device "PC"
    }

    # Tablet
    if { [regexp {\ywindows\y} $user_agent]
         && [regexp {\ytouch\y} $user_agent]
         && ! [regexp {\ytablet pc\y} $user_agent] } {
        set device "Tablet"
    } elseif { [regexp {\yandroid\y} $user_agent]
               && ! [regexp {\ymobile\y} $user_agent] } {
        set device "Tablet"
    } elseif { [regexp {\yipad\y} $user_agent] } {
        set device "Tablet"
    }

    # Mobile
    if { [regexp {\ywindows\y} $user_agent]
         && [regexp {\yphone\y} $user_agent] } {
        set device "Mobile"
    } elseif { [regexp {\yandroid\y} $user_agent]
               && [regexp {\ymobile\y} $user_agent] } {
        set device "Mobile"
    } elseif { [regexp {\yip(hone|od)\y} $user_agent] } {
        set device "Mobile"
    }

    # OS
    set os "Unknown"
    # Windows
    if { [regexp {\ywindows nt\y} $user_agent] } {
        set os "Windows"
    }
    # macOS
    if { [regexp {\ymacintosh\y} $user_agent] } {
        set os "macOS"
    }
    # ChromeOS
    if { [regexp {\ycros\y} $user_agent] } {
        set os "ChromeOS"
    }
    # Linux
    if { [regexp {\ylinux\y} $user_agent] } {
        set os "Linux"
    }
    # Android
    if { [regexp {\yandroid\y} $user_agent] } {
        set os "Android"
    }
    # Windows Phone
    if { [regexp {\ywindows phone\y} $user_agent] } {
        set os "Windows Phone"
    }
    # iOS
    if { [regexp {\yip(hone|ad|od)\y} $user_agent] } {
        set os "iOS"
    }

    # Browser
    set browser "Unknown"
    # Internet Explorer
    if { [regexp {\ytrident/7\y} $user_agent] } {
        set browser "Internet Explorer"
    }

    # Safari
    if { [regexp {\ysafari\y} $user_agent]
         && ! [regexp {\y(opr|chrome|presto)\y} $user_agent] } {
        set browser "Safari"
    }

    # Chrome
    if { [regexp {\ychrome\y} $user_agent]
         && ! [regexp {\y(edge|opr)\y} $user_agent] } {
        set browser "Chrome"
    } elseif { [regexp {\ycrios\y} $user_agent] } {
        set browser "Chrome"
    }

    # Firefox
    if { [regexp {\y(firefox|fxios)\y} $user_agent] } {
        set browser "Firefox"
    }

    # Microsoft Edge
    if { [regexp {\yedge?\y} $user_agent] } {
        set browser "Microsoft Edge"
    }

    return [dict create \
                device $device \
                os $os \
                browser $browser \
               ]
}

namespace eval qc {
    namespace export filter_validate filter_authenticate filter_http_request_validate filter_file_alias_path file_alias_path_exists file_alias_path_new file_alias_path2file_id file_alias_path_update file_alias_path_delete filter_handler_exists filter_fastpath_gzip filter_set_expires filter_form_variables_validate
}

proc qc::filter_validate {event {error_handler qc::error_handler}} {
    #| Matches the URL to a handler and validates the form data if the method is a type of POST.
    # Default to qc::error_handler
    if {$error_handler eq ""} {
        set error_handler qc::error_handler
    }
    ::try {
        set url_path [qc::conn_path]
        set method [string toupper [qc::conn_method]]

        if { $method eq "VALIDATE" } {
            # Client requested only a validation service
            
            if { [qc::registered VALIDATE $url_path] } {
                set validate_method VALIDATE
            } elseif { [qc::registered GET $url_path] } {
                # Check if we can use a GET handler as a substitute
                # TODO check for any registered handler to be used as substitute.
                set validate_method GET
            } elseif { [qc::registered POST $url_path] } {
                set validate_method POST
            } else {
                error "Cannot validate \"$url_path\" because there is no registered service for the path." {} NOT_FOUND
            }

            if { [qc::handlers validate2model $validate_method $url_path]
                 && [qc::handlers validation exists $validate_method $url_path] } {
                # Passed data model validation check custom validation
                qc::handlers validation call $validate_method $url_path
            }

            # Return the results of validation to the client
            qc::return_response
            return "filter_return"
        } elseif { [qc::registered $method $url_path] } {
            # Validate and if invalid return the response otherwise continue as normal

            if {[qc::handlers validate2model $method $url_path]
                 && [qc::handlers validation exists $method $url_path] } {
                # Passed data model validation check custom validation
                qc::handlers validation call $method $url_path
            }
            
            if { [qc::response status get] eq "invalid" } {
                # Validation failed
                if { ![qc::conn_served] } {
                    # Inform the client
                    qc::return_response
                }
                return "filter_return"
            }
        }
        
        # Validation successful or url_path doesn't require validation.
        return "filter_ok"
    } on error [list error_message options] {
        $error_handler $error_message $options
        return "filter_return"
    }
}

proc qc::filter_authenticate {event args} {
    #| Checks if session and authentication token are valid.
    # Default to qc::error_handler
    args $args -login_url "/user/login" -relogin_url "/user/session-expired" -reauth_expired_session -error_handler qc::error_handler -- 

    set url_path [qc::conn_path]
    set method [string toupper [qc::conn_method]]
    ::try {
        if { [qc::registered $method $url_path] } {
            # The method url_path has been registered
            if { ![qc::cookie_exists session_id] || ![qc::session_exists [qc::session_id]] } {
                # No session cookie or session doesn't exist - implicitly log in as anonymous user.
                global session_id current_user_id
                set current_user_id [qc::anonymous_user_id]
                set session_id [qc::anonymous_session_id]
                qc::cookie_set session_id $session_id
                qc::cookie_set authenticity_token [qc::session_authenticity_token $session_id] http_only false
            }

            if { $method in [list "GET" "HEAD"] && [qc::session_user_id [qc::session_id]] == [qc::anonymous_user_id] && [qc::session_id] != [qc::anonymous_session_id] } {
                # GET/HEAD request for anonymous session - roll session after 1 hour
                global session_id current_user_id
                set current_user_id [qc::anonymous_user_id]
                set session_id [qc::anonymous_session_id]
                qc::cookie_set session_id $session_id
                qc::cookie_set authenticity_token [qc::session_authenticity_token $session_id] http_only false
            }

            if { [qc::session_valid [qc::session_id]] } {
                # Valid session - refresh session
                qc::session_update [qc::session_id]
            } else {
                # Expired session (normal user) - implicitly log in as anonymous user and optionally redirect to login forms
                set expired_user_id [qc::session_user_id [qc::session_id]]
                global session_id current_user_id
                set current_user_id [qc::anonymous_user_id]
                set session_id [qc::anonymous_session_id]
                qc::cookie_set session_id $session_id
                qc::cookie_set authenticity_token [qc::session_authenticity_token $session_id] http_only false
                qc::session_update [qc::session_id]
                
                if { [qc::registered authenticate $method $url_path] && [info exists reauth_expired_session] && $expired_user_id != [qc::anonymous_user_id] } {
                    # Request (normal user) registered for authentication redirect to reauthentication page
                    db_1row {
                        select 
                        email 
                        from users
                        where user_id=:expired_user_id
                    }
                    set relogin_url [url $relogin_url login_code $email]
                    if { $method in [list "GET" "HEAD"] } {
                        # Redirect GET/HEAD requests back to this url after login
                        return_next [url $relogin_url next_url [url_here]]
                    } else {
                        # POST requests 
                        qc::response action redirect $relogin_url
                        qc::return_response
                    }
                    return "filter_return"    
                } elseif { [qc::registered authenticate $method $url_path] && $method ni [list "GET" "HEAD"]  && $expired_user_id != [qc::anonymous_user_id] } {
                    # POST request (normal user) registered for authentication - redirect to login page.
                    qc::response action redirect $login_url
                    qc::return_response
                    return "filter_return"
                }
            }

            if { [qc::registered authenticate $method $url_path] && $method ni [list "GET" "HEAD"] } {
                # POST request registered for authentication - check authenticity token
                set header_authenticity_token [qc::http_header_get X-Authenticity-Token]
                set form_authenticity_token ""
                set authenticity_token [qc::session_authenticity_token [qc::session_id]]
                set form_dict [qc::form2dict]
                if { [dict exists $form_dict _authenticity_token] } {
                    # form variable for authenticity token was given
                    set form_authenticity_token [dict get $form_dict _authenticity_token]
                }
                
                if { $header_authenticity_token eq $authenticity_token || $form_authenticity_token eq $authenticity_token } {
                    # authenticity token is valid
                    return "filter_ok"
                } elseif { $header_authenticity_token eq "" && $form_authenticity_token eq "" } {
                    # authenticity token wasn't given
                    error "Authenticity token was not found." {} AUTH
                } else {
                    # authenticity token is invalid
                    error "Authenticity token was invalid." {} AUTH
                }
            }            
        }
    } on error [list error_message options] {
        $error_handler $error_message $options
        return "filter_return"
    }
    return "filter_ok"
}

proc qc::filter_http_request_validate {event {error_handler "qc::error_handler"}} {
    #| Check that request string and connection url are valid.
    qc::setif error_handler "" "qc::error_handler"
    ::try {
        set request [ns_conn request]
        set url [qc::conn_path]
        if { ![qc::conn_request_is_valid $request] } {
            ns_returnbadrequest "\"$request\" is not a valid request."
            return filter_return
        }
        if { ![qc::is uri $url] } {
            return [ns_returnbadrequest "\"$url\" is not a valid URL."]
            return filter_return
        }
        
        return filter_ok
    } on error {error_message options} {
        $error_handler $error_message $options
        return filter_return
    }
}

proc qc::filter_file_alias_paths {event http_method {error_handler qc::error_handler}} {
    #| Preauth filter that handles file aliases
    qc::setif error_handler "" qc::error_handler
    ::try {
        set url_path [qc::conn_path]
        if { [qc::file_alias_path_exists [string trimright $url_path /]] } {
            # URL is an alias for a file
            set alias_path [string trimright $url_path /]
            set alias_file [ns_pagepath]$alias_path
            set file_id [qc::file_alias_path2file_id $alias_path]
            
            qc::db_trans {
                # Obtain db lock to prevent simultaneous requests trying to create a sym link at the same time
                qc::db_0or1row {select file_id from file where file_id=:file_id for update} {
                    return [ns_returnnotfound]
                }
                
                # Create disk cache for target file (if it doesn't exist) and get it's canonical url
                set target_path [dict get [qc::file_data [ns_pagepath]/file $file_id] url]
                set target_file [ns_pagepath]$target_path
                
                if  { ! [file exists $alias_file] } {
                    # Alias sym link doesn't exist - create it now
                    file mkdir [file dirname $alias_file]
                    file link $alias_file $target_file
                }  elseif { [file type $alias_file] eq "link" && [file type [file link $alias_file]] eq "file" && [file link $alias_file] ne $target_file } {
                    # Alias sym link exists but does not point to the correct target file - update sym link
                    file delete $alias_file
                    file mkdir [file dirname $alias_file]
                    file link $alias_file $target_file
                } 
            }

            # Register fastpath for alias_path
            ns_register_fastpath $http_method $alias_path
        }
        return filter_ok
    } on error {error_message options} {
        $error_handler $error_message $options
        return filter_return
    }
}     

proc qc::file_alias_path_exists {url_path} {
    #| Check if this url path is an alias for a file
    qc::db_0or1row {
        select 
        file_id
        from file_alias_path
        where 
        url_path=:url_path
    } {
        return false
    } {
        return true
    }
}

proc qc::file_alias_path2file_id {url_path} {
    #| Convert alias_path to file_id
    qc::db_1row {
        select 
        file_id
        from file_alias_path
        where 
        url_path=:url_path
    } 
    return $file_id
}

proc qc::file_alias_path_new {url_path file_id} {
    #| Create new file_alias_path record and it's corresponding sym link
    qc::db_trans {
        qc::db_dml {lock table file_alias_path in exclusive mode}
        qc::db_dml "insert into file_alias_path [qc::sql_insert url_path file_id]"
        # Create disk cache for target file (if it doesn't exist) and get it's canonical url
        set target_path [dict get [qc::file_data [ns_pagepath]/file $file_id] url]
        set target_file [ns_pagepath]$target_path
        # Create sym link
        set alias_file [ns_pagepath]$url_path
        if { ![file exists $alias_file] } {
            # file doesn't exist - create sym link
            file mkdir [file dirname $alias_file]
            file link $alias_file $target_file
        } elseif { [file type $alias_file] ne "link" } {
            # file exists but is not a sym link - error as we can't create/update sym link
            error "Unable to create sym link \"$alias_file\" as a regular file at this file location already exists"
        } elseif { [file type [file link $alias_file]] ne "file"} {
            # sym link exists but doesn't point to a file -  error as we can't create/update sym link
            error "Unable to update sym link \"$alias_file\" as the current link target is not a regular file"
        } elseif { [file type [file link $alias_file]] eq "file" && [file link $alias_file] ne $target_file} {
            # sym link exists but doesn't point to the correct file - remove sym link and create a new one
            file delete $alias_file
            file mkdir [file dirname $alias_file]
            file link $alias_file $target_file
        } else {
            # sym link exists and already points to the correct file - nothing to do
        }
    }
}

proc qc::file_alias_path_update {url_path file_id} {
    #| Update an exisiting file alias_path record and it's corresponding sym link
    qc::db_trans {
        qc::db_1row {select url_path from file_alias_path where url_path=:url_path for update}
        qc::db_dml "update file_alias_path set [qc::sql_set file_id] where url_path=:url_path"
        # Create disk cache for target file (if it doesn't exist) and get it's canonical url
        set target_path [dict get [qc::file_data [ns_pagepath]/file $file_id] url]
        set target_file [ns_pagepath]$target_path
        # Update sym link
        set alias_file [ns_pagepath]$url_path
        if { ![file exists $alias_file] } {
            # file doesn't exist - create sym link
            file mkdir [file dirname $alias_file]
            file link $alias_file $target_file
        } elseif { [file type $alias_file] ne "link" } {
            # file exists but is not a sym link - error as we can't create/update sym link
            error "Unable to create sym link \"$alias_file\" as a regular file at this file location already exists"
        } elseif { [file type [file link $alias_file]] ne "file"} {
            # sym link exists but doesn't point to a file -  error as we can't create/update sym link
            error "Unable to update sym link \"$alias_file\" as the current link target is not a regular file"
        } elseif { [file type [file link $alias_file]] eq "file" && [file link $alias_file] ne $target_file} {
            # sym link exists but doesn't point to the correct file - remove sym link and create a new one
            file delete $alias_file
            file mkdir [file dirname $alias_file]
            file link $alias_file $target_file
        } else {
            # sym link exists and already points to the correct file - nothing to do
        }
    }
}

proc qc::file_alias_path_delete {url_path} {
    #| Delete a file alias_path record and it's corresponding sym link
    qc::db_trans {
        qc::db_1row {select url_path from file_alias_path where url_path=:url_path for update}
        qc::db_dml "delete from file_alias_path where url_path=:url_path"
        # remove sym link
        set alias_file [ns_pagepath]$url_path
        if { [file exists $alias_file] && [file type $alias_file] eq "link" && [file type [file link $alias_file]] eq "file" } {
            file delete $alias_file
        }
    }
}

proc qc::filter_fastpath_gzip {filter_when file_extensions} {
    #| Postauth filter to seed filesystem with .gz versions of static files if the client accepts gzip-ed content.
    set conn_path [qc::conn_path]
    set file_path [ns_pagepath]$conn_path
    set gzipped_file_path [ns_pagepath]${conn_path}.gz
    
    if { [ns_conn zipaccepted] \
             && ![file readable $gzipped_file_path] \
             && [file readable $file_path] \
             && ( [llength $file_extensions] == 0 || [file extension $file_path] in $file_extensions ) \
             && [file type $file_path] in [list "link" "file"] \
         } {
        ns_gzipfile $file_path $gzipped_file_path
    }
    return "filter_ok"
}

proc qc::filter_set_expires {filter_when seconds {cache_response_directive ""}} {
    #| Postauth filter to set cache control headers: Expires, & Cache-Control. 
    #| If "cache_response_directive" is specified the function adds the "max-age" header field to the response "Cache-Control" header.
    # cache_response_directive: public, private, no-cache, no-store, no-transform, must-revalidate, or proxy-revalidate
    if { $cache_response_directive ne "" } {
        ns_setexpires -cache-control $cache_response_directive $seconds
    } else {
        ns_setexpires $seconds
    }
    return "filter_ok"
}

proc qc::filter_set_headers {filter_when args} {
    #| Filter to set header values.
    foreach {name value} $args {
        ns_set update [ns_conn outputheaders] $name $value
    }
    return "filter_ok"
}

proc qc::filter_form_variables_validate {event args} {
    #| Check form variable names to prevent Tcl namespaced variables from being
    #| set or overwritten.
    qc::args $args -log -error_handler "qc::error_handler" --

    ::try {
        set names [qc::form_var_names]
        foreach name $names {
            if { [regexp {::} $name] } {
                set escaped [qc::html_escape $name]
                ns_returnbadrequest "Invalid form variable name: \"$escaped\""
                
                if { [info exists log] } {
                    set request [ns_conn request]
                    log "qc::filter_http_request_validate:\
                         Invalid Form Variable Name: \"$name\"\
                         for request \"$request\""
                    qc::email_support \
                        subject "Invalid HTTP Request" \
                        html "Invalid form variable name \"$escaped\"\
                              in request \"[qc::html_escape $request]\""
                }
                
                return filter_return
            }
        }
        
        return filter_ok
    } on error {error_message options} {
        $error_handler $error_message $options
        return filter_return
    }
}


namespace eval qc {
    namespace export html_table_paged
}

proc qc::html_table_paged { args } {
    #| Split a table up into parts that fit on a printed page.
    # Usage: html_table_paged page_row_count 20 
    # Usage: html_table_paged first_page_row_count 18 page_row_count 20
    # Usage: html_table_paged single_page_row_count 15 first_page_row_count 18 middle_page_row_count 20 last_page_row_count 17
    set args [args2dict $args] 
    dict2vars $args data table tbody cols class page_row_count first_page_row_count middle_page_row_count last_page_row_count single_page_row_count
   
    # tbody from table
    if { [info exists table] && ![info exists tbody] } {
	set tbody [qc::html_table_tbody_from_table $table $cols]
    }
    # tbody from data
    if { [info exists data] && ![info exists tbody] } {
	set tbody [qc::html_table_tbody_from_ldict $data $cols]
        dict unset args data
    }

    # Maximum number of rows to display on a page (excluding thead and tfoot rows)
    if { [info exists page_row_count] } {
        default middle_page_row_count $page_row_count first_page_row_count $page_row_count last_page_row_count $page_row_count
        default single_page_row_count [expr {$first_page_row_count + $last_page_row_count - $middle_page_row_count}]
    }
    
    if { [llength $tbody]<=$single_page_row_count } {
        set tbodies [list $tbody]
    } else {
        # page 1
        lappend tbodies [lrange $tbody 0 [expr $first_page_row_count - 1]]
        for {set index $first_page_row_count} {([llength $tbody]- $index) > $middle_page_row_count} {incr index $middle_page_row_count} {
            # page n
            lappend tbodies [lrange $tbody $index [expr {$index+$middle_page_row_count-1}]]
        }
        # last page
        lappend tbodies [lrange $tbody $index end]
        if {$index < [llength $tbody]-$last_page_row_count} {
            lappend tbodies {}
        }
    }

    # Add class paged
    if { [info exists class] } {
        lappend class paged
    } else {
        set class paged
    }
    dict set args class $class

    set list {}
    foreach tbody $tbodies {
        dict set args tbody $tbody
        lappend list [html_table {*}$args]
    }
    return [join $list \n]
}

namespace eval qc {
    namespace export cookie_*
}

proc qc::cookie_string_is_valid {cookie_string} {
    #| Test if the cookie string conforms to the RFC
    set re {
        ^
        ([^[:cntrl:]()<>@,;:\\\"/\[\]?={}\ \t]+
        =
        (\"[^[:cntrl:]\s\",;\\]*\"|[^[:cntrl:]\s\",;\\]*)
        (
         ;\ [^[:cntrl:]()<>@,;:\\\"/\[\]?={}\ \t]+
         =
         (\"[^[:cntrl:]\s\",;\\]*\"|[^[:cntrl:]\s\",;\\]*)
         )*)?
        $
    }
    return [regexp -expanded $re $cookie_string]
}

proc qc::cookie_string2multimap {cookie_string} {
    #| Extracts a multimap of name-value pairs from a cookie string
    set cookie_map {}
    set cookie_string [trim $cookie_string]
    if { [string first ";" $cookie_string] == -1 && [string first = $cookie_string] == -1 } {
        return $cookie_map
    }
    while { $cookie_string ne "" } {
        set index [string first ";" $cookie_string]
        if { $index == -1 } {
            set cookie_fragment $cookie_string
            set cookie_string ""
        } else {
            set cookie_fragment [string range $cookie_string 0 [expr {$index - 1}]]
            set cookie_string [string range $cookie_string [expr {$index + 1}] end]
        }
        set index2 [string first = $cookie_fragment]
        if { $index2 == -1 } {
            set name $cookie_fragment
            set value ""
        } else {
            set name [string range $cookie_fragment 0 [expr {$index2 - 1}]]
            set value [string range $cookie_fragment [expr {$index2 + 1}] end]
        }
        set name [string trim $name]
        set value [string trim $value]
        if { [string index $value 0] eq "\"" && [string index $value end] eq "\"" } {
            set value [string range $value 1 end-1]
        }
        if { $name ne "" } {
            lappend cookie_map [qc::url_decode $name] [qc::url_decode $value]
        }
    }
    return $cookie_map
}

proc qc::cookie_get { search_name } {
    #| Get a cookie value or throw an error
    set headers [ns_conn headers]
    set cookie_string [ns_set iget $headers Cookie]
    set cookie_map [qc::cookie_string2multimap $cookie_string]
    if { [multimap_exists $cookie_map $search_name] } {
        return [multimap_get_first $cookie_map $search_name]
    } else {
	error "Cookie [qc::url_decode $search_name] does not exist"
    }
}

proc qc::cookie_exists { name } {
    #| Test if the cookie exists
    if { [info command ns_conn] ne "ns_conn" } {
       return false
    }
    if { ![ns_conn isconnected] } {
        return false
    }
    set headers [ns_conn headers]
    set cookie_string [ns_set iget $headers Cookie]
    if { ! [qc::cookie_string_is_valid $cookie_string] } {
        return false
    }
    set cookie_map [qc::cookie_string2multimap $cookie_string]
    return [multimap_exists $cookie_map [qc::url_encode $name]]
}

proc qc::cookie_set {name value args} {
    #| Set a cookie in outgoing headers for the current conection.
    #| Optional named args:
    #| expires datetime, max_age seconds
    #| domain url, path path ,secure boolean
    # All browsers seem to support Expires but not all support max-age
    # which is supposed to supersede Expires RFC 2109
    array set option $args
    default option(secure) false
    # http_only means cookie is not available to JavaScript.
    default option(http_only) true
    default option(path) /
    set headers [ns_conn outputheaders]
    set cookie "[url_encode $name]=[url_encode $value]"

    if { [info exists option(path)] } {
        append cookie "; path=$option(path)"
    }

    if { [info exists option(max_age)] } {
        append cookie "; max-age=$option(max_age)"
    }

    if { [info exists option(expires)] } {
        append cookie "; expires=[ns_httptime [cast epoch $option(expires)]]"
    }

    if { [info exists option(domain)] } {
        append cookie "; domain=$option(domain)"
    }

    if { [string is true $option(secure)] } {
        append cookie "; secure"
    }

    if { [string is true $option(http_only)] } {
        append cookie "; HttpOnly"
    }

    ns_set put $headers "Set-Cookie" $cookie
}


namespace eval qc::nsv_dict {
    namespace export exists set unset get lappend
    namespace ensemble create

    proc exists {variable key args} {
        #| Returns Boolean indicating whether the given key (or path of keys through a set of nested dictionaries) exists in the given nsv array. 
        ::set keys [list $key {*}$args]
        if { ! [nsv_array exists $variable] } {
            # nsv array does not exist
            return false
        } elseif { ! [nsv_exists $variable [lindex $keys 0]] } {
            # Key doesn't exist in nsv_array
            return false
        } elseif { [llength $keys] == 1 } {
            # 1 key - we have already checked that the key exists in nsv_array
            return true
        } else {
            # Multiple keys - check the key path exists in the dictionary stored in nsv array
            return [dict exists [nsv_get $variable [lindex $keys 0]] {*}[lrange $keys 1 end]]
        }
    }

    proc set {args} {
        #| Sets/updates dictionary value corresponding to a given key in a nsv_array.
        #| When multiple keys are present, this operation creates or updates a chain of nested dictionaries.
        if { [llength $args] < 3 } {
            error "wrong # args: should be \"nsv_dict set variable key ... value\""
        }
        ::set variable [lindex $args 0]
        ::set keys [lrange $args 1 end-1]
        ::set value [lindex $args end]
        
        if { [llength $keys] == 1 } {
            # 1 key - set the value in an nsv array.
            nsv_set $variable [lindex $keys 0] $value
        } else {
            # Multiple keys - set/update dictionary in nsv array
            if { [nsv_exists $variable [lindex $keys 0]] } {
                ::set temp [nsv_get $variable [lindex $keys 0]]
            } else {
                ::set temp {}
            }
            dict set temp {*}[lrange $keys 1 end] $value
            nsv_set $variable [lindex $keys 0] $temp            
        } 
        return 1
    }

    proc unset {variable key args} {
        #| Unsets a given key in dictionary stored in a nsv_array.
        #| Where multiple keys are present, this describes a path through nested dictionaries to the mapping to remove.
        ::set keys [list $key {*}$args]
        if { ! [exists $variable {*}$keys] } {
            error "Key \"$keys\" not known in dictionary"
        } elseif { [llength $keys] == 1 } {
            # 1 key - unset the key in an nsv array.
            nsv_unset $variable [lindex $keys 0]
        } else {
            # Multiple keys - unset key path from dictionary stored in nsv_array
            ::set temp [nsv_get $variable [lindex $keys 0]]
            dict unset temp {*}[lrange $keys 1 end]
            nsv_set $variable [lindex $keys 0] $temp
        }     
        return 1   
    }

    proc get {variable key args} {
        #| Retrieve the value corresponding to a dictionary key stored in a nsv_array.
        ::set keys [list $key {*}$args]
        if { ! [exists $variable {*}$keys] } {
            error "Key \"$keys\" not known in dictionary"
        } elseif { [llength $keys] == 1 } {
            # 1 key - get value of the key in the nsv_array
            return [nsv_get $variable [lindex $keys 0]]
        } else {
            # Multiple keys - get the value corresponding to the key path of dictionary stored in nsv_array
            ::set temp [nsv_get $variable [lindex $keys 0]]
            return [dict get $temp {*}[lrange $keys 1 end]]
        }        
    }

    proc lappend {variable key values} {
        #| Append values to the list corresponding to the dictionary key stored in a nsv_array
        nsv_lappend $variable $key {*}$values
    }
}

namespace eval qc {
    namespace export sftp_put
}

proc qc::sftp_put {args} {
    #| Write data to the filename on the remote host using sftp
    args $args -port 22 username password host filename data
    if { [regexp {^/} $filename] } {
	set url sftp://${host}:${port}${filename}
    } else {
	set url sftp://${host}:${port}/~/${filename}
    }
    
    set file [open "|[qc::which curl] -k -T - -u $username:$password $url 2>/dev/null" w]
    puts -nonewline $file $data
    close $file
}

